package io.netty.bootstrap;

import io.netty.channel.Channel;
import io.netty.channel.Channel.Unsafe;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelFutureListener;
import io.netty.channel.ChannelHandler;
import io.netty.channel.ChannelOption;
import io.netty.channel.ChannelPromise;
import io.netty.channel.DefaultChannelPromise;
import io.netty.channel.EventLoop;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.ReflectiveChannelFactory;
import io.netty.util.AttributeKey;
import io.netty.util.concurrent.EventExecutor;
import io.netty.util.concurrent.GlobalEventExecutor;
import io.netty.util.internal.StringUtil;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;

public abstract class AbstractBootstrap<B extends AbstractBootstrap<B, C>, C extends Channel>
  implements Cloneable
{
  volatile EventLoopGroup group;
  private volatile ChannelFactory<? extends C> channelFactory;
  private volatile SocketAddress localAddress;
  private final Map<ChannelOption<?>, Object> options = new LinkedHashMap();
  private final Map<AttributeKey<?>, Object> attrs = new LinkedHashMap();
  private volatile ChannelHandler handler;
  
  AbstractBootstrap() {}
  
  AbstractBootstrap(AbstractBootstrap<B, C> bootstrap)
  {
    group = group;
    channelFactory = channelFactory;
    handler = handler;
    localAddress = localAddress;
    synchronized (options)
    {
      options.putAll(options);
    }
    synchronized (attrs)
    {
      attrs.putAll(attrs);
    }
  }
  
  public B group(EventLoopGroup group)
  {
    if (group == null) {
      throw new NullPointerException("group");
    }
    if (this.group != null) {
      throw new IllegalStateException("group set already");
    }
    this.group = group;
    return this;
  }
  
  public B channel(Class<? extends C> channelClass)
  {
    if (channelClass == null) {
      throw new NullPointerException("channelClass");
    }
    return channelFactory(new ReflectiveChannelFactory(channelClass));
  }
  
  @Deprecated
  public B channelFactory(ChannelFactory<? extends C> channelFactory)
  {
    if (channelFactory == null) {
      throw new NullPointerException("channelFactory");
    }
    if (this.channelFactory != null) {
      throw new IllegalStateException("channelFactory set already");
    }
    this.channelFactory = channelFactory;
    return this;
  }
  
  public B channelFactory(io.netty.channel.ChannelFactory<? extends C> channelFactory)
  {
    return channelFactory(channelFactory);
  }
  
  public B localAddress(SocketAddress localAddress)
  {
    this.localAddress = localAddress;
    return this;
  }
  
  public B localAddress(int inetPort)
  {
    return localAddress(new InetSocketAddress(inetPort));
  }
  
  public B localAddress(String inetHost, int inetPort)
  {
    return localAddress(new InetSocketAddress(inetHost, inetPort));
  }
  
  public B localAddress(InetAddress inetHost, int inetPort)
  {
    return localAddress(new InetSocketAddress(inetHost, inetPort));
  }
  
  public <T> B option(ChannelOption<T> option, T value)
  {
    if (option == null) {
      throw new NullPointerException("option");
    }
    if (value == null) {
      synchronized (options)
      {
        options.remove(option);
      }
    } else {
      synchronized (options)
      {
        options.put(option, value);
      }
    }
    return this;
  }
  
  public <T> B attr(AttributeKey<T> key, T value)
  {
    if (key == null) {
      throw new NullPointerException("key");
    }
    if (value == null) {
      synchronized (attrs)
      {
        attrs.remove(key);
      }
    } else {
      synchronized (attrs)
      {
        attrs.put(key, value);
      }
    }
    return this;
  }
  
  public B validate()
  {
    if (group == null) {
      throw new IllegalStateException("group not set");
    }
    if (channelFactory == null) {
      throw new IllegalStateException("channel or channelFactory not set");
    }
    return this;
  }
  
  public abstract B clone();
  
  public ChannelFuture register()
  {
    validate();
    return initAndRegister();
  }
  
  public ChannelFuture bind()
  {
    validate();
    SocketAddress localAddress = this.localAddress;
    if (localAddress == null) {
      throw new IllegalStateException("localAddress not set");
    }
    return doBind(localAddress);
  }
  
  public ChannelFuture bind(int inetPort)
  {
    return bind(new InetSocketAddress(inetPort));
  }
  
  public ChannelFuture bind(String inetHost, int inetPort)
  {
    return bind(new InetSocketAddress(inetHost, inetPort));
  }
  
  public ChannelFuture bind(InetAddress inetHost, int inetPort)
  {
    return bind(new InetSocketAddress(inetHost, inetPort));
  }
  
  public ChannelFuture bind(SocketAddress localAddress)
  {
    validate();
    if (localAddress == null) {
      throw new NullPointerException("localAddress");
    }
    return doBind(localAddress);
  }
  
  private ChannelFuture doBind(final SocketAddress localAddress)
  {
    final ChannelFuture regFuture = initAndRegister();
    final Channel channel = regFuture.channel();
    if (regFuture.cause() != null) {
      return regFuture;
    }
    if (regFuture.isDone())
    {
      ChannelPromise promise = channel.newPromise();
      doBind0(regFuture, channel, localAddress, promise);
      return promise;
    }
    final PendingRegistrationPromise promise = new PendingRegistrationPromise(channel);
    regFuture.addListener(new ChannelFutureListener()
    {
      public void operationComplete(ChannelFuture future)
        throws Exception
      {
        Throwable cause = future.cause();
        if (cause != null)
        {
          promise.setFailure(cause);
        }
        else
        {
          promise.registered();
          
          AbstractBootstrap.doBind0(regFuture, channel, localAddress, promise);
        }
      }
    });
    return promise;
  }
  
  final ChannelFuture initAndRegister()
  {
    Channel channel = null;
    try
    {
      channel = channelFactory.newChannel();
      init(channel);
    }
    catch (Throwable t)
    {
      if (channel != null) {
        channel.unsafe().closeForcibly();
      }
      return new DefaultChannelPromise(channel, GlobalEventExecutor.INSTANCE).setFailure(t);
    }
    ChannelFuture regFuture = config().group().register(channel);
    if (regFuture.cause() != null) {
      if (channel.isRegistered()) {
        channel.close();
      } else {
        channel.unsafe().closeForcibly();
      }
    }
    return regFuture;
  }
  
  abstract void init(Channel paramChannel)
    throws Exception;
  
  private static void doBind0(ChannelFuture regFuture, final Channel channel, final SocketAddress localAddress, final ChannelPromise promise)
  {
    channel.eventLoop().execute(new Runnable()
    {
      public void run()
      {
        if (val$regFuture.isSuccess()) {
          channel.bind(localAddress, promise).addListener(ChannelFutureListener.CLOSE_ON_FAILURE);
        } else {
          promise.setFailure(val$regFuture.cause());
        }
      }
    });
  }
  
  public B handler(ChannelHandler handler)
  {
    if (handler == null) {
      throw new NullPointerException("handler");
    }
    this.handler = handler;
    return this;
  }
  
  @Deprecated
  public final EventLoopGroup group()
  {
    return group;
  }
  
  public abstract AbstractBootstrapConfig<B, C> config();
  
  static <K, V> Map<K, V> copiedMap(Map<K, V> map)
  {
    Map<K, V> copied;
    synchronized (map)
    {
      if (map.isEmpty()) {
        return Collections.emptyMap();
      }
      copied = new LinkedHashMap(map);
    }
    Map<K, V> copied;
    return Collections.unmodifiableMap(copied);
  }
  
  final Map<ChannelOption<?>, Object> options0()
  {
    return options;
  }
  
  final Map<AttributeKey<?>, Object> attrs0()
  {
    return attrs;
  }
  
  final SocketAddress localAddress()
  {
    return localAddress;
  }
  
  final ChannelFactory<? extends C> channelFactory()
  {
    return channelFactory;
  }
  
  final ChannelHandler handler()
  {
    return handler;
  }
  
  final Map<ChannelOption<?>, Object> options()
  {
    return copiedMap(options);
  }
  
  final Map<AttributeKey<?>, Object> attrs()
  {
    return copiedMap(attrs);
  }
  
  public String toString()
  {
    StringBuilder buf = new StringBuilder().append(StringUtil.simpleClassName(this)).append('(').append(config()).append(')');
    
    return buf.toString();
  }
  
  static final class PendingRegistrationPromise
    extends DefaultChannelPromise
  {
    private volatile boolean registered;
    
    PendingRegistrationPromise(Channel channel)
    {
      super();
    }
    
    void registered()
    {
      registered = true;
    }
    
    protected EventExecutor executor()
    {
      if (registered) {
        return super.executor();
      }
      return GlobalEventExecutor.INSTANCE;
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.bootstrap.AbstractBootstrap
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */