package io.netty.buffer;

import io.netty.util.AsciiString;
import io.netty.util.ByteProcessor;
import io.netty.util.ByteProcessor.IndexOfProcessor;
import io.netty.util.CharsetUtil;
import io.netty.util.Recycler;
import io.netty.util.Recycler.Handle;
import io.netty.util.concurrent.FastThreadLocal;
import io.netty.util.internal.MathUtil;
import io.netty.util.internal.ObjectUtil;
import io.netty.util.internal.PlatformDependent;
import io.netty.util.internal.StringUtil;
import io.netty.util.internal.SystemPropertyUtil;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.CharBuffer;
import java.nio.charset.CharacterCodingException;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CoderResult;
import java.nio.charset.CodingErrorAction;
import java.util.Arrays;
import java.util.Locale;

public final class ByteBufUtil
{
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(ByteBufUtil.class);
  private static final FastThreadLocal<CharBuffer> CHAR_BUFFERS = new FastThreadLocal()
  {
    protected CharBuffer initialValue()
      throws Exception
    {
      return CharBuffer.allocate(1024);
    }
  };
  private static final byte WRITE_UTF_UNKNOWN = 63;
  private static final int MAX_CHAR_BUFFER_SIZE;
  private static final int THREAD_LOCAL_BUFFER_SIZE;
  private static final int MAX_BYTES_PER_CHAR_UTF8 = (int)CharsetUtil.encoder(CharsetUtil.UTF_8).maxBytesPerChar();
  static final ByteBufAllocator DEFAULT_ALLOCATOR;
  
  static
  {
    String allocType = SystemPropertyUtil.get("io.netty.allocator.type", PlatformDependent.isAndroid() ? "unpooled" : "pooled");
    
    allocType = allocType.toLowerCase(Locale.US).trim();
    ByteBufAllocator alloc;
    if ("unpooled".equals(allocType))
    {
      ByteBufAllocator alloc = UnpooledByteBufAllocator.DEFAULT;
      logger.debug("-Dio.netty.allocator.type: {}", allocType);
    }
    else if ("pooled".equals(allocType))
    {
      ByteBufAllocator alloc = PooledByteBufAllocator.DEFAULT;
      logger.debug("-Dio.netty.allocator.type: {}", allocType);
    }
    else
    {
      alloc = PooledByteBufAllocator.DEFAULT;
      logger.debug("-Dio.netty.allocator.type: pooled (unknown: {})", allocType);
    }
    DEFAULT_ALLOCATOR = alloc;
    
    THREAD_LOCAL_BUFFER_SIZE = SystemPropertyUtil.getInt("io.netty.threadLocalDirectBufferSize", 65536);
    logger.debug("-Dio.netty.threadLocalDirectBufferSize: {}", Integer.valueOf(THREAD_LOCAL_BUFFER_SIZE));
    
    MAX_CHAR_BUFFER_SIZE = SystemPropertyUtil.getInt("io.netty.maxThreadLocalCharBufferSize", 16384);
    logger.debug("-Dio.netty.maxThreadLocalCharBufferSize: {}", Integer.valueOf(MAX_CHAR_BUFFER_SIZE));
  }
  
  public static String hexDump(ByteBuf buffer)
  {
    return hexDump(buffer, buffer.readerIndex(), buffer.readableBytes());
  }
  
  public static String hexDump(ByteBuf buffer, int fromIndex, int length)
  {
    return HexUtil.hexDump(buffer, fromIndex, length);
  }
  
  public static String hexDump(byte[] array)
  {
    return hexDump(array, 0, array.length);
  }
  
  public static String hexDump(byte[] array, int fromIndex, int length)
  {
    return HexUtil.hexDump(array, fromIndex, length);
  }
  
  public static int hashCode(ByteBuf buffer)
  {
    int aLen = buffer.readableBytes();
    int intCount = aLen >>> 2;
    int byteCount = aLen & 0x3;
    
    int hashCode = 1;
    int arrayIndex = buffer.readerIndex();
    if (buffer.order() == ByteOrder.BIG_ENDIAN) {
      for (int i = intCount; i > 0; i--)
      {
        hashCode = 31 * hashCode + buffer.getInt(arrayIndex);
        arrayIndex += 4;
      }
    } else {
      for (int i = intCount; i > 0; i--)
      {
        hashCode = 31 * hashCode + swapInt(buffer.getInt(arrayIndex));
        arrayIndex += 4;
      }
    }
    for (int i = byteCount; i > 0; i--) {
      hashCode = 31 * hashCode + buffer.getByte(arrayIndex++);
    }
    if (hashCode == 0) {
      hashCode = 1;
    }
    return hashCode;
  }
  
  public static boolean equals(ByteBuf a, int aStartIndex, ByteBuf b, int bStartIndex, int length)
  {
    if ((aStartIndex < 0) || (bStartIndex < 0) || (length < 0)) {
      throw new IllegalArgumentException("All indexes and lengths must be non-negative");
    }
    if ((a.writerIndex() - length < aStartIndex) || (b.writerIndex() - length < bStartIndex)) {
      return false;
    }
    int longCount = length >>> 3;
    int byteCount = length & 0x7;
    if (a.order() == b.order()) {
      for (int i = longCount; i > 0; i--)
      {
        if (a.getLong(aStartIndex) != b.getLong(bStartIndex)) {
          return false;
        }
        aStartIndex += 8;
        bStartIndex += 8;
      }
    } else {
      for (int i = longCount; i > 0; i--)
      {
        if (a.getLong(aStartIndex) != swapLong(b.getLong(bStartIndex))) {
          return false;
        }
        aStartIndex += 8;
        bStartIndex += 8;
      }
    }
    for (int i = byteCount; i > 0; i--)
    {
      if (a.getByte(aStartIndex) != b.getByte(bStartIndex)) {
        return false;
      }
      aStartIndex++;
      bStartIndex++;
    }
    return true;
  }
  
  public static boolean equals(ByteBuf bufferA, ByteBuf bufferB)
  {
    int aLen = bufferA.readableBytes();
    if (aLen != bufferB.readableBytes()) {
      return false;
    }
    return equals(bufferA, bufferA.readerIndex(), bufferB, bufferB.readerIndex(), aLen);
  }
  
  public static int compare(ByteBuf bufferA, ByteBuf bufferB)
  {
    int aLen = bufferA.readableBytes();
    int bLen = bufferB.readableBytes();
    int minLength = Math.min(aLen, bLen);
    int uintCount = minLength >>> 2;
    int byteCount = minLength & 0x3;
    int aIndex = bufferA.readerIndex();
    int bIndex = bufferB.readerIndex();
    if (uintCount > 0)
    {
      boolean bufferAIsBigEndian = bufferA.order() == ByteOrder.BIG_ENDIAN;
      
      int uintCountIncrement = uintCount << 2;
      long res;
      long res;
      if (bufferA.order() == bufferB.order()) {
        res = bufferAIsBigEndian ? compareUintBigEndian(bufferA, bufferB, aIndex, bIndex, uintCountIncrement) : compareUintLittleEndian(bufferA, bufferB, aIndex, bIndex, uintCountIncrement);
      } else {
        res = bufferAIsBigEndian ? compareUintBigEndianA(bufferA, bufferB, aIndex, bIndex, uintCountIncrement) : compareUintBigEndianB(bufferA, bufferB, aIndex, bIndex, uintCountIncrement);
      }
      if (res != 0L) {
        return (int)Math.min(2147483647L, res);
      }
      aIndex += uintCountIncrement;
      bIndex += uintCountIncrement;
    }
    for (int aEnd = aIndex + byteCount; aIndex < aEnd; bIndex++)
    {
      int comp = bufferA.getUnsignedByte(aIndex) - bufferB.getUnsignedByte(bIndex);
      if (comp != 0) {
        return comp;
      }
      aIndex++;
    }
    return aLen - bLen;
  }
  
  private static long compareUintBigEndian(ByteBuf bufferA, ByteBuf bufferB, int aIndex, int bIndex, int uintCountIncrement)
  {
    for (int aEnd = aIndex + uintCountIncrement; aIndex < aEnd; bIndex += 4)
    {
      long comp = bufferA.getUnsignedInt(aIndex) - bufferB.getUnsignedInt(bIndex);
      if (comp != 0L) {
        return comp;
      }
      aIndex += 4;
    }
    return 0L;
  }
  
  private static long compareUintLittleEndian(ByteBuf bufferA, ByteBuf bufferB, int aIndex, int bIndex, int uintCountIncrement)
  {
    for (int aEnd = aIndex + uintCountIncrement; aIndex < aEnd; bIndex += 4)
    {
      long comp = bufferA.getUnsignedIntLE(aIndex) - bufferB.getUnsignedIntLE(bIndex);
      if (comp != 0L) {
        return comp;
      }
      aIndex += 4;
    }
    return 0L;
  }
  
  private static long compareUintBigEndianA(ByteBuf bufferA, ByteBuf bufferB, int aIndex, int bIndex, int uintCountIncrement)
  {
    for (int aEnd = aIndex + uintCountIncrement; aIndex < aEnd; bIndex += 4)
    {
      long comp = bufferA.getUnsignedInt(aIndex) - bufferB.getUnsignedIntLE(bIndex);
      if (comp != 0L) {
        return comp;
      }
      aIndex += 4;
    }
    return 0L;
  }
  
  private static long compareUintBigEndianB(ByteBuf bufferA, ByteBuf bufferB, int aIndex, int bIndex, int uintCountIncrement)
  {
    for (int aEnd = aIndex + uintCountIncrement; aIndex < aEnd; bIndex += 4)
    {
      long comp = bufferA.getUnsignedIntLE(aIndex) - bufferB.getUnsignedInt(bIndex);
      if (comp != 0L) {
        return comp;
      }
      aIndex += 4;
    }
    return 0L;
  }
  
  public static int indexOf(ByteBuf buffer, int fromIndex, int toIndex, byte value)
  {
    if (fromIndex <= toIndex) {
      return firstIndexOf(buffer, fromIndex, toIndex, value);
    }
    return lastIndexOf(buffer, fromIndex, toIndex, value);
  }
  
  public static short swapShort(short value)
  {
    return Short.reverseBytes(value);
  }
  
  public static int swapMedium(int value)
  {
    int swapped = value << 16 & 0xFF0000 | value & 0xFF00 | value >>> 16 & 0xFF;
    if ((swapped & 0x800000) != 0) {
      swapped |= 0xFF000000;
    }
    return swapped;
  }
  
  public static int swapInt(int value)
  {
    return Integer.reverseBytes(value);
  }
  
  public static long swapLong(long value)
  {
    return Long.reverseBytes(value);
  }
  
  public static ByteBuf readBytes(ByteBufAllocator alloc, ByteBuf buffer, int length)
  {
    boolean release = true;
    ByteBuf dst = alloc.buffer(length);
    try
    {
      buffer.readBytes(dst);
      release = false;
      return dst;
    }
    finally
    {
      if (release) {
        dst.release();
      }
    }
  }
  
  private static int firstIndexOf(ByteBuf buffer, int fromIndex, int toIndex, byte value)
  {
    fromIndex = Math.max(fromIndex, 0);
    if ((fromIndex >= toIndex) || (buffer.capacity() == 0)) {
      return -1;
    }
    return buffer.forEachByte(fromIndex, toIndex - fromIndex, new ByteProcessor.IndexOfProcessor(value));
  }
  
  private static int lastIndexOf(ByteBuf buffer, int fromIndex, int toIndex, byte value)
  {
    fromIndex = Math.min(fromIndex, buffer.capacity());
    if ((fromIndex < 0) || (buffer.capacity() == 0)) {
      return -1;
    }
    return buffer.forEachByteDesc(toIndex, fromIndex - toIndex, new ByteProcessor.IndexOfProcessor(value));
  }
  
  public static ByteBuf writeUtf8(ByteBufAllocator alloc, CharSequence seq)
  {
    ByteBuf buf = alloc.buffer(utf8MaxBytes(seq));
    writeUtf8(buf, seq);
    return buf;
  }
  
  public static int writeUtf8(ByteBuf buf, CharSequence seq)
  {
    int len = seq.length();
    buf.ensureWritable(utf8MaxBytes(seq));
    for (;;)
    {
      if ((buf instanceof AbstractByteBuf))
      {
        AbstractByteBuf byteBuf = (AbstractByteBuf)buf;
        int written = writeUtf8(byteBuf, writerIndex, seq, len);
        writerIndex += written;
        return written;
      }
      if (!(buf instanceof WrappedByteBuf)) {
        break;
      }
      buf = buf.unwrap();
    }
    byte[] bytes = seq.toString().getBytes(CharsetUtil.UTF_8);
    buf.writeBytes(bytes);
    return bytes.length;
  }
  
  static int writeUtf8(AbstractByteBuf buffer, int writerIndex, CharSequence seq, int len)
  {
    int oldWriterIndex = writerIndex;
    for (int i = 0; i < len; i++)
    {
      char c = seq.charAt(i);
      if (c < '?')
      {
        buffer._setByte(writerIndex++, (byte)c);
      }
      else if (c < '?')
      {
        buffer._setByte(writerIndex++, (byte)(0xC0 | c >> '\006'));
        buffer._setByte(writerIndex++, (byte)(0x80 | c & 0x3F));
      }
      else if (StringUtil.isSurrogate(c))
      {
        if (!Character.isHighSurrogate(c))
        {
          buffer._setByte(writerIndex++, 63);
        }
        else
        {
          try
          {
            c2 = seq.charAt(++i);
          }
          catch (IndexOutOfBoundsException e)
          {
            char c2;
            buffer._setByte(writerIndex++, 63);
            break;
          }
          char c2;
          if (!Character.isLowSurrogate(c2))
          {
            buffer._setByte(writerIndex++, 63);
            buffer._setByte(writerIndex++, Character.isHighSurrogate(c2) ? '?' : c2);
          }
          else
          {
            int codePoint = Character.toCodePoint(c, c2);
            
            buffer._setByte(writerIndex++, (byte)(0xF0 | codePoint >> 18));
            buffer._setByte(writerIndex++, (byte)(0x80 | codePoint >> 12 & 0x3F));
            buffer._setByte(writerIndex++, (byte)(0x80 | codePoint >> 6 & 0x3F));
            buffer._setByte(writerIndex++, (byte)(0x80 | codePoint & 0x3F));
          }
        }
      }
      else
      {
        buffer._setByte(writerIndex++, (byte)(0xE0 | c >> '\f'));
        buffer._setByte(writerIndex++, (byte)(0x80 | c >> '\006' & 0x3F));
        buffer._setByte(writerIndex++, (byte)(0x80 | c & 0x3F));
      }
    }
    return writerIndex - oldWriterIndex;
  }
  
  public static int utf8MaxBytes(CharSequence seq)
  {
    return seq.length() * MAX_BYTES_PER_CHAR_UTF8;
  }
  
  public static ByteBuf writeAscii(ByteBufAllocator alloc, CharSequence seq)
  {
    ByteBuf buf = alloc.buffer(seq.length());
    writeAscii(buf, seq);
    return buf;
  }
  
  public static int writeAscii(ByteBuf buf, CharSequence seq)
  {
    int len = seq.length();
    buf.ensureWritable(len);
    if ((seq instanceof AsciiString))
    {
      AsciiString asciiString = (AsciiString)seq;
      buf.writeBytes(asciiString.array(), asciiString.arrayOffset(), asciiString.length());
    }
    else
    {
      for (;;)
      {
        if ((buf instanceof AbstractByteBuf))
        {
          AbstractByteBuf byteBuf = (AbstractByteBuf)buf;
          int written = writeAscii(byteBuf, writerIndex, seq, len);
          writerIndex += written;
          return written;
        }
        if ((buf instanceof WrappedByteBuf)) {
          buf = buf.unwrap();
        } else {
          buf.writeBytes(seq.toString().getBytes(CharsetUtil.US_ASCII));
        }
      }
    }
    return len;
  }
  
  static int writeAscii(AbstractByteBuf buffer, int writerIndex, CharSequence seq, int len)
  {
    for (int i = 0; i < len; i++) {
      buffer._setByte(writerIndex++, (byte)seq.charAt(i));
    }
    return len;
  }
  
  public static ByteBuf encodeString(ByteBufAllocator alloc, CharBuffer src, Charset charset)
  {
    return encodeString0(alloc, false, src, charset, 0);
  }
  
  public static ByteBuf encodeString(ByteBufAllocator alloc, CharBuffer src, Charset charset, int extraCapacity)
  {
    return encodeString0(alloc, false, src, charset, extraCapacity);
  }
  
  static ByteBuf encodeString0(ByteBufAllocator alloc, boolean enforceHeap, CharBuffer src, Charset charset, int extraCapacity)
  {
    CharsetEncoder encoder = CharsetUtil.encoder(charset);
    int length = (int)(src.remaining() * encoder.maxBytesPerChar()) + extraCapacity;
    boolean release = true;
    ByteBuf dst;
    ByteBuf dst;
    if (enforceHeap) {
      dst = alloc.heapBuffer(length);
    } else {
      dst = alloc.buffer(length);
    }
    try
    {
      ByteBuffer dstBuf = dst.internalNioBuffer(0, length);
      int pos = dstBuf.position();
      CoderResult cr = encoder.encode(src, dstBuf, true);
      if (!cr.isUnderflow()) {
        cr.throwException();
      }
      cr = encoder.flush(dstBuf);
      if (!cr.isUnderflow()) {
        cr.throwException();
      }
      dst.writerIndex(dst.writerIndex() + dstBuf.position() - pos);
      release = false;
      return dst;
    }
    catch (CharacterCodingException x)
    {
      throw new IllegalStateException(x);
    }
    finally
    {
      if (release) {
        dst.release();
      }
    }
  }
  
  static String decodeString(ByteBuf src, int readerIndex, int len, Charset charset)
  {
    if (len == 0) {
      return "";
    }
    CharsetDecoder decoder = CharsetUtil.decoder(charset);
    int maxLength = (int)(len * decoder.maxCharsPerByte());
    CharBuffer dst = (CharBuffer)CHAR_BUFFERS.get();
    if (dst.length() < maxLength)
    {
      dst = CharBuffer.allocate(maxLength);
      if (maxLength <= MAX_CHAR_BUFFER_SIZE) {
        CHAR_BUFFERS.set(dst);
      }
    }
    else
    {
      dst.clear();
    }
    if (src.nioBufferCount() == 1)
    {
      decodeString(decoder, src.internalNioBuffer(readerIndex, len), dst);
    }
    else
    {
      ByteBuf buffer = src.alloc().heapBuffer(len);
      try
      {
        buffer.writeBytes(src, readerIndex, len);
        
        decodeString(decoder, buffer.internalNioBuffer(0, len), dst);
      }
      finally
      {
        buffer.release();
      }
    }
    return dst.flip().toString();
  }
  
  private static void decodeString(CharsetDecoder decoder, ByteBuffer src, CharBuffer dst)
  {
    try
    {
      CoderResult cr = decoder.decode(src, dst, true);
      if (!cr.isUnderflow()) {
        cr.throwException();
      }
      cr = decoder.flush(dst);
      if (!cr.isUnderflow()) {
        cr.throwException();
      }
    }
    catch (CharacterCodingException x)
    {
      throw new IllegalStateException(x);
    }
  }
  
  public static ByteBuf threadLocalDirectBuffer()
  {
    if (THREAD_LOCAL_BUFFER_SIZE <= 0) {
      return null;
    }
    if (PlatformDependent.hasUnsafe()) {
      return ThreadLocalUnsafeDirectByteBuf.newInstance();
    }
    return ThreadLocalDirectByteBuf.newInstance();
  }
  
  public static byte[] getBytes(ByteBuf buf)
  {
    return getBytes(buf, buf.readerIndex(), buf.readableBytes());
  }
  
  public static byte[] getBytes(ByteBuf buf, int start, int length)
  {
    return getBytes(buf, start, length, true);
  }
  
  public static byte[] getBytes(ByteBuf buf, int start, int length, boolean copy)
  {
    if (MathUtil.isOutOfBounds(start, length, buf.capacity())) {
      throw new IndexOutOfBoundsException("expected: 0 <= start(" + start + ") <= start + length(" + length + ") <= " + "buf.capacity(" + buf.capacity() + ')');
    }
    if (buf.hasArray())
    {
      if ((copy) || (start != 0) || (length != buf.capacity()))
      {
        int baseOffset = buf.arrayOffset() + start;
        return Arrays.copyOfRange(buf.array(), baseOffset, baseOffset + length);
      }
      return buf.array();
    }
    byte[] v = new byte[length];
    buf.getBytes(start, v);
    return v;
  }
  
  public static void copy(AsciiString src, int srcIdx, ByteBuf dst, int dstIdx, int length)
  {
    if (MathUtil.isOutOfBounds(srcIdx, length, src.length())) {
      throw new IndexOutOfBoundsException("expected: 0 <= srcIdx(" + srcIdx + ") <= srcIdx + length(" + length + ") <= srcLen(" + src.length() + ')');
    }
    ((ByteBuf)ObjectUtil.checkNotNull(dst, "dst")).setBytes(dstIdx, src.array(), srcIdx + src.arrayOffset(), length);
  }
  
  public static void copy(AsciiString src, int srcIdx, ByteBuf dst, int length)
  {
    if (MathUtil.isOutOfBounds(srcIdx, length, src.length())) {
      throw new IndexOutOfBoundsException("expected: 0 <= srcIdx(" + srcIdx + ") <= srcIdx + length(" + length + ") <= srcLen(" + src.length() + ')');
    }
    ((ByteBuf)ObjectUtil.checkNotNull(dst, "dst")).writeBytes(src.array(), srcIdx + src.arrayOffset(), length);
  }
  
  public static String prettyHexDump(ByteBuf buffer)
  {
    return prettyHexDump(buffer, buffer.readerIndex(), buffer.readableBytes());
  }
  
  public static String prettyHexDump(ByteBuf buffer, int offset, int length)
  {
    return HexUtil.prettyHexDump(buffer, offset, length);
  }
  
  public static void appendPrettyHexDump(StringBuilder dump, ByteBuf buf)
  {
    appendPrettyHexDump(dump, buf, buf.readerIndex(), buf.readableBytes());
  }
  
  public static void appendPrettyHexDump(StringBuilder dump, ByteBuf buf, int offset, int length)
  {
    HexUtil.appendPrettyHexDump(dump, buf, offset, length);
  }
  
  private static final class HexUtil
  {
    private static final char[] BYTE2CHAR = new char['?'];
    private static final char[] HEXDUMP_TABLE = new char['?'];
    private static final String[] HEXPADDING = new String[16];
    private static final String[] HEXDUMP_ROWPREFIXES = new String['?'];
    private static final String[] BYTE2HEX = new String['?'];
    private static final String[] BYTEPADDING = new String[16];
    
    static
    {
      char[] DIGITS = "0123456789abcdef".toCharArray();
      for (int i = 0; i < 256; i++)
      {
        HEXDUMP_TABLE[(i << 1)] = DIGITS[(i >>> 4 & 0xF)];
        HEXDUMP_TABLE[((i << 1) + 1)] = DIGITS[(i & 0xF)];
      }
      for (int i = 0; i < HEXPADDING.length; i++)
      {
        int padding = HEXPADDING.length - i;
        StringBuilder buf = new StringBuilder(padding * 3);
        for (int j = 0; j < padding; j++) {
          buf.append("   ");
        }
        HEXPADDING[i] = buf.toString();
      }
      for (i = 0; i < HEXDUMP_ROWPREFIXES.length; i++)
      {
        StringBuilder buf = new StringBuilder(12);
        buf.append(StringUtil.NEWLINE);
        buf.append(Long.toHexString(i << 4 & 0xFFFFFFFF | 0x100000000));
        buf.setCharAt(buf.length() - 9, '|');
        buf.append('|');
        HEXDUMP_ROWPREFIXES[i] = buf.toString();
      }
      for (i = 0; i < BYTE2HEX.length; i++) {
        BYTE2HEX[i] = (' ' + StringUtil.byteToHexStringPadded(i));
      }
      for (i = 0; i < BYTEPADDING.length; i++)
      {
        int padding = BYTEPADDING.length - i;
        StringBuilder buf = new StringBuilder(padding);
        for (int j = 0; j < padding; j++) {
          buf.append(' ');
        }
        BYTEPADDING[i] = buf.toString();
      }
      for (i = 0; i < BYTE2CHAR.length; i++) {
        if ((i <= 31) || (i >= 127)) {
          BYTE2CHAR[i] = '.';
        } else {
          BYTE2CHAR[i] = ((char)i);
        }
      }
    }
    
    private static String hexDump(ByteBuf buffer, int fromIndex, int length)
    {
      if (length < 0) {
        throw new IllegalArgumentException("length: " + length);
      }
      if (length == 0) {
        return "";
      }
      int endIndex = fromIndex + length;
      char[] buf = new char[length << 1];
      
      int srcIdx = fromIndex;
      for (int dstIdx = 0; srcIdx < endIndex; dstIdx += 2)
      {
        System.arraycopy(HEXDUMP_TABLE, buffer.getUnsignedByte(srcIdx) << 1, buf, dstIdx, 2);srcIdx++;
      }
      return new String(buf);
    }
    
    private static String hexDump(byte[] array, int fromIndex, int length)
    {
      if (length < 0) {
        throw new IllegalArgumentException("length: " + length);
      }
      if (length == 0) {
        return "";
      }
      int endIndex = fromIndex + length;
      char[] buf = new char[length << 1];
      
      int srcIdx = fromIndex;
      for (int dstIdx = 0; srcIdx < endIndex; dstIdx += 2)
      {
        System.arraycopy(HEXDUMP_TABLE, (array[srcIdx] & 0xFF) << 1, buf, dstIdx, 2);srcIdx++;
      }
      return new String(buf);
    }
    
    private static String prettyHexDump(ByteBuf buffer, int offset, int length)
    {
      if (length == 0) {
        return "";
      }
      int rows = length / 16 + (length % 15 == 0 ? 0 : 1) + 4;
      StringBuilder buf = new StringBuilder(rows * 80);
      appendPrettyHexDump(buf, buffer, offset, length);
      return buf.toString();
    }
    
    private static void appendPrettyHexDump(StringBuilder dump, ByteBuf buf, int offset, int length)
    {
      if (MathUtil.isOutOfBounds(offset, length, buf.capacity())) {
        throw new IndexOutOfBoundsException("expected: 0 <= offset(" + offset + ") <= offset + length(" + length + ") <= " + "buf.capacity(" + buf.capacity() + ')');
      }
      if (length == 0) {
        return;
      }
      dump.append("         +-------------------------------------------------+" + StringUtil.NEWLINE + "         |  0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f |" + StringUtil.NEWLINE + "+--------+-------------------------------------------------+----------------+");
      
      int startIndex = offset;
      int fullRows = length >>> 4;
      int remainder = length & 0xF;
      for (int row = 0; row < fullRows; row++)
      {
        int rowStartIndex = (row << 4) + startIndex;
        
        appendHexDumpRowPrefix(dump, row, rowStartIndex);
        
        int rowEndIndex = rowStartIndex + 16;
        for (int j = rowStartIndex; j < rowEndIndex; j++) {
          dump.append(BYTE2HEX[buf.getUnsignedByte(j)]);
        }
        dump.append(" |");
        for (int j = rowStartIndex; j < rowEndIndex; j++) {
          dump.append(BYTE2CHAR[buf.getUnsignedByte(j)]);
        }
        dump.append('|');
      }
      if (remainder != 0)
      {
        int rowStartIndex = (fullRows << 4) + startIndex;
        appendHexDumpRowPrefix(dump, fullRows, rowStartIndex);
        
        int rowEndIndex = rowStartIndex + remainder;
        for (int j = rowStartIndex; j < rowEndIndex; j++) {
          dump.append(BYTE2HEX[buf.getUnsignedByte(j)]);
        }
        dump.append(HEXPADDING[remainder]);
        dump.append(" |");
        for (int j = rowStartIndex; j < rowEndIndex; j++) {
          dump.append(BYTE2CHAR[buf.getUnsignedByte(j)]);
        }
        dump.append(BYTEPADDING[remainder]);
        dump.append('|');
      }
      dump.append(StringUtil.NEWLINE + "+--------+-------------------------------------------------+----------------+");
    }
    
    private static void appendHexDumpRowPrefix(StringBuilder dump, int row, int rowStartIndex)
    {
      if (row < HEXDUMP_ROWPREFIXES.length)
      {
        dump.append(HEXDUMP_ROWPREFIXES[row]);
      }
      else
      {
        dump.append(StringUtil.NEWLINE);
        dump.append(Long.toHexString(rowStartIndex & 0xFFFFFFFF | 0x100000000));
        dump.setCharAt(dump.length() - 9, '|');
        dump.append('|');
      }
    }
  }
  
  static final class ThreadLocalUnsafeDirectByteBuf
    extends UnpooledUnsafeDirectByteBuf
  {
    private static final Recycler<ThreadLocalUnsafeDirectByteBuf> RECYCLER = new Recycler()
    {
      protected ByteBufUtil.ThreadLocalUnsafeDirectByteBuf newObject(Recycler.Handle<ByteBufUtil.ThreadLocalUnsafeDirectByteBuf> handle)
      {
        return new ByteBufUtil.ThreadLocalUnsafeDirectByteBuf(handle, null);
      }
    };
    private final Recycler.Handle<ThreadLocalUnsafeDirectByteBuf> handle;
    
    static ThreadLocalUnsafeDirectByteBuf newInstance()
    {
      ThreadLocalUnsafeDirectByteBuf buf = (ThreadLocalUnsafeDirectByteBuf)RECYCLER.get();
      buf.setRefCnt(1);
      return buf;
    }
    
    private ThreadLocalUnsafeDirectByteBuf(Recycler.Handle<ThreadLocalUnsafeDirectByteBuf> handle)
    {
      super(256, Integer.MAX_VALUE);
      this.handle = handle;
    }
    
    protected void deallocate()
    {
      if (capacity() > ByteBufUtil.THREAD_LOCAL_BUFFER_SIZE)
      {
        super.deallocate();
      }
      else
      {
        clear();
        handle.recycle(this);
      }
    }
  }
  
  static final class ThreadLocalDirectByteBuf
    extends UnpooledDirectByteBuf
  {
    private static final Recycler<ThreadLocalDirectByteBuf> RECYCLER = new Recycler()
    {
      protected ByteBufUtil.ThreadLocalDirectByteBuf newObject(Recycler.Handle<ByteBufUtil.ThreadLocalDirectByteBuf> handle)
      {
        return new ByteBufUtil.ThreadLocalDirectByteBuf(handle, null);
      }
    };
    private final Recycler.Handle<ThreadLocalDirectByteBuf> handle;
    
    static ThreadLocalDirectByteBuf newInstance()
    {
      ThreadLocalDirectByteBuf buf = (ThreadLocalDirectByteBuf)RECYCLER.get();
      buf.setRefCnt(1);
      return buf;
    }
    
    private ThreadLocalDirectByteBuf(Recycler.Handle<ThreadLocalDirectByteBuf> handle)
    {
      super(256, Integer.MAX_VALUE);
      this.handle = handle;
    }
    
    protected void deallocate()
    {
      if (capacity() > ByteBufUtil.THREAD_LOCAL_BUFFER_SIZE)
      {
        super.deallocate();
      }
      else
      {
        clear();
        handle.recycle(this);
      }
    }
  }
  
  public static boolean isText(ByteBuf buf, Charset charset)
  {
    return isText(buf, buf.readerIndex(), buf.readableBytes(), charset);
  }
  
  public static boolean isText(ByteBuf buf, int index, int length, Charset charset)
  {
    ObjectUtil.checkNotNull(buf, "buf");
    ObjectUtil.checkNotNull(charset, "charset");
    int maxIndex = buf.readerIndex() + buf.readableBytes();
    if ((index < 0) || (length < 0) || (index > maxIndex - length)) {
      throw new IndexOutOfBoundsException("index: " + index + " length: " + length);
    }
    if (charset.equals(CharsetUtil.UTF_8)) {
      return isUtf8(buf, index, length);
    }
    if (charset.equals(CharsetUtil.US_ASCII)) {
      return isAscii(buf, index, length);
    }
    CharsetDecoder decoder = CharsetUtil.decoder(charset, CodingErrorAction.REPORT, CodingErrorAction.REPORT);
    try
    {
      if (buf.nioBufferCount() == 1)
      {
        decoder.decode(buf.internalNioBuffer(index, length));
      }
      else
      {
        ByteBuf heapBuffer = buf.alloc().heapBuffer(length);
        try
        {
          heapBuffer.writeBytes(buf, index, length);
          decoder.decode(heapBuffer.internalNioBuffer(0, length));
        }
        finally
        {
          heapBuffer.release();
        }
      }
      return true;
    }
    catch (CharacterCodingException ignore) {}
    return false;
  }
  
  private static final ByteProcessor FIND_NON_ASCII = new ByteProcessor()
  {
    public boolean process(byte value)
    {
      return value >= 0;
    }
  };
  
  private static boolean isAscii(ByteBuf buf, int index, int length)
  {
    return buf.forEachByte(index, length, FIND_NON_ASCII) == -1;
  }
  
  private static boolean isUtf8(ByteBuf buf, int index, int length)
  {
    int endIndex = index + length;
    while (index < endIndex)
    {
      byte b1 = buf.getByte(index++);
      if ((b1 & 0x80) != 0) {
        if ((b1 & 0xE0) == 192)
        {
          if (index >= endIndex) {
            return false;
          }
          byte b2 = buf.getByte(index++);
          if ((b2 & 0xC0) != 128) {
            return false;
          }
          if ((b1 & 0xFF) < 194) {
            return false;
          }
        }
        else if ((b1 & 0xF0) == 224)
        {
          if (index > endIndex - 2) {
            return false;
          }
          byte b2 = buf.getByte(index++);
          byte b3 = buf.getByte(index++);
          if (((b2 & 0xC0) != 128) || ((b3 & 0xC0) != 128)) {
            return false;
          }
          if (((b1 & 0xF) == 0) && ((b2 & 0xFF) < 160)) {
            return false;
          }
          if (((b1 & 0xF) == 13) && ((b2 & 0xFF) > 159)) {
            return false;
          }
        }
        else if ((b1 & 0xF8) == 240)
        {
          if (index > endIndex - 3) {
            return false;
          }
          byte b2 = buf.getByte(index++);
          byte b3 = buf.getByte(index++);
          byte b4 = buf.getByte(index++);
          if (((b2 & 0xC0) != 128) || ((b3 & 0xC0) != 128) || ((b4 & 0xC0) != 128)) {
            return false;
          }
          if (((b1 & 0xFF) > 244) || (((b1 & 0xFF) == 240) && ((b2 & 0xFF) < 144)) || (((b1 & 0xFF) == 244) && ((b2 & 0xFF) > 143))) {
            return false;
          }
        }
        else
        {
          return false;
        }
      }
    }
    return true;
  }
}

/* Location:
 * Qualified Name:     io.netty.buffer.ByteBufUtil
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */