package io.netty.buffer;

import io.netty.util.internal.LongCounter;
import io.netty.util.internal.PlatformDependent;
import io.netty.util.internal.StringUtil;
import java.nio.Buffer;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

abstract class PoolArena<T>
  implements PoolArenaMetric
{
  static final boolean HAS_UNSAFE = PlatformDependent.hasUnsafe();
  static final int numTinySubpagePools = 32;
  final PooledByteBufAllocator parent;
  private final int maxOrder;
  final int pageSize;
  final int pageShifts;
  final int chunkSize;
  final int subpageOverflowMask;
  final int numSmallSubpagePools;
  private final PoolSubpage<T>[] tinySubpagePools;
  private final PoolSubpage<T>[] smallSubpagePools;
  private final PoolChunkList<T> q050;
  private final PoolChunkList<T> q025;
  private final PoolChunkList<T> q000;
  private final PoolChunkList<T> qInit;
  private final PoolChunkList<T> q075;
  private final PoolChunkList<T> q100;
  private final List<PoolChunkListMetric> chunkListMetrics;
  private long allocationsNormal;
  
  static enum SizeClass
  {
    Tiny,  Small,  Normal;
    
    private SizeClass() {}
  }
  
  private final LongCounter allocationsTiny = PlatformDependent.newLongCounter();
  private final LongCounter allocationsSmall = PlatformDependent.newLongCounter();
  private final LongCounter allocationsHuge = PlatformDependent.newLongCounter();
  private final LongCounter activeBytesHuge = PlatformDependent.newLongCounter();
  private long deallocationsTiny;
  private long deallocationsSmall;
  private long deallocationsNormal;
  private final LongCounter deallocationsHuge = PlatformDependent.newLongCounter();
  final AtomicInteger numThreadCaches = new AtomicInteger();
  
  protected PoolArena(PooledByteBufAllocator parent, int pageSize, int maxOrder, int pageShifts, int chunkSize)
  {
    this.parent = parent;
    this.pageSize = pageSize;
    this.maxOrder = maxOrder;
    this.pageShifts = pageShifts;
    this.chunkSize = chunkSize;
    subpageOverflowMask = (pageSize - 1 ^ 0xFFFFFFFF);
    tinySubpagePools = newSubpagePoolArray(32);
    for (int i = 0; i < tinySubpagePools.length; i++) {
      tinySubpagePools[i] = newSubpagePoolHead(pageSize);
    }
    numSmallSubpagePools = (pageShifts - 9);
    smallSubpagePools = newSubpagePoolArray(numSmallSubpagePools);
    for (int i = 0; i < smallSubpagePools.length; i++) {
      smallSubpagePools[i] = newSubpagePoolHead(pageSize);
    }
    q100 = new PoolChunkList(null, 100, Integer.MAX_VALUE, chunkSize);
    q075 = new PoolChunkList(q100, 75, 100, chunkSize);
    q050 = new PoolChunkList(q075, 50, 100, chunkSize);
    q025 = new PoolChunkList(q050, 25, 75, chunkSize);
    q000 = new PoolChunkList(q025, 1, 50, chunkSize);
    qInit = new PoolChunkList(q000, Integer.MIN_VALUE, 25, chunkSize);
    
    q100.prevList(q075);
    q075.prevList(q050);
    q050.prevList(q025);
    q025.prevList(q000);
    q000.prevList(null);
    qInit.prevList(qInit);
    
    List<PoolChunkListMetric> metrics = new ArrayList(6);
    metrics.add(qInit);
    metrics.add(q000);
    metrics.add(q025);
    metrics.add(q050);
    metrics.add(q075);
    metrics.add(q100);
    chunkListMetrics = Collections.unmodifiableList(metrics);
  }
  
  private PoolSubpage<T> newSubpagePoolHead(int pageSize)
  {
    PoolSubpage<T> head = new PoolSubpage(pageSize);
    prev = head;
    next = head;
    return head;
  }
  
  private PoolSubpage<T>[] newSubpagePoolArray(int size)
  {
    return new PoolSubpage[size];
  }
  
  abstract boolean isDirect();
  
  PooledByteBuf<T> allocate(PoolThreadCache cache, int reqCapacity, int maxCapacity)
  {
    PooledByteBuf<T> buf = newByteBuf(maxCapacity);
    allocate(cache, buf, reqCapacity);
    return buf;
  }
  
  static int tinyIdx(int normCapacity)
  {
    return normCapacity >>> 4;
  }
  
  static int smallIdx(int normCapacity)
  {
    int tableIdx = 0;
    int i = normCapacity >>> 10;
    while (i != 0)
    {
      i >>>= 1;
      tableIdx++;
    }
    return tableIdx;
  }
  
  boolean isTinyOrSmall(int normCapacity)
  {
    return (normCapacity & subpageOverflowMask) == 0;
  }
  
  static boolean isTiny(int normCapacity)
  {
    return (normCapacity & 0xFE00) == 0;
  }
  
  private void allocate(PoolThreadCache cache, PooledByteBuf<T> buf, int reqCapacity)
  {
    int normCapacity = normalizeCapacity(reqCapacity);
    if (isTinyOrSmall(normCapacity))
    {
      boolean tiny = isTiny(normCapacity);
      PoolSubpage<T>[] table;
      int tableIdx;
      PoolSubpage<T>[] table;
      if (tiny)
      {
        if (cache.allocateTiny(this, buf, reqCapacity, normCapacity)) {
          return;
        }
        int tableIdx = tinyIdx(normCapacity);
        table = tinySubpagePools;
      }
      else
      {
        if (cache.allocateSmall(this, buf, reqCapacity, normCapacity)) {
          return;
        }
        tableIdx = smallIdx(normCapacity);
        table = smallSubpagePools;
      }
      PoolSubpage<T> head = table[tableIdx];
      synchronized (head)
      {
        PoolSubpage<T> s = next;
        if (s != head)
        {
          assert ((doNotDestroy) && (elemSize == normCapacity));
          long handle = s.allocate();
          assert (handle >= 0L);
          chunk.initBufWithSubpage(buf, handle, reqCapacity);
          if (tiny) {
            allocationsTiny.increment();
          } else {
            allocationsSmall.increment();
          }
          return;
        }
      }
      allocateNormal(buf, reqCapacity, normCapacity);
      return;
    }
    if (normCapacity <= chunkSize)
    {
      if (cache.allocateNormal(this, buf, reqCapacity, normCapacity)) {
        return;
      }
      allocateNormal(buf, reqCapacity, normCapacity);
    }
    else
    {
      allocateHuge(buf, reqCapacity);
    }
  }
  
  private synchronized void allocateNormal(PooledByteBuf<T> buf, int reqCapacity, int normCapacity)
  {
    if ((q050.allocate(buf, reqCapacity, normCapacity)) || (q025.allocate(buf, reqCapacity, normCapacity)) || (q000.allocate(buf, reqCapacity, normCapacity)) || (qInit.allocate(buf, reqCapacity, normCapacity)) || (q075.allocate(buf, reqCapacity, normCapacity)))
    {
      allocationsNormal += 1L;
      return;
    }
    PoolChunk<T> c = newChunk(pageSize, maxOrder, pageShifts, chunkSize);
    long handle = c.allocate(normCapacity);
    allocationsNormal += 1L;
    assert (handle > 0L);
    c.initBuf(buf, handle, reqCapacity);
    qInit.add(c);
  }
  
  private void allocateHuge(PooledByteBuf<T> buf, int reqCapacity)
  {
    PoolChunk<T> chunk = newUnpooledChunk(reqCapacity);
    activeBytesHuge.add(chunk.chunkSize());
    buf.initUnpooled(chunk, reqCapacity);
    allocationsHuge.increment();
  }
  
  void free(PoolChunk<T> chunk, long handle, int normCapacity, PoolThreadCache cache)
  {
    if (unpooled)
    {
      int size = chunk.chunkSize();
      destroyChunk(chunk);
      activeBytesHuge.add(-size);
      deallocationsHuge.increment();
    }
    else
    {
      SizeClass sizeClass = sizeClass(normCapacity);
      if ((cache != null) && (cache.add(this, chunk, handle, normCapacity, sizeClass))) {
        return;
      }
      freeChunk(chunk, handle, sizeClass);
    }
  }
  
  private SizeClass sizeClass(int normCapacity)
  {
    if (!isTinyOrSmall(normCapacity)) {
      return SizeClass.Normal;
    }
    return isTiny(normCapacity) ? SizeClass.Tiny : SizeClass.Small;
  }
  
  void freeChunk(PoolChunk<T> chunk, long handle, SizeClass sizeClass)
  {
    boolean destroyChunk;
    synchronized (this)
    {
      switch (sizeClass)
      {
      case Normal: 
        deallocationsNormal += 1L;
        break;
      case Small: 
        deallocationsSmall += 1L;
        break;
      case Tiny: 
        deallocationsTiny += 1L;
        break;
      default: 
        throw new Error();
      }
      destroyChunk = !parent.free(chunk, handle);
    }
    boolean destroyChunk;
    if (destroyChunk) {
      destroyChunk(chunk);
    }
  }
  
  PoolSubpage<T> findSubpagePoolHead(int elemSize)
  {
    PoolSubpage<T>[] table;
    int tableIdx;
    PoolSubpage<T>[] table;
    if (isTiny(elemSize))
    {
      int tableIdx = elemSize >>> 4;
      table = tinySubpagePools;
    }
    else
    {
      tableIdx = 0;
      elemSize >>>= 10;
      while (elemSize != 0)
      {
        elemSize >>>= 1;
        tableIdx++;
      }
      table = smallSubpagePools;
    }
    return table[tableIdx];
  }
  
  int normalizeCapacity(int reqCapacity)
  {
    if (reqCapacity < 0) {
      throw new IllegalArgumentException("capacity: " + reqCapacity + " (expected: 0+)");
    }
    if (reqCapacity >= chunkSize) {
      return reqCapacity;
    }
    if (!isTiny(reqCapacity))
    {
      int normalizedCapacity = reqCapacity;
      normalizedCapacity--;
      normalizedCapacity |= normalizedCapacity >>> 1;
      normalizedCapacity |= normalizedCapacity >>> 2;
      normalizedCapacity |= normalizedCapacity >>> 4;
      normalizedCapacity |= normalizedCapacity >>> 8;
      normalizedCapacity |= normalizedCapacity >>> 16;
      normalizedCapacity++;
      if (normalizedCapacity < 0) {
        normalizedCapacity >>>= 1;
      }
      return normalizedCapacity;
    }
    if ((reqCapacity & 0xF) == 0) {
      return reqCapacity;
    }
    return (reqCapacity & 0xFFFFFFF0) + 16;
  }
  
  void reallocate(PooledByteBuf<T> buf, int newCapacity, boolean freeOldMemory)
  {
    if ((newCapacity < 0) || (newCapacity > buf.maxCapacity())) {
      throw new IllegalArgumentException("newCapacity: " + newCapacity);
    }
    int oldCapacity = length;
    if (oldCapacity == newCapacity) {
      return;
    }
    PoolChunk<T> oldChunk = chunk;
    long oldHandle = handle;
    T oldMemory = memory;
    int oldOffset = offset;
    int oldMaxLength = maxLength;
    int readerIndex = buf.readerIndex();
    int writerIndex = buf.writerIndex();
    
    allocate(parent.threadCache(), buf, newCapacity);
    if (newCapacity > oldCapacity) {
      memoryCopy(oldMemory, oldOffset, memory, offset, oldCapacity);
    } else if (newCapacity < oldCapacity) {
      if (readerIndex < newCapacity)
      {
        if (writerIndex > newCapacity) {
          writerIndex = newCapacity;
        }
        memoryCopy(oldMemory, oldOffset + readerIndex, memory, offset + readerIndex, writerIndex - readerIndex);
      }
      else
      {
        readerIndex = writerIndex = newCapacity;
      }
    }
    buf.setIndex(readerIndex, writerIndex);
    if (freeOldMemory) {
      free(oldChunk, oldHandle, oldMaxLength, cache);
    }
  }
  
  public int numThreadCaches()
  {
    return numThreadCaches.get();
  }
  
  public int numTinySubpages()
  {
    return tinySubpagePools.length;
  }
  
  public int numSmallSubpages()
  {
    return smallSubpagePools.length;
  }
  
  public int numChunkLists()
  {
    return chunkListMetrics.size();
  }
  
  public List<PoolSubpageMetric> tinySubpages()
  {
    return subPageMetricList(tinySubpagePools);
  }
  
  public List<PoolSubpageMetric> smallSubpages()
  {
    return subPageMetricList(smallSubpagePools);
  }
  
  public List<PoolChunkListMetric> chunkLists()
  {
    return chunkListMetrics;
  }
  
  private static List<PoolSubpageMetric> subPageMetricList(PoolSubpage<?>[] pages)
  {
    List<PoolSubpageMetric> metrics = new ArrayList();
    for (int i = 0; i < pages.length; i++)
    {
      PoolSubpage<?> head = pages[i];
      if (next != head)
      {
        PoolSubpage<?> s = next;
        for (;;)
        {
          metrics.add(s);
          s = next;
          if (s == head) {
            break;
          }
        }
      }
    }
    return metrics;
  }
  
  public long numAllocations()
  {
    long allocsNormal;
    synchronized (this)
    {
      allocsNormal = allocationsNormal;
    }
    long allocsNormal;
    return allocationsTiny.value() + allocationsSmall.value() + allocsNormal + allocationsHuge.value();
  }
  
  public long numTinyAllocations()
  {
    return allocationsTiny.value();
  }
  
  public long numSmallAllocations()
  {
    return allocationsSmall.value();
  }
  
  public synchronized long numNormalAllocations()
  {
    return allocationsNormal;
  }
  
  public long numDeallocations()
  {
    long deallocs;
    synchronized (this)
    {
      deallocs = deallocationsTiny + deallocationsSmall + deallocationsNormal;
    }
    long deallocs;
    return deallocs + deallocationsHuge.value();
  }
  
  public synchronized long numTinyDeallocations()
  {
    return deallocationsTiny;
  }
  
  public synchronized long numSmallDeallocations()
  {
    return deallocationsSmall;
  }
  
  public synchronized long numNormalDeallocations()
  {
    return deallocationsNormal;
  }
  
  public long numHugeAllocations()
  {
    return allocationsHuge.value();
  }
  
  public long numHugeDeallocations()
  {
    return deallocationsHuge.value();
  }
  
  public long numActiveAllocations()
  {
    long val = allocationsTiny.value() + allocationsSmall.value() + allocationsHuge.value() - deallocationsHuge.value();
    synchronized (this)
    {
      val += allocationsNormal - (deallocationsTiny + deallocationsSmall + deallocationsNormal);
    }
    return Math.max(val, 0L);
  }
  
  public long numActiveTinyAllocations()
  {
    return Math.max(numTinyAllocations() - numTinyDeallocations(), 0L);
  }
  
  public long numActiveSmallAllocations()
  {
    return Math.max(numSmallAllocations() - numSmallDeallocations(), 0L);
  }
  
  public long numActiveNormalAllocations()
  {
    long val;
    synchronized (this)
    {
      val = allocationsNormal - deallocationsNormal;
    }
    long val;
    return Math.max(val, 0L);
  }
  
  public long numActiveHugeAllocations()
  {
    return Math.max(numHugeAllocations() - numHugeDeallocations(), 0L);
  }
  
  public long numActiveBytes()
  {
    long val = activeBytesHuge.value();
    synchronized (this)
    {
      for (int i = 0; i < chunkListMetrics.size(); i++) {
        for (PoolChunkMetric m : (PoolChunkListMetric)chunkListMetrics.get(i)) {
          val += m.chunkSize();
        }
      }
    }
    return Math.max(0L, val);
  }
  
  protected abstract PoolChunk<T> newChunk(int paramInt1, int paramInt2, int paramInt3, int paramInt4);
  
  protected abstract PoolChunk<T> newUnpooledChunk(int paramInt);
  
  protected abstract PooledByteBuf<T> newByteBuf(int paramInt);
  
  protected abstract void memoryCopy(T paramT1, int paramInt1, T paramT2, int paramInt2, int paramInt3);
  
  protected abstract void destroyChunk(PoolChunk<T> paramPoolChunk);
  
  public synchronized String toString()
  {
    StringBuilder buf = new StringBuilder().append("Chunk(s) at 0~25%:").append(StringUtil.NEWLINE).append(qInit).append(StringUtil.NEWLINE).append("Chunk(s) at 0~50%:").append(StringUtil.NEWLINE).append(q000).append(StringUtil.NEWLINE).append("Chunk(s) at 25~75%:").append(StringUtil.NEWLINE).append(q025).append(StringUtil.NEWLINE).append("Chunk(s) at 50~100%:").append(StringUtil.NEWLINE).append(q050).append(StringUtil.NEWLINE).append("Chunk(s) at 75~100%:").append(StringUtil.NEWLINE).append(q075).append(StringUtil.NEWLINE).append("Chunk(s) at 100%:").append(StringUtil.NEWLINE).append(q100).append(StringUtil.NEWLINE).append("tiny subpages:");
    
    appendPoolSubPages(buf, tinySubpagePools);
    buf.append(StringUtil.NEWLINE).append("small subpages:");
    
    appendPoolSubPages(buf, smallSubpagePools);
    buf.append(StringUtil.NEWLINE);
    
    return buf.toString();
  }
  
  private static void appendPoolSubPages(StringBuilder buf, PoolSubpage<?>[] subpages)
  {
    for (int i = 0; i < subpages.length; i++)
    {
      PoolSubpage<?> head = subpages[i];
      if (next != head)
      {
        buf.append(StringUtil.NEWLINE).append(i).append(": ");
        
        PoolSubpage<?> s = next;
        for (;;)
        {
          buf.append(s);
          s = next;
          if (s == head) {
            break;
          }
        }
      }
    }
  }
  
  static final class HeapArena
    extends PoolArena<byte[]>
  {
    HeapArena(PooledByteBufAllocator parent, int pageSize, int maxOrder, int pageShifts, int chunkSize)
    {
      super(pageSize, maxOrder, pageShifts, chunkSize);
    }
    
    boolean isDirect()
    {
      return false;
    }
    
    protected PoolChunk<byte[]> newChunk(int pageSize, int maxOrder, int pageShifts, int chunkSize)
    {
      return new PoolChunk(this, new byte[chunkSize], pageSize, maxOrder, pageShifts, chunkSize);
    }
    
    protected PoolChunk<byte[]> newUnpooledChunk(int capacity)
    {
      return new PoolChunk(this, new byte[capacity], capacity);
    }
    
    protected void destroyChunk(PoolChunk<byte[]> chunk) {}
    
    protected PooledByteBuf<byte[]> newByteBuf(int maxCapacity)
    {
      return HAS_UNSAFE ? PooledUnsafeHeapByteBuf.newUnsafeInstance(maxCapacity) : PooledHeapByteBuf.newInstance(maxCapacity);
    }
    
    protected void memoryCopy(byte[] src, int srcOffset, byte[] dst, int dstOffset, int length)
    {
      if (length == 0) {
        return;
      }
      System.arraycopy(src, srcOffset, dst, dstOffset, length);
    }
  }
  
  static final class DirectArena
    extends PoolArena<ByteBuffer>
  {
    DirectArena(PooledByteBufAllocator parent, int pageSize, int maxOrder, int pageShifts, int chunkSize)
    {
      super(pageSize, maxOrder, pageShifts, chunkSize);
    }
    
    boolean isDirect()
    {
      return true;
    }
    
    protected PoolChunk<ByteBuffer> newChunk(int pageSize, int maxOrder, int pageShifts, int chunkSize)
    {
      return new PoolChunk(this, allocateDirect(chunkSize), pageSize, maxOrder, pageShifts, chunkSize);
    }
    
    protected PoolChunk<ByteBuffer> newUnpooledChunk(int capacity)
    {
      return new PoolChunk(this, allocateDirect(capacity), capacity);
    }
    
    private static ByteBuffer allocateDirect(int capacity)
    {
      return PlatformDependent.useDirectBufferNoCleaner() ? PlatformDependent.allocateDirectNoCleaner(capacity) : ByteBuffer.allocateDirect(capacity);
    }
    
    protected void destroyChunk(PoolChunk<ByteBuffer> chunk)
    {
      if (PlatformDependent.useDirectBufferNoCleaner()) {
        PlatformDependent.freeDirectNoCleaner((ByteBuffer)memory);
      } else {
        PlatformDependent.freeDirectBuffer((ByteBuffer)memory);
      }
    }
    
    protected PooledByteBuf<ByteBuffer> newByteBuf(int maxCapacity)
    {
      if (HAS_UNSAFE) {
        return PooledUnsafeDirectByteBuf.newInstance(maxCapacity);
      }
      return PooledDirectByteBuf.newInstance(maxCapacity);
    }
    
    protected void memoryCopy(ByteBuffer src, int srcOffset, ByteBuffer dst, int dstOffset, int length)
    {
      if (length == 0) {
        return;
      }
      if (HAS_UNSAFE)
      {
        PlatformDependent.copyMemory(PlatformDependent.directBufferAddress(src) + srcOffset, PlatformDependent.directBufferAddress(dst) + dstOffset, length);
      }
      else
      {
        src = src.duplicate();
        dst = dst.duplicate();
        src.position(srcOffset).limit(srcOffset + length);
        dst.position(dstOffset);
        dst.put(src);
      }
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.buffer.PoolArena
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */