package io.netty.buffer;

import io.netty.util.internal.StringUtil;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

final class PoolChunkList<T>
  implements PoolChunkListMetric
{
  private static final Iterator<PoolChunkMetric> EMPTY_METRICS = Collections.emptyList().iterator();
  private final PoolChunkList<T> nextList;
  private final int minUsage;
  private final int maxUsage;
  private final int maxCapacity;
  private PoolChunk<T> head;
  private PoolChunkList<T> prevList;
  
  PoolChunkList(PoolChunkList<T> nextList, int minUsage, int maxUsage, int chunkSize)
  {
    assert (minUsage <= maxUsage);
    this.nextList = nextList;
    this.minUsage = minUsage;
    this.maxUsage = maxUsage;
    maxCapacity = calculateMaxCapacity(minUsage, chunkSize);
  }
  
  private static int calculateMaxCapacity(int minUsage, int chunkSize)
  {
    minUsage = minUsage0(minUsage);
    if (minUsage == 100) {
      return 0;
    }
    return (int)(chunkSize * (100L - minUsage) / 100L);
  }
  
  void prevList(PoolChunkList<T> prevList)
  {
    assert (this.prevList == null);
    this.prevList = prevList;
  }
  
  boolean allocate(PooledByteBuf<T> buf, int reqCapacity, int normCapacity)
  {
    if ((head == null) || (normCapacity > maxCapacity)) {
      return false;
    }
    PoolChunk<T> cur = head;
    for (;;)
    {
      long handle = cur.allocate(normCapacity);
      if (handle < 0L)
      {
        cur = next;
        if (cur == null) {
          return false;
        }
      }
      else
      {
        cur.initBuf(buf, handle, reqCapacity);
        if (cur.usage() >= maxUsage)
        {
          remove(cur);
          nextList.add(cur);
        }
        return true;
      }
    }
  }
  
  boolean free(PoolChunk<T> chunk, long handle)
  {
    chunk.free(handle);
    if (chunk.usage() < minUsage)
    {
      remove(chunk);
      
      return move0(chunk);
    }
    return true;
  }
  
  private boolean move(PoolChunk<T> chunk)
  {
    assert (chunk.usage() < maxUsage);
    if (chunk.usage() < minUsage) {
      return move0(chunk);
    }
    add0(chunk);
    return true;
  }
  
  private boolean move0(PoolChunk<T> chunk)
  {
    if (prevList == null)
    {
      assert (chunk.usage() == 0);
      return false;
    }
    return prevList.move(chunk);
  }
  
  void add(PoolChunk<T> chunk)
  {
    if (chunk.usage() >= maxUsage)
    {
      nextList.add(chunk);
      return;
    }
    add0(chunk);
  }
  
  void add0(PoolChunk<T> chunk)
  {
    parent = this;
    if (head == null)
    {
      head = chunk;
      prev = null;
      next = null;
    }
    else
    {
      prev = null;
      next = head;
      head.prev = chunk;
      head = chunk;
    }
  }
  
  private void remove(PoolChunk<T> cur)
  {
    if (cur == head)
    {
      head = next;
      if (head != null) {
        head.prev = null;
      }
    }
    else
    {
      PoolChunk<T> next = next;
      prev.next = next;
      if (next != null) {
        prev = prev;
      }
    }
  }
  
  public int minUsage()
  {
    return minUsage0(minUsage);
  }
  
  public int maxUsage()
  {
    return Math.min(maxUsage, 100);
  }
  
  private static int minUsage0(int value)
  {
    return Math.max(1, value);
  }
  
  public Iterator<PoolChunkMetric> iterator()
  {
    if (head == null) {
      return EMPTY_METRICS;
    }
    List<PoolChunkMetric> metrics = new ArrayList();
    PoolChunk<T> cur = head;
    for (;;)
    {
      metrics.add(cur);
      cur = next;
      if (cur == null) {
        break;
      }
    }
    return metrics.iterator();
  }
  
  public String toString()
  {
    if (head == null) {
      return "none";
    }
    StringBuilder buf = new StringBuilder();
    PoolChunk<T> cur = head;
    for (;;)
    {
      buf.append(cur);
      cur = next;
      if (cur == null) {
        break;
      }
      buf.append(StringUtil.NEWLINE);
    }
    return buf.toString();
  }
}

/* Location:
 * Qualified Name:     io.netty.buffer.PoolChunkList
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */