package io.netty.buffer;

import io.netty.util.Recycler;
import io.netty.util.Recycler.Handle;
import io.netty.util.ThreadDeathWatcher;
import io.netty.util.internal.MathUtil;
import io.netty.util.internal.PlatformDependent;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;
import java.nio.ByteBuffer;
import java.util.Queue;
import java.util.concurrent.atomic.AtomicInteger;

final class PoolThreadCache
{
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(PoolThreadCache.class);
  final PoolArena<byte[]> heapArena;
  final PoolArena<ByteBuffer> directArena;
  private final MemoryRegionCache<byte[]>[] tinySubPageHeapCaches;
  private final MemoryRegionCache<byte[]>[] smallSubPageHeapCaches;
  private final MemoryRegionCache<ByteBuffer>[] tinySubPageDirectCaches;
  private final MemoryRegionCache<ByteBuffer>[] smallSubPageDirectCaches;
  private final MemoryRegionCache<byte[]>[] normalHeapCaches;
  private final MemoryRegionCache<ByteBuffer>[] normalDirectCaches;
  private final int numShiftsNormalDirect;
  private final int numShiftsNormalHeap;
  private final int freeSweepAllocationThreshold;
  private int allocations;
  private final Thread thread = Thread.currentThread();
  private final Runnable freeTask = new Runnable()
  {
    public void run()
    {
      PoolThreadCache.this.free0();
    }
  };
  
  PoolThreadCache(PoolArena<byte[]> heapArena, PoolArena<ByteBuffer> directArena, int tinyCacheSize, int smallCacheSize, int normalCacheSize, int maxCachedBufferCapacity, int freeSweepAllocationThreshold)
  {
    if (maxCachedBufferCapacity < 0) {
      throw new IllegalArgumentException("maxCachedBufferCapacity: " + maxCachedBufferCapacity + " (expected: >= 0)");
    }
    if (freeSweepAllocationThreshold < 1) {
      throw new IllegalArgumentException("freeSweepAllocationThreshold: " + freeSweepAllocationThreshold + " (expected: > 0)");
    }
    this.freeSweepAllocationThreshold = freeSweepAllocationThreshold;
    this.heapArena = heapArena;
    this.directArena = directArena;
    if (directArena != null)
    {
      tinySubPageDirectCaches = createSubPageCaches(tinyCacheSize, 32, PoolArena.SizeClass.Tiny);
      
      smallSubPageDirectCaches = createSubPageCaches(smallCacheSize, numSmallSubpagePools, PoolArena.SizeClass.Small);
      
      numShiftsNormalDirect = log2(pageSize);
      normalDirectCaches = createNormalCaches(normalCacheSize, maxCachedBufferCapacity, directArena);
      
      numThreadCaches.getAndIncrement();
    }
    else
    {
      tinySubPageDirectCaches = null;
      smallSubPageDirectCaches = null;
      normalDirectCaches = null;
      numShiftsNormalDirect = -1;
    }
    if (heapArena != null)
    {
      tinySubPageHeapCaches = createSubPageCaches(tinyCacheSize, 32, PoolArena.SizeClass.Tiny);
      
      smallSubPageHeapCaches = createSubPageCaches(smallCacheSize, numSmallSubpagePools, PoolArena.SizeClass.Small);
      
      numShiftsNormalHeap = log2(pageSize);
      normalHeapCaches = createNormalCaches(normalCacheSize, maxCachedBufferCapacity, heapArena);
      
      numThreadCaches.getAndIncrement();
    }
    else
    {
      tinySubPageHeapCaches = null;
      smallSubPageHeapCaches = null;
      normalHeapCaches = null;
      numShiftsNormalHeap = -1;
    }
    ThreadDeathWatcher.watch(thread, freeTask);
  }
  
  private static <T> MemoryRegionCache<T>[] createSubPageCaches(int cacheSize, int numCaches, PoolArena.SizeClass sizeClass)
  {
    if (cacheSize > 0)
    {
      MemoryRegionCache<T>[] cache = new MemoryRegionCache[numCaches];
      for (int i = 0; i < cache.length; i++) {
        cache[i] = new SubPageMemoryRegionCache(cacheSize, sizeClass);
      }
      return cache;
    }
    return null;
  }
  
  private static <T> MemoryRegionCache<T>[] createNormalCaches(int cacheSize, int maxCachedBufferCapacity, PoolArena<T> area)
  {
    if (cacheSize > 0)
    {
      int max = Math.min(chunkSize, maxCachedBufferCapacity);
      int arraySize = Math.max(1, log2(max / pageSize) + 1);
      
      MemoryRegionCache<T>[] cache = new MemoryRegionCache[arraySize];
      for (int i = 0; i < cache.length; i++) {
        cache[i] = new NormalMemoryRegionCache(cacheSize);
      }
      return cache;
    }
    return null;
  }
  
  private static int log2(int val)
  {
    int res = 0;
    while (val > 1)
    {
      val >>= 1;
      res++;
    }
    return res;
  }
  
  boolean allocateTiny(PoolArena<?> area, PooledByteBuf<?> buf, int reqCapacity, int normCapacity)
  {
    return allocate(cacheForTiny(area, normCapacity), buf, reqCapacity);
  }
  
  boolean allocateSmall(PoolArena<?> area, PooledByteBuf<?> buf, int reqCapacity, int normCapacity)
  {
    return allocate(cacheForSmall(area, normCapacity), buf, reqCapacity);
  }
  
  boolean allocateNormal(PoolArena<?> area, PooledByteBuf<?> buf, int reqCapacity, int normCapacity)
  {
    return allocate(cacheForNormal(area, normCapacity), buf, reqCapacity);
  }
  
  private boolean allocate(MemoryRegionCache<?> cache, PooledByteBuf buf, int reqCapacity)
  {
    if (cache == null) {
      return false;
    }
    boolean allocated = cache.allocate(buf, reqCapacity);
    if (++allocations >= freeSweepAllocationThreshold)
    {
      allocations = 0;
      trim();
    }
    return allocated;
  }
  
  boolean add(PoolArena<?> area, PoolChunk chunk, long handle, int normCapacity, PoolArena.SizeClass sizeClass)
  {
    MemoryRegionCache<?> cache = cache(area, normCapacity, sizeClass);
    if (cache == null) {
      return false;
    }
    return cache.add(chunk, handle);
  }
  
  private MemoryRegionCache<?> cache(PoolArena<?> area, int normCapacity, PoolArena.SizeClass sizeClass)
  {
    switch (sizeClass)
    {
    case Normal: 
      return cacheForNormal(area, normCapacity);
    case Small: 
      return cacheForSmall(area, normCapacity);
    case Tiny: 
      return cacheForTiny(area, normCapacity);
    }
    throw new Error();
  }
  
  void free()
  {
    ThreadDeathWatcher.unwatch(thread, freeTask);
    free0();
  }
  
  private void free0()
  {
    int numFreed = free(tinySubPageDirectCaches) + free(smallSubPageDirectCaches) + free(normalDirectCaches) + free(tinySubPageHeapCaches) + free(smallSubPageHeapCaches) + free(normalHeapCaches);
    if ((numFreed > 0) && (logger.isDebugEnabled())) {
      logger.debug("Freed {} thread-local buffer(s) from thread: {}", Integer.valueOf(numFreed), thread.getName());
    }
    if (directArena != null) {
      directArena.numThreadCaches.getAndDecrement();
    }
    if (heapArena != null) {
      heapArena.numThreadCaches.getAndDecrement();
    }
  }
  
  private static int free(MemoryRegionCache<?>[] caches)
  {
    if (caches == null) {
      return 0;
    }
    int numFreed = 0;
    for (MemoryRegionCache<?> c : caches) {
      numFreed += free(c);
    }
    return numFreed;
  }
  
  private static int free(MemoryRegionCache<?> cache)
  {
    if (cache == null) {
      return 0;
    }
    return cache.free();
  }
  
  void trim()
  {
    trim(tinySubPageDirectCaches);
    trim(smallSubPageDirectCaches);
    trim(normalDirectCaches);
    trim(tinySubPageHeapCaches);
    trim(smallSubPageHeapCaches);
    trim(normalHeapCaches);
  }
  
  private static void trim(MemoryRegionCache<?>[] caches)
  {
    if (caches == null) {
      return;
    }
    for (MemoryRegionCache<?> c : caches) {
      trim(c);
    }
  }
  
  private static void trim(MemoryRegionCache<?> cache)
  {
    if (cache == null) {
      return;
    }
    cache.trim();
  }
  
  private MemoryRegionCache<?> cacheForTiny(PoolArena<?> area, int normCapacity)
  {
    int idx = PoolArena.tinyIdx(normCapacity);
    if (area.isDirect()) {
      return cache(tinySubPageDirectCaches, idx);
    }
    return cache(tinySubPageHeapCaches, idx);
  }
  
  private MemoryRegionCache<?> cacheForSmall(PoolArena<?> area, int normCapacity)
  {
    int idx = PoolArena.smallIdx(normCapacity);
    if (area.isDirect()) {
      return cache(smallSubPageDirectCaches, idx);
    }
    return cache(smallSubPageHeapCaches, idx);
  }
  
  private MemoryRegionCache<?> cacheForNormal(PoolArena<?> area, int normCapacity)
  {
    if (area.isDirect())
    {
      int idx = log2(normCapacity >> numShiftsNormalDirect);
      return cache(normalDirectCaches, idx);
    }
    int idx = log2(normCapacity >> numShiftsNormalHeap);
    return cache(normalHeapCaches, idx);
  }
  
  private static <T> MemoryRegionCache<T> cache(MemoryRegionCache<T>[] cache, int idx)
  {
    if ((cache == null) || (idx > cache.length - 1)) {
      return null;
    }
    return cache[idx];
  }
  
  private static final class SubPageMemoryRegionCache<T>
    extends PoolThreadCache.MemoryRegionCache<T>
  {
    SubPageMemoryRegionCache(int size, PoolArena.SizeClass sizeClass)
    {
      super(sizeClass);
    }
    
    protected void initBuf(PoolChunk<T> chunk, long handle, PooledByteBuf<T> buf, int reqCapacity)
    {
      chunk.initBufWithSubpage(buf, handle, reqCapacity);
    }
  }
  
  private static final class NormalMemoryRegionCache<T>
    extends PoolThreadCache.MemoryRegionCache<T>
  {
    NormalMemoryRegionCache(int size)
    {
      super(PoolArena.SizeClass.Normal);
    }
    
    protected void initBuf(PoolChunk<T> chunk, long handle, PooledByteBuf<T> buf, int reqCapacity)
    {
      chunk.initBuf(buf, handle, reqCapacity);
    }
  }
  
  private static abstract class MemoryRegionCache<T>
  {
    private final int size;
    private final Queue<Entry<T>> queue;
    private final PoolArena.SizeClass sizeClass;
    private int allocations;
    
    MemoryRegionCache(int size, PoolArena.SizeClass sizeClass)
    {
      this.size = MathUtil.safeFindNextPositivePowerOfTwo(size);
      queue = PlatformDependent.newFixedMpscQueue(this.size);
      this.sizeClass = sizeClass;
    }
    
    protected abstract void initBuf(PoolChunk<T> paramPoolChunk, long paramLong, PooledByteBuf<T> paramPooledByteBuf, int paramInt);
    
    public final boolean add(PoolChunk<T> chunk, long handle)
    {
      Entry<T> entry = newEntry(chunk, handle);
      boolean queued = queue.offer(entry);
      if (!queued) {
        entry.recycle();
      }
      return queued;
    }
    
    public final boolean allocate(PooledByteBuf<T> buf, int reqCapacity)
    {
      Entry<T> entry = (Entry)queue.poll();
      if (entry == null) {
        return false;
      }
      initBuf(chunk, handle, buf, reqCapacity);
      entry.recycle();
      
      allocations += 1;
      return true;
    }
    
    public final int free()
    {
      return free(Integer.MAX_VALUE);
    }
    
    private int free(int max)
    {
      for (int numFreed = 0; numFreed < max; numFreed++)
      {
        Entry<T> entry = (Entry)queue.poll();
        if (entry != null) {
          freeEntry(entry);
        } else {
          return numFreed;
        }
      }
      return numFreed;
    }
    
    public final void trim()
    {
      int free = size - allocations;
      allocations = 0;
      if (free > 0) {
        free(free);
      }
    }
    
    private void freeEntry(Entry entry)
    {
      PoolChunk chunk = chunk;
      long handle = handle;
      
      entry.recycle();
      
      arena.freeChunk(chunk, handle, sizeClass);
    }
    
    static final class Entry<T>
    {
      final Recycler.Handle<Entry<?>> recyclerHandle;
      PoolChunk<T> chunk;
      long handle = -1L;
      
      Entry(Recycler.Handle<Entry<?>> recyclerHandle)
      {
        this.recyclerHandle = recyclerHandle;
      }
      
      void recycle()
      {
        chunk = null;
        handle = -1L;
        recyclerHandle.recycle(this);
      }
    }
    
    private static Entry newEntry(PoolChunk<?> chunk, long handle)
    {
      Entry entry = (Entry)RECYCLER.get();
      chunk = chunk;
      handle = handle;
      return entry;
    }
    
    private static final Recycler<Entry> RECYCLER = new Recycler()
    {
      protected PoolThreadCache.MemoryRegionCache.Entry newObject(Recycler.Handle<PoolThreadCache.MemoryRegionCache.Entry> handle)
      {
        return new PoolThreadCache.MemoryRegionCache.Entry(handle);
      }
    };
  }
}

/* Location:
 * Qualified Name:     io.netty.buffer.PoolThreadCache
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */