package io.netty.buffer;

import io.netty.util.concurrent.FastThreadLocal;
import io.netty.util.internal.PlatformDependent;
import io.netty.util.internal.StringUtil;
import io.netty.util.internal.SystemPropertyUtil;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

public class PooledByteBufAllocator
  extends AbstractByteBufAllocator
{
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(PooledByteBufAllocator.class);
  private static final int DEFAULT_NUM_HEAP_ARENA;
  private static final int DEFAULT_NUM_DIRECT_ARENA;
  private static final int DEFAULT_PAGE_SIZE;
  private static final int DEFAULT_MAX_ORDER;
  private static final int DEFAULT_TINY_CACHE_SIZE;
  private static final int DEFAULT_SMALL_CACHE_SIZE;
  private static final int DEFAULT_NORMAL_CACHE_SIZE;
  private static final int DEFAULT_MAX_CACHED_BUFFER_CAPACITY;
  private static final int DEFAULT_CACHE_TRIM_INTERVAL;
  private static final int MIN_PAGE_SIZE = 4096;
  private static final int MAX_CHUNK_SIZE = 1073741824;
  
  static
  {
    int defaultPageSize = SystemPropertyUtil.getInt("io.netty.allocator.pageSize", 8192);
    Throwable pageSizeFallbackCause = null;
    try
    {
      validateAndCalculatePageShifts(defaultPageSize);
    }
    catch (Throwable t)
    {
      pageSizeFallbackCause = t;
      defaultPageSize = 8192;
    }
    DEFAULT_PAGE_SIZE = defaultPageSize;
    
    int defaultMaxOrder = SystemPropertyUtil.getInt("io.netty.allocator.maxOrder", 11);
    Throwable maxOrderFallbackCause = null;
    try
    {
      validateAndCalculateChunkSize(DEFAULT_PAGE_SIZE, defaultMaxOrder);
    }
    catch (Throwable t)
    {
      maxOrderFallbackCause = t;
      defaultMaxOrder = 11;
    }
    DEFAULT_MAX_ORDER = defaultMaxOrder;
    
    Runtime runtime = Runtime.getRuntime();
    
    int defaultMinNumArena = runtime.availableProcessors() * 2;
    int defaultChunkSize = DEFAULT_PAGE_SIZE << DEFAULT_MAX_ORDER;
    DEFAULT_NUM_HEAP_ARENA = Math.max(0, SystemPropertyUtil.getInt("io.netty.allocator.numHeapArenas", (int)Math.min(defaultMinNumArena, runtime.maxMemory() / defaultChunkSize / 2L / 3L)));
    
    DEFAULT_NUM_DIRECT_ARENA = Math.max(0, SystemPropertyUtil.getInt("io.netty.allocator.numDirectArenas", (int)Math.min(defaultMinNumArena, PlatformDependent.maxDirectMemory() / defaultChunkSize / 2L / 3L)));
    
    DEFAULT_TINY_CACHE_SIZE = SystemPropertyUtil.getInt("io.netty.allocator.tinyCacheSize", 512);
    DEFAULT_SMALL_CACHE_SIZE = SystemPropertyUtil.getInt("io.netty.allocator.smallCacheSize", 256);
    DEFAULT_NORMAL_CACHE_SIZE = SystemPropertyUtil.getInt("io.netty.allocator.normalCacheSize", 64);
    
    DEFAULT_MAX_CACHED_BUFFER_CAPACITY = SystemPropertyUtil.getInt("io.netty.allocator.maxCachedBufferCapacity", 32768);
    
    DEFAULT_CACHE_TRIM_INTERVAL = SystemPropertyUtil.getInt("io.netty.allocator.cacheTrimInterval", 8192);
    if (logger.isDebugEnabled())
    {
      logger.debug("-Dio.netty.allocator.numHeapArenas: {}", Integer.valueOf(DEFAULT_NUM_HEAP_ARENA));
      logger.debug("-Dio.netty.allocator.numDirectArenas: {}", Integer.valueOf(DEFAULT_NUM_DIRECT_ARENA));
      if (pageSizeFallbackCause == null) {
        logger.debug("-Dio.netty.allocator.pageSize: {}", Integer.valueOf(DEFAULT_PAGE_SIZE));
      } else {
        logger.debug("-Dio.netty.allocator.pageSize: {}", Integer.valueOf(DEFAULT_PAGE_SIZE), pageSizeFallbackCause);
      }
      if (maxOrderFallbackCause == null) {
        logger.debug("-Dio.netty.allocator.maxOrder: {}", Integer.valueOf(DEFAULT_MAX_ORDER));
      } else {
        logger.debug("-Dio.netty.allocator.maxOrder: {}", Integer.valueOf(DEFAULT_MAX_ORDER), maxOrderFallbackCause);
      }
      logger.debug("-Dio.netty.allocator.chunkSize: {}", Integer.valueOf(DEFAULT_PAGE_SIZE << DEFAULT_MAX_ORDER));
      logger.debug("-Dio.netty.allocator.tinyCacheSize: {}", Integer.valueOf(DEFAULT_TINY_CACHE_SIZE));
      logger.debug("-Dio.netty.allocator.smallCacheSize: {}", Integer.valueOf(DEFAULT_SMALL_CACHE_SIZE));
      logger.debug("-Dio.netty.allocator.normalCacheSize: {}", Integer.valueOf(DEFAULT_NORMAL_CACHE_SIZE));
      logger.debug("-Dio.netty.allocator.maxCachedBufferCapacity: {}", Integer.valueOf(DEFAULT_MAX_CACHED_BUFFER_CAPACITY));
      logger.debug("-Dio.netty.allocator.cacheTrimInterval: {}", Integer.valueOf(DEFAULT_CACHE_TRIM_INTERVAL));
    }
  }
  
  public static final PooledByteBufAllocator DEFAULT = new PooledByteBufAllocator(PlatformDependent.directBufferPreferred());
  private final PoolArena<byte[]>[] heapArenas;
  private final PoolArena<ByteBuffer>[] directArenas;
  private final int tinyCacheSize;
  private final int smallCacheSize;
  private final int normalCacheSize;
  private final List<PoolArenaMetric> heapArenaMetrics;
  private final List<PoolArenaMetric> directArenaMetrics;
  private final PoolThreadLocalCache threadCache;
  
  public PooledByteBufAllocator()
  {
    this(false);
  }
  
  public PooledByteBufAllocator(boolean preferDirect)
  {
    this(preferDirect, DEFAULT_NUM_HEAP_ARENA, DEFAULT_NUM_DIRECT_ARENA, DEFAULT_PAGE_SIZE, DEFAULT_MAX_ORDER);
  }
  
  public PooledByteBufAllocator(int nHeapArena, int nDirectArena, int pageSize, int maxOrder)
  {
    this(false, nHeapArena, nDirectArena, pageSize, maxOrder);
  }
  
  public PooledByteBufAllocator(boolean preferDirect, int nHeapArena, int nDirectArena, int pageSize, int maxOrder)
  {
    this(preferDirect, nHeapArena, nDirectArena, pageSize, maxOrder, DEFAULT_TINY_CACHE_SIZE, DEFAULT_SMALL_CACHE_SIZE, DEFAULT_NORMAL_CACHE_SIZE);
  }
  
  public PooledByteBufAllocator(boolean preferDirect, int nHeapArena, int nDirectArena, int pageSize, int maxOrder, int tinyCacheSize, int smallCacheSize, int normalCacheSize)
  {
    super(preferDirect);
    threadCache = new PoolThreadLocalCache();
    this.tinyCacheSize = tinyCacheSize;
    this.smallCacheSize = smallCacheSize;
    this.normalCacheSize = normalCacheSize;
    int chunkSize = validateAndCalculateChunkSize(pageSize, maxOrder);
    if (nHeapArena < 0) {
      throw new IllegalArgumentException("nHeapArena: " + nHeapArena + " (expected: >= 0)");
    }
    if (nDirectArena < 0) {
      throw new IllegalArgumentException("nDirectArea: " + nDirectArena + " (expected: >= 0)");
    }
    int pageShifts = validateAndCalculatePageShifts(pageSize);
    if (nHeapArena > 0)
    {
      heapArenas = newArenaArray(nHeapArena);
      List<PoolArenaMetric> metrics = new ArrayList(heapArenas.length);
      for (int i = 0; i < heapArenas.length; i++)
      {
        PoolArena.HeapArena arena = new PoolArena.HeapArena(this, pageSize, maxOrder, pageShifts, chunkSize);
        heapArenas[i] = arena;
        metrics.add(arena);
      }
      heapArenaMetrics = Collections.unmodifiableList(metrics);
    }
    else
    {
      heapArenas = null;
      heapArenaMetrics = Collections.emptyList();
    }
    if (nDirectArena > 0)
    {
      directArenas = newArenaArray(nDirectArena);
      List<PoolArenaMetric> metrics = new ArrayList(directArenas.length);
      for (int i = 0; i < directArenas.length; i++)
      {
        PoolArena.DirectArena arena = new PoolArena.DirectArena(this, pageSize, maxOrder, pageShifts, chunkSize);
        
        directArenas[i] = arena;
        metrics.add(arena);
      }
      directArenaMetrics = Collections.unmodifiableList(metrics);
    }
    else
    {
      directArenas = null;
      directArenaMetrics = Collections.emptyList();
    }
  }
  
  private static <T> PoolArena<T>[] newArenaArray(int size)
  {
    return new PoolArena[size];
  }
  
  private static int validateAndCalculatePageShifts(int pageSize)
  {
    if (pageSize < 4096) {
      throw new IllegalArgumentException("pageSize: " + pageSize + " (expected: " + 4096 + ")");
    }
    if ((pageSize & pageSize - 1) != 0) {
      throw new IllegalArgumentException("pageSize: " + pageSize + " (expected: power of 2)");
    }
    return 31 - Integer.numberOfLeadingZeros(pageSize);
  }
  
  private static int validateAndCalculateChunkSize(int pageSize, int maxOrder)
  {
    if (maxOrder > 14) {
      throw new IllegalArgumentException("maxOrder: " + maxOrder + " (expected: 0-14)");
    }
    int chunkSize = pageSize;
    for (int i = maxOrder; i > 0; i--)
    {
      if (chunkSize > 536870912) {
        throw new IllegalArgumentException(String.format("pageSize (%d) << maxOrder (%d) must not exceed %d", new Object[] { Integer.valueOf(pageSize), Integer.valueOf(maxOrder), Integer.valueOf(1073741824) }));
      }
      chunkSize <<= 1;
    }
    return chunkSize;
  }
  
  protected ByteBuf newHeapBuffer(int initialCapacity, int maxCapacity)
  {
    PoolThreadCache cache = (PoolThreadCache)threadCache.get();
    PoolArena<byte[]> heapArena = heapArena;
    ByteBuf buf;
    ByteBuf buf;
    if (heapArena != null) {
      buf = heapArena.allocate(cache, initialCapacity, maxCapacity);
    } else {
      buf = new UnpooledHeapByteBuf(this, initialCapacity, maxCapacity);
    }
    return toLeakAwareBuffer(buf);
  }
  
  protected ByteBuf newDirectBuffer(int initialCapacity, int maxCapacity)
  {
    PoolThreadCache cache = (PoolThreadCache)threadCache.get();
    PoolArena<ByteBuffer> directArena = directArena;
    ByteBuf buf;
    ByteBuf buf;
    if (directArena != null)
    {
      buf = directArena.allocate(cache, initialCapacity, maxCapacity);
    }
    else
    {
      ByteBuf buf;
      if (PlatformDependent.hasUnsafe()) {
        buf = UnsafeByteBufUtil.newUnsafeDirectByteBuf(this, initialCapacity, maxCapacity);
      } else {
        buf = new UnpooledDirectByteBuf(this, initialCapacity, maxCapacity);
      }
    }
    return toLeakAwareBuffer(buf);
  }
  
  public static int defaultNumHeapArena()
  {
    return DEFAULT_NUM_HEAP_ARENA;
  }
  
  public static int defaultNumDirectArena()
  {
    return DEFAULT_NUM_DIRECT_ARENA;
  }
  
  public static int defaultPageSize()
  {
    return DEFAULT_PAGE_SIZE;
  }
  
  public static int defaultMaxOrder()
  {
    return DEFAULT_MAX_ORDER;
  }
  
  public static int defaultTinyCacheSize()
  {
    return DEFAULT_TINY_CACHE_SIZE;
  }
  
  public static int defaultSmallCacheSize()
  {
    return DEFAULT_SMALL_CACHE_SIZE;
  }
  
  public static int defaultNormalCacheSize()
  {
    return DEFAULT_NORMAL_CACHE_SIZE;
  }
  
  public boolean isDirectBufferPooled()
  {
    return directArenas != null;
  }
  
  @Deprecated
  public boolean hasThreadLocalCache()
  {
    return threadCache.isSet();
  }
  
  @Deprecated
  public void freeThreadLocalCache()
  {
    threadCache.remove();
  }
  
  final class PoolThreadLocalCache
    extends FastThreadLocal<PoolThreadCache>
  {
    PoolThreadLocalCache() {}
    
    protected synchronized PoolThreadCache initialValue()
    {
      PoolArena<byte[]> heapArena = leastUsedArena(heapArenas);
      PoolArena<ByteBuffer> directArena = leastUsedArena(directArenas);
      
      return new PoolThreadCache(heapArena, directArena, tinyCacheSize, smallCacheSize, normalCacheSize, PooledByteBufAllocator.DEFAULT_MAX_CACHED_BUFFER_CAPACITY, PooledByteBufAllocator.DEFAULT_CACHE_TRIM_INTERVAL);
    }
    
    protected void onRemoval(PoolThreadCache threadCache)
    {
      threadCache.free();
    }
    
    private <T> PoolArena<T> leastUsedArena(PoolArena<T>[] arenas)
    {
      if ((arenas == null) || (arenas.length == 0)) {
        return null;
      }
      PoolArena<T> minArena = arenas[0];
      for (int i = 1; i < arenas.length; i++)
      {
        PoolArena<T> arena = arenas[i];
        if (numThreadCaches.get() < numThreadCaches.get()) {
          minArena = arena;
        }
      }
      return minArena;
    }
  }
  
  public int numHeapArenas()
  {
    return heapArenaMetrics.size();
  }
  
  public int numDirectArenas()
  {
    return directArenaMetrics.size();
  }
  
  public List<PoolArenaMetric> heapArenas()
  {
    return heapArenaMetrics;
  }
  
  public List<PoolArenaMetric> directArenas()
  {
    return directArenaMetrics;
  }
  
  public int numThreadLocalCaches()
  {
    PoolArena<?>[] arenas = heapArenas != null ? heapArenas : directArenas;
    if (arenas == null) {
      return 0;
    }
    int total = 0;
    for (int i = 0; i < arenas.length; i++) {
      total += numThreadCaches.get();
    }
    return total;
  }
  
  public int tinyCacheSize()
  {
    return tinyCacheSize;
  }
  
  public int smallCacheSize()
  {
    return smallCacheSize;
  }
  
  public int normalCacheSize()
  {
    return normalCacheSize;
  }
  
  final PoolThreadCache threadCache()
  {
    return (PoolThreadCache)threadCache.get();
  }
  
  public String dumpStats()
  {
    int heapArenasLen = heapArenas == null ? 0 : heapArenas.length;
    StringBuilder buf = new StringBuilder(512).append(heapArenasLen).append(" heap arena(s):").append(StringUtil.NEWLINE);
    if (heapArenasLen > 0) {
      for (PoolArena<byte[]> a : heapArenas) {
        buf.append(a);
      }
    }
    int directArenasLen = directArenas == null ? 0 : directArenas.length;
    
    buf.append(directArenasLen).append(" direct arena(s):").append(StringUtil.NEWLINE);
    if (directArenasLen > 0) {
      for (PoolArena<ByteBuffer> a : directArenas) {
        buf.append(a);
      }
    }
    return buf.toString();
  }
}

/* Location:
 * Qualified Name:     io.netty.buffer.PooledByteBufAllocator
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */