package io.netty.channel;

import io.netty.buffer.ByteBufAllocator;
import io.netty.util.DefaultAttributeMap;
import io.netty.util.ReferenceCountUtil;
import io.netty.util.internal.PlatformDependent;
import io.netty.util.internal.ThrowableUtil;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;
import java.io.IOException;
import java.net.ConnectException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NoRouteToHostException;
import java.net.SocketAddress;
import java.net.SocketException;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.NotYetConnectedException;
import java.util.concurrent.Executor;
import java.util.concurrent.RejectedExecutionException;

public abstract class AbstractChannel
  extends DefaultAttributeMap
  implements Channel
{
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(AbstractChannel.class);
  private static final ClosedChannelException FLUSH0_CLOSED_CHANNEL_EXCEPTION = (ClosedChannelException)ThrowableUtil.unknownStackTrace(new ClosedChannelException(), AbstractUnsafe.class, "flush0()");
  private static final ClosedChannelException ENSURE_OPEN_CLOSED_CHANNEL_EXCEPTION = (ClosedChannelException)ThrowableUtil.unknownStackTrace(new ClosedChannelException(), AbstractUnsafe.class, "ensureOpen(...)");
  private static final ClosedChannelException CLOSE_CLOSED_CHANNEL_EXCEPTION = (ClosedChannelException)ThrowableUtil.unknownStackTrace(new ClosedChannelException(), AbstractUnsafe.class, "close(...)");
  private static final ClosedChannelException WRITE_CLOSED_CHANNEL_EXCEPTION = (ClosedChannelException)ThrowableUtil.unknownStackTrace(new ClosedChannelException(), AbstractUnsafe.class, "write(...)");
  private static final NotYetConnectedException FLUSH0_NOT_YET_CONNECTED_EXCEPTION = (NotYetConnectedException)ThrowableUtil.unknownStackTrace(new NotYetConnectedException(), AbstractUnsafe.class, "flush0()");
  private final Channel parent;
  private final ChannelId id;
  private final Channel.Unsafe unsafe;
  private final DefaultChannelPipeline pipeline;
  private final VoidChannelPromise unsafeVoidPromise = new VoidChannelPromise(this, false);
  private final CloseFuture closeFuture = new CloseFuture(this);
  private volatile SocketAddress localAddress;
  private volatile SocketAddress remoteAddress;
  private volatile EventLoop eventLoop;
  private volatile boolean registered;
  private boolean strValActive;
  private String strVal;
  
  protected AbstractChannel(Channel parent)
  {
    this.parent = parent;
    id = newId();
    unsafe = newUnsafe();
    pipeline = newChannelPipeline();
  }
  
  protected AbstractChannel(Channel parent, ChannelId id)
  {
    this.parent = parent;
    this.id = id;
    unsafe = newUnsafe();
    pipeline = newChannelPipeline();
  }
  
  public final ChannelId id()
  {
    return id;
  }
  
  protected ChannelId newId()
  {
    return DefaultChannelId.newInstance();
  }
  
  protected DefaultChannelPipeline newChannelPipeline()
  {
    return new DefaultChannelPipeline(this);
  }
  
  public boolean isWritable()
  {
    ChannelOutboundBuffer buf = unsafe.outboundBuffer();
    return (buf != null) && (buf.isWritable());
  }
  
  public long bytesBeforeUnwritable()
  {
    ChannelOutboundBuffer buf = unsafe.outboundBuffer();
    
    return buf != null ? buf.bytesBeforeUnwritable() : 0L;
  }
  
  public long bytesBeforeWritable()
  {
    ChannelOutboundBuffer buf = unsafe.outboundBuffer();
    
    return buf != null ? buf.bytesBeforeWritable() : Long.MAX_VALUE;
  }
  
  public Channel parent()
  {
    return parent;
  }
  
  public ChannelPipeline pipeline()
  {
    return pipeline;
  }
  
  public ByteBufAllocator alloc()
  {
    return config().getAllocator();
  }
  
  public EventLoop eventLoop()
  {
    EventLoop eventLoop = this.eventLoop;
    if (eventLoop == null) {
      throw new IllegalStateException("channel not registered to an event loop");
    }
    return eventLoop;
  }
  
  public SocketAddress localAddress()
  {
    SocketAddress localAddress = this.localAddress;
    if (localAddress == null) {
      try
      {
        this.localAddress = (localAddress = unsafe().localAddress());
      }
      catch (Throwable t)
      {
        return null;
      }
    }
    return localAddress;
  }
  
  @Deprecated
  protected void invalidateLocalAddress()
  {
    localAddress = null;
  }
  
  public SocketAddress remoteAddress()
  {
    SocketAddress remoteAddress = this.remoteAddress;
    if (remoteAddress == null) {
      try
      {
        this.remoteAddress = (remoteAddress = unsafe().remoteAddress());
      }
      catch (Throwable t)
      {
        return null;
      }
    }
    return remoteAddress;
  }
  
  @Deprecated
  protected void invalidateRemoteAddress()
  {
    remoteAddress = null;
  }
  
  public boolean isRegistered()
  {
    return registered;
  }
  
  public ChannelFuture bind(SocketAddress localAddress)
  {
    return pipeline.bind(localAddress);
  }
  
  public ChannelFuture connect(SocketAddress remoteAddress)
  {
    return pipeline.connect(remoteAddress);
  }
  
  public ChannelFuture connect(SocketAddress remoteAddress, SocketAddress localAddress)
  {
    return pipeline.connect(remoteAddress, localAddress);
  }
  
  public ChannelFuture disconnect()
  {
    return pipeline.disconnect();
  }
  
  public ChannelFuture close()
  {
    return pipeline.close();
  }
  
  public ChannelFuture deregister()
  {
    return pipeline.deregister();
  }
  
  public Channel flush()
  {
    pipeline.flush();
    return this;
  }
  
  public ChannelFuture bind(SocketAddress localAddress, ChannelPromise promise)
  {
    return pipeline.bind(localAddress, promise);
  }
  
  public ChannelFuture connect(SocketAddress remoteAddress, ChannelPromise promise)
  {
    return pipeline.connect(remoteAddress, promise);
  }
  
  public ChannelFuture connect(SocketAddress remoteAddress, SocketAddress localAddress, ChannelPromise promise)
  {
    return pipeline.connect(remoteAddress, localAddress, promise);
  }
  
  public ChannelFuture disconnect(ChannelPromise promise)
  {
    return pipeline.disconnect(promise);
  }
  
  public ChannelFuture close(ChannelPromise promise)
  {
    return pipeline.close(promise);
  }
  
  public ChannelFuture deregister(ChannelPromise promise)
  {
    return pipeline.deregister(promise);
  }
  
  public Channel read()
  {
    pipeline.read();
    return this;
  }
  
  public ChannelFuture write(Object msg)
  {
    return pipeline.write(msg);
  }
  
  public ChannelFuture write(Object msg, ChannelPromise promise)
  {
    return pipeline.write(msg, promise);
  }
  
  public ChannelFuture writeAndFlush(Object msg)
  {
    return pipeline.writeAndFlush(msg);
  }
  
  public ChannelFuture writeAndFlush(Object msg, ChannelPromise promise)
  {
    return pipeline.writeAndFlush(msg, promise);
  }
  
  public ChannelPromise newPromise()
  {
    return pipeline.newPromise();
  }
  
  public ChannelProgressivePromise newProgressivePromise()
  {
    return pipeline.newProgressivePromise();
  }
  
  public ChannelFuture newSucceededFuture()
  {
    return pipeline.newSucceededFuture();
  }
  
  public ChannelFuture newFailedFuture(Throwable cause)
  {
    return pipeline.newFailedFuture(cause);
  }
  
  public ChannelFuture closeFuture()
  {
    return closeFuture;
  }
  
  public Channel.Unsafe unsafe()
  {
    return unsafe;
  }
  
  protected abstract AbstractUnsafe newUnsafe();
  
  public final int hashCode()
  {
    return id.hashCode();
  }
  
  public final boolean equals(Object o)
  {
    return this == o;
  }
  
  public final int compareTo(Channel o)
  {
    if (this == o) {
      return 0;
    }
    return id().compareTo(o.id());
  }
  
  public String toString()
  {
    boolean active = isActive();
    if ((strValActive == active) && (strVal != null)) {
      return strVal;
    }
    SocketAddress remoteAddr = remoteAddress();
    SocketAddress localAddr = localAddress();
    if (remoteAddr != null)
    {
      StringBuilder buf = new StringBuilder(96).append("[id: 0x").append(id.asShortText()).append(", L:").append(localAddr).append(active ? " - " : " ! ").append("R:").append(remoteAddr).append(']');
      
      strVal = buf.toString();
    }
    else if (localAddr != null)
    {
      StringBuilder buf = new StringBuilder(64).append("[id: 0x").append(id.asShortText()).append(", L:").append(localAddr).append(']');
      
      strVal = buf.toString();
    }
    else
    {
      StringBuilder buf = new StringBuilder(16).append("[id: 0x").append(id.asShortText()).append(']');
      
      strVal = buf.toString();
    }
    strValActive = active;
    return strVal;
  }
  
  public final ChannelPromise voidPromise()
  {
    return pipeline.voidPromise();
  }
  
  protected abstract boolean isCompatible(EventLoop paramEventLoop);
  
  protected abstract SocketAddress localAddress0();
  
  protected abstract SocketAddress remoteAddress0();
  
  protected void doRegister()
    throws Exception
  {}
  
  protected abstract void doBind(SocketAddress paramSocketAddress)
    throws Exception;
  
  protected abstract void doDisconnect()
    throws Exception;
  
  protected abstract void doClose()
    throws Exception;
  
  protected void doDeregister()
    throws Exception
  {}
  
  protected abstract void doBeginRead()
    throws Exception;
  
  protected abstract void doWrite(ChannelOutboundBuffer paramChannelOutboundBuffer)
    throws Exception;
  
  protected abstract class AbstractUnsafe
    implements Channel.Unsafe
  {
    private volatile ChannelOutboundBuffer outboundBuffer = new ChannelOutboundBuffer(AbstractChannel.this);
    private RecvByteBufAllocator.Handle recvHandle;
    private boolean inFlush0;
    private boolean neverRegistered = true;
    
    protected AbstractUnsafe() {}
    
    private void assertEventLoop()
    {
      assert ((!registered) || (eventLoop.inEventLoop()));
    }
    
    public RecvByteBufAllocator.Handle recvBufAllocHandle()
    {
      if (recvHandle == null) {
        recvHandle = config().getRecvByteBufAllocator().newHandle();
      }
      return recvHandle;
    }
    
    public final ChannelOutboundBuffer outboundBuffer()
    {
      return outboundBuffer;
    }
    
    public final SocketAddress localAddress()
    {
      return localAddress0();
    }
    
    public final SocketAddress remoteAddress()
    {
      return remoteAddress0();
    }
    
    public final void register(EventLoop eventLoop, final ChannelPromise promise)
    {
      if (eventLoop == null) {
        throw new NullPointerException("eventLoop");
      }
      if (isRegistered())
      {
        promise.setFailure(new IllegalStateException("registered to an event loop already"));
        return;
      }
      if (!isCompatible(eventLoop))
      {
        promise.setFailure(new IllegalStateException("incompatible event loop type: " + eventLoop.getClass().getName()));
        
        return;
      }
      AbstractChannel.this.eventLoop = eventLoop;
      if (eventLoop.inEventLoop()) {
        register0(promise);
      } else {
        try
        {
          eventLoop.execute(new Runnable()
          {
            public void run()
            {
              AbstractChannel.AbstractUnsafe.this.register0(promise);
            }
          });
        }
        catch (Throwable t)
        {
          AbstractChannel.logger.warn("Force-closing a channel whose registration task was not accepted by an event loop: {}", AbstractChannel.this, t);
          
          closeForcibly();
          closeFuture.setClosed();
          safeSetFailure(promise, t);
        }
      }
    }
    
    private void register0(ChannelPromise promise)
    {
      try
      {
        if ((!promise.setUncancellable()) || (!ensureOpen(promise))) {
          return;
        }
        boolean firstRegistration = neverRegistered;
        doRegister();
        neverRegistered = false;
        registered = true;
        
        pipeline.invokeHandlerAddedIfNeeded();
        
        safeSetSuccess(promise);
        pipeline.fireChannelRegistered();
        if (isActive()) {
          if (firstRegistration) {
            pipeline.fireChannelActive();
          } else if (config().isAutoRead()) {
            beginRead();
          }
        }
      }
      catch (Throwable t)
      {
        closeForcibly();
        closeFuture.setClosed();
        safeSetFailure(promise, t);
      }
    }
    
    public final void bind(SocketAddress localAddress, ChannelPromise promise)
    {
      assertEventLoop();
      if ((!promise.setUncancellable()) || (!ensureOpen(promise))) {
        return;
      }
      if ((Boolean.TRUE.equals(config().getOption(ChannelOption.SO_BROADCAST))) && ((localAddress instanceof InetSocketAddress)) && (!((InetSocketAddress)localAddress).getAddress().isAnyLocalAddress()) && (!PlatformDependent.isWindows()) && (!PlatformDependent.isRoot())) {
        AbstractChannel.logger.warn("A non-root user can't receive a broadcast packet if the socket is not bound to a wildcard address; binding to a non-wildcard address (" + localAddress + ") anyway as requested.");
      }
      boolean wasActive = isActive();
      try
      {
        doBind(localAddress);
      }
      catch (Throwable t)
      {
        safeSetFailure(promise, t);
        closeIfClosed();
        return;
      }
      if ((!wasActive) && (isActive())) {
        invokeLater(new Runnable()
        {
          public void run()
          {
            pipeline.fireChannelActive();
          }
        });
      }
      safeSetSuccess(promise);
    }
    
    public final void disconnect(ChannelPromise promise)
    {
      assertEventLoop();
      if (!promise.setUncancellable()) {
        return;
      }
      boolean wasActive = isActive();
      try
      {
        doDisconnect();
      }
      catch (Throwable t)
      {
        safeSetFailure(promise, t);
        closeIfClosed();
        return;
      }
      if ((wasActive) && (!isActive())) {
        invokeLater(new Runnable()
        {
          public void run()
          {
            pipeline.fireChannelInactive();
          }
        });
      }
      safeSetSuccess(promise);
      closeIfClosed();
    }
    
    public final void close(ChannelPromise promise)
    {
      assertEventLoop();
      
      close(promise, AbstractChannel.CLOSE_CLOSED_CHANNEL_EXCEPTION, AbstractChannel.CLOSE_CLOSED_CHANNEL_EXCEPTION, false);
    }
    
    private void close(final ChannelPromise promise, final Throwable cause, final ClosedChannelException closeCause, final boolean notify)
    {
      if (!promise.setUncancellable()) {
        return;
      }
      final ChannelOutboundBuffer outboundBuffer = this.outboundBuffer;
      if (outboundBuffer == null)
      {
        if (!(promise instanceof VoidChannelPromise)) {
          closeFuture.addListener(new ChannelFutureListener()
          {
            public void operationComplete(ChannelFuture future)
              throws Exception
            {
              promise.setSuccess();
            }
          });
        }
        return;
      }
      if (closeFuture.isDone())
      {
        safeSetSuccess(promise);
        return;
      }
      final boolean wasActive = isActive();
      this.outboundBuffer = null;
      Executor closeExecutor = prepareToClose();
      if (closeExecutor != null)
      {
        closeExecutor.execute(new Runnable()
        {
          public void run()
          {
            try
            {
              AbstractChannel.AbstractUnsafe.this.doClose0(promise);
            }
            finally
            {
              AbstractChannel.AbstractUnsafe.this.invokeLater(new Runnable()
              {
                public void run()
                {
                  val$outboundBuffer.failFlushed(val$cause, val$notify);
                  val$outboundBuffer.close(val$closeCause);
                  AbstractChannel.AbstractUnsafe.this.fireChannelInactiveAndDeregister(val$wasActive);
                }
              });
            }
          }
        });
      }
      else
      {
        try
        {
          doClose0(promise);
        }
        finally
        {
          outboundBuffer.failFlushed(cause, notify);
          outboundBuffer.close(closeCause);
        }
        if (inFlush0) {
          invokeLater(new Runnable()
          {
            public void run()
            {
              AbstractChannel.AbstractUnsafe.this.fireChannelInactiveAndDeregister(wasActive);
            }
          });
        } else {
          fireChannelInactiveAndDeregister(wasActive);
        }
      }
    }
    
    private void doClose0(ChannelPromise promise)
    {
      try
      {
        doClose();
        closeFuture.setClosed();
        safeSetSuccess(promise);
      }
      catch (Throwable t)
      {
        closeFuture.setClosed();
        safeSetFailure(promise, t);
      }
    }
    
    private void fireChannelInactiveAndDeregister(boolean wasActive)
    {
      deregister(voidPromise(), (wasActive) && (!isActive()));
    }
    
    public final void closeForcibly()
    {
      assertEventLoop();
      try
      {
        doClose();
      }
      catch (Exception e)
      {
        AbstractChannel.logger.warn("Failed to close a channel.", e);
      }
    }
    
    public final void deregister(ChannelPromise promise)
    {
      assertEventLoop();
      
      deregister(promise, false);
    }
    
    private void deregister(final ChannelPromise promise, final boolean fireChannelInactive)
    {
      if (!promise.setUncancellable()) {
        return;
      }
      if (!registered)
      {
        safeSetSuccess(promise);
        return;
      }
      invokeLater(new Runnable()
      {
        public void run()
        {
          try
          {
            doDeregister();
          }
          catch (Throwable t)
          {
            AbstractChannel.logger.warn("Unexpected exception occurred while deregistering a channel.", t);
          }
          finally
          {
            if (fireChannelInactive) {
              pipeline.fireChannelInactive();
            }
            if (registered)
            {
              registered = false;
              pipeline.fireChannelUnregistered();
            }
            safeSetSuccess(promise);
          }
        }
      });
    }
    
    public final void beginRead()
    {
      assertEventLoop();
      if (!isActive()) {
        return;
      }
      try
      {
        doBeginRead();
      }
      catch (Exception e)
      {
        invokeLater(new Runnable()
        {
          public void run()
          {
            pipeline.fireExceptionCaught(e);
          }
        });
        close(voidPromise());
      }
    }
    
    public final void write(Object msg, ChannelPromise promise)
    {
      assertEventLoop();
      
      ChannelOutboundBuffer outboundBuffer = this.outboundBuffer;
      if (outboundBuffer == null)
      {
        safeSetFailure(promise, AbstractChannel.WRITE_CLOSED_CHANNEL_EXCEPTION);
        
        ReferenceCountUtil.release(msg);
        return;
      }
      try
      {
        msg = filterOutboundMessage(msg);
        int size = pipeline.estimatorHandle().size(msg);
        if (size < 0) {
          size = 0;
        }
      }
      catch (Throwable t)
      {
        safeSetFailure(promise, t);
        ReferenceCountUtil.release(msg); return;
      }
      int size;
      outboundBuffer.addMessage(msg, size, promise);
    }
    
    public final void flush()
    {
      assertEventLoop();
      
      ChannelOutboundBuffer outboundBuffer = this.outboundBuffer;
      if (outboundBuffer == null) {
        return;
      }
      outboundBuffer.addFlush();
      flush0();
    }
    
    protected void flush0()
    {
      if (inFlush0) {
        return;
      }
      ChannelOutboundBuffer outboundBuffer = this.outboundBuffer;
      if ((outboundBuffer == null) || (outboundBuffer.isEmpty())) {
        return;
      }
      inFlush0 = true;
      if (!isActive())
      {
        try
        {
          if (isOpen()) {
            outboundBuffer.failFlushed(AbstractChannel.FLUSH0_NOT_YET_CONNECTED_EXCEPTION, true);
          } else {
            outboundBuffer.failFlushed(AbstractChannel.FLUSH0_CLOSED_CHANNEL_EXCEPTION, false);
          }
        }
        finally
        {
          inFlush0 = false;
        }
        return;
      }
      try
      {
        doWrite(outboundBuffer);
      }
      catch (Throwable t)
      {
        if (((t instanceof IOException)) && (config().isAutoClose())) {
          close(voidPromise(), t, AbstractChannel.FLUSH0_CLOSED_CHANNEL_EXCEPTION, false);
        } else {
          outboundBuffer.failFlushed(t, true);
        }
      }
      finally
      {
        inFlush0 = false;
      }
    }
    
    public final ChannelPromise voidPromise()
    {
      assertEventLoop();
      
      return unsafeVoidPromise;
    }
    
    @Deprecated
    protected final boolean ensureOpen(ChannelPromise promise)
    {
      if (isOpen()) {
        return true;
      }
      safeSetFailure(promise, AbstractChannel.ENSURE_OPEN_CLOSED_CHANNEL_EXCEPTION);
      return false;
    }
    
    protected final void safeSetSuccess(ChannelPromise promise)
    {
      if ((!(promise instanceof VoidChannelPromise)) && (!promise.trySuccess())) {
        AbstractChannel.logger.warn("Failed to mark a promise as success because it is done already: {}", promise);
      }
    }
    
    protected final void safeSetFailure(ChannelPromise promise, Throwable cause)
    {
      if ((!(promise instanceof VoidChannelPromise)) && (!promise.tryFailure(cause))) {
        AbstractChannel.logger.warn("Failed to mark a promise as failure because it's done already: {}", promise, cause);
      }
    }
    
    protected final void closeIfClosed()
    {
      if (isOpen()) {
        return;
      }
      close(voidPromise());
    }
    
    private void invokeLater(Runnable task)
    {
      try
      {
        eventLoop().execute(task);
      }
      catch (RejectedExecutionException e)
      {
        AbstractChannel.logger.warn("Can't invoke task later as EventLoop rejected it", e);
      }
    }
    
    protected final Throwable annotateConnectException(Throwable cause, SocketAddress remoteAddress)
    {
      if ((cause instanceof ConnectException)) {
        return new AbstractChannel.AnnotatedConnectException((ConnectException)cause, remoteAddress);
      }
      if ((cause instanceof NoRouteToHostException)) {
        return new AbstractChannel.AnnotatedNoRouteToHostException((NoRouteToHostException)cause, remoteAddress);
      }
      if ((cause instanceof SocketException)) {
        return new AbstractChannel.AnnotatedSocketException((SocketException)cause, remoteAddress);
      }
      return cause;
    }
    
    protected Executor prepareToClose()
    {
      return null;
    }
  }
  
  protected Object filterOutboundMessage(Object msg)
    throws Exception
  {
    return msg;
  }
  
  static final class CloseFuture
    extends DefaultChannelPromise
  {
    CloseFuture(AbstractChannel ch)
    {
      super();
    }
    
    public ChannelPromise setSuccess()
    {
      throw new IllegalStateException();
    }
    
    public ChannelPromise setFailure(Throwable cause)
    {
      throw new IllegalStateException();
    }
    
    public boolean trySuccess()
    {
      throw new IllegalStateException();
    }
    
    public boolean tryFailure(Throwable cause)
    {
      throw new IllegalStateException();
    }
    
    boolean setClosed()
    {
      return super.trySuccess();
    }
  }
  
  private static final class AnnotatedConnectException
    extends ConnectException
  {
    private static final long serialVersionUID = 3901958112696433556L;
    
    AnnotatedConnectException(ConnectException exception, SocketAddress remoteAddress)
    {
      super();
      setStackTrace(exception.getStackTrace());
    }
    
    public Throwable fillInStackTrace()
    {
      return this;
    }
  }
  
  private static final class AnnotatedNoRouteToHostException
    extends NoRouteToHostException
  {
    private static final long serialVersionUID = -6801433937592080623L;
    
    AnnotatedNoRouteToHostException(NoRouteToHostException exception, SocketAddress remoteAddress)
    {
      super();
      setStackTrace(exception.getStackTrace());
    }
    
    public Throwable fillInStackTrace()
    {
      return this;
    }
  }
  
  private static final class AnnotatedSocketException
    extends SocketException
  {
    private static final long serialVersionUID = 3896743275010454039L;
    
    AnnotatedSocketException(SocketException exception, SocketAddress remoteAddress)
    {
      super();
      setStackTrace(exception.getStackTrace());
    }
    
    public Throwable fillInStackTrace()
    {
      return this;
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.channel.AbstractChannel
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */