package io.netty.channel;

import io.netty.util.Recycler;
import io.netty.util.Recycler.Handle;
import io.netty.util.ReferenceCountUtil;
import io.netty.util.concurrent.EventExecutor;
import io.netty.util.concurrent.PromiseCombiner;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;

public final class PendingWriteQueue
{
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(PendingWriteQueue.class);
  private final ChannelHandlerContext ctx;
  private final ChannelOutboundBuffer buffer;
  private final MessageSizeEstimator.Handle estimatorHandle;
  private PendingWrite head;
  private PendingWrite tail;
  private int size;
  private long bytes;
  
  public PendingWriteQueue(ChannelHandlerContext ctx)
  {
    if (ctx == null) {
      throw new NullPointerException("ctx");
    }
    this.ctx = ctx;
    buffer = ctx.channel().unsafe().outboundBuffer();
    estimatorHandle = ctx.channel().config().getMessageSizeEstimator().newHandle();
  }
  
  public boolean isEmpty()
  {
    assert (ctx.executor().inEventLoop());
    return head == null;
  }
  
  public int size()
  {
    assert (ctx.executor().inEventLoop());
    return size;
  }
  
  public long bytes()
  {
    assert (ctx.executor().inEventLoop());
    return bytes;
  }
  
  public void add(Object msg, ChannelPromise promise)
  {
    assert (ctx.executor().inEventLoop());
    if (msg == null) {
      throw new NullPointerException("msg");
    }
    if (promise == null) {
      throw new NullPointerException("promise");
    }
    int messageSize = estimatorHandle.size(msg);
    if (messageSize < 0) {
      messageSize = 0;
    }
    PendingWrite write = PendingWrite.newInstance(msg, messageSize, promise);
    PendingWrite currentTail = tail;
    if (currentTail == null)
    {
      tail = (head = write);
    }
    else
    {
      next = write;
      tail = write;
    }
    size += 1;
    bytes += messageSize;
    if (buffer != null) {
      buffer.incrementPendingOutboundBytes(size);
    }
  }
  
  public ChannelFuture removeAndWriteAll()
  {
    assert (ctx.executor().inEventLoop());
    if (isEmpty()) {
      return null;
    }
    ChannelPromise p = ctx.newPromise();
    PromiseCombiner combiner = new PromiseCombiner();
    try
    {
      for (PendingWrite write = head; write != null; write = head)
      {
        head = (tail = null);
        size = 0;
        bytes = 0L;
        while (write != null)
        {
          PendingWrite next = next;
          Object msg = msg;
          ChannelPromise promise = promise;
          recycle(write, false);
          combiner.add(promise);
          ctx.write(msg, promise);
          write = next;
        }
      }
      combiner.finish(p);
    }
    catch (Throwable cause)
    {
      p.setFailure(cause);
    }
    assertEmpty();
    return p;
  }
  
  public void removeAndFailAll(Throwable cause)
  {
    assert (ctx.executor().inEventLoop());
    if (cause == null) {
      throw new NullPointerException("cause");
    }
    for (PendingWrite write = head; write != null; write = head)
    {
      head = (tail = null);
      size = 0;
      bytes = 0L;
      while (write != null)
      {
        PendingWrite next = next;
        ReferenceCountUtil.safeRelease(msg);
        ChannelPromise promise = promise;
        recycle(write, false);
        safeFail(promise, cause);
        write = next;
      }
    }
    assertEmpty();
  }
  
  public void removeAndFail(Throwable cause)
  {
    assert (ctx.executor().inEventLoop());
    if (cause == null) {
      throw new NullPointerException("cause");
    }
    PendingWrite write = head;
    if (write == null) {
      return;
    }
    ReferenceCountUtil.safeRelease(msg);
    ChannelPromise promise = promise;
    safeFail(promise, cause);
    recycle(write, true);
  }
  
  private void assertEmpty()
  {
    assert ((tail == null) && (head == null) && (size == 0));
  }
  
  public ChannelFuture removeAndWrite()
  {
    assert (ctx.executor().inEventLoop());
    PendingWrite write = head;
    if (write == null) {
      return null;
    }
    Object msg = msg;
    ChannelPromise promise = promise;
    recycle(write, true);
    return ctx.write(msg, promise);
  }
  
  public ChannelPromise remove()
  {
    assert (ctx.executor().inEventLoop());
    PendingWrite write = head;
    if (write == null) {
      return null;
    }
    ChannelPromise promise = promise;
    ReferenceCountUtil.safeRelease(msg);
    recycle(write, true);
    return promise;
  }
  
  public Object current()
  {
    assert (ctx.executor().inEventLoop());
    PendingWrite write = head;
    if (write == null) {
      return null;
    }
    return msg;
  }
  
  private void recycle(PendingWrite write, boolean update)
  {
    PendingWrite next = next;
    long writeSize = size;
    if (update) {
      if (next == null)
      {
        head = (tail = null);
        size = 0;
        bytes = 0L;
      }
      else
      {
        head = next;
        size -= 1;
        bytes -= writeSize;
        assert ((size > 0) && (bytes >= 0L));
      }
    }
    write.recycle();
    if (buffer != null) {
      buffer.decrementPendingOutboundBytes(writeSize);
    }
  }
  
  private static void safeFail(ChannelPromise promise, Throwable cause)
  {
    if ((!(promise instanceof VoidChannelPromise)) && (!promise.tryFailure(cause))) {
      logger.warn("Failed to mark a promise as failure because it's done already: {}", promise, cause);
    }
  }
  
  static final class PendingWrite
  {
    private static final Recycler<PendingWrite> RECYCLER = new Recycler()
    {
      protected PendingWriteQueue.PendingWrite newObject(Recycler.Handle<PendingWriteQueue.PendingWrite> handle)
      {
        return new PendingWriteQueue.PendingWrite(handle, null);
      }
    };
    private final Recycler.Handle<PendingWrite> handle;
    private PendingWrite next;
    private long size;
    private ChannelPromise promise;
    private Object msg;
    
    private PendingWrite(Recycler.Handle<PendingWrite> handle)
    {
      this.handle = handle;
    }
    
    static PendingWrite newInstance(Object msg, int size, ChannelPromise promise)
    {
      PendingWrite write = (PendingWrite)RECYCLER.get();
      size = size;
      msg = msg;
      promise = promise;
      return write;
    }
    
    private void recycle()
    {
      size = 0L;
      next = null;
      msg = null;
      promise = null;
      handle.recycle(this);
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.channel.PendingWriteQueue
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */