package io.netty.channel.epoll;

import io.netty.buffer.ByteBuf;
import io.netty.buffer.ByteBufAllocator;
import io.netty.buffer.ByteBufUtil;
import io.netty.buffer.Unpooled;
import io.netty.channel.AbstractChannel;
import io.netty.channel.AbstractChannel.AbstractUnsafe;
import io.netty.channel.Channel;
import io.netty.channel.Channel.Unsafe;
import io.netty.channel.ChannelConfig;
import io.netty.channel.ChannelException;
import io.netty.channel.ChannelMetadata;
import io.netty.channel.ChannelOption;
import io.netty.channel.ChannelPipeline;
import io.netty.channel.EventLoop;
import io.netty.channel.RecvByteBufAllocator.Handle;
import io.netty.channel.socket.ChannelInputShutdownEvent;
import io.netty.channel.unix.Socket;
import io.netty.channel.unix.UnixChannel;
import io.netty.util.ReferenceCountUtil;
import io.netty.util.internal.ObjectUtil;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.NotYetConnectedException;
import java.nio.channels.UnresolvedAddressException;

abstract class AbstractEpollChannel
  extends AbstractChannel
  implements UnixChannel
{
  private static final ChannelMetadata METADATA = new ChannelMetadata(false);
  private final int readFlag;
  private final Socket fileDescriptor;
  protected int flags = Native.EPOLLET;
  protected volatile boolean active;
  
  AbstractEpollChannel(Socket fd, int flag)
  {
    this(null, fd, flag, false);
  }
  
  AbstractEpollChannel(Channel parent, Socket fd, int flag, boolean active)
  {
    super(parent);
    fileDescriptor = ((Socket)ObjectUtil.checkNotNull(fd, "fd"));
    readFlag = flag;
    flags |= flag;
    this.active = active;
  }
  
  static boolean isSoErrorZero(Socket fd)
  {
    try
    {
      return fd.getSoError() == 0;
    }
    catch (IOException e)
    {
      throw new ChannelException(e);
    }
  }
  
  void setFlag(int flag)
    throws IOException
  {
    if (!isFlagSet(flag))
    {
      flags |= flag;
      modifyEvents();
    }
  }
  
  void clearFlag(int flag)
    throws IOException
  {
    if (isFlagSet(flag))
    {
      flags &= (flag ^ 0xFFFFFFFF);
      modifyEvents();
    }
  }
  
  boolean isFlagSet(int flag)
  {
    return (flags & flag) != 0;
  }
  
  public final Socket fd()
  {
    return fileDescriptor;
  }
  
  public abstract EpollChannelConfig config();
  
  public boolean isActive()
  {
    return active;
  }
  
  public ChannelMetadata metadata()
  {
    return METADATA;
  }
  
  protected void doClose()
    throws Exception
  {
    active = false;
    try
    {
      doDeregister();
    }
    finally
    {
      fileDescriptor.close();
    }
  }
  
  protected void doDisconnect()
    throws Exception
  {
    doClose();
  }
  
  protected boolean isCompatible(EventLoop loop)
  {
    return loop instanceof EpollEventLoop;
  }
  
  public boolean isOpen()
  {
    return fileDescriptor.isOpen();
  }
  
  protected void doDeregister()
    throws Exception
  {
    ((EpollEventLoop)eventLoop()).remove(this);
  }
  
  protected final void doBeginRead()
    throws Exception
  {
    AbstractEpollUnsafe unsafe = (AbstractEpollUnsafe)unsafe();
    readPending = true;
    
    setFlag(readFlag);
    if (maybeMoreDataToRead) {
      unsafe.executeEpollInReadyRunnable();
    }
  }
  
  final void clearEpollIn()
  {
    if (isRegistered())
    {
      EventLoop loop = eventLoop();
      final AbstractEpollUnsafe unsafe = (AbstractEpollUnsafe)unsafe();
      if (loop.inEventLoop()) {
        unsafe.clearEpollIn0();
      } else {
        loop.execute(new Runnable()
        {
          public void run()
          {
            if ((!unsafereadPending) && (!config().isAutoRead())) {
              unsafe.clearEpollIn0();
            }
          }
        });
      }
    }
    else
    {
      flags &= (readFlag ^ 0xFFFFFFFF);
    }
  }
  
  private void modifyEvents()
    throws IOException
  {
    if ((isOpen()) && (isRegistered())) {
      ((EpollEventLoop)eventLoop()).modify(this);
    }
  }
  
  protected void doRegister()
    throws Exception
  {
    EpollEventLoop loop = (EpollEventLoop)eventLoop();
    
    unsafeepollInReadyRunnablePending = false;
    loop.add(this);
  }
  
  protected abstract AbstractEpollUnsafe newUnsafe();
  
  protected final ByteBuf newDirectBuffer(ByteBuf buf)
  {
    return newDirectBuffer(buf, buf);
  }
  
  protected final ByteBuf newDirectBuffer(Object holder, ByteBuf buf)
  {
    int readableBytes = buf.readableBytes();
    if (readableBytes == 0)
    {
      ReferenceCountUtil.safeRelease(holder);
      return Unpooled.EMPTY_BUFFER;
    }
    ByteBufAllocator alloc = alloc();
    if (alloc.isDirectBufferPooled()) {
      return newDirectBuffer0(holder, buf, alloc, readableBytes);
    }
    ByteBuf directBuf = ByteBufUtil.threadLocalDirectBuffer();
    if (directBuf == null) {
      return newDirectBuffer0(holder, buf, alloc, readableBytes);
    }
    directBuf.writeBytes(buf, buf.readerIndex(), readableBytes);
    ReferenceCountUtil.safeRelease(holder);
    return directBuf;
  }
  
  private static ByteBuf newDirectBuffer0(Object holder, ByteBuf buf, ByteBufAllocator alloc, int capacity)
  {
    ByteBuf directBuf = alloc.directBuffer(capacity);
    directBuf.writeBytes(buf, buf.readerIndex(), capacity);
    ReferenceCountUtil.safeRelease(holder);
    return directBuf;
  }
  
  protected static void checkResolvable(InetSocketAddress addr)
  {
    if (addr.isUnresolved()) {
      throw new UnresolvedAddressException();
    }
  }
  
  protected final int doReadBytes(ByteBuf byteBuf)
    throws Exception
  {
    int writerIndex = byteBuf.writerIndex();
    
    unsafe().recvBufAllocHandle().attemptedBytesRead(byteBuf.writableBytes());
    int localReadAmount;
    int localReadAmount;
    if (byteBuf.hasMemoryAddress())
    {
      localReadAmount = fileDescriptor.readAddress(byteBuf.memoryAddress(), writerIndex, byteBuf.capacity());
    }
    else
    {
      ByteBuffer buf = byteBuf.internalNioBuffer(writerIndex, byteBuf.writableBytes());
      localReadAmount = fileDescriptor.read(buf, buf.position(), buf.limit());
    }
    if (localReadAmount > 0) {
      byteBuf.writerIndex(writerIndex + localReadAmount);
    }
    return localReadAmount;
  }
  
  protected final int doWriteBytes(ByteBuf buf, int writeSpinCount)
    throws Exception
  {
    int readableBytes = buf.readableBytes();
    int writtenBytes = 0;
    if (buf.hasMemoryAddress())
    {
      long memoryAddress = buf.memoryAddress();
      int readerIndex = buf.readerIndex();
      int writerIndex = buf.writerIndex();
      for (int i = writeSpinCount - 1; i >= 0; i--)
      {
        int localFlushedAmount = fileDescriptor.writeAddress(memoryAddress, readerIndex, writerIndex);
        if (localFlushedAmount <= 0) {
          break;
        }
        writtenBytes += localFlushedAmount;
        if (writtenBytes == readableBytes) {
          return writtenBytes;
        }
        readerIndex += localFlushedAmount;
      }
    }
    else
    {
      ByteBuffer nioBuf;
      ByteBuffer nioBuf;
      if (buf.nioBufferCount() == 1) {
        nioBuf = buf.internalNioBuffer(buf.readerIndex(), buf.readableBytes());
      } else {
        nioBuf = buf.nioBuffer();
      }
      for (int i = writeSpinCount - 1; i >= 0; i--)
      {
        int pos = nioBuf.position();
        int limit = nioBuf.limit();
        int localFlushedAmount = fileDescriptor.write(nioBuf, pos, limit);
        if (localFlushedAmount <= 0) {
          break;
        }
        nioBuf.position(pos + localFlushedAmount);
        writtenBytes += localFlushedAmount;
        if (writtenBytes == readableBytes) {
          return writtenBytes;
        }
      }
    }
    if (writtenBytes < readableBytes) {
      setFlag(Native.EPOLLOUT);
    }
    return writtenBytes;
  }
  
  protected abstract class AbstractEpollUnsafe
    extends AbstractChannel.AbstractUnsafe
  {
    boolean readPending;
    boolean maybeMoreDataToRead;
    boolean epollInReadyRunnablePending;
    private EpollRecvByteAllocatorHandle allocHandle;
    private Runnable epollInReadyRunnable;
    
    protected AbstractEpollUnsafe()
    {
      super();
    }
    
    abstract void epollInReady();
    
    final void epollInBefore()
    {
      maybeMoreDataToRead = false;
    }
    
    final void epollInFinally(ChannelConfig config)
    {
      maybeMoreDataToRead = allocHandle.maybeMoreDataToRead();
      if ((!readPending) && (!config.isAutoRead())) {
        clearEpollIn();
      } else if ((readPending) && (maybeMoreDataToRead) && (!fd().isInputShutdown())) {
        executeEpollInReadyRunnable();
      }
    }
    
    final void executeEpollInReadyRunnable()
    {
      if (epollInReadyRunnablePending) {
        return;
      }
      epollInReadyRunnablePending = true;
      if (epollInReadyRunnable == null) {
        epollInReadyRunnable = new Runnable()
        {
          public void run()
          {
            epollInReadyRunnablePending = false;
            epollInReady();
          }
        };
      }
      eventLoop().execute(epollInReadyRunnable);
    }
    
    final void epollRdHupReady()
    {
      recvBufAllocHandle().receivedRdHup();
      if (isActive())
      {
        epollInReady();
        
        clearEpollRdHup();
      }
      shutdownInput();
    }
    
    private void clearEpollRdHup()
    {
      try
      {
        clearFlag(Native.EPOLLRDHUP);
      }
      catch (IOException e)
      {
        pipeline().fireExceptionCaught(e);
        close(voidPromise());
      }
    }
    
    void shutdownInput()
    {
      if (!fd().isInputShutdown()) {
        if (Boolean.TRUE.equals(config().getOption(ChannelOption.ALLOW_HALF_CLOSURE)))
        {
          try
          {
            fd().shutdown(true, false);
            clearEpollIn0();
          }
          catch (IOException ignored)
          {
            fireEventAndClose(ChannelInputShutdownEvent.INSTANCE);
            return;
          }
          catch (NotYetConnectedException ignore)
          {
            fireEventAndClose(ChannelInputShutdownEvent.INSTANCE);
            return;
          }
          pipeline().fireUserEventTriggered(ChannelInputShutdownEvent.INSTANCE);
        }
        else
        {
          close(voidPromise());
        }
      }
    }
    
    private void fireEventAndClose(Object evt)
    {
      pipeline().fireUserEventTriggered(evt);
      close(voidPromise());
    }
    
    public EpollRecvByteAllocatorHandle recvBufAllocHandle()
    {
      if (allocHandle == null) {
        allocHandle = newEpollHandle(super.recvBufAllocHandle());
      }
      return allocHandle;
    }
    
    EpollRecvByteAllocatorHandle newEpollHandle(RecvByteBufAllocator.Handle handle)
    {
      return new EpollRecvByteAllocatorHandle(handle, config());
    }
    
    protected void flush0()
    {
      if (isFlagSet(Native.EPOLLOUT)) {
        return;
      }
      super.flush0();
    }
    
    void epollOutReady()
    {
      if (fd().isOutputShutdown()) {
        return;
      }
      super.flush0();
    }
    
    protected final void clearEpollIn0()
    {
      assert (eventLoop().inEventLoop());
      try
      {
        readPending = false;
        clearFlag(readFlag);
      }
      catch (IOException e)
      {
        pipeline().fireExceptionCaught(e);
        unsafe().close(unsafe().voidPromise());
      }
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.channel.epoll.AbstractEpollChannel
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */