package io.netty.channel.epoll;

import io.netty.buffer.ByteBuf;
import io.netty.buffer.ByteBufAllocator;
import io.netty.buffer.CompositeByteBuf;
import io.netty.channel.AddressedEnvelope;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelMetadata;
import io.netty.channel.ChannelOption;
import io.netty.channel.ChannelOutboundBuffer;
import io.netty.channel.ChannelPipeline;
import io.netty.channel.ChannelPromise;
import io.netty.channel.DefaultAddressedEnvelope;
import io.netty.channel.EventLoop;
import io.netty.channel.socket.DatagramChannel;
import io.netty.channel.socket.DatagramChannelConfig;
import io.netty.channel.socket.DatagramPacket;
import io.netty.channel.unix.DatagramSocketAddress;
import io.netty.channel.unix.FileDescriptor;
import io.netty.channel.unix.Socket;
import io.netty.util.internal.PlatformDependent;
import io.netty.util.internal.StringUtil;
import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NetworkInterface;
import java.net.SocketAddress;
import java.net.SocketException;
import java.nio.ByteBuffer;
import java.nio.channels.NotYetConnectedException;
import java.util.ArrayList;
import java.util.List;

public final class EpollDatagramChannel
  extends AbstractEpollChannel
  implements DatagramChannel
{
  private static final ChannelMetadata METADATA = new ChannelMetadata(true);
  private static final String EXPECTED_TYPES = " (expected: " + StringUtil.simpleClassName(DatagramPacket.class) + ", " + StringUtil.simpleClassName(AddressedEnvelope.class) + '<' + StringUtil.simpleClassName(ByteBuf.class) + ", " + StringUtil.simpleClassName(InetSocketAddress.class) + ">, " + StringUtil.simpleClassName(ByteBuf.class) + ')';
  private volatile InetSocketAddress local;
  private volatile InetSocketAddress remote;
  private volatile boolean connected;
  private final EpollDatagramChannelConfig config;
  
  public EpollDatagramChannel()
  {
    super(Socket.newSocketDgram(), Native.EPOLLIN);
    config = new EpollDatagramChannelConfig(this);
  }
  
  @Deprecated
  public EpollDatagramChannel(FileDescriptor fd)
  {
    this(new Socket(fd.intValue()));
  }
  
  public EpollDatagramChannel(Socket fd)
  {
    super(null, fd, Native.EPOLLIN, true);
    
    local = fd.localAddress();
    config = new EpollDatagramChannelConfig(this);
  }
  
  public InetSocketAddress remoteAddress()
  {
    return (InetSocketAddress)super.remoteAddress();
  }
  
  public InetSocketAddress localAddress()
  {
    return (InetSocketAddress)super.localAddress();
  }
  
  public ChannelMetadata metadata()
  {
    return METADATA;
  }
  
  public boolean isActive()
  {
    return (fd().isOpen()) && (((((Boolean)config.getOption(ChannelOption.DATAGRAM_CHANNEL_ACTIVE_ON_REGISTRATION)).booleanValue()) && (isRegistered())) || (active));
  }
  
  public boolean isConnected()
  {
    return connected;
  }
  
  public ChannelFuture joinGroup(InetAddress multicastAddress)
  {
    return joinGroup(multicastAddress, newPromise());
  }
  
  public ChannelFuture joinGroup(InetAddress multicastAddress, ChannelPromise promise)
  {
    try
    {
      return joinGroup(multicastAddress, NetworkInterface.getByInetAddress(localAddress().getAddress()), null, promise);
    }
    catch (SocketException e)
    {
      promise.setFailure(e);
    }
    return promise;
  }
  
  public ChannelFuture joinGroup(InetSocketAddress multicastAddress, NetworkInterface networkInterface)
  {
    return joinGroup(multicastAddress, networkInterface, newPromise());
  }
  
  public ChannelFuture joinGroup(InetSocketAddress multicastAddress, NetworkInterface networkInterface, ChannelPromise promise)
  {
    return joinGroup(multicastAddress.getAddress(), networkInterface, null, promise);
  }
  
  public ChannelFuture joinGroup(InetAddress multicastAddress, NetworkInterface networkInterface, InetAddress source)
  {
    return joinGroup(multicastAddress, networkInterface, source, newPromise());
  }
  
  public ChannelFuture joinGroup(InetAddress multicastAddress, NetworkInterface networkInterface, InetAddress source, ChannelPromise promise)
  {
    if (multicastAddress == null) {
      throw new NullPointerException("multicastAddress");
    }
    if (networkInterface == null) {
      throw new NullPointerException("networkInterface");
    }
    promise.setFailure(new UnsupportedOperationException("Multicast not supported"));
    return promise;
  }
  
  public ChannelFuture leaveGroup(InetAddress multicastAddress)
  {
    return leaveGroup(multicastAddress, newPromise());
  }
  
  public ChannelFuture leaveGroup(InetAddress multicastAddress, ChannelPromise promise)
  {
    try
    {
      return leaveGroup(multicastAddress, NetworkInterface.getByInetAddress(localAddress().getAddress()), null, promise);
    }
    catch (SocketException e)
    {
      promise.setFailure(e);
    }
    return promise;
  }
  
  public ChannelFuture leaveGroup(InetSocketAddress multicastAddress, NetworkInterface networkInterface)
  {
    return leaveGroup(multicastAddress, networkInterface, newPromise());
  }
  
  public ChannelFuture leaveGroup(InetSocketAddress multicastAddress, NetworkInterface networkInterface, ChannelPromise promise)
  {
    return leaveGroup(multicastAddress.getAddress(), networkInterface, null, promise);
  }
  
  public ChannelFuture leaveGroup(InetAddress multicastAddress, NetworkInterface networkInterface, InetAddress source)
  {
    return leaveGroup(multicastAddress, networkInterface, source, newPromise());
  }
  
  public ChannelFuture leaveGroup(InetAddress multicastAddress, NetworkInterface networkInterface, InetAddress source, ChannelPromise promise)
  {
    if (multicastAddress == null) {
      throw new NullPointerException("multicastAddress");
    }
    if (networkInterface == null) {
      throw new NullPointerException("networkInterface");
    }
    promise.setFailure(new UnsupportedOperationException("Multicast not supported"));
    
    return promise;
  }
  
  public ChannelFuture block(InetAddress multicastAddress, NetworkInterface networkInterface, InetAddress sourceToBlock)
  {
    return block(multicastAddress, networkInterface, sourceToBlock, newPromise());
  }
  
  public ChannelFuture block(InetAddress multicastAddress, NetworkInterface networkInterface, InetAddress sourceToBlock, ChannelPromise promise)
  {
    if (multicastAddress == null) {
      throw new NullPointerException("multicastAddress");
    }
    if (sourceToBlock == null) {
      throw new NullPointerException("sourceToBlock");
    }
    if (networkInterface == null) {
      throw new NullPointerException("networkInterface");
    }
    promise.setFailure(new UnsupportedOperationException("Multicast not supported"));
    return promise;
  }
  
  public ChannelFuture block(InetAddress multicastAddress, InetAddress sourceToBlock)
  {
    return block(multicastAddress, sourceToBlock, newPromise());
  }
  
  public ChannelFuture block(InetAddress multicastAddress, InetAddress sourceToBlock, ChannelPromise promise)
  {
    try
    {
      return block(multicastAddress, NetworkInterface.getByInetAddress(localAddress().getAddress()), sourceToBlock, promise);
    }
    catch (Throwable e)
    {
      promise.setFailure(e);
    }
    return promise;
  }
  
  protected AbstractEpollChannel.AbstractEpollUnsafe newUnsafe()
  {
    return new EpollDatagramChannelUnsafe();
  }
  
  protected InetSocketAddress localAddress0()
  {
    return local;
  }
  
  protected InetSocketAddress remoteAddress0()
  {
    return remote;
  }
  
  protected void doBind(SocketAddress localAddress)
    throws Exception
  {
    InetSocketAddress addr = (InetSocketAddress)localAddress;
    checkResolvable(addr);
    fd().bind(addr);
    local = fd().localAddress();
    active = true;
  }
  
  protected void doWrite(ChannelOutboundBuffer in)
    throws Exception
  {
    for (;;)
    {
      Object msg = in.current();
      if (msg == null)
      {
        clearFlag(Native.EPOLLOUT);
        break;
      }
      try
      {
        if ((Native.IS_SUPPORTING_SENDMMSG) && (in.size() > 1))
        {
          NativeDatagramPacketArray array = NativeDatagramPacketArray.getInstance(in);
          int cnt = array.count();
          if (cnt >= 1)
          {
            int offset = 0;
            NativeDatagramPacketArray.NativeDatagramPacket[] packets = array.packets();
            while (cnt > 0)
            {
              int send = Native.sendmmsg(fd().intValue(), packets, offset, cnt);
              if (send == 0)
              {
                setFlag(Native.EPOLLOUT);
                return;
              }
              for (int i = 0; i < send; i++) {
                in.remove();
              }
              cnt -= send;
              offset += send;
            }
            continue;
          }
        }
        boolean done = false;
        for (int i = config().getWriteSpinCount() - 1; i >= 0; i--) {
          if (doWriteMessage(msg))
          {
            done = true;
            break;
          }
        }
        if (done)
        {
          in.remove();
        }
        else
        {
          setFlag(Native.EPOLLOUT);
          break;
        }
      }
      catch (IOException e)
      {
        in.remove(e);
      }
    }
  }
  
  private boolean doWriteMessage(Object msg)
    throws Exception
  {
    InetSocketAddress remoteAddress;
    ByteBuf data;
    InetSocketAddress remoteAddress;
    if ((msg instanceof AddressedEnvelope))
    {
      AddressedEnvelope<ByteBuf, InetSocketAddress> envelope = (AddressedEnvelope)msg;
      
      ByteBuf data = (ByteBuf)envelope.content();
      remoteAddress = (InetSocketAddress)envelope.recipient();
    }
    else
    {
      data = (ByteBuf)msg;
      remoteAddress = null;
    }
    int dataLen = data.readableBytes();
    if (dataLen == 0) {
      return true;
    }
    if (remoteAddress == null)
    {
      remoteAddress = remote;
      if (remoteAddress == null) {
        throw new NotYetConnectedException();
      }
    }
    int writtenBytes;
    int writtenBytes;
    if (data.hasMemoryAddress())
    {
      long memoryAddress = data.memoryAddress();
      writtenBytes = fd().sendToAddress(memoryAddress, data.readerIndex(), data.writerIndex(), remoteAddress.getAddress(), remoteAddress.getPort());
    }
    else
    {
      int writtenBytes;
      if ((data instanceof CompositeByteBuf))
      {
        IovArray array = ((EpollEventLoop)eventLoop()).cleanArray();
        array.add(data);
        int cnt = array.count();
        assert (cnt != 0);
        
        writtenBytes = fd().sendToAddresses(array.memoryAddress(0), cnt, remoteAddress.getAddress(), remoteAddress.getPort());
      }
      else
      {
        ByteBuffer nioData = data.internalNioBuffer(data.readerIndex(), data.readableBytes());
        writtenBytes = fd().sendTo(nioData, nioData.position(), nioData.limit(), remoteAddress.getAddress(), remoteAddress.getPort());
      }
    }
    return writtenBytes > 0;
  }
  
  protected Object filterOutboundMessage(Object msg)
  {
    if ((msg instanceof DatagramPacket))
    {
      DatagramPacket packet = (DatagramPacket)msg;
      ByteBuf content = (ByteBuf)packet.content();
      if (content.hasMemoryAddress()) {
        return msg;
      }
      if ((content.isDirect()) && ((content instanceof CompositeByteBuf)))
      {
        CompositeByteBuf comp = (CompositeByteBuf)content;
        if ((comp.isDirect()) && (comp.nioBufferCount() <= Native.IOV_MAX)) {
          return msg;
        }
      }
      return new DatagramPacket(newDirectBuffer(packet, content), (InetSocketAddress)packet.recipient());
    }
    if ((msg instanceof ByteBuf))
    {
      ByteBuf buf = (ByteBuf)msg;
      if ((!buf.hasMemoryAddress()) && ((PlatformDependent.hasUnsafe()) || (!buf.isDirect()))) {
        if ((buf instanceof CompositeByteBuf))
        {
          CompositeByteBuf comp = (CompositeByteBuf)buf;
          if ((!comp.isDirect()) || (comp.nioBufferCount() > Native.IOV_MAX))
          {
            buf = newDirectBuffer(buf);
            assert (buf.hasMemoryAddress());
          }
        }
        else
        {
          buf = newDirectBuffer(buf);
          assert (buf.hasMemoryAddress());
        }
      }
      return buf;
    }
    if ((msg instanceof AddressedEnvelope))
    {
      AddressedEnvelope<Object, SocketAddress> e = (AddressedEnvelope)msg;
      if (((e.content() instanceof ByteBuf)) && ((e.recipient() == null) || ((e.recipient() instanceof InetSocketAddress))))
      {
        ByteBuf content = (ByteBuf)e.content();
        if (content.hasMemoryAddress()) {
          return e;
        }
        if ((content instanceof CompositeByteBuf))
        {
          CompositeByteBuf comp = (CompositeByteBuf)content;
          if ((comp.isDirect()) && (comp.nioBufferCount() <= Native.IOV_MAX)) {
            return e;
          }
        }
        return new DefaultAddressedEnvelope(newDirectBuffer(e, content), (InetSocketAddress)e.recipient());
      }
    }
    throw new UnsupportedOperationException("unsupported message type: " + StringUtil.simpleClassName(msg) + EXPECTED_TYPES);
  }
  
  public EpollDatagramChannelConfig config()
  {
    return config;
  }
  
  protected void doDisconnect()
    throws Exception
  {
    connected = false;
  }
  
  final class EpollDatagramChannelUnsafe
    extends AbstractEpollChannel.AbstractEpollUnsafe
  {
    EpollDatagramChannelUnsafe()
    {
      super();
    }
    
    private final List<Object> readBuf = new ArrayList();
    
    public void connect(SocketAddress remote, SocketAddress local, ChannelPromise channelPromise)
    {
      boolean success = false;
      try
      {
        try
        {
          boolean wasActive = isActive();
          InetSocketAddress remoteAddress = (InetSocketAddress)remote;
          if (local != null)
          {
            InetSocketAddress localAddress = (InetSocketAddress)local;
            doBind(localAddress);
          }
          AbstractEpollChannel.checkResolvable(remoteAddress);
          EpollDatagramChannel.this.remote = remoteAddress;
          EpollDatagramChannel.this.local = fd().localAddress();
          success = true;
          
          channelPromise.trySuccess();
          if ((!wasActive) && (isActive())) {
            pipeline().fireChannelActive();
          }
        }
        finally
        {
          if (!success) {
            doClose();
          } else {
            connected = true;
          }
        }
      }
      catch (Throwable cause)
      {
        channelPromise.tryFailure(cause);
      }
    }
    
    void epollInReady()
    {
      assert (eventLoop().inEventLoop());
      if (fd().isInputShutdown())
      {
        clearEpollIn0();
        return;
      }
      DatagramChannelConfig config = config();
      EpollRecvByteAllocatorHandle allocHandle = recvBufAllocHandle();
      allocHandle.edgeTriggered(isFlagSet(Native.EPOLLET));
      
      ChannelPipeline pipeline = pipeline();
      ByteBufAllocator allocator = config.getAllocator();
      allocHandle.reset(config);
      epollInBefore();
      
      Throwable exception = null;
      try
      {
        ByteBuf data = null;
        try
        {
          do
          {
            data = allocHandle.allocate(allocator);
            allocHandle.attemptedBytesRead(data.writableBytes());
            DatagramSocketAddress remoteAddress;
            DatagramSocketAddress remoteAddress;
            if (data.hasMemoryAddress())
            {
              remoteAddress = fd().recvFromAddress(data.memoryAddress(), data.writerIndex(), data.capacity());
            }
            else
            {
              ByteBuffer nioData = data.internalNioBuffer(data.writerIndex(), data.writableBytes());
              remoteAddress = fd().recvFrom(nioData, nioData.position(), nioData.limit());
            }
            if (remoteAddress == null)
            {
              allocHandle.lastBytesRead(-1);
              data.release();
              data = null;
              break;
            }
            allocHandle.incMessagesRead(1);
            allocHandle.lastBytesRead(remoteAddress.receivedAmount());
            data.writerIndex(data.writerIndex() + allocHandle.lastBytesRead());
            
            readBuf.add(new DatagramPacket(data, (InetSocketAddress)localAddress(), remoteAddress));
            data = null;
          } while (allocHandle.continueReading());
        }
        catch (Throwable t)
        {
          if (data != null) {
            data.release();
          }
          exception = t;
        }
        int size = readBuf.size();
        for (int i = 0; i < size; i++)
        {
          readPending = false;
          pipeline.fireChannelRead(readBuf.get(i));
        }
        readBuf.clear();
        allocHandle.readComplete();
        pipeline.fireChannelReadComplete();
        if (exception != null) {
          pipeline.fireExceptionCaught(exception);
        }
      }
      finally
      {
        epollInFinally(config);
      }
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.channel.epoll.EpollDatagramChannel
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */