package io.netty.channel.epoll;

import io.netty.channel.Channel.Unsafe;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.SelectStrategy;
import io.netty.channel.SingleThreadEventLoop;
import io.netty.channel.unix.FileDescriptor;
import io.netty.channel.unix.Socket;
import io.netty.util.IntSupplier;
import io.netty.util.collection.IntObjectHashMap;
import io.netty.util.collection.IntObjectMap;
import io.netty.util.concurrent.Future;
import io.netty.util.concurrent.RejectedExecutionHandler;
import io.netty.util.internal.ObjectUtil;
import io.netty.util.internal.PlatformDependent;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Queue;
import java.util.concurrent.Callable;
import java.util.concurrent.Executor;
import java.util.concurrent.atomic.AtomicIntegerFieldUpdater;

final class EpollEventLoop
  extends SingleThreadEventLoop
{
  private static final InternalLogger logger;
  private static final AtomicIntegerFieldUpdater<EpollEventLoop> WAKEN_UP_UPDATER;
  private final FileDescriptor epollFd;
  private final FileDescriptor eventFd;
  
  static
  {
    logger = InternalLoggerFactory.getInstance(EpollEventLoop.class);
    
    AtomicIntegerFieldUpdater<EpollEventLoop> updater = PlatformDependent.newAtomicIntegerFieldUpdater(EpollEventLoop.class, "wakenUp");
    if (updater == null) {
      updater = AtomicIntegerFieldUpdater.newUpdater(EpollEventLoop.class, "wakenUp");
    }
    WAKEN_UP_UPDATER = updater;
  }
  
  private final IntObjectMap<AbstractEpollChannel> channels = new IntObjectHashMap(4096);
  private final boolean allowGrowing;
  private final EpollEventArray events;
  private final IovArray iovArray = new IovArray();
  private final SelectStrategy selectStrategy;
  private final IntSupplier selectNowSupplier = new IntSupplier()
  {
    public int get()
      throws Exception
    {
      return Native.epollWait(epollFd.intValue(), events, 0);
    }
  };
  private final Callable<Integer> pendingTasksCallable = new Callable()
  {
    public Integer call()
      throws Exception
    {
      return Integer.valueOf(EpollEventLoop.this.pendingTasks());
    }
  };
  private volatile int wakenUp;
  private volatile int ioRatio = 50;
  
  EpollEventLoop(EventLoopGroup parent, Executor executor, int maxEvents, SelectStrategy strategy, RejectedExecutionHandler rejectedExecutionHandler)
  {
    super(parent, executor, false, DEFAULT_MAX_PENDING_TASKS, rejectedExecutionHandler);
    selectStrategy = ((SelectStrategy)ObjectUtil.checkNotNull(strategy, "strategy"));
    if (maxEvents == 0)
    {
      allowGrowing = true;
      events = new EpollEventArray(4096);
    }
    else
    {
      allowGrowing = false;
      events = new EpollEventArray(maxEvents);
    }
    boolean success = false;
    FileDescriptor epollFd = null;
    FileDescriptor eventFd = null;
    try
    {
      this.epollFd = (epollFd = Native.newEpollCreate());
      this.eventFd = (eventFd = Native.newEventFd());
      try
      {
        Native.epollCtlAdd(epollFd.intValue(), eventFd.intValue(), Native.EPOLLIN);
      }
      catch (IOException e)
      {
        throw new IllegalStateException("Unable to add eventFd filedescriptor to epoll", e);
      }
      success = true; return;
    }
    finally
    {
      if (!success)
      {
        if (epollFd != null) {
          try
          {
            epollFd.close();
          }
          catch (Exception localException2) {}
        }
        if (eventFd != null) {
          try
          {
            eventFd.close();
          }
          catch (Exception localException3) {}
        }
      }
    }
  }
  
  IovArray cleanArray()
  {
    iovArray.clear();
    return iovArray;
  }
  
  protected void wakeup(boolean inEventLoop)
  {
    if ((!inEventLoop) && (WAKEN_UP_UPDATER.compareAndSet(this, 0, 1))) {
      Native.eventFdWrite(eventFd.intValue(), 1L);
    }
  }
  
  void add(AbstractEpollChannel ch)
    throws IOException
  {
    assert (inEventLoop());
    int fd = ch.fd().intValue();
    Native.epollCtlAdd(epollFd.intValue(), fd, flags);
    channels.put(fd, ch);
  }
  
  void modify(AbstractEpollChannel ch)
    throws IOException
  {
    assert (inEventLoop());
    Native.epollCtlMod(epollFd.intValue(), ch.fd().intValue(), flags);
  }
  
  void remove(AbstractEpollChannel ch)
    throws IOException
  {
    assert (inEventLoop());
    if (ch.isOpen())
    {
      int fd = ch.fd().intValue();
      if (channels.remove(fd) != null) {
        Native.epollCtlDel(epollFd.intValue(), ch.fd().intValue());
      }
    }
  }
  
  protected Queue<Runnable> newTaskQueue(int maxPendingTasks)
  {
    return PlatformDependent.newMpscQueue(maxPendingTasks);
  }
  
  public int pendingTasks()
  {
    if (inEventLoop()) {
      return super.pendingTasks();
    }
    return ((Integer)submit(pendingTasksCallable).syncUninterruptibly().getNow()).intValue();
  }
  
  public int getIoRatio()
  {
    return ioRatio;
  }
  
  public void setIoRatio(int ioRatio)
  {
    if ((ioRatio <= 0) || (ioRatio > 100)) {
      throw new IllegalArgumentException("ioRatio: " + ioRatio + " (expected: 0 < ioRatio <= 100)");
    }
    this.ioRatio = ioRatio;
  }
  
  private int epollWait(boolean oldWakenUp)
    throws IOException
  {
    int selectCnt = 0;
    long currentTimeNanos = System.nanoTime();
    long selectDeadLineNanos = currentTimeNanos + delayNanos(currentTimeNanos);
    for (;;)
    {
      long timeoutMillis = (selectDeadLineNanos - currentTimeNanos + 500000L) / 1000000L;
      if (timeoutMillis <= 0L)
      {
        if (selectCnt != 0) {
          break;
        }
        int ready = Native.epollWait(epollFd.intValue(), events, 0);
        if (ready > 0) {
          return ready;
        }
        break;
      }
      if ((hasTasks()) && (WAKEN_UP_UPDATER.compareAndSet(this, 0, 1))) {
        return Native.epollWait(epollFd.intValue(), events, 0);
      }
      int selectedKeys = Native.epollWait(epollFd.intValue(), events, (int)timeoutMillis);
      selectCnt++;
      if ((selectedKeys != 0) || (oldWakenUp) || (wakenUp == 1) || (hasTasks()) || (hasScheduledTasks())) {
        return selectedKeys;
      }
      currentTimeNanos = System.nanoTime();
    }
    return 0;
  }
  
  protected void run()
  {
    try
    {
      for (;;)
      {
        int strategy = selectStrategy.calculateStrategy(selectNowSupplier, hasTasks());
        switch (strategy)
        {
        case -2: 
          break;
        case -1: 
          strategy = epollWait(WAKEN_UP_UPDATER.getAndSet(this, 0) == 1);
          if (wakenUp == 1) {
            Native.eventFdWrite(eventFd.intValue(), 1L);
          }
        default: 
          int ioRatio = this.ioRatio;
          if (ioRatio == 100)
          {
            if (strategy > 0) {
              processReady(events, strategy);
            }
            runAllTasks();
          }
          else
          {
            long ioStartTime = System.nanoTime();
            if (strategy > 0) {
              processReady(events, strategy);
            }
            long ioTime = System.nanoTime() - ioStartTime;
            runAllTasks(ioTime * (100 - ioRatio) / ioRatio);
          }
          if ((allowGrowing) && (strategy == events.length())) {
            events.increase();
          }
          if (isShuttingDown())
          {
            closeAll();
            if (confirmShutdown()) {
              return;
            }
          }
        }
      }
    }
    catch (Throwable t)
    {
      logger.warn("Unexpected exception in the selector loop.", t);
      try
      {
        Thread.sleep(1000L);
      }
      catch (InterruptedException localInterruptedException) {}
    }
  }
  
  private void closeAll()
  {
    try
    {
      Native.epollWait(epollFd.intValue(), events, 0);
    }
    catch (IOException localIOException) {}
    Collection<AbstractEpollChannel> array = new ArrayList(channels.size());
    for (AbstractEpollChannel channel : channels.values()) {
      array.add(channel);
    }
    for (AbstractEpollChannel ch : array) {
      ch.unsafe().close(ch.unsafe().voidPromise());
    }
  }
  
  private void processReady(EpollEventArray events, int ready)
  {
    for (int i = 0; i < ready; i++)
    {
      int fd = events.fd(i);
      if (fd == eventFd.intValue())
      {
        Native.eventFdRead(eventFd.intValue());
      }
      else
      {
        long ev = events.events(i);
        
        AbstractEpollChannel ch = (AbstractEpollChannel)channels.get(fd);
        if (ch != null)
        {
          AbstractEpollChannel.AbstractEpollUnsafe unsafe = (AbstractEpollChannel.AbstractEpollUnsafe)ch.unsafe();
          if ((ev & (Native.EPOLLERR | Native.EPOLLOUT)) != 0L) {
            unsafe.epollOutReady();
          }
          if ((ev & (Native.EPOLLERR | Native.EPOLLIN)) != 0L) {
            unsafe.epollInReady();
          }
          if ((ev & Native.EPOLLRDHUP) != 0L) {
            unsafe.epollRdHupReady();
          }
        }
        else
        {
          try
          {
            Native.epollCtlDel(epollFd.intValue(), fd);
          }
          catch (IOException localIOException) {}
        }
      }
    }
  }
  
  protected void cleanup()
  {
    try
    {
      try
      {
        epollFd.close();
      }
      catch (IOException e)
      {
        logger.warn("Failed to close the epoll fd.", e);
      }
      try
      {
        eventFd.close();
      }
      catch (IOException e)
      {
        logger.warn("Failed to close the event fd.", e);
      }
    }
    finally
    {
      iovArray.release();
      events.free();
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.channel.epoll.EpollEventLoop
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */