package io.netty.channel.epoll;

import io.netty.channel.Channel;
import io.netty.channel.ChannelException;
import io.netty.channel.socket.ServerSocketChannel;
import io.netty.channel.socket.SocketChannel;
import io.netty.channel.unix.FileDescriptor;
import io.netty.channel.unix.Socket;
import io.netty.util.concurrent.GlobalEventExecutor;
import io.netty.util.internal.PlatformDependent;
import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.net.UnknownHostException;
import java.nio.channels.AlreadyConnectedException;
import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.concurrent.Executor;

public final class EpollSocketChannel
  extends AbstractEpollStreamChannel
  implements SocketChannel
{
  private final EpollSocketChannelConfig config;
  private volatile InetSocketAddress local;
  private volatile InetSocketAddress remote;
  private InetSocketAddress requestedRemote;
  private volatile Collection<InetAddress> tcpMd5SigAddresses = Collections.emptyList();
  
  EpollSocketChannel(Channel parent, Socket fd, InetSocketAddress remote)
  {
    super(parent, fd);
    config = new EpollSocketChannelConfig(this);
    
    this.remote = remote;
    local = fd.localAddress();
    if ((parent instanceof EpollServerSocketChannel)) {
      tcpMd5SigAddresses = ((EpollServerSocketChannel)parent).tcpMd5SigAddresses();
    }
  }
  
  public EpollSocketChannel()
  {
    super(Socket.newSocketStream(), false);
    config = new EpollSocketChannelConfig(this);
  }
  
  @Deprecated
  public EpollSocketChannel(FileDescriptor fd)
  {
    super(fd);
    
    remote = fd().remoteAddress();
    local = fd().localAddress();
    config = new EpollSocketChannelConfig(this);
  }
  
  public EpollSocketChannel(Socket fd, boolean active)
  {
    super(fd, active);
    
    remote = fd.remoteAddress();
    local = fd.localAddress();
    config = new EpollSocketChannelConfig(this);
  }
  
  public EpollTcpInfo tcpInfo()
  {
    return tcpInfo(new EpollTcpInfo());
  }
  
  public EpollTcpInfo tcpInfo(EpollTcpInfo info)
  {
    try
    {
      Native.tcpInfo(fd().intValue(), info);
      return info;
    }
    catch (IOException e)
    {
      throw new ChannelException(e);
    }
  }
  
  public InetSocketAddress remoteAddress()
  {
    return (InetSocketAddress)super.remoteAddress();
  }
  
  public InetSocketAddress localAddress()
  {
    return (InetSocketAddress)super.localAddress();
  }
  
  protected SocketAddress localAddress0()
  {
    return local;
  }
  
  protected SocketAddress remoteAddress0()
  {
    return remote;
  }
  
  protected void doBind(SocketAddress local)
    throws Exception
  {
    InetSocketAddress localAddress = (InetSocketAddress)local;
    fd().bind(localAddress);
    this.local = fd().localAddress();
  }
  
  public EpollSocketChannelConfig config()
  {
    return config;
  }
  
  public ServerSocketChannel parent()
  {
    return (ServerSocketChannel)super.parent();
  }
  
  protected AbstractEpollChannel.AbstractEpollUnsafe newUnsafe()
  {
    return new EpollSocketChannelUnsafe(null);
  }
  
  private static InetSocketAddress computeRemoteAddr(InetSocketAddress remoteAddr, InetSocketAddress osRemoteAddr)
  {
    if (osRemoteAddr != null)
    {
      if (PlatformDependent.javaVersion() >= 7) {
        try
        {
          return new InetSocketAddress(InetAddress.getByAddress(remoteAddr.getHostString(), osRemoteAddr.getAddress().getAddress()), osRemoteAddr.getPort());
        }
        catch (UnknownHostException localUnknownHostException) {}
      }
      return osRemoteAddr;
    }
    return remoteAddr;
  }
  
  protected boolean doConnect(SocketAddress remoteAddress, SocketAddress localAddress)
    throws Exception
  {
    if (localAddress != null) {
      checkResolvable((InetSocketAddress)localAddress);
    }
    InetSocketAddress remoteAddr = (InetSocketAddress)remoteAddress;
    checkResolvable(remoteAddr);
    if (remote != null) {
      throw new AlreadyConnectedException();
    }
    boolean connected = super.doConnect(remoteAddress, localAddress);
    if (connected) {
      remote = computeRemoteAddr(remoteAddr, fd().remoteAddress());
    } else {
      requestedRemote = remoteAddr;
    }
    local = fd().localAddress();
    return connected;
  }
  
  private final class EpollSocketChannelUnsafe
    extends AbstractEpollStreamChannel.EpollStreamUnsafe
  {
    private EpollSocketChannelUnsafe()
    {
      super();
    }
    
    protected Executor prepareToClose()
    {
      try
      {
        if ((isOpen()) && (config().getSoLinger() > 0))
        {
          ((EpollEventLoop)eventLoop()).remove(EpollSocketChannel.this);
          return GlobalEventExecutor.INSTANCE;
        }
      }
      catch (Throwable localThrowable) {}
      return null;
    }
    
    boolean doFinishConnect()
      throws Exception
    {
      if (super.doFinishConnect())
      {
        remote = EpollSocketChannel.computeRemoteAddr(requestedRemote, fd().remoteAddress());
        requestedRemote = null;
        return true;
      }
      return false;
    }
  }
  
  void setTcpMd5Sig(Map<InetAddress, byte[]> keys)
    throws IOException
  {
    tcpMd5SigAddresses = TcpMd5Util.newTcpMd5Sigs(this, tcpMd5SigAddresses, keys);
  }
}

/* Location:
 * Qualified Name:     io.netty.channel.epoll.EpollSocketChannel
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */