package io.netty.channel.epoll;

import io.netty.channel.DefaultFileRegion;
import io.netty.channel.unix.Errors;
import io.netty.channel.unix.Errors.NativeIoException;
import io.netty.channel.unix.FileDescriptor;
import io.netty.channel.unix.NativeInetAddress;
import io.netty.util.internal.NativeLibraryLoader;
import io.netty.util.internal.PlatformDependent;
import io.netty.util.internal.SystemPropertyUtil;
import io.netty.util.internal.ThrowableUtil;
import java.io.IOException;
import java.net.InetAddress;
import java.nio.channels.ClosedChannelException;
import java.util.Locale;

public final class Native
{
  public static final int EPOLLIN;
  public static final int EPOLLOUT;
  public static final int EPOLLRDHUP;
  public static final int EPOLLET;
  public static final int EPOLLERR;
  public static final int IOV_MAX;
  public static final int UIO_MAX_IOV;
  public static final boolean IS_SUPPORTING_SENDMMSG;
  public static final boolean IS_SUPPORTING_TCP_FASTOPEN;
  public static final long SSIZE_MAX;
  public static final int TCP_MD5SIG_MAXKEYLEN;
  public static final String KERNEL_VERSION;
  
  static
  {
    try
    {
      offsetofEpollData();
    }
    catch (UnsatisfiedLinkError ignore)
    {
      loadNativeLibrary();
    }
    EPOLLIN = NativeStaticallyReferencedJniMethods.epollin();
    EPOLLOUT = NativeStaticallyReferencedJniMethods.epollout();
    EPOLLRDHUP = NativeStaticallyReferencedJniMethods.epollrdhup();
    EPOLLET = NativeStaticallyReferencedJniMethods.epollet();
    EPOLLERR = NativeStaticallyReferencedJniMethods.epollerr();
    
    IOV_MAX = NativeStaticallyReferencedJniMethods.iovMax();
    UIO_MAX_IOV = NativeStaticallyReferencedJniMethods.uioMaxIov();
    IS_SUPPORTING_SENDMMSG = NativeStaticallyReferencedJniMethods.isSupportingSendmmsg();
    IS_SUPPORTING_TCP_FASTOPEN = NativeStaticallyReferencedJniMethods.isSupportingTcpFastopen();
    SSIZE_MAX = NativeStaticallyReferencedJniMethods.ssizeMax();
    TCP_MD5SIG_MAXKEYLEN = NativeStaticallyReferencedJniMethods.tcpMd5SigMaxKeyLen();
    KERNEL_VERSION = NativeStaticallyReferencedJniMethods.kernelVersion();
    
    SENDFILE_CLOSED_CHANNEL_EXCEPTION = (ClosedChannelException)ThrowableUtil.unknownStackTrace(new ClosedChannelException(), Native.class, "sendfile(...)");
    
    SENDMMSG_CLOSED_CHANNEL_EXCEPTION = (ClosedChannelException)ThrowableUtil.unknownStackTrace(new ClosedChannelException(), Native.class, "sendmmsg(...)");
    
    SPLICE_CLOSED_CHANNEL_EXCEPTION = (ClosedChannelException)ThrowableUtil.unknownStackTrace(new ClosedChannelException(), Native.class, "splice(...)");
  }
  
  private static final Errors.NativeIoException SENDFILE_CONNECTION_RESET_EXCEPTION = Errors.newConnectionResetException("syscall:sendfile(...)", Errors.ERRNO_EPIPE_NEGATIVE);
  private static final Errors.NativeIoException SENDMMSG_CONNECTION_RESET_EXCEPTION = Errors.newConnectionResetException("syscall:sendmmsg(...)", Errors.ERRNO_EPIPE_NEGATIVE);
  private static final Errors.NativeIoException SPLICE_CONNECTION_RESET_EXCEPTION = Errors.newConnectionResetException("syscall:splice(...)", Errors.ERRNO_EPIPE_NEGATIVE);
  private static final ClosedChannelException SENDFILE_CLOSED_CHANNEL_EXCEPTION;
  private static final ClosedChannelException SENDMMSG_CLOSED_CHANNEL_EXCEPTION;
  private static final ClosedChannelException SPLICE_CLOSED_CHANNEL_EXCEPTION;
  
  public static FileDescriptor newEventFd()
  {
    return new FileDescriptor(eventFd());
  }
  
  public static FileDescriptor newEpollCreate()
  {
    return new FileDescriptor(epollCreate());
  }
  
  public static int epollWait(int efd, EpollEventArray events, int timeout)
    throws IOException
  {
    int ready = epollWait0(efd, events.memoryAddress(), events.length(), timeout);
    if (ready < 0) {
      throw Errors.newIOException("epoll_wait", ready);
    }
    return ready;
  }
  
  public static void epollCtlAdd(int efd, int fd, int flags)
    throws IOException
  {
    int res = epollCtlAdd0(efd, fd, flags);
    if (res < 0) {
      throw Errors.newIOException("epoll_ctl", res);
    }
  }
  
  public static void epollCtlMod(int efd, int fd, int flags)
    throws IOException
  {
    int res = epollCtlMod0(efd, fd, flags);
    if (res < 0) {
      throw Errors.newIOException("epoll_ctl", res);
    }
  }
  
  public static void epollCtlDel(int efd, int fd)
    throws IOException
  {
    int res = epollCtlDel0(efd, fd);
    if (res < 0) {
      throw Errors.newIOException("epoll_ctl", res);
    }
  }
  
  public static int splice(int fd, long offIn, int fdOut, long offOut, long len)
    throws IOException
  {
    int res = splice0(fd, offIn, fdOut, offOut, len);
    if (res >= 0) {
      return res;
    }
    return Errors.ioResult("splice", res, SPLICE_CONNECTION_RESET_EXCEPTION, SPLICE_CLOSED_CHANNEL_EXCEPTION);
  }
  
  public static long sendfile(int dest, DefaultFileRegion src, long baseOffset, long offset, long length)
    throws IOException
  {
    src.open();
    
    long res = sendfile0(dest, src, baseOffset, offset, length);
    if (res >= 0L) {
      return res;
    }
    return Errors.ioResult("sendfile", (int)res, SENDFILE_CONNECTION_RESET_EXCEPTION, SENDFILE_CLOSED_CHANNEL_EXCEPTION);
  }
  
  public static int sendmmsg(int fd, NativeDatagramPacketArray.NativeDatagramPacket[] msgs, int offset, int len)
    throws IOException
  {
    int res = sendmmsg0(fd, msgs, offset, len);
    if (res >= 0) {
      return res;
    }
    return Errors.ioResult("sendmmsg", res, SENDMMSG_CONNECTION_RESET_EXCEPTION, SENDMMSG_CLOSED_CHANNEL_EXCEPTION);
  }
  
  public static int recvFd(int fd)
    throws IOException
  {
    int res = recvFd0(fd);
    if (res > 0) {
      return res;
    }
    if (res == 0) {
      return -1;
    }
    if ((res == Errors.ERRNO_EAGAIN_NEGATIVE) || (res == Errors.ERRNO_EWOULDBLOCK_NEGATIVE)) {
      return 0;
    }
    throw Errors.newIOException("recvFd", res);
  }
  
  public static int sendFd(int socketFd, int fd)
    throws IOException
  {
    int res = sendFd0(socketFd, fd);
    if (res >= 0) {
      return res;
    }
    if ((res == Errors.ERRNO_EAGAIN_NEGATIVE) || (res == Errors.ERRNO_EWOULDBLOCK_NEGATIVE)) {
      return -1;
    }
    throw Errors.newIOException("sendFd", res);
  }
  
  public static void tcpInfo(int fd, EpollTcpInfo info)
    throws IOException
  {
    tcpInfo0(fd, info);
  }
  
  public static void setTcpMd5Sig(int fd, InetAddress address, byte[] key)
    throws IOException
  {
    NativeInetAddress a = NativeInetAddress.newInstance(address);
    setTcpMd5Sig0(fd, a.address(), a.scopeId(), key);
  }
  
  private static void loadNativeLibrary()
  {
    String name = SystemPropertyUtil.get("os.name").toLowerCase(Locale.UK).trim();
    if (!name.startsWith("linux")) {
      throw new IllegalStateException("Only supported on Linux");
    }
    NativeLibraryLoader.load(SystemPropertyUtil.get("io.netty.packagePrefix", "").replace('.', '-') + "netty-transport-native-epoll", PlatformDependent.getClassLoader(Native.class));
  }
  
  private static native int eventFd();
  
  public static native void eventFdWrite(int paramInt, long paramLong);
  
  public static native void eventFdRead(int paramInt);
  
  private static native int epollCreate();
  
  private static native int epollWait0(int paramInt1, long paramLong, int paramInt2, int paramInt3);
  
  private static native int epollCtlAdd0(int paramInt1, int paramInt2, int paramInt3);
  
  private static native int epollCtlMod0(int paramInt1, int paramInt2, int paramInt3);
  
  private static native int epollCtlDel0(int paramInt1, int paramInt2);
  
  private static native int splice0(int paramInt1, long paramLong1, int paramInt2, long paramLong2, long paramLong3);
  
  private static native long sendfile0(int paramInt, DefaultFileRegion paramDefaultFileRegion, long paramLong1, long paramLong2, long paramLong3)
    throws IOException;
  
  private static native int sendmmsg0(int paramInt1, NativeDatagramPacketArray.NativeDatagramPacket[] paramArrayOfNativeDatagramPacket, int paramInt2, int paramInt3);
  
  private static native int recvFd0(int paramInt);
  
  private static native int sendFd0(int paramInt1, int paramInt2);
  
  public static native int isReuseAddress(int paramInt)
    throws IOException;
  
  public static native int isReusePort(int paramInt)
    throws IOException;
  
  public static native int getTcpNotSentLowAt(int paramInt)
    throws IOException;
  
  public static native int getTrafficClass(int paramInt)
    throws IOException;
  
  public static native int isBroadcast(int paramInt)
    throws IOException;
  
  public static native int getTcpKeepIdle(int paramInt)
    throws IOException;
  
  public static native int getTcpKeepIntvl(int paramInt)
    throws IOException;
  
  public static native int getTcpKeepCnt(int paramInt)
    throws IOException;
  
  public static native int getTcpUserTimeout(int paramInt)
    throws IOException;
  
  public static native int isIpFreeBind(int paramInt)
    throws IOException;
  
  public static native void setReuseAddress(int paramInt1, int paramInt2)
    throws IOException;
  
  public static native void setReusePort(int paramInt1, int paramInt2)
    throws IOException;
  
  public static native void setTcpFastopen(int paramInt1, int paramInt2)
    throws IOException;
  
  public static native void setTcpNotSentLowAt(int paramInt1, int paramInt2)
    throws IOException;
  
  public static native void setTrafficClass(int paramInt1, int paramInt2)
    throws IOException;
  
  public static native void setBroadcast(int paramInt1, int paramInt2)
    throws IOException;
  
  public static native void setTcpKeepIdle(int paramInt1, int paramInt2)
    throws IOException;
  
  public static native void setTcpKeepIntvl(int paramInt1, int paramInt2)
    throws IOException;
  
  public static native void setTcpKeepCnt(int paramInt1, int paramInt2)
    throws IOException;
  
  public static native void setTcpUserTimeout(int paramInt1, int paramInt2)
    throws IOException;
  
  public static native void setIpFreeBind(int paramInt1, int paramInt2)
    throws IOException;
  
  private static native void tcpInfo0(int paramInt, int[] paramArrayOfInt)
    throws IOException;
  
  private static native void setTcpMd5Sig0(int paramInt1, byte[] paramArrayOfByte1, int paramInt2, byte[] paramArrayOfByte2)
    throws IOException;
  
  public static native int sizeofEpollEvent();
  
  public static native int offsetofEpollData();
}

/* Location:
 * Qualified Name:     io.netty.channel.epoll.Native
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */