package io.netty.channel.local;

import io.netty.channel.AbstractChannel;
import io.netty.channel.AbstractChannel.AbstractUnsafe;
import io.netty.channel.Channel;
import io.netty.channel.Channel.Unsafe;
import io.netty.channel.ChannelConfig;
import io.netty.channel.ChannelMetadata;
import io.netty.channel.ChannelOutboundBuffer;
import io.netty.channel.ChannelPipeline;
import io.netty.channel.ChannelPromise;
import io.netty.channel.DefaultChannelConfig;
import io.netty.channel.EventLoop;
import io.netty.channel.SingleThreadEventLoop;
import io.netty.util.ReferenceCountUtil;
import io.netty.util.concurrent.Future;
import io.netty.util.concurrent.SingleThreadEventExecutor;
import io.netty.util.internal.InternalThreadLocalMap;
import io.netty.util.internal.PlatformDependent;
import io.netty.util.internal.ThrowableUtil;
import java.net.ConnectException;
import java.net.SocketAddress;
import java.nio.channels.AlreadyConnectedException;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.ConnectionPendingException;
import java.nio.channels.NotYetConnectedException;
import java.util.Queue;
import java.util.concurrent.atomic.AtomicReferenceFieldUpdater;

public class LocalChannel
  extends AbstractChannel
{
  private static final AtomicReferenceFieldUpdater<LocalChannel, Future> FINISH_READ_FUTURE_UPDATER;
  
  private static enum State
  {
    OPEN,  BOUND,  CONNECTED,  CLOSED;
    
    private State() {}
  }
  
  private static final ChannelMetadata METADATA = new ChannelMetadata(false);
  private static final int MAX_READER_STACK_DEPTH = 8;
  private static final ClosedChannelException DO_WRITE_CLOSED_CHANNEL_EXCEPTION = (ClosedChannelException)ThrowableUtil.unknownStackTrace(new ClosedChannelException(), LocalChannel.class, "doWrite(...)");
  private static final ClosedChannelException DO_CLOSE_CLOSED_CHANNEL_EXCEPTION = (ClosedChannelException)ThrowableUtil.unknownStackTrace(new ClosedChannelException(), LocalChannel.class, "doClose()");
  private final ChannelConfig config = new DefaultChannelConfig(this);
  private final Queue<Object> inboundBuffer = PlatformDependent.newSpscQueue();
  private final Runnable readTask = new Runnable()
  {
    public void run()
    {
      ChannelPipeline pipeline = pipeline();
      for (;;)
      {
        Object m = inboundBuffer.poll();
        if (m == null) {
          break;
        }
        pipeline.fireChannelRead(m);
      }
      pipeline.fireChannelReadComplete();
    }
  };
  private final Runnable shutdownHook = new Runnable()
  {
    public void run()
    {
      unsafe().close(unsafe().voidPromise());
    }
  };
  private volatile State state;
  private volatile LocalChannel peer;
  private volatile LocalAddress localAddress;
  private volatile LocalAddress remoteAddress;
  private volatile ChannelPromise connectPromise;
  private volatile boolean readInProgress;
  private volatile boolean registerInProgress;
  private volatile boolean writeInProgress;
  private volatile Future<?> finishReadFuture;
  
  static
  {
    AtomicReferenceFieldUpdater<LocalChannel, Future> finishReadFutureUpdater = PlatformDependent.newAtomicReferenceFieldUpdater(LocalChannel.class, "finishReadFuture");
    if (finishReadFutureUpdater == null) {
      finishReadFutureUpdater = AtomicReferenceFieldUpdater.newUpdater(LocalChannel.class, Future.class, "finishReadFuture");
    }
    FINISH_READ_FUTURE_UPDATER = finishReadFutureUpdater;
  }
  
  public LocalChannel()
  {
    super(null);
  }
  
  LocalChannel(LocalServerChannel parent, LocalChannel peer)
  {
    super(parent);
    this.peer = peer;
    localAddress = parent.localAddress();
    remoteAddress = peer.localAddress();
  }
  
  public ChannelMetadata metadata()
  {
    return METADATA;
  }
  
  public ChannelConfig config()
  {
    return config;
  }
  
  public LocalServerChannel parent()
  {
    return (LocalServerChannel)super.parent();
  }
  
  public LocalAddress localAddress()
  {
    return (LocalAddress)super.localAddress();
  }
  
  public LocalAddress remoteAddress()
  {
    return (LocalAddress)super.remoteAddress();
  }
  
  public boolean isOpen()
  {
    return state != State.CLOSED;
  }
  
  public boolean isActive()
  {
    return state == State.CONNECTED;
  }
  
  protected AbstractChannel.AbstractUnsafe newUnsafe()
  {
    return new LocalUnsafe(null);
  }
  
  protected boolean isCompatible(EventLoop loop)
  {
    return loop instanceof SingleThreadEventLoop;
  }
  
  protected SocketAddress localAddress0()
  {
    return localAddress;
  }
  
  protected SocketAddress remoteAddress0()
  {
    return remoteAddress;
  }
  
  protected void doRegister()
    throws Exception
  {
    if ((this.peer != null) && (parent() != null))
    {
      final LocalChannel peer = this.peer;
      registerInProgress = true;
      state = State.CONNECTED;
      
      remoteAddress = (parent() == null ? null : parent().localAddress());
      state = State.CONNECTED;
      
      peer.eventLoop().execute(new Runnable()
      {
        public void run()
        {
          registerInProgress = false;
          ChannelPromise promise = peerconnectPromise;
          if ((promise != null) && (promise.trySuccess())) {
            peer.pipeline().fireChannelActive();
          }
        }
      });
    }
    ((SingleThreadEventExecutor)eventLoop()).addShutdownHook(shutdownHook);
  }
  
  protected void doBind(SocketAddress localAddress)
    throws Exception
  {
    this.localAddress = LocalChannelRegistry.register(this, this.localAddress, localAddress);
    
    state = State.BOUND;
  }
  
  protected void doDisconnect()
    throws Exception
  {
    doClose();
  }
  
  protected void doClose()
    throws Exception
  {
    final LocalChannel peer = this.peer;
    if (state != State.CLOSED)
    {
      if (localAddress != null)
      {
        if (parent() == null) {
          LocalChannelRegistry.unregister(localAddress);
        }
        localAddress = null;
      }
      state = State.CLOSED;
      
      ChannelPromise promise = connectPromise;
      if (promise != null)
      {
        promise.tryFailure(DO_CLOSE_CLOSED_CHANNEL_EXCEPTION);
        connectPromise = null;
      }
      if ((writeInProgress) && (peer != null)) {
        finishPeerRead(peer);
      }
    }
    if ((peer != null) && (peer.isActive()))
    {
      if ((peer.eventLoop().inEventLoop()) && (!registerInProgress))
      {
        doPeerClose(peer, writeInProgress);
      }
      else
      {
        final boolean peerWriteInProgress = writeInProgress;
        try
        {
          peer.eventLoop().execute(new Runnable()
          {
            public void run()
            {
              LocalChannel.this.doPeerClose(peer, peerWriteInProgress);
            }
          });
        }
        catch (RuntimeException e)
        {
          releaseInboundBuffers();
          throw e;
        }
      }
      this.peer = null;
    }
  }
  
  private void doPeerClose(LocalChannel peer, boolean peerWriteInProgress)
  {
    if (peerWriteInProgress) {
      finishPeerRead0(this);
    }
    peer.unsafe().close(peer.unsafe().voidPromise());
  }
  
  protected void doDeregister()
    throws Exception
  {
    ((SingleThreadEventExecutor)eventLoop()).removeShutdownHook(shutdownHook);
  }
  
  protected void doBeginRead()
    throws Exception
  {
    if (readInProgress) {
      return;
    }
    ChannelPipeline pipeline = pipeline();
    Queue<Object> inboundBuffer = this.inboundBuffer;
    if (inboundBuffer.isEmpty())
    {
      readInProgress = true;
      return;
    }
    InternalThreadLocalMap threadLocals = InternalThreadLocalMap.get();
    Integer stackDepth = Integer.valueOf(threadLocals.localChannelReaderStackDepth());
    if (stackDepth.intValue() < 8)
    {
      threadLocals.setLocalChannelReaderStackDepth(stackDepth.intValue() + 1);
      try
      {
        for (;;)
        {
          Object received = inboundBuffer.poll();
          if (received == null) {
            break;
          }
          pipeline.fireChannelRead(received);
        }
        pipeline.fireChannelReadComplete();
      }
      finally
      {
        threadLocals.setLocalChannelReaderStackDepth(stackDepth.intValue());
      }
    }
    else
    {
      try
      {
        eventLoop().execute(readTask);
      }
      catch (RuntimeException e)
      {
        releaseInboundBuffers();
        throw e;
      }
    }
  }
  
  protected void doWrite(ChannelOutboundBuffer in)
    throws Exception
  {
    switch (state)
    {
    case OPEN: 
    case BOUND: 
      throw new NotYetConnectedException();
    case CLOSED: 
      throw DO_WRITE_CLOSED_CHANNEL_EXCEPTION;
    }
    LocalChannel peer = this.peer;
    
    writeInProgress = true;
    try
    {
      for (;;)
      {
        Object msg = in.current();
        if (msg == null) {
          break;
        }
        try
        {
          if (state == State.CONNECTED)
          {
            inboundBuffer.add(ReferenceCountUtil.retain(msg));
            in.remove();
          }
          else
          {
            in.remove(DO_WRITE_CLOSED_CHANNEL_EXCEPTION);
          }
        }
        catch (Throwable cause)
        {
          in.remove(cause);
        }
      }
    }
    finally
    {
      writeInProgress = false;
    }
    finishPeerRead(peer);
  }
  
  private void finishPeerRead(LocalChannel peer)
  {
    if ((peer.eventLoop() == eventLoop()) && (!writeInProgress)) {
      finishPeerRead0(peer);
    } else {
      runFinishPeerReadTask(peer);
    }
  }
  
  private void runFinishPeerReadTask(final LocalChannel peer)
  {
    Runnable finishPeerReadTask = new Runnable()
    {
      public void run()
      {
        LocalChannel.this.finishPeerRead0(peer);
      }
    };
    try
    {
      if (writeInProgress) {
        finishReadFuture = peer.eventLoop().submit(finishPeerReadTask);
      } else {
        peer.eventLoop().execute(finishPeerReadTask);
      }
    }
    catch (RuntimeException e)
    {
      peer.releaseInboundBuffers();
      throw e;
    }
  }
  
  private void releaseInboundBuffers()
  {
    for (;;)
    {
      Object o = inboundBuffer.poll();
      if (o == null) {
        break;
      }
      ReferenceCountUtil.release(o);
    }
  }
  
  private void finishPeerRead0(LocalChannel peer)
  {
    Future<?> peerFinishReadFuture = finishReadFuture;
    if (peerFinishReadFuture != null)
    {
      if (!peerFinishReadFuture.isDone())
      {
        runFinishPeerReadTask(peer);
        return;
      }
      FINISH_READ_FUTURE_UPDATER.compareAndSet(peer, peerFinishReadFuture, null);
    }
    ChannelPipeline peerPipeline = peer.pipeline();
    if (readInProgress)
    {
      readInProgress = false;
      for (;;)
      {
        Object received = inboundBuffer.poll();
        if (received == null) {
          break;
        }
        peerPipeline.fireChannelRead(received);
      }
      peerPipeline.fireChannelReadComplete();
    }
  }
  
  private class LocalUnsafe
    extends AbstractChannel.AbstractUnsafe
  {
    private LocalUnsafe()
    {
      super();
    }
    
    public void connect(SocketAddress remoteAddress, SocketAddress localAddress, ChannelPromise promise)
    {
      if ((!promise.setUncancellable()) || (!ensureOpen(promise))) {
        return;
      }
      if (state == LocalChannel.State.CONNECTED)
      {
        Exception cause = new AlreadyConnectedException();
        safeSetFailure(promise, cause);
        pipeline().fireExceptionCaught(cause);
        return;
      }
      if (connectPromise != null) {
        throw new ConnectionPendingException();
      }
      connectPromise = promise;
      if (state != LocalChannel.State.BOUND) {
        if (localAddress == null) {
          localAddress = new LocalAddress(LocalChannel.this);
        }
      }
      if (localAddress != null) {
        try
        {
          doBind(localAddress);
        }
        catch (Throwable t)
        {
          safeSetFailure(promise, t);
          close(voidPromise());
          return;
        }
      }
      Channel boundChannel = LocalChannelRegistry.get(remoteAddress);
      if (!(boundChannel instanceof LocalServerChannel))
      {
        Exception cause = new ConnectException("connection refused: " + remoteAddress);
        safeSetFailure(promise, cause);
        close(voidPromise());
        return;
      }
      LocalServerChannel serverChannel = (LocalServerChannel)boundChannel;
      peer = serverChannel.serve(LocalChannel.this);
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.channel.local.LocalChannel
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */