package io.netty.channel.nio;

import io.netty.buffer.ByteBuf;
import io.netty.buffer.ByteBufAllocator;
import io.netty.channel.Channel;
import io.netty.channel.ChannelConfig;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelMetadata;
import io.netty.channel.ChannelOption;
import io.netty.channel.ChannelOutboundBuffer;
import io.netty.channel.ChannelPipeline;
import io.netty.channel.FileRegion;
import io.netty.channel.RecvByteBufAllocator.Handle;
import io.netty.channel.socket.ChannelInputShutdownEvent;
import io.netty.util.internal.StringUtil;
import java.io.IOException;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;

public abstract class AbstractNioByteChannel
  extends AbstractNioChannel
{
  private static final ChannelMetadata METADATA = new ChannelMetadata(false, 16);
  private static final String EXPECTED_TYPES = " (expected: " + StringUtil.simpleClassName(ByteBuf.class) + ", " + StringUtil.simpleClassName(FileRegion.class) + ')';
  private Runnable flushTask;
  
  protected AbstractNioByteChannel(Channel parent, SelectableChannel ch)
  {
    super(parent, ch, 1);
  }
  
  protected abstract ChannelFuture shutdownInput();
  
  protected AbstractNioChannel.AbstractNioUnsafe newUnsafe()
  {
    return new NioByteUnsafe();
  }
  
  public ChannelMetadata metadata()
  {
    return METADATA;
  }
  
  protected class NioByteUnsafe
    extends AbstractNioChannel.AbstractNioUnsafe
  {
    protected NioByteUnsafe()
    {
      super();
    }
    
    private void closeOnRead(ChannelPipeline pipeline)
    {
      if (isOpen()) {
        if (Boolean.TRUE.equals(config().getOption(ChannelOption.ALLOW_HALF_CLOSURE)))
        {
          shutdownInput();
          SelectionKey key = selectionKey();
          key.interestOps(key.interestOps() & (readInterestOp ^ 0xFFFFFFFF));
          pipeline.fireUserEventTriggered(ChannelInputShutdownEvent.INSTANCE);
        }
        else
        {
          close(voidPromise());
        }
      }
    }
    
    private void handleReadException(ChannelPipeline pipeline, ByteBuf byteBuf, Throwable cause, boolean close, RecvByteBufAllocator.Handle allocHandle)
    {
      if (byteBuf != null) {
        if (byteBuf.isReadable())
        {
          readPending = false;
          pipeline.fireChannelRead(byteBuf);
        }
        else
        {
          byteBuf.release();
        }
      }
      allocHandle.readComplete();
      pipeline.fireChannelReadComplete();
      pipeline.fireExceptionCaught(cause);
      if ((close) || ((cause instanceof IOException))) {
        closeOnRead(pipeline);
      }
    }
    
    public final void read()
    {
      ChannelConfig config = config();
      ChannelPipeline pipeline = pipeline();
      ByteBufAllocator allocator = config.getAllocator();
      RecvByteBufAllocator.Handle allocHandle = recvBufAllocHandle();
      allocHandle.reset(config);
      
      ByteBuf byteBuf = null;
      boolean close = false;
      try
      {
        do
        {
          byteBuf = allocHandle.allocate(allocator);
          allocHandle.lastBytesRead(doReadBytes(byteBuf));
          if (allocHandle.lastBytesRead() <= 0)
          {
            byteBuf.release();
            byteBuf = null;
            close = allocHandle.lastBytesRead() < 0;
            break;
          }
          allocHandle.incMessagesRead(1);
          readPending = false;
          pipeline.fireChannelRead(byteBuf);
          byteBuf = null;
        } while (allocHandle.continueReading());
        allocHandle.readComplete();
        pipeline.fireChannelReadComplete();
        if (close) {
          closeOnRead(pipeline);
        }
      }
      catch (Throwable t)
      {
        handleReadException(pipeline, byteBuf, t, close, allocHandle);
      }
      finally
      {
        if ((!readPending) && (!config.isAutoRead())) {
          removeReadOp();
        }
      }
    }
  }
  
  protected void doWrite(ChannelOutboundBuffer in)
    throws Exception
  {
    int writeSpinCount = -1;
    
    boolean setOpWrite = false;
    for (;;)
    {
      Object msg = in.current();
      if (msg == null)
      {
        clearOpWrite();
        
        return;
      }
      if ((msg instanceof ByteBuf))
      {
        ByteBuf buf = (ByteBuf)msg;
        int readableBytes = buf.readableBytes();
        if (readableBytes == 0)
        {
          in.remove();
        }
        else
        {
          boolean done = false;
          long flushedAmount = 0L;
          if (writeSpinCount == -1) {
            writeSpinCount = config().getWriteSpinCount();
          }
          for (int i = writeSpinCount - 1; i >= 0; i--)
          {
            int localFlushedAmount = doWriteBytes(buf);
            if (localFlushedAmount == 0)
            {
              setOpWrite = true;
              break;
            }
            flushedAmount += localFlushedAmount;
            if (!buf.isReadable())
            {
              done = true;
              break;
            }
          }
          in.progress(flushedAmount);
          if (!done) {
            break;
          }
          in.remove();
        }
      }
      else if ((msg instanceof FileRegion))
      {
        FileRegion region = (FileRegion)msg;
        boolean done = region.transferred() >= region.count();
        if (!done)
        {
          long flushedAmount = 0L;
          if (writeSpinCount == -1) {
            writeSpinCount = config().getWriteSpinCount();
          }
          for (int i = writeSpinCount - 1; i >= 0; i--)
          {
            long localFlushedAmount = doWriteFileRegion(region);
            if (localFlushedAmount == 0L)
            {
              setOpWrite = true;
              break;
            }
            flushedAmount += localFlushedAmount;
            if (region.transferred() >= region.count())
            {
              done = true;
              break;
            }
          }
          in.progress(flushedAmount);
        }
        if (!done) {
          break;
        }
        in.remove();
      }
      else
      {
        throw new Error();
      }
    }
    incompleteWrite(setOpWrite);
  }
  
  protected final Object filterOutboundMessage(Object msg)
  {
    if ((msg instanceof ByteBuf))
    {
      ByteBuf buf = (ByteBuf)msg;
      if (buf.isDirect()) {
        return msg;
      }
      return newDirectBuffer(buf);
    }
    if ((msg instanceof FileRegion)) {
      return msg;
    }
    throw new UnsupportedOperationException("unsupported message type: " + StringUtil.simpleClassName(msg) + EXPECTED_TYPES);
  }
  
  protected final void incompleteWrite(boolean setOpWrite)
  {
    if (setOpWrite)
    {
      setOpWrite();
    }
    else
    {
      Runnable flushTask = this.flushTask;
      if (flushTask == null) {
        flushTask = this. = new Runnable()
        {
          public void run()
          {
            flush();
          }
        };
      }
      eventLoop().execute(flushTask);
    }
  }
  
  protected abstract long doWriteFileRegion(FileRegion paramFileRegion)
    throws Exception;
  
  protected abstract int doReadBytes(ByteBuf paramByteBuf)
    throws Exception;
  
  protected abstract int doWriteBytes(ByteBuf paramByteBuf)
    throws Exception;
  
  protected final void setOpWrite()
  {
    SelectionKey key = selectionKey();
    if (!key.isValid()) {
      return;
    }
    int interestOps = key.interestOps();
    if ((interestOps & 0x4) == 0) {
      key.interestOps(interestOps | 0x4);
    }
  }
  
  protected final void clearOpWrite()
  {
    SelectionKey key = selectionKey();
    if (!key.isValid()) {
      return;
    }
    int interestOps = key.interestOps();
    if ((interestOps & 0x4) != 0) {
      key.interestOps(interestOps & 0xFFFFFFFB);
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.channel.nio.AbstractNioByteChannel
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */