package io.netty.channel.nio;

import io.netty.channel.ChannelException;
import io.netty.channel.EventLoop;
import io.netty.channel.EventLoopException;
import io.netty.channel.SelectStrategy;
import io.netty.channel.SingleThreadEventLoop;
import io.netty.util.IntSupplier;
import io.netty.util.concurrent.Future;
import io.netty.util.concurrent.RejectedExecutionHandler;
import io.netty.util.internal.PlatformDependent;
import io.netty.util.internal.SystemPropertyUtil;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;
import java.io.IOException;
import java.lang.reflect.Field;
import java.nio.channels.CancelledKeyException;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.spi.SelectorProvider;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.Executor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

public final class NioEventLoop
  extends SingleThreadEventLoop
{
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(NioEventLoop.class);
  private static final int CLEANUP_INTERVAL = 256;
  private static final boolean DISABLE_KEYSET_OPTIMIZATION = SystemPropertyUtil.getBoolean("io.netty.noKeySetOptimization", false);
  private static final int MIN_PREMATURE_SELECTOR_RETURNS = 3;
  private static final int SELECTOR_AUTO_REBUILD_THRESHOLD;
  private final IntSupplier selectNowSupplier = new IntSupplier()
  {
    public int get()
      throws Exception
    {
      return selectNow();
    }
  };
  private final Callable<Integer> pendingTasksCallable = new Callable()
  {
    public Integer call()
      throws Exception
    {
      return Integer.valueOf(NioEventLoop.this.pendingTasks());
    }
  };
  Selector selector;
  private SelectedSelectionKeySet selectedKeys;
  private final SelectorProvider provider;
  
  static
  {
    String key = "sun.nio.ch.bugLevel";
    String buglevel = SystemPropertyUtil.get("sun.nio.ch.bugLevel");
    if (buglevel == null) {
      try
      {
        AccessController.doPrivileged(new PrivilegedAction()
        {
          public Void run()
          {
            System.setProperty("sun.nio.ch.bugLevel", "");
            return null;
          }
        });
      }
      catch (SecurityException e)
      {
        logger.debug("Unable to get/set System Property: sun.nio.ch.bugLevel", e);
      }
    }
    int selectorAutoRebuildThreshold = SystemPropertyUtil.getInt("io.netty.selectorAutoRebuildThreshold", 512);
    if (selectorAutoRebuildThreshold < 3) {
      selectorAutoRebuildThreshold = 0;
    }
    SELECTOR_AUTO_REBUILD_THRESHOLD = selectorAutoRebuildThreshold;
    if (logger.isDebugEnabled())
    {
      logger.debug("-Dio.netty.noKeySetOptimization: {}", Boolean.valueOf(DISABLE_KEYSET_OPTIMIZATION));
      logger.debug("-Dio.netty.selectorAutoRebuildThreshold: {}", Integer.valueOf(SELECTOR_AUTO_REBUILD_THRESHOLD));
    }
  }
  
  private final AtomicBoolean wakenUp = new AtomicBoolean();
  private final SelectStrategy selectStrategy;
  private volatile int ioRatio = 50;
  private int cancelledKeys;
  private boolean needsToSelectAgain;
  
  NioEventLoop(NioEventLoopGroup parent, Executor executor, SelectorProvider selectorProvider, SelectStrategy strategy, RejectedExecutionHandler rejectedExecutionHandler)
  {
    super(parent, executor, false, DEFAULT_MAX_PENDING_TASKS, rejectedExecutionHandler);
    if (selectorProvider == null) {
      throw new NullPointerException("selectorProvider");
    }
    if (strategy == null) {
      throw new NullPointerException("selectStrategy");
    }
    provider = selectorProvider;
    selector = openSelector();
    selectStrategy = strategy;
  }
  
  private Selector openSelector()
  {
    try
    {
      selector = provider.openSelector();
    }
    catch (IOException e)
    {
      Selector selector;
      throw new ChannelException("failed to open a new selector", e);
    }
    final Selector selector;
    if (DISABLE_KEYSET_OPTIMIZATION) {
      return selector;
    }
    final SelectedSelectionKeySet selectedKeySet = new SelectedSelectionKeySet();
    
    Object maybeSelectorImplClass = AccessController.doPrivileged(new PrivilegedAction()
    {
      public Object run()
      {
        try
        {
          return Class.forName("sun.nio.ch.SelectorImpl", false, PlatformDependent.getSystemClassLoader());
        }
        catch (ClassNotFoundException e)
        {
          return e;
        }
        catch (SecurityException e)
        {
          return e;
        }
      }
    });
    if ((!(maybeSelectorImplClass instanceof Class)) || (!((Class)maybeSelectorImplClass).isAssignableFrom(selector.getClass())))
    {
      if ((maybeSelectorImplClass instanceof Exception))
      {
        Exception e = (Exception)maybeSelectorImplClass;
        logger.trace("failed to instrument a special java.util.Set into: {}", selector, e);
      }
      return selector;
    }
    final Class<?> selectorImplClass = (Class)maybeSelectorImplClass;
    
    Object maybeException = AccessController.doPrivileged(new PrivilegedAction()
    {
      public Object run()
      {
        try
        {
          Field selectedKeysField = selectorImplClass.getDeclaredField("selectedKeys");
          Field publicSelectedKeysField = selectorImplClass.getDeclaredField("publicSelectedKeys");
          
          selectedKeysField.setAccessible(true);
          publicSelectedKeysField.setAccessible(true);
          
          selectedKeysField.set(selector, selectedKeySet);
          publicSelectedKeysField.set(selector, selectedKeySet);
          return null;
        }
        catch (NoSuchFieldException e)
        {
          return e;
        }
        catch (IllegalAccessException e)
        {
          return e;
        }
      }
    });
    if ((maybeException instanceof Exception))
    {
      selectedKeys = null;
      Exception e = (Exception)maybeException;
      logger.trace("failed to instrument a special java.util.Set into: {}", selector, e);
    }
    else
    {
      selectedKeys = selectedKeySet;
      logger.trace("instrumented a special java.util.Set into: {}", selector);
    }
    return selector;
  }
  
  public SelectorProvider selectorProvider()
  {
    return provider;
  }
  
  protected Queue<Runnable> newTaskQueue(int maxPendingTasks)
  {
    return PlatformDependent.newMpscQueue(maxPendingTasks);
  }
  
  public int pendingTasks()
  {
    if (inEventLoop()) {
      return super.pendingTasks();
    }
    return ((Integer)submit(pendingTasksCallable).syncUninterruptibly().getNow()).intValue();
  }
  
  public void register(SelectableChannel ch, int interestOps, NioTask<?> task)
  {
    if (ch == null) {
      throw new NullPointerException("ch");
    }
    if (interestOps == 0) {
      throw new IllegalArgumentException("interestOps must be non-zero.");
    }
    if ((interestOps & (ch.validOps() ^ 0xFFFFFFFF)) != 0) {
      throw new IllegalArgumentException("invalid interestOps: " + interestOps + "(validOps: " + ch.validOps() + ')');
    }
    if (task == null) {
      throw new NullPointerException("task");
    }
    if (isShutdown()) {
      throw new IllegalStateException("event loop shut down");
    }
    try
    {
      ch.register(selector, interestOps, task);
    }
    catch (Exception e)
    {
      throw new EventLoopException("failed to register a channel", e);
    }
  }
  
  public int getIoRatio()
  {
    return ioRatio;
  }
  
  public void setIoRatio(int ioRatio)
  {
    if ((ioRatio <= 0) || (ioRatio > 100)) {
      throw new IllegalArgumentException("ioRatio: " + ioRatio + " (expected: 0 < ioRatio <= 100)");
    }
    this.ioRatio = ioRatio;
  }
  
  public void rebuildSelector()
  {
    if (!inEventLoop())
    {
      execute(new Runnable()
      {
        public void run()
        {
          rebuildSelector();
        }
      });
      return;
    }
    Selector oldSelector = selector;
    if (oldSelector == null) {
      return;
    }
    try
    {
      newSelector = openSelector();
    }
    catch (Exception e)
    {
      Selector newSelector;
      logger.warn("Failed to create a new Selector.", e); return;
    }
    Selector newSelector;
    int nChannels = 0;
    for (;;)
    {
      try
      {
        Iterator i$ = oldSelector.keys().iterator();
        if (i$.hasNext())
        {
          SelectionKey key = (SelectionKey)i$.next();
          Object a = key.attachment();
          try
          {
            if ((key.isValid()) && (key.channel().keyFor(newSelector) != null)) {
              continue;
            }
            int interestOps = key.interestOps();
            key.cancel();
            SelectionKey newKey = key.channel().register(newSelector, interestOps, a);
            if ((a instanceof AbstractNioChannel)) {
              selectionKey = newKey;
            }
            nChannels++;
          }
          catch (Exception e)
          {
            logger.warn("Failed to re-register a Channel to the new Selector.", e);
            if ((a instanceof AbstractNioChannel))
            {
              AbstractNioChannel ch = (AbstractNioChannel)a;
              ch.unsafe().close(ch.unsafe().voidPromise());
            }
            else
            {
              NioTask<SelectableChannel> task = (NioTask)a;
              invokeChannelUnregistered(task, key, e);
            }
          }
        }
      }
      catch (ConcurrentModificationException e) {}
    }
    selector = newSelector;
    try
    {
      oldSelector.close();
    }
    catch (Throwable t)
    {
      if (logger.isWarnEnabled()) {
        logger.warn("Failed to close the old Selector.", t);
      }
    }
    logger.info("Migrated " + nChannels + " channel(s) to the new Selector.");
  }
  
  protected void run()
  {
    try
    {
      for (;;)
      {
        switch (selectStrategy.calculateStrategy(selectNowSupplier, hasTasks()))
        {
        case -2: 
          break;
        case -1: 
          select(wakenUp.getAndSet(false));
          if (wakenUp.get()) {
            selector.wakeup();
          }
        default: 
          cancelledKeys = 0;
          needsToSelectAgain = false;
          int ioRatio = this.ioRatio;
          if (ioRatio == 100)
          {
            processSelectedKeys();
            runAllTasks();
          }
          else
          {
            long ioStartTime = System.nanoTime();
            
            processSelectedKeys();
            
            long ioTime = System.nanoTime() - ioStartTime;
            runAllTasks(ioTime * (100 - ioRatio) / ioRatio);
          }
          if (isShuttingDown())
          {
            closeAll();
            if (confirmShutdown()) {
              return;
            }
          }
        }
      }
    }
    catch (Throwable t)
    {
      logger.warn("Unexpected exception in the selector loop.", t);
      try
      {
        Thread.sleep(1000L);
      }
      catch (InterruptedException localInterruptedException) {}
    }
  }
  
  private void processSelectedKeys()
  {
    if (selectedKeys != null) {
      processSelectedKeysOptimized(selectedKeys.flip());
    } else {
      processSelectedKeysPlain(selector.selectedKeys());
    }
  }
  
  protected void cleanup()
  {
    try
    {
      selector.close();
    }
    catch (IOException e)
    {
      logger.warn("Failed to close a selector.", e);
    }
  }
  
  void cancel(SelectionKey key)
  {
    key.cancel();
    cancelledKeys += 1;
    if (cancelledKeys >= 256)
    {
      cancelledKeys = 0;
      needsToSelectAgain = true;
    }
  }
  
  protected Runnable pollTask()
  {
    Runnable task = super.pollTask();
    if (needsToSelectAgain) {
      selectAgain();
    }
    return task;
  }
  
  private void processSelectedKeysPlain(Set<SelectionKey> selectedKeys)
  {
    if (selectedKeys.isEmpty()) {
      return;
    }
    Iterator<SelectionKey> i = selectedKeys.iterator();
    for (;;)
    {
      SelectionKey k = (SelectionKey)i.next();
      Object a = k.attachment();
      i.remove();
      if ((a instanceof AbstractNioChannel))
      {
        processSelectedKey(k, (AbstractNioChannel)a);
      }
      else
      {
        NioTask<SelectableChannel> task = (NioTask)a;
        processSelectedKey(k, task);
      }
      if (!i.hasNext()) {
        break;
      }
      if (needsToSelectAgain)
      {
        selectAgain();
        selectedKeys = selector.selectedKeys();
        if (selectedKeys.isEmpty()) {
          break;
        }
        i = selectedKeys.iterator();
      }
    }
  }
  
  private void processSelectedKeysOptimized(SelectionKey[] selectedKeys)
  {
    for (int i = 0;; i++)
    {
      SelectionKey k = selectedKeys[i];
      if (k == null) {
        break;
      }
      selectedKeys[i] = null;
      
      Object a = k.attachment();
      if ((a instanceof AbstractNioChannel))
      {
        processSelectedKey(k, (AbstractNioChannel)a);
      }
      else
      {
        NioTask<SelectableChannel> task = (NioTask)a;
        processSelectedKey(k, task);
      }
      if (needsToSelectAgain)
      {
        for (;;)
        {
          i++;
          if (selectedKeys[i] == null) {
            break;
          }
          selectedKeys[i] = null;
        }
        selectAgain();
        
        selectedKeys = this.selectedKeys.flip();
        i = -1;
      }
    }
  }
  
  private void processSelectedKey(SelectionKey k, AbstractNioChannel ch)
  {
    AbstractNioChannel.NioUnsafe unsafe = ch.unsafe();
    if (!k.isValid())
    {
      try
      {
        eventLoop = ch.eventLoop();
      }
      catch (Throwable ignored)
      {
        EventLoop eventLoop;
        return;
      }
      EventLoop eventLoop;
      if ((eventLoop != this) || (eventLoop == null)) {
        return;
      }
      unsafe.close(unsafe.voidPromise());
      return;
    }
    try
    {
      int readyOps = k.readyOps();
      if (((readyOps & 0x11) != 0) || (readyOps == 0))
      {
        unsafe.read();
        if (!ch.isOpen()) {
          return;
        }
      }
      if ((readyOps & 0x4) != 0) {
        ch.unsafe().forceFlush();
      }
      if ((readyOps & 0x8) != 0)
      {
        int ops = k.interestOps();
        ops &= 0xFFFFFFF7;
        k.interestOps(ops);
        
        unsafe.finishConnect();
      }
    }
    catch (CancelledKeyException ignored)
    {
      unsafe.close(unsafe.voidPromise());
    }
  }
  
  private static void processSelectedKey(SelectionKey k, NioTask<SelectableChannel> task)
  {
    int state = 0;
    try
    {
      task.channelReady(k.channel(), k);
      state = 1;
    }
    catch (Exception e)
    {
      k.cancel();
      invokeChannelUnregistered(task, k, e);
      state = 2;
    }
    finally
    {
      switch (state)
      {
      case 0: 
        k.cancel();
        invokeChannelUnregistered(task, k, null);
        break;
      case 1: 
        if (!k.isValid()) {
          invokeChannelUnregistered(task, k, null);
        }
        break;
      }
    }
  }
  
  private void closeAll()
  {
    selectAgain();
    Set<SelectionKey> keys = selector.keys();
    Collection<AbstractNioChannel> channels = new ArrayList(keys.size());
    for (SelectionKey k : keys)
    {
      Object a = k.attachment();
      if ((a instanceof AbstractNioChannel))
      {
        channels.add((AbstractNioChannel)a);
      }
      else
      {
        k.cancel();
        
        NioTask<SelectableChannel> task = (NioTask)a;
        invokeChannelUnregistered(task, k, null);
      }
    }
    for (AbstractNioChannel ch : channels) {
      ch.unsafe().close(ch.unsafe().voidPromise());
    }
  }
  
  private static void invokeChannelUnregistered(NioTask<SelectableChannel> task, SelectionKey k, Throwable cause)
  {
    try
    {
      task.channelUnregistered(k.channel(), cause);
    }
    catch (Exception e)
    {
      logger.warn("Unexpected exception while running NioTask.channelUnregistered()", e);
    }
  }
  
  protected void wakeup(boolean inEventLoop)
  {
    if ((!inEventLoop) && (wakenUp.compareAndSet(false, true))) {
      selector.wakeup();
    }
  }
  
  int selectNow()
    throws IOException
  {
    try
    {
      return selector.selectNow();
    }
    finally
    {
      if (wakenUp.get()) {
        selector.wakeup();
      }
    }
  }
  
  private void select(boolean oldWakenUp)
    throws IOException
  {
    Selector selector = this.selector;
    try
    {
      int selectCnt = 0;
      long currentTimeNanos = System.nanoTime();
      long selectDeadLineNanos = currentTimeNanos + delayNanos(currentTimeNanos);
      for (;;)
      {
        long timeoutMillis = (selectDeadLineNanos - currentTimeNanos + 500000L) / 1000000L;
        if (timeoutMillis <= 0L)
        {
          if (selectCnt != 0) {
            break;
          }
          selector.selectNow();
          selectCnt = 1; break;
        }
        if ((hasTasks()) && (wakenUp.compareAndSet(false, true)))
        {
          selector.selectNow();
          selectCnt = 1;
          break;
        }
        int selectedKeys = selector.select(timeoutMillis);
        selectCnt++;
        if ((selectedKeys != 0) || (oldWakenUp) || (wakenUp.get()) || (hasTasks()) || (hasScheduledTasks())) {
          break;
        }
        if (Thread.interrupted())
        {
          if (logger.isDebugEnabled()) {
            logger.debug("Selector.select() returned prematurely because Thread.currentThread().interrupt() was called. Use NioEventLoop.shutdownGracefully() to shutdown the NioEventLoop.");
          }
          selectCnt = 1;
          break;
        }
        long time = System.nanoTime();
        if (time - TimeUnit.MILLISECONDS.toNanos(timeoutMillis) >= currentTimeNanos)
        {
          selectCnt = 1;
        }
        else if ((SELECTOR_AUTO_REBUILD_THRESHOLD > 0) && (selectCnt >= SELECTOR_AUTO_REBUILD_THRESHOLD))
        {
          logger.warn("Selector.select() returned prematurely {} times in a row; rebuilding Selector {}.", Integer.valueOf(selectCnt), selector);
          
          rebuildSelector();
          selector = this.selector;
          
          selector.selectNow();
          selectCnt = 1;
          break;
        }
        currentTimeNanos = time;
      }
      if ((selectCnt > 3) && 
        (logger.isDebugEnabled())) {
        logger.debug("Selector.select() returned prematurely {} times in a row for Selector {}.", Integer.valueOf(selectCnt - 1), selector);
      }
    }
    catch (CancelledKeyException e)
    {
      if (logger.isDebugEnabled()) {
        logger.debug(CancelledKeyException.class.getSimpleName() + " raised by a Selector {} - JDK bug?", selector, e);
      }
    }
  }
  
  private void selectAgain()
  {
    needsToSelectAgain = false;
    try
    {
      selector.selectNow();
    }
    catch (Throwable t)
    {
      logger.warn("Failed to update SelectionKeys.", t);
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.channel.nio.NioEventLoop
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */