package io.netty.handler.codec;

import io.netty.util.internal.ObjectUtil;
import java.util.AbstractList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;

public final class HeadersUtils
{
  public static <K, V> List<String> getAllAsString(Headers<K, V, ?> headers, K name)
  {
    List<V> allNames = headers.getAll(name);
    new AbstractList()
    {
      public String get(int index)
      {
        V value = val$allNames.get(index);
        return value != null ? value.toString() : null;
      }
      
      public int size()
      {
        return val$allNames.size();
      }
    };
  }
  
  public static <K, V> String getAsString(Headers<K, V, ?> headers, K name)
  {
    V orig = headers.get(name);
    return orig != null ? orig.toString() : null;
  }
  
  public static Iterator<Map.Entry<String, String>> iteratorAsString(Iterable<Map.Entry<CharSequence, CharSequence>> headers)
  {
    return new StringEntryIterator(headers.iterator());
  }
  
  public static Set<String> namesAsString(Headers<CharSequence, CharSequence, ?> headers)
  {
    return new CharSequenceDelegatingStringSet(headers.names());
  }
  
  private static final class StringEntryIterator
    implements Iterator<Map.Entry<String, String>>
  {
    private final Iterator<Map.Entry<CharSequence, CharSequence>> iter;
    
    public StringEntryIterator(Iterator<Map.Entry<CharSequence, CharSequence>> iter)
    {
      this.iter = iter;
    }
    
    public boolean hasNext()
    {
      return iter.hasNext();
    }
    
    public Map.Entry<String, String> next()
    {
      return new HeadersUtils.StringEntry((Map.Entry)iter.next());
    }
    
    public void remove()
    {
      iter.remove();
    }
  }
  
  private static final class StringEntry
    implements Map.Entry<String, String>
  {
    private final Map.Entry<CharSequence, CharSequence> entry;
    private String name;
    private String value;
    
    StringEntry(Map.Entry<CharSequence, CharSequence> entry)
    {
      this.entry = entry;
    }
    
    public String getKey()
    {
      if (name == null) {
        name = ((CharSequence)entry.getKey()).toString();
      }
      return name;
    }
    
    public String getValue()
    {
      if ((value == null) && (entry.getValue() != null)) {
        value = ((CharSequence)entry.getValue()).toString();
      }
      return value;
    }
    
    public String setValue(String value)
    {
      String old = getValue();
      entry.setValue(value);
      return old;
    }
    
    public String toString()
    {
      return entry.toString();
    }
  }
  
  private static final class StringIterator<T>
    implements Iterator<String>
  {
    private final Iterator<T> iter;
    
    public StringIterator(Iterator<T> iter)
    {
      this.iter = iter;
    }
    
    public boolean hasNext()
    {
      return iter.hasNext();
    }
    
    public String next()
    {
      T next = iter.next();
      return next != null ? next.toString() : null;
    }
    
    public void remove()
    {
      iter.remove();
    }
  }
  
  private static final class CharSequenceDelegatingStringSet
    extends HeadersUtils.DelegatingStringSet<CharSequence>
  {
    public CharSequenceDelegatingStringSet(Set<CharSequence> allNames)
    {
      super();
    }
    
    public boolean add(String e)
    {
      return allNames.add(e);
    }
    
    public boolean addAll(Collection<? extends String> c)
    {
      return allNames.addAll(c);
    }
  }
  
  private static abstract class DelegatingStringSet<T>
    implements Set<String>
  {
    protected final Set<T> allNames;
    
    public DelegatingStringSet(Set<T> allNames)
    {
      this.allNames = ((Set)ObjectUtil.checkNotNull(allNames, "allNames"));
    }
    
    public int size()
    {
      return allNames.size();
    }
    
    public boolean isEmpty()
    {
      return allNames.isEmpty();
    }
    
    public boolean contains(Object o)
    {
      return allNames.contains(o.toString());
    }
    
    public Iterator<String> iterator()
    {
      return new HeadersUtils.StringIterator(allNames.iterator());
    }
    
    public Object[] toArray()
    {
      Object[] arr = new Object[size()];
      fillArray(arr);
      return arr;
    }
    
    public <X> X[] toArray(X[] a)
    {
      if ((a == null) || (a.length < size()))
      {
        X[] arr = (Object[])new Object[size()];
        fillArray(arr);
        return arr;
      }
      fillArray(a);
      return a;
    }
    
    private void fillArray(Object[] arr)
    {
      Iterator<T> itr = allNames.iterator();
      for (int i = 0; i < size(); i++) {
        arr[i] = itr.next();
      }
    }
    
    public boolean remove(Object o)
    {
      return allNames.remove(o);
    }
    
    public boolean containsAll(Collection<?> c)
    {
      for (Object o : c) {
        if (!contains(o)) {
          return false;
        }
      }
      return true;
    }
    
    public boolean removeAll(Collection<?> c)
    {
      boolean modified = false;
      for (Object o : c) {
        if (remove(o)) {
          modified = true;
        }
      }
      return modified;
    }
    
    public boolean retainAll(Collection<?> c)
    {
      boolean modified = false;
      Iterator<String> it = iterator();
      while (it.hasNext()) {
        if (!c.contains(it.next()))
        {
          it.remove();
          modified = true;
        }
      }
      return modified;
    }
    
    public void clear()
    {
      allNames.clear();
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.handler.codec.HeadersUtils
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */