package io.netty.handler.codec.http;

import io.netty.buffer.ByteBuf;
import io.netty.buffer.Unpooled;
import io.netty.channel.ChannelHandlerContext;
import io.netty.handler.codec.ByteToMessageDecoder;
import io.netty.handler.codec.DecoderResult;
import io.netty.handler.codec.PrematureChannelClosureException;
import io.netty.handler.codec.TooLongFrameException;
import io.netty.util.AsciiString;
import io.netty.util.ByteProcessor;
import io.netty.util.internal.AppendableCharSequence;
import java.util.List;

public abstract class HttpObjectDecoder
  extends ByteToMessageDecoder
{
  private static final String EMPTY_VALUE = "";
  private final int maxChunkSize;
  private final boolean chunkedSupported;
  protected final boolean validateHeaders;
  private final HeaderParser headerParser;
  private final LineParser lineParser;
  private HttpMessage message;
  private long chunkSize;
  private long contentLength = Long.MIN_VALUE;
  private volatile boolean resetRequested;
  private CharSequence name;
  private CharSequence value;
  private LastHttpContent trailer;
  
  private static enum State
  {
    SKIP_CONTROL_CHARS,  READ_INITIAL,  READ_HEADER,  READ_VARIABLE_LENGTH_CONTENT,  READ_FIXED_LENGTH_CONTENT,  READ_CHUNK_SIZE,  READ_CHUNKED_CONTENT,  READ_CHUNK_DELIMITER,  READ_CHUNK_FOOTER,  BAD_MESSAGE,  UPGRADED;
    
    private State() {}
  }
  
  private State currentState = State.SKIP_CONTROL_CHARS;
  
  protected HttpObjectDecoder()
  {
    this(4096, 8192, 8192, true);
  }
  
  protected HttpObjectDecoder(int maxInitialLineLength, int maxHeaderSize, int maxChunkSize, boolean chunkedSupported)
  {
    this(maxInitialLineLength, maxHeaderSize, maxChunkSize, chunkedSupported, true);
  }
  
  protected HttpObjectDecoder(int maxInitialLineLength, int maxHeaderSize, int maxChunkSize, boolean chunkedSupported, boolean validateHeaders)
  {
    this(maxInitialLineLength, maxHeaderSize, maxChunkSize, chunkedSupported, validateHeaders, 128);
  }
  
  protected HttpObjectDecoder(int maxInitialLineLength, int maxHeaderSize, int maxChunkSize, boolean chunkedSupported, boolean validateHeaders, int initialBufferSize)
  {
    if (maxInitialLineLength <= 0) {
      throw new IllegalArgumentException("maxInitialLineLength must be a positive integer: " + maxInitialLineLength);
    }
    if (maxHeaderSize <= 0) {
      throw new IllegalArgumentException("maxHeaderSize must be a positive integer: " + maxHeaderSize);
    }
    if (maxChunkSize <= 0) {
      throw new IllegalArgumentException("maxChunkSize must be a positive integer: " + maxChunkSize);
    }
    AppendableCharSequence seq = new AppendableCharSequence(initialBufferSize);
    lineParser = new LineParser(seq, maxInitialLineLength);
    headerParser = new HeaderParser(seq, maxHeaderSize);
    this.maxChunkSize = maxChunkSize;
    this.chunkedSupported = chunkedSupported;
    this.validateHeaders = validateHeaders;
  }
  
  protected void decode(ChannelHandlerContext ctx, ByteBuf buffer, List<Object> out)
    throws Exception
  {
    if (resetRequested) {
      resetNow();
    }
    switch (currentState)
    {
    case SKIP_CONTROL_CHARS: 
      if (!skipControlCharacters(buffer)) {
        return;
      }
      currentState = State.READ_INITIAL;
    case READ_INITIAL: 
      try
      {
        AppendableCharSequence line = lineParser.parse(buffer);
        if (line == null) {
          return;
        }
        String[] initialLine = splitInitialLine(line);
        if (initialLine.length < 3)
        {
          currentState = State.SKIP_CONTROL_CHARS;
          return;
        }
        message = createMessage(initialLine);
        currentState = State.READ_HEADER;
      }
      catch (Exception e)
      {
        out.add(invalidMessage(buffer, e));
        return;
      }
    case READ_HEADER: 
      try
      {
        State nextState = readHeaders(buffer);
        if (nextState == null) {
          return;
        }
        currentState = nextState;
        switch (nextState)
        {
        case SKIP_CONTROL_CHARS: 
          out.add(message);
          out.add(LastHttpContent.EMPTY_LAST_CONTENT);
          resetNow();
          return;
        case READ_CHUNK_SIZE: 
          if (!chunkedSupported) {
            throw new IllegalArgumentException("Chunked messages not supported");
          }
          out.add(message);
          return;
        }
        long contentLength = contentLength();
        if ((contentLength == 0L) || ((contentLength == -1L) && (isDecodingRequest())))
        {
          out.add(message);
          out.add(LastHttpContent.EMPTY_LAST_CONTENT);
          resetNow();
          return;
        }
        assert ((nextState == State.READ_FIXED_LENGTH_CONTENT) || (nextState == State.READ_VARIABLE_LENGTH_CONTENT));
        
        out.add(message);
        if (nextState == State.READ_FIXED_LENGTH_CONTENT) {
          this.chunkSize = contentLength;
        }
        return;
      }
      catch (Exception e)
      {
        out.add(invalidMessage(buffer, e));
        return;
      }
    case READ_VARIABLE_LENGTH_CONTENT: 
      int toRead = Math.min(buffer.readableBytes(), maxChunkSize);
      if (toRead > 0)
      {
        ByteBuf content = buffer.readRetainedSlice(toRead);
        out.add(new DefaultHttpContent(content));
      }
      return;
    case READ_FIXED_LENGTH_CONTENT: 
      int readLimit = buffer.readableBytes();
      if (readLimit == 0) {
        return;
      }
      int toRead = Math.min(readLimit, maxChunkSize);
      if (toRead > this.chunkSize) {
        toRead = (int)this.chunkSize;
      }
      ByteBuf content = buffer.readRetainedSlice(toRead);
      this.chunkSize -= toRead;
      if (this.chunkSize == 0L)
      {
        out.add(new DefaultLastHttpContent(content, validateHeaders));
        resetNow();
      }
      else
      {
        out.add(new DefaultHttpContent(content));
      }
      return;
    case READ_CHUNK_SIZE: 
      try
      {
        AppendableCharSequence line = lineParser.parse(buffer);
        if (line == null) {
          return;
        }
        int chunkSize = getChunkSize(line.toString());
        this.chunkSize = chunkSize;
        if (chunkSize == 0)
        {
          currentState = State.READ_CHUNK_FOOTER;
          return;
        }
        currentState = State.READ_CHUNKED_CONTENT;
      }
      catch (Exception e)
      {
        out.add(invalidChunk(buffer, e));
        return;
      }
    case READ_CHUNKED_CONTENT: 
      assert (this.chunkSize <= 2147483647L);
      int toRead = Math.min((int)this.chunkSize, maxChunkSize);
      toRead = Math.min(toRead, buffer.readableBytes());
      if (toRead == 0) {
        return;
      }
      HttpContent chunk = new DefaultHttpContent(buffer.readRetainedSlice(toRead));
      this.chunkSize -= toRead;
      
      out.add(chunk);
      if (this.chunkSize != 0L) {
        return;
      }
      currentState = State.READ_CHUNK_DELIMITER;
    case READ_CHUNK_DELIMITER: 
      int wIdx = buffer.writerIndex();
      int rIdx = buffer.readerIndex();
      while (wIdx > rIdx)
      {
        byte next = buffer.getByte(rIdx++);
        if (next == 10)
        {
          currentState = State.READ_CHUNK_SIZE;
          break;
        }
      }
      buffer.readerIndex(rIdx);
      return;
    case READ_CHUNK_FOOTER: 
      try
      {
        LastHttpContent trailer = readTrailingHeaders(buffer);
        if (trailer == null) {
          return;
        }
        out.add(trailer);
        resetNow();
        return;
      }
      catch (Exception e)
      {
        out.add(invalidChunk(buffer, e));
        return;
      }
    case BAD_MESSAGE: 
      buffer.skipBytes(buffer.readableBytes());
      break;
    case UPGRADED: 
      int readableBytes = buffer.readableBytes();
      if (readableBytes > 0) {
        out.add(buffer.readBytes(readableBytes));
      }
      break;
    }
  }
  
  protected void decodeLast(ChannelHandlerContext ctx, ByteBuf in, List<Object> out)
    throws Exception
  {
    super.decodeLast(ctx, in, out);
    if (resetRequested) {
      resetNow();
    }
    if (message != null)
    {
      boolean chunked = HttpUtil.isTransferEncodingChunked(message);
      if ((currentState == State.READ_VARIABLE_LENGTH_CONTENT) && (!in.isReadable()) && (!chunked))
      {
        out.add(LastHttpContent.EMPTY_LAST_CONTENT);
        resetNow();
        return;
      }
      if (currentState == State.READ_HEADER)
      {
        out.add(invalidMessage(Unpooled.EMPTY_BUFFER, new PrematureChannelClosureException("Connection closed before received headers")));
        
        resetNow(); return;
      }
      boolean prematureClosure;
      boolean prematureClosure;
      if ((isDecodingRequest()) || (chunked)) {
        prematureClosure = true;
      } else {
        prematureClosure = contentLength() > 0L;
      }
      if (!prematureClosure) {
        out.add(LastHttpContent.EMPTY_LAST_CONTENT);
      }
      resetNow();
    }
  }
  
  public void userEventTriggered(ChannelHandlerContext ctx, Object evt)
    throws Exception
  {
    if ((evt instanceof HttpExpectationFailedEvent)) {
      switch (currentState)
      {
      case READ_CHUNK_SIZE: 
      case READ_VARIABLE_LENGTH_CONTENT: 
      case READ_FIXED_LENGTH_CONTENT: 
        reset();
        break;
      }
    }
    super.userEventTriggered(ctx, evt);
  }
  
  protected boolean isContentAlwaysEmpty(HttpMessage msg)
  {
    if ((msg instanceof HttpResponse))
    {
      HttpResponse res = (HttpResponse)msg;
      int code = res.status().code();
      if ((code >= 100) && (code < 200)) {
        return (code != 101) || (res.headers().contains(HttpHeaderNames.SEC_WEBSOCKET_ACCEPT)) || (!res.headers().contains(HttpHeaderNames.UPGRADE, HttpHeaderValues.WEBSOCKET, true));
      }
      switch (code)
      {
      case 204: 
      case 205: 
      case 304: 
        return true;
      }
    }
    return false;
  }
  
  public void reset()
  {
    resetRequested = true;
  }
  
  private void resetNow()
  {
    HttpMessage message = this.message;
    this.message = null;
    name = null;
    value = null;
    contentLength = Long.MIN_VALUE;
    lineParser.reset();
    headerParser.reset();
    trailer = null;
    if (!isDecodingRequest())
    {
      HttpResponse res = (HttpResponse)message;
      if ((res != null) && (res.status().code() == 101))
      {
        currentState = State.UPGRADED;
        return;
      }
    }
    currentState = State.SKIP_CONTROL_CHARS;
  }
  
  private HttpMessage invalidMessage(ByteBuf in, Exception cause)
  {
    currentState = State.BAD_MESSAGE;
    
    in.skipBytes(in.readableBytes());
    if (message != null)
    {
      message.setDecoderResult(DecoderResult.failure(cause));
    }
    else
    {
      message = createInvalidMessage();
      message.setDecoderResult(DecoderResult.failure(cause));
    }
    HttpMessage ret = message;
    message = null;
    return ret;
  }
  
  private HttpContent invalidChunk(ByteBuf in, Exception cause)
  {
    currentState = State.BAD_MESSAGE;
    
    in.skipBytes(in.readableBytes());
    
    HttpContent chunk = new DefaultLastHttpContent(Unpooled.EMPTY_BUFFER);
    chunk.setDecoderResult(DecoderResult.failure(cause));
    message = null;
    trailer = null;
    return chunk;
  }
  
  private static boolean skipControlCharacters(ByteBuf buffer)
  {
    boolean skiped = false;
    int wIdx = buffer.writerIndex();
    int rIdx = buffer.readerIndex();
    while (wIdx > rIdx)
    {
      int c = buffer.getUnsignedByte(rIdx++);
      if ((!Character.isISOControl(c)) && (!Character.isWhitespace(c)))
      {
        rIdx--;
        skiped = true;
        break;
      }
    }
    buffer.readerIndex(rIdx);
    return skiped;
  }
  
  private State readHeaders(ByteBuf buffer)
  {
    HttpMessage message = this.message;
    HttpHeaders headers = message.headers();
    
    AppendableCharSequence line = headerParser.parse(buffer);
    if (line == null) {
      return null;
    }
    if (line.length() > 0) {
      do
      {
        char firstChar = line.charAt(0);
        if ((name != null) && ((firstChar == ' ') || (firstChar == '\t')))
        {
          StringBuilder buf = new StringBuilder(value.length() + line.length() + 1);
          buf.append(value).append(' ').append(line.toString().trim());
          
          value = buf.toString();
        }
        else
        {
          if (name != null) {
            headers.add(name, value);
          }
          splitHeader(line);
        }
        line = headerParser.parse(buffer);
        if (line == null) {
          return null;
        }
      } while (line.length() > 0);
    }
    if (name != null) {
      headers.add(name, value);
    }
    name = null;
    value = null;
    State nextState;
    State nextState;
    if (isContentAlwaysEmpty(message))
    {
      HttpUtil.setTransferEncodingChunked(message, false);
      nextState = State.SKIP_CONTROL_CHARS;
    }
    else
    {
      State nextState;
      if (HttpUtil.isTransferEncodingChunked(message))
      {
        nextState = State.READ_CHUNK_SIZE;
      }
      else
      {
        State nextState;
        if (contentLength() >= 0L) {
          nextState = State.READ_FIXED_LENGTH_CONTENT;
        } else {
          nextState = State.READ_VARIABLE_LENGTH_CONTENT;
        }
      }
    }
    return nextState;
  }
  
  private long contentLength()
  {
    if (contentLength == Long.MIN_VALUE) {
      contentLength = HttpUtil.getContentLength(message, -1L);
    }
    return contentLength;
  }
  
  private LastHttpContent readTrailingHeaders(ByteBuf buffer)
  {
    AppendableCharSequence line = headerParser.parse(buffer);
    if (line == null) {
      return null;
    }
    CharSequence lastHeader = null;
    if (line.length() > 0)
    {
      LastHttpContent trailer = this.trailer;
      if (trailer == null) {
        trailer = this.trailer = new DefaultLastHttpContent(Unpooled.EMPTY_BUFFER, validateHeaders);
      }
      do
      {
        char firstChar = line.charAt(0);
        if ((lastHeader != null) && ((firstChar == ' ') || (firstChar == '\t')))
        {
          List<String> current = trailer.trailingHeaders().getAll(lastHeader);
          if (!current.isEmpty())
          {
            int lastPos = current.size() - 1;
            String lineTrimmed = line.toString().trim();
            CharSequence currentLastPos = (CharSequence)current.get(lastPos);
            StringBuilder b = new StringBuilder(currentLastPos.length() + lineTrimmed.length());
            b.append(currentLastPos).append(lineTrimmed);
            
            current.set(lastPos, b.toString());
          }
        }
        else
        {
          splitHeader(line);
          CharSequence headerName = name;
          if ((!HttpHeaderNames.CONTENT_LENGTH.contentEqualsIgnoreCase(headerName)) && (!HttpHeaderNames.TRANSFER_ENCODING.contentEqualsIgnoreCase(headerName)) && (!HttpHeaderNames.TRAILER.contentEqualsIgnoreCase(headerName))) {
            trailer.trailingHeaders().add(headerName, value);
          }
          lastHeader = name;
          
          name = null;
          value = null;
        }
        line = headerParser.parse(buffer);
        if (line == null) {
          return null;
        }
      } while (line.length() > 0);
      this.trailer = null;
      return trailer;
    }
    return LastHttpContent.EMPTY_LAST_CONTENT;
  }
  
  protected abstract boolean isDecodingRequest();
  
  protected abstract HttpMessage createMessage(String[] paramArrayOfString)
    throws Exception;
  
  protected abstract HttpMessage createInvalidMessage();
  
  private static int getChunkSize(String hex)
  {
    hex = hex.trim();
    for (int i = 0; i < hex.length(); i++)
    {
      char c = hex.charAt(i);
      if ((c == ';') || (Character.isWhitespace(c)) || (Character.isISOControl(c)))
      {
        hex = hex.substring(0, i);
        break;
      }
    }
    return Integer.parseInt(hex, 16);
  }
  
  private static String[] splitInitialLine(AppendableCharSequence sb)
  {
    int aStart = findNonWhitespace(sb, 0);
    int aEnd = findWhitespace(sb, aStart);
    
    int bStart = findNonWhitespace(sb, aEnd);
    int bEnd = findWhitespace(sb, bStart);
    
    int cStart = findNonWhitespace(sb, bEnd);
    int cEnd = findEndOfString(sb);
    
    return new String[] { sb.subStringUnsafe(aStart, aEnd), sb.subStringUnsafe(bStart, bEnd), cStart < cEnd ? sb.subStringUnsafe(cStart, cEnd) : "" };
  }
  
  private void splitHeader(AppendableCharSequence sb)
  {
    int length = sb.length();
    
    int nameStart = findNonWhitespace(sb, 0);
    for (int nameEnd = nameStart; nameEnd < length; nameEnd++)
    {
      char ch = sb.charAt(nameEnd);
      if ((ch == ':') || (Character.isWhitespace(ch))) {
        break;
      }
    }
    for (int colonEnd = nameEnd; colonEnd < length; colonEnd++) {
      if (sb.charAt(colonEnd) == ':')
      {
        colonEnd++;
        break;
      }
    }
    name = sb.subStringUnsafe(nameStart, nameEnd);
    int valueStart = findNonWhitespace(sb, colonEnd);
    if (valueStart == length)
    {
      value = "";
    }
    else
    {
      int valueEnd = findEndOfString(sb);
      value = sb.subStringUnsafe(valueStart, valueEnd);
    }
  }
  
  private static int findNonWhitespace(AppendableCharSequence sb, int offset)
  {
    for (int result = offset; result < sb.length(); result++) {
      if (!Character.isWhitespace(sb.charAtUnsafe(result))) {
        return result;
      }
    }
    return sb.length();
  }
  
  private static int findWhitespace(AppendableCharSequence sb, int offset)
  {
    for (int result = offset; result < sb.length(); result++) {
      if (Character.isWhitespace(sb.charAtUnsafe(result))) {
        return result;
      }
    }
    return sb.length();
  }
  
  private static int findEndOfString(AppendableCharSequence sb)
  {
    for (int result = sb.length() - 1; result > 0; result--) {
      if (!Character.isWhitespace(sb.charAtUnsafe(result))) {
        return result + 1;
      }
    }
    return 0;
  }
  
  private static class HeaderParser
    implements ByteProcessor
  {
    private final AppendableCharSequence seq;
    private final int maxLength;
    private int size;
    
    HeaderParser(AppendableCharSequence seq, int maxLength)
    {
      this.seq = seq;
      this.maxLength = maxLength;
    }
    
    public AppendableCharSequence parse(ByteBuf buffer)
    {
      int oldSize = size;
      seq.reset();
      int i = buffer.forEachByte(this);
      if (i == -1)
      {
        size = oldSize;
        return null;
      }
      buffer.readerIndex(i + 1);
      return seq;
    }
    
    public void reset()
    {
      size = 0;
    }
    
    public boolean process(byte value)
      throws Exception
    {
      char nextByte = (char)value;
      if (nextByte == '\r') {
        return true;
      }
      if (nextByte == '\n') {
        return false;
      }
      if (++size > maxLength) {
        throw newException(maxLength);
      }
      seq.append(nextByte);
      return true;
    }
    
    protected TooLongFrameException newException(int maxLength)
    {
      return new TooLongFrameException("HTTP header is larger than " + maxLength + " bytes.");
    }
  }
  
  private static final class LineParser
    extends HttpObjectDecoder.HeaderParser
  {
    LineParser(AppendableCharSequence seq, int maxLength)
    {
      super(maxLength);
    }
    
    public AppendableCharSequence parse(ByteBuf buffer)
    {
      reset();
      return super.parse(buffer);
    }
    
    protected TooLongFrameException newException(int maxLength)
    {
      return new TooLongFrameException("An HTTP line is larger than " + maxLength + " bytes.");
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.handler.codec.http.HttpObjectDecoder
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */