package io.netty.handler.codec.http2;

import io.netty.channel.Channel;
import io.netty.channel.Channel.Unsafe;
import io.netty.channel.ChannelDuplexHandler;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelHandler;
import io.netty.channel.ChannelHandler.Sharable;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.ChannelPipeline;
import io.netty.channel.ChannelPromise;
import io.netty.channel.EventLoop;
import io.netty.channel.EventLoopGroup;
import io.netty.handler.codec.UnsupportedMessageTypeException;
import io.netty.util.ReferenceCountUtil;
import io.netty.util.collection.IntObjectHashMap;
import io.netty.util.collection.IntObjectMap;
import io.netty.util.collection.IntObjectMap.PrimitiveEntry;
import io.netty.util.concurrent.EventExecutor;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;
import java.util.ArrayList;
import java.util.List;

public final class Http2MultiplexCodec
  extends ChannelDuplexHandler
{
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(Http2MultiplexCodec.class);
  private final ChannelHandler streamHandler;
  private final EventLoopGroup streamGroup;
  private final List<Http2StreamChannel> channelsToFireChildReadComplete = new ArrayList();
  private final boolean server;
  private ChannelHandlerContext ctx;
  private volatile Runnable flushTask;
  private final IntObjectMap<Http2StreamChannel> childChannels = new IntObjectHashMap();
  
  public Http2MultiplexCodec(boolean server, EventLoopGroup streamGroup, ChannelHandler streamHandler)
  {
    if (!streamHandler.getClass().isAnnotationPresent(ChannelHandler.Sharable.class)) {
      throw new IllegalArgumentException("streamHandler must be Sharable");
    }
    this.server = server;
    this.streamHandler = streamHandler;
    this.streamGroup = streamGroup;
  }
  
  public void handlerAdded(ChannelHandlerContext ctx)
    throws Exception
  {
    this.ctx = ctx;
  }
  
  public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause)
  {
    if (!(cause instanceof Http2Exception.StreamException))
    {
      ctx.fireExceptionCaught(cause);
      return;
    }
    Http2Exception.StreamException streamEx = (Http2Exception.StreamException)cause;
    try
    {
      Http2StreamChannel childChannel = (Http2StreamChannel)childChannels.get(streamEx.streamId());
      if (childChannel != null) {
        childChannel.pipeline().fireExceptionCaught(streamEx);
      } else {
        logger.warn(String.format("Exception caught for unknown HTTP/2 stream '%d'", new Object[] { Integer.valueOf(streamEx.streamId()) }), streamEx);
      }
    }
    finally
    {
      onStreamClosed(streamEx.streamId());
    }
  }
  
  public void flush(ChannelHandlerContext ctx)
  {
    ctx.flush();
  }
  
  public void channelRead(ChannelHandlerContext ctx, Object msg)
    throws Exception
  {
    if (!(msg instanceof Http2Frame))
    {
      ctx.fireChannelRead(msg);
      return;
    }
    if ((msg instanceof Http2StreamFrame))
    {
      Http2StreamFrame frame = (Http2StreamFrame)msg;
      int streamId = frame.streamId();
      Http2StreamChannel childChannel = (Http2StreamChannel)childChannels.get(streamId);
      if (childChannel == null)
      {
        ReferenceCountUtil.release(msg);
        throw new Http2Exception.StreamException(streamId, Http2Error.STREAM_CLOSED, String.format("Received %s frame for an unknown stream %d", new Object[] { frame.name(), Integer.valueOf(streamId) }));
      }
      fireChildReadAndRegister(childChannel, frame);
    }
    else if ((msg instanceof Http2GoAwayFrame))
    {
      Http2GoAwayFrame goAwayFrame = (Http2GoAwayFrame)msg;
      for (IntObjectMap.PrimitiveEntry<Http2StreamChannel> entry : childChannels.entries())
      {
        Http2StreamChannel childChannel = (Http2StreamChannel)entry.value();
        int streamId = entry.key();
        if ((streamId > goAwayFrame.lastStreamId()) && (isLocalStream(streamId))) {
          childChannel.pipeline().fireUserEventTriggered(goAwayFrame.retainedDuplicate());
        }
      }
      goAwayFrame.release();
    }
    else
    {
      ReferenceCountUtil.release(msg);
      throw new UnsupportedMessageTypeException(msg, new Class[0]);
    }
  }
  
  private void fireChildReadAndRegister(Http2StreamChannel childChannel, Http2StreamFrame frame)
  {
    childChannel.fireChildRead(frame);
    if (!inStreamsToFireChildReadComplete)
    {
      channelsToFireChildReadComplete.add(childChannel);
      inStreamsToFireChildReadComplete = true;
    }
  }
  
  public void userEventTriggered(ChannelHandlerContext ctx, Object evt)
    throws Exception
  {
    if (!(evt instanceof Http2StreamStateEvent))
    {
      ctx.fireUserEventTriggered(evt);
      return;
    }
    try
    {
      int streamId = ((Http2StreamStateEvent)evt).streamId();
      if ((evt instanceof Http2StreamActiveEvent)) {
        onStreamActive(streamId);
      } else if ((evt instanceof Http2StreamClosedEvent)) {
        onStreamClosed(streamId);
      } else {
        throw new UnsupportedMessageTypeException(evt, new Class[0]);
      }
    }
    finally
    {
      ReferenceCountUtil.release(evt);
    }
  }
  
  private void onStreamActive(int streamId)
  {
    ChannelFuture future = createStreamChannel(ctx, streamId, streamHandler);
    Http2StreamChannel childChannel = (Http2StreamChannel)future.channel();
    Http2StreamChannel oldChannel = (Http2StreamChannel)childChannels.put(streamId, childChannel);
    assert (oldChannel == null);
  }
  
  private void onStreamClosed(int streamId)
  {
    final Http2StreamChannel childChannel = (Http2StreamChannel)childChannels.remove(streamId);
    if (childChannel != null)
    {
      EventLoop eventLoop = childChannel.eventLoop();
      if (eventLoop.inEventLoop()) {
        onStreamClosed0(childChannel);
      } else {
        eventLoop.execute(new Runnable()
        {
          public void run()
          {
            Http2MultiplexCodec.this.onStreamClosed0(childChannel);
          }
        });
      }
    }
  }
  
  private void onStreamClosed0(Http2StreamChannel childChannel)
  {
    assert (childChannel.eventLoop().inEventLoop());
    
    onStreamClosedFired = true;
    childChannel.fireChildRead(AbstractHttp2StreamChannel.CLOSE_MESSAGE);
  }
  
  void flushFromStreamChannel()
  {
    EventExecutor executor = ctx.executor();
    if (executor.inEventLoop())
    {
      flush(ctx);
    }
    else
    {
      Runnable task = flushTask;
      if (task == null) {
        task =  = new Runnable()
        {
          public void run()
          {
            flush(ctx);
          }
        };
      }
      executor.execute(task);
    }
  }
  
  void writeFromStreamChannel(final Object msg, final boolean flush)
  {
    final ChannelPromise promise = ctx.newPromise();
    EventExecutor executor = ctx.executor();
    if (executor.inEventLoop()) {
      writeFromStreamChannel0(msg, flush, promise);
    } else {
      try
      {
        executor.execute(new Runnable()
        {
          public void run()
          {
            Http2MultiplexCodec.this.writeFromStreamChannel0(msg, flush, promise);
          }
        });
      }
      catch (Throwable cause)
      {
        promise.setFailure(cause);
      }
    }
  }
  
  private void writeFromStreamChannel0(Object msg, boolean flush, ChannelPromise promise)
  {
    try
    {
      write(ctx, msg, promise);
    }
    catch (Throwable cause)
    {
      promise.tryFailure(cause);
    }
    if (flush) {
      flush(ctx);
    }
  }
  
  private ChannelFuture createStreamChannel(ChannelHandlerContext ctx, int streamId, ChannelHandler handler)
  {
    EventLoopGroup group = streamGroup != null ? streamGroup : ctx.channel().eventLoop();
    Http2StreamChannel channel = new Http2StreamChannel(streamId);
    channel.pipeline().addLast(new ChannelHandler[] { handler });
    ChannelFuture future = group.register(channel);
    if (future.cause() != null) {
      if (channel.isRegistered()) {
        channel.close();
      } else {
        channel.unsafe().closeForcibly();
      }
    }
    return future;
  }
  
  public void channelReadComplete(ChannelHandlerContext ctx)
  {
    for (int i = 0; i < channelsToFireChildReadComplete.size(); i++)
    {
      Http2StreamChannel childChannel = (Http2StreamChannel)channelsToFireChildReadComplete.get(i);
      
      inStreamsToFireChildReadComplete = false;
      childChannel.fireChildReadComplete();
    }
    channelsToFireChildReadComplete.clear();
  }
  
  final class Http2StreamChannel
    extends AbstractHttp2StreamChannel
  {
    private final int streamId;
    boolean onStreamClosedFired;
    boolean inStreamsToFireChildReadComplete;
    
    Http2StreamChannel(int streamId)
    {
      super();
      this.streamId = streamId;
    }
    
    protected void doClose()
      throws Exception
    {
      if (!onStreamClosedFired)
      {
        Http2StreamFrame resetFrame = new DefaultHttp2ResetFrame(Http2Error.CANCEL).setStreamId(streamId);
        writeFromStreamChannel(resetFrame, true);
      }
      super.doClose();
    }
    
    protected void doWrite(Object msg)
    {
      if (!(msg instanceof Http2StreamFrame))
      {
        ReferenceCountUtil.release(msg);
        throw new IllegalArgumentException("Message must be an Http2StreamFrame: " + msg);
      }
      Http2StreamFrame frame = (Http2StreamFrame)msg;
      if (frame.streamId() != -1)
      {
        ReferenceCountUtil.release(frame);
        throw new IllegalArgumentException("Stream must not be set on the frame");
      }
      frame.setStreamId(streamId);
      
      writeFromStreamChannel(msg, false);
    }
    
    protected void doWriteComplete()
    {
      flushFromStreamChannel();
    }
    
    protected EventExecutor preferredEventExecutor()
    {
      return ctx.executor();
    }
    
    protected void bytesConsumed(int bytes)
    {
      ctx.write(new DefaultHttp2WindowUpdateFrame(bytes).setStreamId(streamId));
    }
  }
  
  private boolean isLocalStream(int streamId)
  {
    boolean even = (streamId & 0x1) == 0;
    return (streamId > 0) && (server == even);
  }
}

/* Location:
 * Qualified Name:     io.netty.handler.codec.http2.Http2MultiplexCodec
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */