package io.netty.handler.codec.http2.internal.hpack;

import io.netty.buffer.ByteBuf;
import io.netty.handler.codec.http2.Http2Error;
import io.netty.handler.codec.http2.Http2Exception;
import io.netty.handler.codec.http2.Http2Headers;
import io.netty.util.AsciiString;
import io.netty.util.internal.ThrowableUtil;

public final class Decoder
{
  private static final Http2Exception DECODE_DECOMPRESSION_EXCEPTION = (Http2Exception)ThrowableUtil.unknownStackTrace(Http2Exception.connectionError(Http2Error.COMPRESSION_ERROR, "HPACK - decompression failure", new Object[0]), Decoder.class, "decode(...)");
  private static final Http2Exception DECODE_ULE_128_DECOMPRESSION_EXCEPTION = (Http2Exception)ThrowableUtil.unknownStackTrace(Http2Exception.connectionError(Http2Error.COMPRESSION_ERROR, "HPACK - decompression failure", new Object[0]), Decoder.class, "decodeULE128(...)");
  private static final Http2Exception DECODE_ILLEGAL_INDEX_VALUE = (Http2Exception)ThrowableUtil.unknownStackTrace(Http2Exception.connectionError(Http2Error.COMPRESSION_ERROR, "HPACK - illegal index value", new Object[0]), Decoder.class, "decode(...)");
  private static final Http2Exception INDEX_HEADER_ILLEGAL_INDEX_VALUE = (Http2Exception)ThrowableUtil.unknownStackTrace(Http2Exception.connectionError(Http2Error.COMPRESSION_ERROR, "HPACK - illegal index value", new Object[0]), Decoder.class, "indexHeader(...)");
  private static final Http2Exception READ_NAME_ILLEGAL_INDEX_VALUE = (Http2Exception)ThrowableUtil.unknownStackTrace(Http2Exception.connectionError(Http2Error.COMPRESSION_ERROR, "HPACK - illegal index value", new Object[0]), Decoder.class, "readName(...)");
  private static final Http2Exception INVALID_MAX_DYNAMIC_TABLE_SIZE = (Http2Exception)ThrowableUtil.unknownStackTrace(Http2Exception.connectionError(Http2Error.COMPRESSION_ERROR, "HPACK - invalid max dynamic table size", new Object[0]), Decoder.class, "setDynamicTableSize(...)");
  private static final Http2Exception MAX_DYNAMIC_TABLE_SIZE_CHANGE_REQUIRED = (Http2Exception)ThrowableUtil.unknownStackTrace(Http2Exception.connectionError(Http2Error.COMPRESSION_ERROR, "HPACK - max dynamic table size change required", new Object[0]), Decoder.class, "decode(...)");
  private static final byte READ_HEADER_REPRESENTATION = 0;
  private static final byte READ_MAX_DYNAMIC_TABLE_SIZE = 1;
  private static final byte READ_INDEXED_HEADER = 2;
  private static final byte READ_INDEXED_HEADER_NAME = 3;
  private static final byte READ_LITERAL_HEADER_NAME_LENGTH_PREFIX = 4;
  private static final byte READ_LITERAL_HEADER_NAME_LENGTH = 5;
  private static final byte READ_LITERAL_HEADER_NAME = 6;
  private static final byte READ_LITERAL_HEADER_VALUE_LENGTH_PREFIX = 7;
  private static final byte READ_LITERAL_HEADER_VALUE_LENGTH = 8;
  private static final byte READ_LITERAL_HEADER_VALUE = 9;
  private final DynamicTable dynamicTable;
  private final HuffmanDecoder huffmanDecoder;
  private final int maxHeadersLength;
  private int maxDynamicTableSize;
  private int encoderMaxDynamicTableSize;
  private boolean maxDynamicTableSizeChangeRequired;
  
  public Decoder(int maxHeadersLength, int maxHeaderTableSize, int initialHuffmanDecodeCapacity)
  {
    dynamicTable = new DynamicTable(maxHeaderTableSize);
    this.maxHeadersLength = maxHeadersLength;
    maxDynamicTableSize = maxHeaderTableSize;
    encoderMaxDynamicTableSize = maxHeaderTableSize;
    maxDynamicTableSizeChangeRequired = false;
    huffmanDecoder = new HuffmanDecoder(initialHuffmanDecodeCapacity);
  }
  
  public void decode(ByteBuf in, Http2Headers headers)
    throws Http2Exception
  {
    int index = 0;
    int headersLength = 0;
    int nameLength = 0;
    int valueLength = 0;
    byte state = 0;
    boolean huffmanEncoded = false;
    CharSequence name = null;
    HpackUtil.IndexType indexType = HpackUtil.IndexType.NONE;
    while (in.isReadable()) {
      switch (state)
      {
      case 0: 
        byte b = in.readByte();
        if ((maxDynamicTableSizeChangeRequired) && ((b & 0xE0) != 32)) {
          throw MAX_DYNAMIC_TABLE_SIZE_CHANGE_REQUIRED;
        }
        if (b < 0)
        {
          index = b & 0x7F;
          switch (index)
          {
          case 0: 
            throw DECODE_ILLEGAL_INDEX_VALUE;
          case 127: 
            state = 2;
            break;
          default: 
            headersLength = indexHeader(index, headers, headersLength);break;
          }
        }
        else if ((b & 0x40) == 64)
        {
          indexType = HpackUtil.IndexType.INCREMENTAL;
          index = b & 0x3F;
          switch (index)
          {
          case 0: 
            state = 4;
            break;
          case 63: 
            state = 3;
            break;
          default: 
            name = readName(index);
            state = 7;break;
          }
        }
        else if ((b & 0x20) == 32)
        {
          index = b & 0x1F;
          if (index == 31)
          {
            state = 1;
          }
          else
          {
            setDynamicTableSize(index);
            state = 0;
          }
        }
        else
        {
          indexType = (b & 0x10) == 16 ? HpackUtil.IndexType.NEVER : HpackUtil.IndexType.NONE;
          index = b & 0xF;
          switch (index)
          {
          case 0: 
            state = 4;
            break;
          case 15: 
            state = 3;
            break;
          default: 
            name = readName(index);
            state = 7;
          }
        }
        break;
      case 1: 
        int maxSize = decodeULE128(in) + index;
        if (maxSize < 0) {
          throw DECODE_DECOMPRESSION_EXCEPTION;
        }
        setDynamicTableSize(maxSize);
        state = 0;
        break;
      case 2: 
        int headerIndex = decodeULE128(in) + index;
        if (headerIndex < 0) {
          throw DECODE_DECOMPRESSION_EXCEPTION;
        }
        headersLength = indexHeader(headerIndex, headers, headersLength);
        state = 0;
        break;
      case 3: 
        int nameIndex = decodeULE128(in) + index;
        if (nameIndex < 0) {
          throw DECODE_DECOMPRESSION_EXCEPTION;
        }
        name = readName(nameIndex);
        state = 7;
        break;
      case 4: 
        byte b = in.readByte();
        huffmanEncoded = (b & 0x80) == 128;
        index = b & 0x7F;
        if (index == 127)
        {
          state = 5;
        }
        else
        {
          if (index > maxHeadersLength - headersLength) {
            maxHeaderSizeExceeded();
          }
          nameLength = index;
          state = 6;
        }
        break;
      case 5: 
        nameLength = decodeULE128(in) + index;
        if (nameLength < 0) {
          throw DECODE_DECOMPRESSION_EXCEPTION;
        }
        if (nameLength > maxHeadersLength - headersLength) {
          maxHeaderSizeExceeded();
        }
        state = 6;
        break;
      case 6: 
        if (in.readableBytes() < nameLength) {
          throw notEnoughDataException(in);
        }
        name = readStringLiteral(in, nameLength, huffmanEncoded);
        
        state = 7;
        break;
      case 7: 
        byte b = in.readByte();
        huffmanEncoded = (b & 0x80) == 128;
        index = b & 0x7F;
        switch (index)
        {
        case 127: 
          state = 8;
          break;
        case 0: 
          headersLength = insertHeader(headers, name, AsciiString.EMPTY_STRING, indexType, headersLength);
          state = 0;
          break;
        default: 
          if (index + nameLength > maxHeadersLength - headersLength) {
            maxHeaderSizeExceeded();
          }
          valueLength = index;
          state = 9;
        }
        break;
      case 8: 
        valueLength = decodeULE128(in) + index;
        if (valueLength < 0) {
          throw DECODE_DECOMPRESSION_EXCEPTION;
        }
        if (valueLength + nameLength > maxHeadersLength - headersLength) {
          maxHeaderSizeExceeded();
        }
        state = 9;
        break;
      case 9: 
        if (in.readableBytes() < valueLength) {
          throw notEnoughDataException(in);
        }
        CharSequence value = readStringLiteral(in, valueLength, huffmanEncoded);
        headersLength = insertHeader(headers, name, value, indexType, headersLength);
        state = 0;
        break;
      default: 
        throw new Error("should not reach here state: " + state);
      }
    }
  }
  
  public void setMaxHeaderTableSize(int maxHeaderTableSize)
  {
    maxDynamicTableSize = maxHeaderTableSize;
    if (maxDynamicTableSize < encoderMaxDynamicTableSize)
    {
      maxDynamicTableSizeChangeRequired = true;
      dynamicTable.setCapacity(maxDynamicTableSize);
    }
  }
  
  public int getMaxHeaderTableSize()
  {
    return dynamicTable.capacity();
  }
  
  int length()
  {
    return dynamicTable.length();
  }
  
  int size()
  {
    return dynamicTable.size();
  }
  
  HeaderField getHeaderField(int index)
  {
    return dynamicTable.getEntry(index + 1);
  }
  
  private void setDynamicTableSize(int dynamicTableSize)
    throws Http2Exception
  {
    if (dynamicTableSize > maxDynamicTableSize) {
      throw INVALID_MAX_DYNAMIC_TABLE_SIZE;
    }
    encoderMaxDynamicTableSize = dynamicTableSize;
    maxDynamicTableSizeChangeRequired = false;
    dynamicTable.setCapacity(dynamicTableSize);
  }
  
  private CharSequence readName(int index)
    throws Http2Exception
  {
    if (index <= StaticTable.length)
    {
      HeaderField headerField = StaticTable.getEntry(index);
      return name;
    }
    if (index - StaticTable.length <= dynamicTable.length())
    {
      HeaderField headerField = dynamicTable.getEntry(index - StaticTable.length);
      return name;
    }
    throw READ_NAME_ILLEGAL_INDEX_VALUE;
  }
  
  private int indexHeader(int index, Http2Headers headers, int headersLength)
    throws Http2Exception
  {
    if (index <= StaticTable.length)
    {
      HeaderField headerField = StaticTable.getEntry(index);
      return addHeader(headers, name, value, headersLength);
    }
    if (index - StaticTable.length <= dynamicTable.length())
    {
      HeaderField headerField = dynamicTable.getEntry(index - StaticTable.length);
      return addHeader(headers, name, value, headersLength);
    }
    throw INDEX_HEADER_ILLEGAL_INDEX_VALUE;
  }
  
  private int insertHeader(Http2Headers headers, CharSequence name, CharSequence value, HpackUtil.IndexType indexType, int headerSize)
    throws Http2Exception
  {
    headerSize = addHeader(headers, name, value, headerSize);
    switch (indexType)
    {
    case NONE: 
    case NEVER: 
      break;
    case INCREMENTAL: 
      dynamicTable.add(new HeaderField(name, value));
      break;
    default: 
      throw new Error("should not reach here");
    }
    return headerSize;
  }
  
  private int addHeader(Http2Headers headers, CharSequence name, CharSequence value, int headersLength)
    throws Http2Exception
  {
    long newHeadersLength = headersLength + name.length() + value.length();
    if (newHeadersLength > maxHeadersLength) {
      maxHeaderSizeExceeded();
    }
    headers.add(name, value);
    return (int)newHeadersLength;
  }
  
  private void maxHeaderSizeExceeded()
    throws Http2Exception
  {
    throw Http2Exception.connectionError(Http2Error.ENHANCE_YOUR_CALM, "Header size exceeded max allowed bytes (%d)", new Object[] { Integer.valueOf(maxHeadersLength) });
  }
  
  private CharSequence readStringLiteral(ByteBuf in, int length, boolean huffmanEncoded)
    throws Http2Exception
  {
    if (huffmanEncoded) {
      return huffmanDecoder.decode(in, length);
    }
    byte[] buf = new byte[length];
    in.readBytes(buf);
    return new AsciiString(buf, false);
  }
  
  private static IllegalArgumentException notEnoughDataException(ByteBuf in)
  {
    return new IllegalArgumentException("decode only works with an entire header block! " + in);
  }
  
  private static int decodeULE128(ByteBuf in)
    throws Http2Exception
  {
    int writerIndex = in.writerIndex();
    int readerIndex = in.readerIndex();int shift = 0;int result = 0;
    for (; readerIndex < writerIndex; shift += 7)
    {
      byte b = in.getByte(readerIndex);
      if ((shift == 28) && ((b & 0xF8) != 0))
      {
        in.readerIndex(readerIndex + 1);
        break;
      }
      if ((b & 0x80) == 0)
      {
        in.readerIndex(readerIndex + 1);
        return result | (b & 0x7F) << shift;
      }
      result |= (b & 0x7F) << shift;readerIndex++;
    }
    throw DECODE_ULE_128_DECOMPRESSION_EXCEPTION;
  }
}

/* Location:
 * Qualified Name:     io.netty.handler.codec.http2.internal.hpack.Decoder
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */