package io.netty.handler.ssl;

import io.netty.buffer.ByteBufAllocator;
import io.netty.util.internal.ObjectUtil;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;
import java.io.File;
import java.io.IOException;
import java.security.InvalidAlgorithmParameterException;
import java.security.KeyException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.Security;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.security.spec.InvalidKeySpecException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import javax.crypto.NoSuchPaddingException;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLEngine;
import javax.net.ssl.SSLSessionContext;

public class JdkSslContext
  extends SslContext
{
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(JdkSslContext.class);
  static final String PROTOCOL = "TLS";
  static final String[] PROTOCOLS;
  static final List<String> DEFAULT_CIPHERS;
  static final Set<String> SUPPORTED_CIPHERS;
  private final String[] cipherSuites;
  private final List<String> unmodifiableCipherSuites;
  private final JdkApplicationProtocolNegotiator apn;
  private final ClientAuth clientAuth;
  private final SSLContext sslContext;
  private final boolean isClient;
  
  static
  {
    try
    {
      SSLContext context = SSLContext.getInstance("TLS");
      context.init(null, null, null);
    }
    catch (Exception e)
    {
      throw new Error("failed to initialize the default SSL context", e);
    }
    SSLContext context;
    SSLEngine engine = context.createSSLEngine();
    
    String[] supportedProtocols = engine.getSupportedProtocols();
    Set<String> supportedProtocolsSet = new HashSet(supportedProtocols.length);
    for (int i = 0; i < supportedProtocols.length; i++) {
      supportedProtocolsSet.add(supportedProtocols[i]);
    }
    List<String> protocols = new ArrayList();
    addIfSupported(supportedProtocolsSet, protocols, new String[] { "TLSv1.2", "TLSv1.1", "TLSv1" });
    if (!protocols.isEmpty()) {
      PROTOCOLS = (String[])protocols.toArray(new String[protocols.size()]);
    } else {
      PROTOCOLS = engine.getEnabledProtocols();
    }
    String[] supportedCiphers = engine.getSupportedCipherSuites();
    SUPPORTED_CIPHERS = new HashSet(supportedCiphers.length);
    for (i = 0; i < supportedCiphers.length; i++) {
      SUPPORTED_CIPHERS.add(supportedCiphers[i]);
    }
    List<String> ciphers = new ArrayList();
    addIfSupported(SUPPORTED_CIPHERS, ciphers, new String[] { "TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256", "TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA", "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA", "TLS_RSA_WITH_AES_128_GCM_SHA256", "TLS_RSA_WITH_AES_128_CBC_SHA", "TLS_RSA_WITH_AES_256_CBC_SHA", "SSL_RSA_WITH_3DES_EDE_CBC_SHA" });
    if (ciphers.isEmpty()) {
      for (String cipher : engine.getEnabledCipherSuites()) {
        if (!cipher.contains("_RC4_")) {
          ciphers.add(cipher);
        }
      }
    }
    DEFAULT_CIPHERS = Collections.unmodifiableList(ciphers);
    if (logger.isDebugEnabled())
    {
      logger.debug("Default protocols (JDK): {} ", Arrays.asList(PROTOCOLS));
      logger.debug("Default cipher suites (JDK): {}", DEFAULT_CIPHERS);
    }
  }
  
  private static void addIfSupported(Set<String> supported, List<String> enabled, String... names)
  {
    for (String n : names) {
      if (supported.contains(n)) {
        enabled.add(n);
      }
    }
  }
  
  public JdkSslContext(SSLContext sslContext, boolean isClient, ClientAuth clientAuth)
  {
    this(sslContext, isClient, null, IdentityCipherSuiteFilter.INSTANCE, JdkDefaultApplicationProtocolNegotiator.INSTANCE, clientAuth);
  }
  
  public JdkSslContext(SSLContext sslContext, boolean isClient, Iterable<String> ciphers, CipherSuiteFilter cipherFilter, ApplicationProtocolConfig apn, ClientAuth clientAuth)
  {
    this(sslContext, isClient, ciphers, cipherFilter, toNegotiator(apn, !isClient), clientAuth);
  }
  
  JdkSslContext(SSLContext sslContext, boolean isClient, Iterable<String> ciphers, CipherSuiteFilter cipherFilter, JdkApplicationProtocolNegotiator apn, ClientAuth clientAuth)
  {
    this.apn = ((JdkApplicationProtocolNegotiator)ObjectUtil.checkNotNull(apn, "apn"));
    this.clientAuth = ((ClientAuth)ObjectUtil.checkNotNull(clientAuth, "clientAuth"));
    cipherSuites = ((CipherSuiteFilter)ObjectUtil.checkNotNull(cipherFilter, "cipherFilter")).filterCipherSuites(ciphers, DEFAULT_CIPHERS, SUPPORTED_CIPHERS);
    
    unmodifiableCipherSuites = Collections.unmodifiableList(Arrays.asList(cipherSuites));
    this.sslContext = ((SSLContext)ObjectUtil.checkNotNull(sslContext, "sslContext"));
    this.isClient = isClient;
  }
  
  public final SSLContext context()
  {
    return sslContext;
  }
  
  public final boolean isClient()
  {
    return isClient;
  }
  
  public final SSLSessionContext sessionContext()
  {
    if (isServer()) {
      return context().getServerSessionContext();
    }
    return context().getClientSessionContext();
  }
  
  public final List<String> cipherSuites()
  {
    return unmodifiableCipherSuites;
  }
  
  public final long sessionCacheSize()
  {
    return sessionContext().getSessionCacheSize();
  }
  
  public final long sessionTimeout()
  {
    return sessionContext().getSessionTimeout();
  }
  
  public final SSLEngine newEngine(ByteBufAllocator alloc)
  {
    return configureAndWrapEngine(context().createSSLEngine());
  }
  
  public final SSLEngine newEngine(ByteBufAllocator alloc, String peerHost, int peerPort)
  {
    return configureAndWrapEngine(context().createSSLEngine(peerHost, peerPort));
  }
  
  private SSLEngine configureAndWrapEngine(SSLEngine engine)
  {
    engine.setEnabledCipherSuites(cipherSuites);
    engine.setEnabledProtocols(PROTOCOLS);
    engine.setUseClientMode(isClient());
    if (isServer()) {
      switch (clientAuth)
      {
      case OPTIONAL: 
        engine.setWantClientAuth(true);
        break;
      case REQUIRE: 
        engine.setNeedClientAuth(true);
      }
    }
    return apn.wrapperFactory().wrapSslEngine(engine, apn, isServer());
  }
  
  public final JdkApplicationProtocolNegotiator applicationProtocolNegotiator()
  {
    return apn;
  }
  
  static JdkApplicationProtocolNegotiator toNegotiator(ApplicationProtocolConfig config, boolean isServer)
  {
    if (config == null) {
      return JdkDefaultApplicationProtocolNegotiator.INSTANCE;
    }
    switch (config.protocol())
    {
    case NONE: 
      return JdkDefaultApplicationProtocolNegotiator.INSTANCE;
    case ALPN: 
      if (isServer)
      {
        switch (config.selectorFailureBehavior())
        {
        case FATAL_ALERT: 
          return new JdkAlpnApplicationProtocolNegotiator(true, config.supportedProtocols());
        case NO_ADVERTISE: 
          return new JdkAlpnApplicationProtocolNegotiator(false, config.supportedProtocols());
        }
        throw new UnsupportedOperationException("JDK provider does not support " + config.selectorFailureBehavior() + " failure behavior");
      }
      switch (config.selectedListenerFailureBehavior())
      {
      case ACCEPT: 
        return new JdkAlpnApplicationProtocolNegotiator(false, config.supportedProtocols());
      case FATAL_ALERT: 
        return new JdkAlpnApplicationProtocolNegotiator(true, config.supportedProtocols());
      }
      throw new UnsupportedOperationException("JDK provider does not support " + config.selectedListenerFailureBehavior() + " failure behavior");
    case NPN: 
      if (isServer)
      {
        switch (config.selectedListenerFailureBehavior())
        {
        case ACCEPT: 
          return new JdkNpnApplicationProtocolNegotiator(false, config.supportedProtocols());
        case FATAL_ALERT: 
          return new JdkNpnApplicationProtocolNegotiator(true, config.supportedProtocols());
        }
        throw new UnsupportedOperationException("JDK provider does not support " + config.selectedListenerFailureBehavior() + " failure behavior");
      }
      switch (config.selectorFailureBehavior())
      {
      case FATAL_ALERT: 
        return new JdkNpnApplicationProtocolNegotiator(true, config.supportedProtocols());
      case NO_ADVERTISE: 
        return new JdkNpnApplicationProtocolNegotiator(false, config.supportedProtocols());
      }
      throw new UnsupportedOperationException("JDK provider does not support " + config.selectorFailureBehavior() + " failure behavior");
    }
    throw new UnsupportedOperationException("JDK provider does not support " + config.protocol() + " protocol");
  }
  
  @Deprecated
  protected static KeyManagerFactory buildKeyManagerFactory(File certChainFile, File keyFile, String keyPassword, KeyManagerFactory kmf)
    throws UnrecoverableKeyException, KeyStoreException, NoSuchAlgorithmException, NoSuchPaddingException, InvalidKeySpecException, InvalidAlgorithmParameterException, CertificateException, KeyException, IOException
  {
    String algorithm = Security.getProperty("ssl.KeyManagerFactory.algorithm");
    if (algorithm == null) {
      algorithm = "SunX509";
    }
    return buildKeyManagerFactory(certChainFile, algorithm, keyFile, keyPassword, kmf);
  }
  
  @Deprecated
  protected static KeyManagerFactory buildKeyManagerFactory(File certChainFile, String keyAlgorithm, File keyFile, String keyPassword, KeyManagerFactory kmf)
    throws KeyStoreException, NoSuchAlgorithmException, NoSuchPaddingException, InvalidKeySpecException, InvalidAlgorithmParameterException, IOException, CertificateException, KeyException, UnrecoverableKeyException
  {
    return buildKeyManagerFactory(toX509Certificates(certChainFile), keyAlgorithm, toPrivateKey(keyFile, keyPassword), keyPassword, kmf);
  }
}

/* Location:
 * Qualified Name:     io.netty.handler.ssl.JdkSslContext
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */