package io.netty.handler.ssl;

import io.netty.buffer.ByteBuf;
import io.netty.handler.ssl.util.SelfSignedCertificate;
import io.netty.util.ReferenceCountUtil;
import io.netty.util.ReferenceCounted;
import io.netty.util.internal.NativeLibraryLoader;
import io.netty.util.internal.SystemPropertyUtil;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;
import java.io.IOException;
import java.io.InputStream;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.cert.X509Certificate;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Locale;
import java.util.Properties;
import java.util.Set;
import org.apache.tomcat.Apr;
import org.apache.tomcat.jni.Buffer;
import org.apache.tomcat.jni.Library;
import org.apache.tomcat.jni.Pool;
import org.apache.tomcat.jni.SSL;
import org.apache.tomcat.jni.SSLContext;

public final class OpenSsl
{
  private static final InternalLogger logger;
  private static final String LINUX = "linux";
  private static final String UNKNOWN = "unknown";
  private static final Throwable UNAVAILABILITY_CAUSE;
  static final Set<String> AVAILABLE_CIPHER_SUITES;
  private static final Set<String> AVAILABLE_OPENSSL_CIPHER_SUITES;
  private static final Set<String> AVAILABLE_JAVA_CIPHER_SUITES;
  private static final boolean SUPPORTS_KEYMANAGER_FACTORY;
  private static final boolean USE_KEYMANAGER_FACTORY;
  static final String PROTOCOL_SSL_V2_HELLO = "SSLv2Hello";
  static final String PROTOCOL_SSL_V2 = "SSLv2";
  static final String PROTOCOL_SSL_V3 = "SSLv3";
  static final String PROTOCOL_TLS_V1 = "TLSv1";
  static final String PROTOCOL_TLS_V1_1 = "TLSv1.1";
  static final String PROTOCOL_TLS_V1_2 = "TLSv1.2";
  private static final String[] SUPPORTED_PROTOCOLS;
  static final Set<String> SUPPORTED_PROTOCOLS_SET;
  
  static
  {
    logger = InternalLoggerFactory.getInstance(OpenSsl.class);
    
    SUPPORTED_PROTOCOLS = new String[] { "SSLv2Hello", "SSLv2", "SSLv3", "TLSv1", "TLSv1.1", "TLSv1.2" };
    
    SUPPORTED_PROTOCOLS_SET = Collections.unmodifiableSet(new HashSet(Arrays.asList(SUPPORTED_PROTOCOLS)));
    
    Throwable cause = null;
    try
    {
      Class.forName("org.apache.tomcat.jni.SSL", false, OpenSsl.class.getClassLoader());
    }
    catch (ClassNotFoundException t)
    {
      cause = t;
      logger.debug("netty-tcnative not in the classpath; " + OpenSslEngine.class.getSimpleName() + " will be unavailable.");
    }
    if (cause == null)
    {
      try
      {
        loadTcNative();
      }
      catch (Throwable t)
      {
        cause = t;
        logger.debug("Failed to load netty-tcnative; " + OpenSslEngine.class.getSimpleName() + " will be unavailable, unless the " + "application has already loaded the symbols by some other means. " + "See http://netty.io/wiki/forked-tomcat-native.html for more information.", t);
      }
      try
      {
        initializeTcNative();
        
        cause = null;
      }
      catch (Throwable t)
      {
        if (cause == null) {
          cause = t;
        }
        logger.debug("Failed to initialize netty-tcnative; " + OpenSslEngine.class.getSimpleName() + " will be unavailable. " + "See http://netty.io/wiki/forked-tomcat-native.html for more information.", t);
      }
    }
    if ((cause == null) && (!isNettyTcnative()))
    {
      logger.debug("incompatible tcnative in the classpath; " + OpenSslEngine.class.getSimpleName() + " will be unavailable.");
      
      cause = new ClassNotFoundException("incompatible tcnative in the classpath");
    }
    UNAVAILABILITY_CAUSE = cause;
    if (cause == null)
    {
      Set<String> availableOpenSslCipherSuites = new LinkedHashSet(128);
      boolean supportsKeyManagerFactory = false;
      boolean useKeyManagerFactory = false;
      long aprPool = Pool.create(0L);
      try
      {
        long sslCtx = SSLContext.make(aprPool, 31, 1);
        long privateKeyBio = 0L;
        long certBio = 0L;
        try
        {
          SSLContext.setOptions(sslCtx, 4095);
          SSLContext.setCipherSuite(sslCtx, "ALL");
          long ssl = SSL.newSSL(sslCtx, true);
          try
          {
            for (String c : SSL.getCiphers(ssl)) {
              if ((c != null) && (c.length() != 0) && (!availableOpenSslCipherSuites.contains(c))) {
                availableOpenSslCipherSuites.add(c);
              }
            }
            try
            {
              SelfSignedCertificate cert = new SelfSignedCertificate();
              certBio = OpenSslContext.toBIO(new X509Certificate[] { cert.cert() });
              SSL.setCertificateChainBio(ssl, certBio, false);
              supportsKeyManagerFactory = true;
              useKeyManagerFactory = ((Boolean)AccessController.doPrivileged(new PrivilegedAction()
              {
                public Boolean run()
                {
                  return Boolean.valueOf(SystemPropertyUtil.getBoolean("io.netty.handler.ssl.openssl.useKeyManagerFactory", true));
                }
              })).booleanValue();
            }
            catch (Throwable ignore)
            {
              logger.debug("KeyManagerFactory not supported.");
            }
          }
          finally
          {
            SSL.freeSSL(ssl);
            if (privateKeyBio != 0L) {
              SSL.freeBIO(privateKeyBio);
            }
            if (certBio == 0L) {}
          }
        }
        finally
        {
          SSLContext.free(sslCtx);
        }
      }
      catch (Exception e)
      {
        logger.warn("Failed to get the list of available OpenSSL cipher suites.", e);
      }
      finally
      {
        Pool.destroy(aprPool);
      }
      AVAILABLE_OPENSSL_CIPHER_SUITES = Collections.unmodifiableSet(availableOpenSslCipherSuites);
      
      Set<String> availableJavaCipherSuites = new LinkedHashSet(AVAILABLE_OPENSSL_CIPHER_SUITES.size() * 2);
      for (String cipher : AVAILABLE_OPENSSL_CIPHER_SUITES)
      {
        availableJavaCipherSuites.add(CipherSuiteConverter.toJava(cipher, "TLS"));
        availableJavaCipherSuites.add(CipherSuiteConverter.toJava(cipher, "SSL"));
      }
      AVAILABLE_JAVA_CIPHER_SUITES = Collections.unmodifiableSet(availableJavaCipherSuites);
      
      Set<String> availableCipherSuites = new LinkedHashSet(AVAILABLE_OPENSSL_CIPHER_SUITES.size() + AVAILABLE_JAVA_CIPHER_SUITES.size());
      for (String cipher : AVAILABLE_OPENSSL_CIPHER_SUITES) {
        availableCipherSuites.add(cipher);
      }
      for (String cipher : AVAILABLE_JAVA_CIPHER_SUITES) {
        availableCipherSuites.add(cipher);
      }
      AVAILABLE_CIPHER_SUITES = availableCipherSuites;
      SUPPORTS_KEYMANAGER_FACTORY = supportsKeyManagerFactory;
      USE_KEYMANAGER_FACTORY = useKeyManagerFactory;
    }
    else
    {
      AVAILABLE_OPENSSL_CIPHER_SUITES = Collections.emptySet();
      AVAILABLE_JAVA_CIPHER_SUITES = Collections.emptySet();
      AVAILABLE_CIPHER_SUITES = Collections.emptySet();
      SUPPORTS_KEYMANAGER_FACTORY = false;
      USE_KEYMANAGER_FACTORY = false;
    }
  }
  
  private static boolean isNettyTcnative()
  {
    ((Boolean)AccessController.doPrivileged(new PrivilegedAction()
    {
      public Boolean run()
      {
        InputStream is = null;
        try
        {
          is = Apr.class.getResourceAsStream("/org/apache/tomcat/apr.properties");
          Properties props = new Properties();
          props.load(is);
          info = props.getProperty("tcn.info");
          return Boolean.valueOf((info != null) && (info.startsWith("netty-tcnative")));
        }
        catch (Throwable ignore)
        {
          String info;
          return Boolean.valueOf(false);
        }
        finally
        {
          if (is != null) {
            try
            {
              is.close();
            }
            catch (IOException localIOException2) {}
          }
        }
      }
    })).booleanValue();
  }
  
  public static boolean isAvailable()
  {
    return UNAVAILABILITY_CAUSE == null;
  }
  
  public static boolean isAlpnSupported()
  {
    return version() >= 268443648L;
  }
  
  public static int version()
  {
    if (isAvailable()) {
      return SSL.version();
    }
    return -1;
  }
  
  public static String versionString()
  {
    if (isAvailable()) {
      return SSL.versionString();
    }
    return null;
  }
  
  public static void ensureAvailability()
  {
    if (UNAVAILABILITY_CAUSE != null) {
      throw ((Error)new UnsatisfiedLinkError("failed to load the required native library").initCause(UNAVAILABILITY_CAUSE));
    }
  }
  
  public static Throwable unavailabilityCause()
  {
    return UNAVAILABILITY_CAUSE;
  }
  
  @Deprecated
  public static Set<String> availableCipherSuites()
  {
    return availableOpenSslCipherSuites();
  }
  
  public static Set<String> availableOpenSslCipherSuites()
  {
    return AVAILABLE_OPENSSL_CIPHER_SUITES;
  }
  
  public static Set<String> availableJavaCipherSuites()
  {
    return AVAILABLE_JAVA_CIPHER_SUITES;
  }
  
  public static boolean isCipherSuiteAvailable(String cipherSuite)
  {
    String converted = CipherSuiteConverter.toOpenSsl(cipherSuite);
    if (converted != null) {
      cipherSuite = converted;
    }
    return AVAILABLE_OPENSSL_CIPHER_SUITES.contains(cipherSuite);
  }
  
  public static boolean supportsKeyManagerFactory()
  {
    return SUPPORTS_KEYMANAGER_FACTORY;
  }
  
  static boolean useKeyManagerFactory()
  {
    return USE_KEYMANAGER_FACTORY;
  }
  
  static boolean isError(long errorCode)
  {
    return errorCode != 0L;
  }
  
  static long memoryAddress(ByteBuf buf)
  {
    assert (buf.isDirect());
    return buf.hasMemoryAddress() ? buf.memoryAddress() : Buffer.address(buf.nioBuffer());
  }
  
  private static void loadTcNative()
    throws Exception
  {
    String os = normalizeOs(SystemPropertyUtil.get("os.name", ""));
    String arch = normalizeArch(SystemPropertyUtil.get("os.arch", ""));
    
    Set<String> libNames = new LinkedHashSet(3);
    
    libNames.add("netty-tcnative-" + os + '-' + arch);
    if ("linux".equalsIgnoreCase(os)) {
      libNames.add("netty-tcnative-" + os + '-' + arch + "-fedora");
    }
    libNames.add("netty-tcnative");
    
    NativeLibraryLoader.loadFirstAvailable(SSL.class.getClassLoader(), (String[])libNames.toArray(new String[libNames.size()]));
  }
  
  private static void initializeTcNative()
    throws Exception
  {
    Library.initialize("provided");
    SSL.initialize(null);
  }
  
  private static String normalizeOs(String value)
  {
    value = normalize(value);
    if (value.startsWith("aix")) {
      return "aix";
    }
    if (value.startsWith("hpux")) {
      return "hpux";
    }
    if (value.startsWith("os400")) {
      if ((value.length() <= 5) || (!Character.isDigit(value.charAt(5)))) {
        return "os400";
      }
    }
    if (value.startsWith("linux")) {
      return "linux";
    }
    if ((value.startsWith("macosx")) || (value.startsWith("osx"))) {
      return "osx";
    }
    if (value.startsWith("freebsd")) {
      return "freebsd";
    }
    if (value.startsWith("openbsd")) {
      return "openbsd";
    }
    if (value.startsWith("netbsd")) {
      return "netbsd";
    }
    if ((value.startsWith("solaris")) || (value.startsWith("sunos"))) {
      return "sunos";
    }
    if (value.startsWith("windows")) {
      return "windows";
    }
    return "unknown";
  }
  
  private static String normalizeArch(String value)
  {
    value = normalize(value);
    if (value.matches("^(x8664|amd64|ia32e|em64t|x64)$")) {
      return "x86_64";
    }
    if (value.matches("^(x8632|x86|i[3-6]86|ia32|x32)$")) {
      return "x86_32";
    }
    if (value.matches("^(ia64|itanium64)$")) {
      return "itanium_64";
    }
    if (value.matches("^(sparc|sparc32)$")) {
      return "sparc_32";
    }
    if (value.matches("^(sparcv9|sparc64)$")) {
      return "sparc_64";
    }
    if (value.matches("^(arm|arm32)$")) {
      return "arm_32";
    }
    if ("aarch64".equals(value)) {
      return "aarch_64";
    }
    if (value.matches("^(ppc|ppc32)$")) {
      return "ppc_32";
    }
    if ("ppc64".equals(value)) {
      return "ppc_64";
    }
    if ("ppc64le".equals(value)) {
      return "ppcle_64";
    }
    if ("s390".equals(value)) {
      return "s390_32";
    }
    if ("s390x".equals(value)) {
      return "s390_64";
    }
    return "unknown";
  }
  
  private static String normalize(String value)
  {
    return value.toLowerCase(Locale.US).replaceAll("[^a-z0-9]+", "");
  }
  
  static void releaseIfNeeded(ReferenceCounted counted)
  {
    if (counted.refCnt() > 0) {
      ReferenceCountUtil.safeRelease(counted);
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.handler.ssl.OpenSsl
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */