package io.netty.handler.ssl;

import io.netty.buffer.ByteBuf;
import io.netty.buffer.ByteBufAllocator;
import io.netty.buffer.ByteBufUtil;
import io.netty.buffer.Unpooled;
import io.netty.channel.Channel;
import io.netty.channel.ChannelConfig;
import io.netty.channel.ChannelException;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelFutureListener;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.ChannelOutboundHandler;
import io.netty.channel.ChannelPromise;
import io.netty.channel.ChannelPromiseNotifier;
import io.netty.channel.PendingWriteQueue;
import io.netty.handler.codec.ByteToMessageDecoder;
import io.netty.handler.codec.UnsupportedMessageTypeException;
import io.netty.util.concurrent.DefaultPromise;
import io.netty.util.concurrent.EventExecutor;
import io.netty.util.concurrent.Future;
import io.netty.util.concurrent.FutureListener;
import io.netty.util.concurrent.ImmediateExecutor;
import io.netty.util.concurrent.Promise;
import io.netty.util.internal.PlatformDependent;
import io.netty.util.internal.ThrowableUtil;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;
import java.io.IOException;
import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.DatagramChannel;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Executor;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.net.ssl.SSLEngine;
import javax.net.ssl.SSLEngineResult;
import javax.net.ssl.SSLEngineResult.HandshakeStatus;
import javax.net.ssl.SSLEngineResult.Status;
import javax.net.ssl.SSLException;
import javax.net.ssl.SSLSession;

public class SslHandler
  extends ByteToMessageDecoder
  implements ChannelOutboundHandler
{
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(SslHandler.class);
  private static final Pattern IGNORABLE_CLASS_IN_STACK = Pattern.compile("^.*(?:Socket|Datagram|Sctp|Udt)Channel.*$");
  private static final Pattern IGNORABLE_ERROR_MESSAGE = Pattern.compile("^.*(?:connection.*(?:reset|closed|abort|broken)|broken.*pipe).*$", 2);
  private static final SSLException SSLENGINE_CLOSED = (SSLException)ThrowableUtil.unknownStackTrace(new SSLException("SSLEngine closed already"), SslHandler.class, "wrap(...)");
  private static final SSLException HANDSHAKE_TIMED_OUT = (SSLException)ThrowableUtil.unknownStackTrace(new SSLException("handshake timed out"), SslHandler.class, "handshake(...)");
  private static final ClosedChannelException CHANNEL_CLOSED = (ClosedChannelException)ThrowableUtil.unknownStackTrace(new ClosedChannelException(), SslHandler.class, "channelInactive(...)");
  private volatile ChannelHandlerContext ctx;
  private final SSLEngine engine;
  private final int maxPacketBufferSize;
  private final Executor delegatedTaskExecutor;
  private final ByteBuffer[] singleBuffer = new ByteBuffer[1];
  private final boolean wantsDirectBuffer;
  private final boolean wantsLargeOutboundNetworkBuffer;
  private final boolean startTls;
  private boolean sentFirstMessage;
  private boolean flushedBeforeHandshake;
  private boolean readDuringHandshake;
  private PendingWriteQueue pendingUnencryptedWrites;
  private Promise<Channel> handshakePromise = new LazyChannelPromise(null);
  private final LazyChannelPromise sslCloseFuture = new LazyChannelPromise(null);
  private boolean needsFlush;
  private boolean outboundClosed;
  private int packetLength;
  private boolean firedChannelRead;
  private volatile long handshakeTimeoutMillis = 10000L;
  private volatile long closeNotifyTimeoutMillis = 3000L;
  
  public SslHandler(SSLEngine engine)
  {
    this(engine, false);
  }
  
  public SslHandler(SSLEngine engine, boolean startTls)
  {
    this(engine, startTls, ImmediateExecutor.INSTANCE);
  }
  
  @Deprecated
  public SslHandler(SSLEngine engine, Executor delegatedTaskExecutor)
  {
    this(engine, false, delegatedTaskExecutor);
  }
  
  @Deprecated
  public SslHandler(SSLEngine engine, boolean startTls, Executor delegatedTaskExecutor)
  {
    if (engine == null) {
      throw new NullPointerException("engine");
    }
    if (delegatedTaskExecutor == null) {
      throw new NullPointerException("delegatedTaskExecutor");
    }
    this.engine = engine;
    this.delegatedTaskExecutor = delegatedTaskExecutor;
    this.startTls = startTls;
    maxPacketBufferSize = engine.getSession().getPacketBufferSize();
    
    boolean opensslEngine = engine instanceof OpenSslEngine;
    wantsDirectBuffer = opensslEngine;
    wantsLargeOutboundNetworkBuffer = (!opensslEngine);
    
    setCumulator(opensslEngine ? COMPOSITE_CUMULATOR : MERGE_CUMULATOR);
  }
  
  public long getHandshakeTimeoutMillis()
  {
    return handshakeTimeoutMillis;
  }
  
  public void setHandshakeTimeout(long handshakeTimeout, TimeUnit unit)
  {
    if (unit == null) {
      throw new NullPointerException("unit");
    }
    setHandshakeTimeoutMillis(unit.toMillis(handshakeTimeout));
  }
  
  public void setHandshakeTimeoutMillis(long handshakeTimeoutMillis)
  {
    if (handshakeTimeoutMillis < 0L) {
      throw new IllegalArgumentException("handshakeTimeoutMillis: " + handshakeTimeoutMillis + " (expected: >= 0)");
    }
    this.handshakeTimeoutMillis = handshakeTimeoutMillis;
  }
  
  public long getCloseNotifyTimeoutMillis()
  {
    return closeNotifyTimeoutMillis;
  }
  
  public void setCloseNotifyTimeout(long closeNotifyTimeout, TimeUnit unit)
  {
    if (unit == null) {
      throw new NullPointerException("unit");
    }
    setCloseNotifyTimeoutMillis(unit.toMillis(closeNotifyTimeout));
  }
  
  public void setCloseNotifyTimeoutMillis(long closeNotifyTimeoutMillis)
  {
    if (closeNotifyTimeoutMillis < 0L) {
      throw new IllegalArgumentException("closeNotifyTimeoutMillis: " + closeNotifyTimeoutMillis + " (expected: >= 0)");
    }
    this.closeNotifyTimeoutMillis = closeNotifyTimeoutMillis;
  }
  
  public SSLEngine engine()
  {
    return engine;
  }
  
  public String applicationProtocol()
  {
    SSLSession sess = engine().getSession();
    if (!(sess instanceof ApplicationProtocolAccessor)) {
      return null;
    }
    return ((ApplicationProtocolAccessor)sess).getApplicationProtocol();
  }
  
  public Future<Channel> handshakeFuture()
  {
    return handshakePromise;
  }
  
  public ChannelFuture close()
  {
    return close(ctx.newPromise());
  }
  
  public ChannelFuture close(final ChannelPromise future)
  {
    final ChannelHandlerContext ctx = this.ctx;
    ctx.executor().execute(new Runnable()
    {
      public void run()
      {
        outboundClosed = true;
        engine.closeOutbound();
        try
        {
          write(ctx, Unpooled.EMPTY_BUFFER, future);
          flush(ctx);
        }
        catch (Exception e)
        {
          if (!future.tryFailure(e)) {
            SslHandler.logger.warn("{} flush() raised a masked exception.", ctx.channel(), e);
          }
        }
      }
    });
    return future;
  }
  
  public Future<Channel> sslCloseFuture()
  {
    return sslCloseFuture;
  }
  
  public void handlerRemoved0(ChannelHandlerContext ctx)
    throws Exception
  {
    if (!pendingUnencryptedWrites.isEmpty()) {
      pendingUnencryptedWrites.removeAndFailAll(new ChannelException("Pending write on removal of SslHandler"));
    }
    if ((engine instanceof ReferenceCountedOpenSslEngine)) {
      ((ReferenceCountedOpenSslEngine)engine).release();
    }
  }
  
  public void bind(ChannelHandlerContext ctx, SocketAddress localAddress, ChannelPromise promise)
    throws Exception
  {
    ctx.bind(localAddress, promise);
  }
  
  public void connect(ChannelHandlerContext ctx, SocketAddress remoteAddress, SocketAddress localAddress, ChannelPromise promise)
    throws Exception
  {
    ctx.connect(remoteAddress, localAddress, promise);
  }
  
  public void deregister(ChannelHandlerContext ctx, ChannelPromise promise)
    throws Exception
  {
    ctx.deregister(promise);
  }
  
  public void disconnect(ChannelHandlerContext ctx, ChannelPromise promise)
    throws Exception
  {
    closeOutboundAndChannel(ctx, promise, true);
  }
  
  public void close(ChannelHandlerContext ctx, ChannelPromise promise)
    throws Exception
  {
    closeOutboundAndChannel(ctx, promise, false);
  }
  
  public void read(ChannelHandlerContext ctx)
    throws Exception
  {
    if (!handshakePromise.isDone()) {
      readDuringHandshake = true;
    }
    ctx.read();
  }
  
  public void write(ChannelHandlerContext ctx, Object msg, ChannelPromise promise)
    throws Exception
  {
    if (!(msg instanceof ByteBuf))
    {
      promise.setFailure(new UnsupportedMessageTypeException(msg, new Class[] { ByteBuf.class }));
      return;
    }
    pendingUnencryptedWrites.add(msg, promise);
  }
  
  public void flush(ChannelHandlerContext ctx)
    throws Exception
  {
    if ((startTls) && (!sentFirstMessage))
    {
      sentFirstMessage = true;
      pendingUnencryptedWrites.removeAndWriteAll();
      ctx.flush();
      return;
    }
    if (pendingUnencryptedWrites.isEmpty()) {
      pendingUnencryptedWrites.add(Unpooled.EMPTY_BUFFER, ctx.newPromise());
    }
    if (!handshakePromise.isDone()) {
      flushedBeforeHandshake = true;
    }
    try
    {
      wrap(ctx, false);
    }
    catch (Throwable cause)
    {
      pendingUnencryptedWrites.removeAndFailAll(cause);
      
      PlatformDependent.throwException(cause);
    }
    finally
    {
      ctx.flush();
    }
  }
  
  private void wrap(ChannelHandlerContext ctx, boolean inUnwrap)
    throws SSLException
  {
    ByteBuf out = null;
    ChannelPromise promise = null;
    ByteBufAllocator alloc = ctx.alloc();
    boolean needUnwrap = false;
    try
    {
      for (;;)
      {
        Object msg = pendingUnencryptedWrites.current();
        if (msg == null) {
          break;
        }
        ByteBuf buf = (ByteBuf)msg;
        if (out == null) {
          out = allocateOutNetBuf(ctx, buf.readableBytes());
        }
        SSLEngineResult result = wrap(alloc, engine, buf, out);
        if (result.getStatus() == SSLEngineResult.Status.CLOSED)
        {
          pendingUnencryptedWrites.removeAndFailAll(SSLENGINE_CLOSED); return;
        }
        if (!buf.isReadable()) {
          promise = pendingUnencryptedWrites.remove();
        } else {
          promise = null;
        }
        switch (result.getHandshakeStatus())
        {
        case NEED_TASK: 
          runDelegatedTasks();
          break;
        case FINISHED: 
          setHandshakeSuccess();
        case NOT_HANDSHAKING: 
          setHandshakeSuccessIfStillHandshaking();
        case NEED_WRAP: 
          finishWrap(ctx, out, promise, inUnwrap, false);
          promise = null;
          out = null;
          break;
        case NEED_UNWRAP: 
          needUnwrap = true; return;
        default: 
          throw new IllegalStateException("Unknown handshake status: " + result.getHandshakeStatus());
        }
      }
    }
    catch (SSLException e)
    {
      setHandshakeFailure(ctx, e);
      throw e;
    }
    finally
    {
      finishWrap(ctx, out, promise, inUnwrap, needUnwrap);
    }
  }
  
  private void finishWrap(ChannelHandlerContext ctx, ByteBuf out, ChannelPromise promise, boolean inUnwrap, boolean needUnwrap)
  {
    if (out == null)
    {
      out = Unpooled.EMPTY_BUFFER;
    }
    else if (!out.isReadable())
    {
      out.release();
      out = Unpooled.EMPTY_BUFFER;
    }
    if (promise != null) {
      ctx.write(out, promise);
    } else {
      ctx.write(out);
    }
    if (inUnwrap) {
      needsFlush = true;
    }
    if (needUnwrap) {
      readIfNeeded(ctx);
    }
  }
  
  private void wrapNonAppData(ChannelHandlerContext ctx, boolean inUnwrap)
    throws SSLException
  {
    ByteBuf out = null;
    ByteBufAllocator alloc = ctx.alloc();
    try
    {
      for (;;)
      {
        if (out == null) {
          out = allocateOutNetBuf(ctx, 0);
        }
        SSLEngineResult result = wrap(alloc, engine, Unpooled.EMPTY_BUFFER, out);
        if (result.bytesProduced() > 0)
        {
          ctx.write(out);
          if (inUnwrap) {
            needsFlush = true;
          }
          out = null;
        }
        switch (result.getHandshakeStatus())
        {
        case FINISHED: 
          setHandshakeSuccess();
          break;
        case NEED_TASK: 
          runDelegatedTasks();
          break;
        case NEED_UNWRAP: 
          if (!inUnwrap) {
            unwrapNonAppData(ctx);
          }
          break;
        case NEED_WRAP: 
          break;
        case NOT_HANDSHAKING: 
          setHandshakeSuccessIfStillHandshaking();
          if (!inUnwrap) {
            unwrapNonAppData(ctx);
          }
          break;
        default: 
          throw new IllegalStateException("Unknown handshake status: " + result.getHandshakeStatus());
        }
        if (result.bytesProduced() == 0) {
          break;
        }
        if ((result.bytesConsumed() == 0) && (result.getHandshakeStatus() == SSLEngineResult.HandshakeStatus.NOT_HANDSHAKING)) {
          break;
        }
      }
    }
    catch (SSLException e)
    {
      setHandshakeFailure(ctx, e);
      
      flushIfNeeded(ctx);
      throw e;
    }
    finally
    {
      if (out != null) {
        out.release();
      }
    }
  }
  
  /* Error */
  private SSLEngineResult wrap(ByteBufAllocator alloc, SSLEngine engine, ByteBuf in, ByteBuf out)
    throws SSLException
  {
    // Byte code:
    //   0: aconst_null
    //   1: astore 5
    //   3: aload_3
    //   4: invokevirtual 125	io/netty/buffer/ByteBuf:readerIndex	()I
    //   7: istore 6
    //   9: aload_3
    //   10: invokevirtual 94	io/netty/buffer/ByteBuf:readableBytes	()I
    //   13: istore 7
    //   15: aload_3
    //   16: invokevirtual 126	io/netty/buffer/ByteBuf:isDirect	()Z
    //   19: ifne +10 -> 29
    //   22: aload_0
    //   23: getfield 35	io/netty/handler/ssl/SslHandler:wantsDirectBuffer	Z
    //   26: ifne +48 -> 74
    //   29: aload_3
    //   30: instanceof 127
    //   33: ifne +32 -> 65
    //   36: aload_3
    //   37: invokevirtual 128	io/netty/buffer/ByteBuf:nioBufferCount	()I
    //   40: iconst_1
    //   41: if_icmpne +24 -> 65
    //   44: aload_0
    //   45: getfield 14	io/netty/handler/ssl/SslHandler:singleBuffer	[Ljava/nio/ByteBuffer;
    //   48: astore 8
    //   50: aload 8
    //   52: iconst_0
    //   53: aload_3
    //   54: iload 6
    //   56: iload 7
    //   58: invokevirtual 129	io/netty/buffer/ByteBuf:internalNioBuffer	(II)Ljava/nio/ByteBuffer;
    //   61: aastore
    //   62: goto +51 -> 113
    //   65: aload_3
    //   66: invokevirtual 130	io/netty/buffer/ByteBuf:nioBuffers	()[Ljava/nio/ByteBuffer;
    //   69: astore 8
    //   71: goto +42 -> 113
    //   74: aload_1
    //   75: iload 7
    //   77: invokeinterface 131 2 0
    //   82: astore 5
    //   84: aload 5
    //   86: aload_3
    //   87: iload 6
    //   89: iload 7
    //   91: invokevirtual 132	io/netty/buffer/ByteBuf:writeBytes	(Lio/netty/buffer/ByteBuf;II)Lio/netty/buffer/ByteBuf;
    //   94: pop
    //   95: aload_0
    //   96: getfield 14	io/netty/handler/ssl/SslHandler:singleBuffer	[Ljava/nio/ByteBuffer;
    //   99: astore 8
    //   101: aload 8
    //   103: iconst_0
    //   104: aload 5
    //   106: iconst_0
    //   107: iload 7
    //   109: invokevirtual 129	io/netty/buffer/ByteBuf:internalNioBuffer	(II)Ljava/nio/ByteBuffer;
    //   112: aastore
    //   113: aload 4
    //   115: aload 4
    //   117: invokevirtual 133	io/netty/buffer/ByteBuf:writerIndex	()I
    //   120: aload 4
    //   122: invokevirtual 134	io/netty/buffer/ByteBuf:writableBytes	()I
    //   125: invokevirtual 135	io/netty/buffer/ByteBuf:nioBuffer	(II)Ljava/nio/ByteBuffer;
    //   128: astore 9
    //   130: aload_2
    //   131: aload 8
    //   133: aload 9
    //   135: invokevirtual 136	javax/net/ssl/SSLEngine:wrap	([Ljava/nio/ByteBuffer;Ljava/nio/ByteBuffer;)Ljavax/net/ssl/SSLEngineResult;
    //   138: astore 10
    //   140: aload_3
    //   141: aload 10
    //   143: invokevirtual 122	javax/net/ssl/SSLEngineResult:bytesConsumed	()I
    //   146: invokevirtual 137	io/netty/buffer/ByteBuf:skipBytes	(I)Lio/netty/buffer/ByteBuf;
    //   149: pop
    //   150: aload 4
    //   152: aload 4
    //   154: invokevirtual 133	io/netty/buffer/ByteBuf:writerIndex	()I
    //   157: aload 10
    //   159: invokevirtual 120	javax/net/ssl/SSLEngineResult:bytesProduced	()I
    //   162: iadd
    //   163: invokevirtual 138	io/netty/buffer/ByteBuf:writerIndex	(I)Lio/netty/buffer/ByteBuf;
    //   166: pop
    //   167: getstatic 139	io/netty/handler/ssl/SslHandler$9:$SwitchMap$javax$net$ssl$SSLEngineResult$Status	[I
    //   170: aload 10
    //   172: invokevirtual 97	javax/net/ssl/SSLEngineResult:getStatus	()Ljavax/net/ssl/SSLEngineResult$Status;
    //   175: invokevirtual 140	javax/net/ssl/SSLEngineResult$Status:ordinal	()I
    //   178: iaload
    //   179: lookupswitch	default:+30->209, 1:+17->196
    //   196: aload 4
    //   198: aload_0
    //   199: getfield 33	io/netty/handler/ssl/SslHandler:maxPacketBufferSize	I
    //   202: invokevirtual 141	io/netty/buffer/ByteBuf:ensureWritable	(I)Lio/netty/buffer/ByteBuf;
    //   205: pop
    //   206: goto +28 -> 234
    //   209: aload 10
    //   211: astore 11
    //   213: aload_0
    //   214: getfield 14	io/netty/handler/ssl/SslHandler:singleBuffer	[Ljava/nio/ByteBuffer;
    //   217: iconst_0
    //   218: aconst_null
    //   219: aastore
    //   220: aload 5
    //   222: ifnull +9 -> 231
    //   225: aload 5
    //   227: invokevirtual 115	io/netty/buffer/ByteBuf:release	()Z
    //   230: pop
    //   231: aload 11
    //   233: areturn
    //   234: goto -121 -> 113
    //   237: astore 12
    //   239: aload_0
    //   240: getfield 14	io/netty/handler/ssl/SslHandler:singleBuffer	[Ljava/nio/ByteBuffer;
    //   243: iconst_0
    //   244: aconst_null
    //   245: aastore
    //   246: aload 5
    //   248: ifnull +9 -> 257
    //   251: aload 5
    //   253: invokevirtual 115	io/netty/buffer/ByteBuf:release	()Z
    //   256: pop
    //   257: aload 12
    //   259: athrow
    // Line number table:
    //   Java source line #663	-> byte code offset #0
    //   Java source line #665	-> byte code offset #3
    //   Java source line #666	-> byte code offset #9
    //   Java source line #671	-> byte code offset #15
    //   Java source line #676	-> byte code offset #29
    //   Java source line #677	-> byte code offset #44
    //   Java source line #680	-> byte code offset #50
    //   Java source line #682	-> byte code offset #65
    //   Java source line #688	-> byte code offset #74
    //   Java source line #689	-> byte code offset #84
    //   Java source line #690	-> byte code offset #95
    //   Java source line #691	-> byte code offset #101
    //   Java source line #695	-> byte code offset #113
    //   Java source line #696	-> byte code offset #130
    //   Java source line #697	-> byte code offset #140
    //   Java source line #698	-> byte code offset #150
    //   Java source line #700	-> byte code offset #167
    //   Java source line #702	-> byte code offset #196
    //   Java source line #703	-> byte code offset #206
    //   Java source line #705	-> byte code offset #209
    //   Java source line #710	-> byte code offset #213
    //   Java source line #712	-> byte code offset #220
    //   Java source line #713	-> byte code offset #225
    //   Java source line #707	-> byte code offset #234
    //   Java source line #710	-> byte code offset #237
    //   Java source line #712	-> byte code offset #246
    //   Java source line #713	-> byte code offset #251
    // Local variable table:
    //   start	length	slot	name	signature
    //   0	260	0	this	SslHandler
    //   0	260	1	alloc	ByteBufAllocator
    //   0	260	2	engine	SSLEngine
    //   0	260	3	in	ByteBuf
    //   0	260	4	out	ByteBuf
    //   1	251	5	newDirectIn	ByteBuf
    //   7	81	6	readerIndex	int
    //   13	95	7	readableBytes	int
    //   48	3	8	in0	ByteBuffer[]
    //   69	3	8	in0	ByteBuffer[]
    //   99	33	8	in0	ByteBuffer[]
    //   128	6	9	out0	ByteBuffer
    //   138	72	10	result	SSLEngineResult
    //   211	21	11	localSSLEngineResult1	SSLEngineResult
    //   237	21	12	localObject	Object
    // Exception table:
    //   from	to	target	type
    //   3	213	237	finally
    //   234	239	237	finally
  }
  
  public void channelInactive(ChannelHandlerContext ctx)
    throws Exception
  {
    setHandshakeFailure(ctx, CHANNEL_CLOSED, !outboundClosed);
    super.channelInactive(ctx);
  }
  
  public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause)
    throws Exception
  {
    if (ignoreException(cause))
    {
      if (logger.isDebugEnabled()) {
        logger.debug("{} Swallowing a harmless 'connection reset by peer / broken pipe' error that occurred while writing close_notify in response to the peer's close_notify", ctx.channel(), cause);
      }
      if (ctx.channel().isActive()) {
        ctx.close();
      }
    }
    else
    {
      ctx.fireExceptionCaught(cause);
    }
  }
  
  private boolean ignoreException(Throwable t)
  {
    if ((!(t instanceof SSLException)) && ((t instanceof IOException)) && (sslCloseFuture.isDone()))
    {
      String message = String.valueOf(t.getMessage()).toLowerCase();
      if (IGNORABLE_ERROR_MESSAGE.matcher(message).matches()) {
        return true;
      }
      StackTraceElement[] elements = t.getStackTrace();
      for (StackTraceElement element : elements)
      {
        String classname = element.getClassName();
        String methodname = element.getMethodName();
        if (!classname.startsWith("io.netty.")) {
          if ("read".equals(methodname))
          {
            if (IGNORABLE_CLASS_IN_STACK.matcher(classname).matches()) {
              return true;
            }
            try
            {
              Class<?> clazz = PlatformDependent.getClassLoader(getClass()).loadClass(classname);
              if ((SocketChannel.class.isAssignableFrom(clazz)) || (DatagramChannel.class.isAssignableFrom(clazz))) {
                return true;
              }
              if ((PlatformDependent.javaVersion() >= 7) && ("com.sun.nio.sctp.SctpChannel".equals(clazz.getSuperclass().getName()))) {
                return true;
              }
            }
            catch (ClassNotFoundException localClassNotFoundException) {}
          }
        }
      }
    }
    return false;
  }
  
  public static boolean isEncrypted(ByteBuf buffer)
  {
    if (buffer.readableBytes() < 5) {
      throw new IllegalArgumentException("buffer must have at least 5 readable bytes");
    }
    return SslUtils.getEncryptedPacketLength(buffer, buffer.readerIndex()) != -1;
  }
  
  protected void decode(ChannelHandlerContext ctx, ByteBuf in, List<Object> out)
    throws SSLException
  {
    int startOffset = in.readerIndex();
    int endOffset = in.writerIndex();
    int offset = startOffset;
    int totalLength = 0;
    if (this.packetLength > 0)
    {
      if (endOffset - startOffset < this.packetLength) {
        return;
      }
      offset += this.packetLength;
      totalLength = this.packetLength;
      this.packetLength = 0;
    }
    boolean nonSslRecord = false;
    while (totalLength < 18713)
    {
      int readableBytes = endOffset - offset;
      if (readableBytes < 5) {
        break;
      }
      int packetLength = SslUtils.getEncryptedPacketLength(in, offset);
      if (packetLength == -1)
      {
        nonSslRecord = true;
        break;
      }
      assert (packetLength > 0);
      if (packetLength > readableBytes)
      {
        this.packetLength = packetLength;
        break;
      }
      int newTotalLength = totalLength + packetLength;
      if (newTotalLength > 18713) {
        break;
      }
      offset += packetLength;
      totalLength = newTotalLength;
    }
    if (totalLength > 0)
    {
      boolean decoded = false;
      
      in.skipBytes(totalLength);
      decoded = unwrap(ctx, in, startOffset, totalLength);
      if (!firedChannelRead) {
        firedChannelRead = decoded;
      }
    }
    if (nonSslRecord)
    {
      NotSslRecordException e = new NotSslRecordException("not an SSL/TLS record: " + ByteBufUtil.hexDump(in));
      
      in.skipBytes(in.readableBytes());
      ctx.fireExceptionCaught(e);
      setHandshakeFailure(ctx, e);
    }
  }
  
  public void channelReadComplete(ChannelHandlerContext ctx)
    throws Exception
  {
    discardSomeReadBytes();
    
    flushIfNeeded(ctx);
    readIfNeeded(ctx);
    
    firedChannelRead = false;
    ctx.fireChannelReadComplete();
  }
  
  private void readIfNeeded(ChannelHandlerContext ctx)
  {
    if ((!ctx.channel().config().isAutoRead()) && ((!firedChannelRead) || (!handshakePromise.isDone()))) {
      ctx.read();
    }
  }
  
  private void flushIfNeeded(ChannelHandlerContext ctx)
  {
    if (needsFlush)
    {
      needsFlush = false;
      ctx.flush();
    }
  }
  
  private void unwrapNonAppData(ChannelHandlerContext ctx)
    throws SSLException
  {
    unwrap(ctx, Unpooled.EMPTY_BUFFER, 0, 0);
  }
  
  private boolean unwrap(ChannelHandlerContext ctx, ByteBuf packet, int offset, int length)
    throws SSLException
  {
    boolean decoded = false;
    boolean wrapLater = false;
    boolean notifyClosure = false;
    ByteBuf decodeOut = allocate(ctx, length);
    try
    {
      for (;;)
      {
        SSLEngineResult result = unwrap(engine, packet, offset, length, decodeOut);
        SSLEngineResult.Status status = result.getStatus();
        SSLEngineResult.HandshakeStatus handshakeStatus = result.getHandshakeStatus();
        int produced = result.bytesProduced();
        int consumed = result.bytesConsumed();
        
        offset += consumed;
        length -= consumed;
        switch (status)
        {
        case BUFFER_OVERFLOW: 
          int readableBytes = decodeOut.readableBytes();
          if (readableBytes > 0)
          {
            decoded = true;
            ctx.fireChannelRead(decodeOut);
          }
          else
          {
            decodeOut.release();
          }
          decodeOut = allocate(ctx, engine.getSession().getApplicationBufferSize() - readableBytes);
          break;
        case CLOSED: 
          notifyClosure = true;
          break;
        default: 
          switch (handshakeStatus)
          {
          case NEED_UNWRAP: 
            break;
          case NEED_WRAP: 
            wrapNonAppData(ctx, true);
            break;
          case NEED_TASK: 
            runDelegatedTasks();
            break;
          case FINISHED: 
            setHandshakeSuccess();
            wrapLater = true;
            
            break;
          case NOT_HANDSHAKING: 
            if (setHandshakeSuccessIfStillHandshaking())
            {
              wrapLater = true;
            }
            else if (flushedBeforeHandshake)
            {
              flushedBeforeHandshake = false;
              wrapLater = true;
            }
            break;
          default: 
            throw new IllegalStateException("unknown handshake status: " + handshakeStatus);
            if ((status == SSLEngineResult.Status.BUFFER_UNDERFLOW) || ((consumed == 0) && (produced == 0)))
            {
              if (handshakeStatus != SSLEngineResult.HandshakeStatus.NEED_UNWRAP) {
                break label341;
              }
              readIfNeeded(ctx);
              break label341;
            }
            break;
          }
          break;
        }
      }
      label341:
      if (wrapLater) {
        wrap(ctx, true);
      }
      if (notifyClosure) {
        sslCloseFuture.trySuccess(ctx.channel());
      }
    }
    catch (SSLException e)
    {
      setHandshakeFailure(ctx, e);
      throw e;
    }
    finally
    {
      if (decodeOut.isReadable())
      {
        decoded = true;
        
        ctx.fireChannelRead(decodeOut);
      }
      else
      {
        decodeOut.release();
      }
    }
    return decoded;
  }
  
  /* Error */
  private SSLEngineResult unwrap(SSLEngine engine, ByteBuf in, int readerIndex, int len, ByteBuf out)
    throws SSLException
  {
    // Byte code:
    //   0: aload_2
    //   1: invokevirtual 128	io/netty/buffer/ByteBuf:nioBufferCount	()I
    //   4: istore 6
    //   6: aload 5
    //   8: invokevirtual 133	io/netty/buffer/ByteBuf:writerIndex	()I
    //   11: istore 7
    //   13: aload_1
    //   14: instanceof 34
    //   17: ifeq +90 -> 107
    //   20: iload 6
    //   22: iconst_1
    //   23: if_icmple +84 -> 107
    //   26: aload_1
    //   27: checkcast 34	io/netty/handler/ssl/OpenSslEngine
    //   30: astore 9
    //   32: aload_0
    //   33: getfield 14	io/netty/handler/ssl/SslHandler:singleBuffer	[Ljava/nio/ByteBuffer;
    //   36: iconst_0
    //   37: aload 5
    //   39: iload 7
    //   41: aload 5
    //   43: invokevirtual 134	io/netty/buffer/ByteBuf:writableBytes	()I
    //   46: invokestatic 205	io/netty/handler/ssl/SslHandler:toByteBuffer	(Lio/netty/buffer/ByteBuf;II)Ljava/nio/ByteBuffer;
    //   49: aastore
    //   50: aload 9
    //   52: aload_2
    //   53: iload_3
    //   54: iload 4
    //   56: invokevirtual 206	io/netty/buffer/ByteBuf:nioBuffers	(II)[Ljava/nio/ByteBuffer;
    //   59: aload_0
    //   60: getfield 14	io/netty/handler/ssl/SslHandler:singleBuffer	[Ljava/nio/ByteBuffer;
    //   63: invokevirtual 207	io/netty/handler/ssl/OpenSslEngine:unwrap	([Ljava/nio/ByteBuffer;[Ljava/nio/ByteBuffer;)Ljavax/net/ssl/SSLEngineResult;
    //   66: astore 8
    //   68: aload 5
    //   70: iload 7
    //   72: aload 8
    //   74: invokevirtual 120	javax/net/ssl/SSLEngineResult:bytesProduced	()I
    //   77: iadd
    //   78: invokevirtual 138	io/netty/buffer/ByteBuf:writerIndex	(I)Lio/netty/buffer/ByteBuf;
    //   81: pop
    //   82: aload_0
    //   83: getfield 14	io/netty/handler/ssl/SslHandler:singleBuffer	[Ljava/nio/ByteBuffer;
    //   86: iconst_0
    //   87: aconst_null
    //   88: aastore
    //   89: goto +15 -> 104
    //   92: astore 10
    //   94: aload_0
    //   95: getfield 14	io/netty/handler/ssl/SslHandler:singleBuffer	[Ljava/nio/ByteBuffer;
    //   98: iconst_0
    //   99: aconst_null
    //   100: aastore
    //   101: aload 10
    //   103: athrow
    //   104: goto +28 -> 132
    //   107: aload_1
    //   108: aload_2
    //   109: iload_3
    //   110: iload 4
    //   112: invokestatic 205	io/netty/handler/ssl/SslHandler:toByteBuffer	(Lio/netty/buffer/ByteBuf;II)Ljava/nio/ByteBuffer;
    //   115: aload 5
    //   117: iload 7
    //   119: aload 5
    //   121: invokevirtual 134	io/netty/buffer/ByteBuf:writableBytes	()I
    //   124: invokestatic 205	io/netty/handler/ssl/SslHandler:toByteBuffer	(Lio/netty/buffer/ByteBuf;II)Ljava/nio/ByteBuffer;
    //   127: invokevirtual 208	javax/net/ssl/SSLEngine:unwrap	(Ljava/nio/ByteBuffer;Ljava/nio/ByteBuffer;)Ljavax/net/ssl/SSLEngineResult;
    //   130: astore 8
    //   132: aload 5
    //   134: iload 7
    //   136: aload 8
    //   138: invokevirtual 120	javax/net/ssl/SSLEngineResult:bytesProduced	()I
    //   141: iadd
    //   142: invokevirtual 138	io/netty/buffer/ByteBuf:writerIndex	(I)Lio/netty/buffer/ByteBuf;
    //   145: pop
    //   146: aload 8
    //   148: areturn
    // Line number table:
    //   Java source line #1076	-> byte code offset #0
    //   Java source line #1077	-> byte code offset #6
    //   Java source line #1079	-> byte code offset #13
    //   Java source line #1085	-> byte code offset #26
    //   Java source line #1087	-> byte code offset #32
    //   Java source line #1088	-> byte code offset #50
    //   Java source line #1089	-> byte code offset #68
    //   Java source line #1091	-> byte code offset #82
    //   Java source line #1092	-> byte code offset #89
    //   Java source line #1091	-> byte code offset #92
    //   Java source line #1093	-> byte code offset #104
    //   Java source line #1094	-> byte code offset #107
    //   Java source line #1097	-> byte code offset #132
    //   Java source line #1098	-> byte code offset #146
    // Local variable table:
    //   start	length	slot	name	signature
    //   0	149	0	this	SslHandler
    //   0	149	1	engine	SSLEngine
    //   0	149	2	in	ByteBuf
    //   0	149	3	readerIndex	int
    //   0	149	4	len	int
    //   0	149	5	out	ByteBuf
    //   4	17	6	nioBufferCount	int
    //   11	124	7	writerIndex	int
    //   66	7	8	result	SSLEngineResult
    //   104	1	8	result	SSLEngineResult
    //   130	17	8	result	SSLEngineResult
    //   30	21	9	opensslEngine	OpenSslEngine
    //   92	10	10	localObject	Object
    // Exception table:
    //   from	to	target	type
    //   32	82	92	finally
    //   92	94	92	finally
  }
  
  private static ByteBuffer toByteBuffer(ByteBuf out, int index, int len)
  {
    return out.nioBufferCount() == 1 ? out.internalNioBuffer(index, len) : out.nioBuffer(index, len);
  }
  
  private void runDelegatedTasks()
  {
    if (delegatedTaskExecutor == ImmediateExecutor.INSTANCE) {
      for (;;)
      {
        Runnable task = engine.getDelegatedTask();
        if (task == null) {
          break;
        }
        task.run();
      }
    }
    final List<Runnable> tasks = new ArrayList(2);
    for (;;)
    {
      Runnable task = engine.getDelegatedTask();
      if (task == null) {
        break;
      }
      tasks.add(task);
    }
    if (tasks.isEmpty()) {
      return;
    }
    final CountDownLatch latch = new CountDownLatch(1);
    delegatedTaskExecutor.execute(new Runnable()
    {
      public void run()
      {
        try
        {
          for (Runnable task : tasks) {
            task.run();
          }
        }
        catch (Exception e)
        {
          ctx.fireExceptionCaught(e);
        }
        finally
        {
          latch.countDown();
        }
      }
    });
    boolean interrupted = false;
    while (latch.getCount() != 0L) {
      try
      {
        latch.await();
      }
      catch (InterruptedException e)
      {
        interrupted = true;
      }
    }
    if (interrupted) {
      Thread.currentThread().interrupt();
    }
  }
  
  private boolean setHandshakeSuccessIfStillHandshaking()
  {
    if (!handshakePromise.isDone())
    {
      setHandshakeSuccess();
      return true;
    }
    return false;
  }
  
  private void setHandshakeSuccess()
  {
    handshakePromise.trySuccess(ctx.channel());
    if (logger.isDebugEnabled()) {
      logger.debug("{} HANDSHAKEN: {}", ctx.channel(), engine.getSession().getCipherSuite());
    }
    ctx.fireUserEventTriggered(SslHandshakeCompletionEvent.SUCCESS);
    if ((readDuringHandshake) && (!ctx.channel().config().isAutoRead()))
    {
      readDuringHandshake = false;
      ctx.read();
    }
  }
  
  private void setHandshakeFailure(ChannelHandlerContext ctx, Throwable cause)
  {
    setHandshakeFailure(ctx, cause, true);
  }
  
  private void setHandshakeFailure(ChannelHandlerContext ctx, Throwable cause, boolean closeInbound)
  {
    engine.closeOutbound();
    if (closeInbound) {
      try
      {
        engine.closeInbound();
      }
      catch (SSLException e)
      {
        String msg = e.getMessage();
        if ((msg == null) || (!msg.contains("possible truncation attack"))) {
          logger.debug("{} SSLEngine.closeInbound() raised an exception.", ctx.channel(), e);
        }
      }
    }
    notifyHandshakeFailure(cause);
    pendingUnencryptedWrites.removeAndFailAll(cause);
  }
  
  private void notifyHandshakeFailure(Throwable cause)
  {
    if (handshakePromise.tryFailure(cause)) {
      SslUtils.notifyHandshakeFailure(ctx, cause);
    }
  }
  
  private void closeOutboundAndChannel(ChannelHandlerContext ctx, ChannelPromise promise, boolean disconnect)
    throws Exception
  {
    if (!ctx.channel().isActive())
    {
      if (disconnect) {
        ctx.disconnect(promise);
      } else {
        ctx.close(promise);
      }
      return;
    }
    outboundClosed = true;
    engine.closeOutbound();
    
    ChannelPromise closeNotifyFuture = ctx.newPromise();
    try
    {
      write(ctx, Unpooled.EMPTY_BUFFER, closeNotifyFuture);
      flush(ctx);
    }
    finally
    {
      safeClose(ctx, closeNotifyFuture, promise);
    }
  }
  
  public void handlerAdded(ChannelHandlerContext ctx)
    throws Exception
  {
    this.ctx = ctx;
    pendingUnencryptedWrites = new PendingWriteQueue(ctx);
    if ((ctx.channel().isActive()) && (engine.getUseClientMode())) {
      handshake(null);
    }
  }
  
  public Future<Channel> renegotiate()
  {
    ChannelHandlerContext ctx = this.ctx;
    if (ctx == null) {
      throw new IllegalStateException();
    }
    return renegotiate(ctx.executor().newPromise());
  }
  
  public Future<Channel> renegotiate(final Promise<Channel> promise)
  {
    if (promise == null) {
      throw new NullPointerException("promise");
    }
    ChannelHandlerContext ctx = this.ctx;
    if (ctx == null) {
      throw new IllegalStateException();
    }
    EventExecutor executor = ctx.executor();
    if (!executor.inEventLoop())
    {
      executor.execute(new Runnable()
      {
        public void run()
        {
          SslHandler.this.handshake(promise);
        }
      });
      return promise;
    }
    handshake(promise);
    return promise;
  }
  
  private void handshake(final Promise<Channel> newHandshakePromise)
  {
    final Promise<Channel> p;
    if (newHandshakePromise != null)
    {
      Promise<Channel> oldHandshakePromise = handshakePromise;
      if (!oldHandshakePromise.isDone())
      {
        oldHandshakePromise.addListener(new FutureListener()
        {
          public void operationComplete(Future<Channel> future)
            throws Exception
          {
            if (future.isSuccess()) {
              newHandshakePromise.setSuccess(future.getNow());
            } else {
              newHandshakePromise.setFailure(future.cause());
            }
          }
        }); return;
      }
      Promise<Channel> p;
      handshakePromise = (p = newHandshakePromise);
    }
    else
    {
      if (engine.getHandshakeStatus() != SSLEngineResult.HandshakeStatus.NOT_HANDSHAKING) {
        return;
      }
      p = handshakePromise;
      assert (!p.isDone());
    }
    ChannelHandlerContext ctx = this.ctx;
    try
    {
      engine.beginHandshake();
      wrapNonAppData(ctx, false);
      ctx.flush();
    }
    catch (Exception e)
    {
      notifyHandshakeFailure(e);
    }
    long handshakeTimeoutMillis = this.handshakeTimeoutMillis;
    if ((handshakeTimeoutMillis <= 0L) || (p.isDone())) {
      return;
    }
    final ScheduledFuture<?> timeoutFuture = ctx.executor().schedule(new Runnable()
    {
      public void run()
      {
        if (p.isDone()) {
          return;
        }
        SslHandler.this.notifyHandshakeFailure(SslHandler.HANDSHAKE_TIMED_OUT);
      }
    }, handshakeTimeoutMillis, TimeUnit.MILLISECONDS);
    
    p.addListener(new FutureListener()
    {
      public void operationComplete(Future<Channel> f)
        throws Exception
      {
        timeoutFuture.cancel(false);
      }
    });
  }
  
  public void channelActive(ChannelHandlerContext ctx)
    throws Exception
  {
    if ((!startTls) && (engine.getUseClientMode())) {
      handshake(null);
    }
    ctx.fireChannelActive();
  }
  
  private void safeClose(final ChannelHandlerContext ctx, ChannelFuture flushFuture, final ChannelPromise promise)
  {
    if (!ctx.channel().isActive())
    {
      ctx.close(promise); return;
    }
    ScheduledFuture<?> timeoutFuture;
    final ScheduledFuture<?> timeoutFuture;
    if (closeNotifyTimeoutMillis > 0L) {
      timeoutFuture = ctx.executor().schedule(new Runnable()
      {
        public void run()
        {
          SslHandler.logger.warn("{} Last write attempt timed out; force-closing the connection.", ctx.channel());
          
          SslHandler.addCloseListener(ctx.close(ctx.newPromise()), promise);
        }
      }, closeNotifyTimeoutMillis, TimeUnit.MILLISECONDS);
    } else {
      timeoutFuture = null;
    }
    flushFuture.addListener(new ChannelFutureListener()
    {
      public void operationComplete(ChannelFuture f)
        throws Exception
      {
        if (timeoutFuture != null) {
          timeoutFuture.cancel(false);
        }
        SslHandler.addCloseListener(ctx.close(ctx.newPromise()), promise);
      }
    });
  }
  
  private static void addCloseListener(ChannelFuture future, ChannelPromise promise)
  {
    future.addListener(new ChannelPromiseNotifier(false, new ChannelPromise[] { promise }));
  }
  
  private ByteBuf allocate(ChannelHandlerContext ctx, int capacity)
  {
    ByteBufAllocator alloc = ctx.alloc();
    if (wantsDirectBuffer) {
      return alloc.directBuffer(capacity);
    }
    return alloc.buffer(capacity);
  }
  
  private ByteBuf allocateOutNetBuf(ChannelHandlerContext ctx, int pendingBytes)
  {
    if (wantsLargeOutboundNetworkBuffer) {
      return allocate(ctx, maxPacketBufferSize);
    }
    return allocate(ctx, Math.min(pendingBytes + 2329, maxPacketBufferSize));
  }
  
  private final class LazyChannelPromise
    extends DefaultPromise<Channel>
  {
    private LazyChannelPromise() {}
    
    protected EventExecutor executor()
    {
      if (ctx == null) {
        throw new IllegalStateException();
      }
      return ctx.executor();
    }
    
    protected void checkDeadLock()
    {
      if (ctx == null) {
        return;
      }
      super.checkDeadLock();
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.handler.ssl.SslHandler
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */