package io.netty.handler.timeout;

import io.netty.channel.Channel;
import io.netty.channel.ChannelDuplexHandler;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelFutureListener;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.ChannelPromise;
import io.netty.util.concurrent.EventExecutor;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

public class IdleStateHandler
  extends ChannelDuplexHandler
{
  private static final long MIN_TIMEOUT_NANOS = TimeUnit.MILLISECONDS.toNanos(1L);
  private final ChannelFutureListener writeListener = new ChannelFutureListener()
  {
    public void operationComplete(ChannelFuture future)
      throws Exception
    {
      lastWriteTime = System.nanoTime();
      firstWriterIdleEvent = IdleStateHandler.access$102(IdleStateHandler.this, true);
    }
  };
  private final long readerIdleTimeNanos;
  private final long writerIdleTimeNanos;
  private final long allIdleTimeNanos;
  volatile ScheduledFuture<?> readerIdleTimeout;
  volatile long lastReadTime;
  private boolean firstReaderIdleEvent = true;
  volatile ScheduledFuture<?> writerIdleTimeout;
  volatile long lastWriteTime;
  private boolean firstWriterIdleEvent = true;
  volatile ScheduledFuture<?> allIdleTimeout;
  private boolean firstAllIdleEvent = true;
  private volatile int state;
  private volatile boolean reading;
  
  public IdleStateHandler(int readerIdleTimeSeconds, int writerIdleTimeSeconds, int allIdleTimeSeconds)
  {
    this(readerIdleTimeSeconds, writerIdleTimeSeconds, allIdleTimeSeconds, TimeUnit.SECONDS);
  }
  
  public IdleStateHandler(long readerIdleTime, long writerIdleTime, long allIdleTime, TimeUnit unit)
  {
    if (unit == null) {
      throw new NullPointerException("unit");
    }
    if (readerIdleTime <= 0L) {
      readerIdleTimeNanos = 0L;
    } else {
      readerIdleTimeNanos = Math.max(unit.toNanos(readerIdleTime), MIN_TIMEOUT_NANOS);
    }
    if (writerIdleTime <= 0L) {
      writerIdleTimeNanos = 0L;
    } else {
      writerIdleTimeNanos = Math.max(unit.toNanos(writerIdleTime), MIN_TIMEOUT_NANOS);
    }
    if (allIdleTime <= 0L) {
      allIdleTimeNanos = 0L;
    } else {
      allIdleTimeNanos = Math.max(unit.toNanos(allIdleTime), MIN_TIMEOUT_NANOS);
    }
  }
  
  public long getReaderIdleTimeInMillis()
  {
    return TimeUnit.NANOSECONDS.toMillis(readerIdleTimeNanos);
  }
  
  public long getWriterIdleTimeInMillis()
  {
    return TimeUnit.NANOSECONDS.toMillis(writerIdleTimeNanos);
  }
  
  public long getAllIdleTimeInMillis()
  {
    return TimeUnit.NANOSECONDS.toMillis(allIdleTimeNanos);
  }
  
  public void handlerAdded(ChannelHandlerContext ctx)
    throws Exception
  {
    if ((ctx.channel().isActive()) && (ctx.channel().isRegistered())) {
      initialize(ctx);
    }
  }
  
  public void handlerRemoved(ChannelHandlerContext ctx)
    throws Exception
  {
    destroy();
  }
  
  public void channelRegistered(ChannelHandlerContext ctx)
    throws Exception
  {
    if (ctx.channel().isActive()) {
      initialize(ctx);
    }
    super.channelRegistered(ctx);
  }
  
  public void channelActive(ChannelHandlerContext ctx)
    throws Exception
  {
    initialize(ctx);
    super.channelActive(ctx);
  }
  
  public void channelInactive(ChannelHandlerContext ctx)
    throws Exception
  {
    destroy();
    super.channelInactive(ctx);
  }
  
  public void channelRead(ChannelHandlerContext ctx, Object msg)
    throws Exception
  {
    if ((readerIdleTimeNanos > 0L) || (allIdleTimeNanos > 0L))
    {
      reading = true;
      firstReaderIdleEvent = (firstAllIdleEvent = 1);
    }
    ctx.fireChannelRead(msg);
  }
  
  public void channelReadComplete(ChannelHandlerContext ctx)
    throws Exception
  {
    if ((readerIdleTimeNanos > 0L) || (allIdleTimeNanos > 0L))
    {
      lastReadTime = System.nanoTime();
      reading = false;
    }
    ctx.fireChannelReadComplete();
  }
  
  public void write(ChannelHandlerContext ctx, Object msg, ChannelPromise promise)
    throws Exception
  {
    if ((writerIdleTimeNanos > 0L) || (allIdleTimeNanos > 0L))
    {
      ChannelPromise unvoid = promise.unvoid();
      unvoid.addListener(writeListener);
      ctx.write(msg, unvoid);
    }
    else
    {
      ctx.write(msg, promise);
    }
  }
  
  private void initialize(ChannelHandlerContext ctx)
  {
    switch (state)
    {
    case 1: 
    case 2: 
      return;
    }
    state = 1;
    
    EventExecutor loop = ctx.executor();
    
    lastReadTime = (lastWriteTime = System.nanoTime());
    if (readerIdleTimeNanos > 0L) {
      readerIdleTimeout = loop.schedule(new ReaderIdleTimeoutTask(ctx), readerIdleTimeNanos, TimeUnit.NANOSECONDS);
    }
    if (writerIdleTimeNanos > 0L) {
      writerIdleTimeout = loop.schedule(new WriterIdleTimeoutTask(ctx), writerIdleTimeNanos, TimeUnit.NANOSECONDS);
    }
    if (allIdleTimeNanos > 0L) {
      allIdleTimeout = loop.schedule(new AllIdleTimeoutTask(ctx), allIdleTimeNanos, TimeUnit.NANOSECONDS);
    }
  }
  
  private void destroy()
  {
    state = 2;
    if (readerIdleTimeout != null)
    {
      readerIdleTimeout.cancel(false);
      readerIdleTimeout = null;
    }
    if (writerIdleTimeout != null)
    {
      writerIdleTimeout.cancel(false);
      writerIdleTimeout = null;
    }
    if (allIdleTimeout != null)
    {
      allIdleTimeout.cancel(false);
      allIdleTimeout = null;
    }
  }
  
  protected void channelIdle(ChannelHandlerContext ctx, IdleStateEvent evt)
    throws Exception
  {
    ctx.fireUserEventTriggered(evt);
  }
  
  protected IdleStateEvent newIdleStateEvent(IdleState state, boolean first)
  {
    switch (state)
    {
    case ALL_IDLE: 
      return first ? IdleStateEvent.FIRST_ALL_IDLE_STATE_EVENT : IdleStateEvent.ALL_IDLE_STATE_EVENT;
    case READER_IDLE: 
      return first ? IdleStateEvent.FIRST_READER_IDLE_STATE_EVENT : IdleStateEvent.READER_IDLE_STATE_EVENT;
    case WRITER_IDLE: 
      return first ? IdleStateEvent.FIRST_WRITER_IDLE_STATE_EVENT : IdleStateEvent.WRITER_IDLE_STATE_EVENT;
    }
    throw new Error();
  }
  
  private final class ReaderIdleTimeoutTask
    implements Runnable
  {
    private final ChannelHandlerContext ctx;
    
    ReaderIdleTimeoutTask(ChannelHandlerContext ctx)
    {
      this.ctx = ctx;
    }
    
    public void run()
    {
      if (!ctx.channel().isOpen()) {
        return;
      }
      long nextDelay = readerIdleTimeNanos;
      if (!reading) {
        nextDelay -= System.nanoTime() - lastReadTime;
      }
      if (nextDelay <= 0L)
      {
        readerIdleTimeout = ctx.executor().schedule(this, readerIdleTimeNanos, TimeUnit.NANOSECONDS);
        try
        {
          IdleStateEvent event = newIdleStateEvent(IdleState.READER_IDLE, firstReaderIdleEvent);
          if (firstReaderIdleEvent) {
            firstReaderIdleEvent = false;
          }
          channelIdle(ctx, event);
        }
        catch (Throwable t)
        {
          ctx.fireExceptionCaught(t);
        }
      }
      else
      {
        readerIdleTimeout = ctx.executor().schedule(this, nextDelay, TimeUnit.NANOSECONDS);
      }
    }
  }
  
  private final class WriterIdleTimeoutTask
    implements Runnable
  {
    private final ChannelHandlerContext ctx;
    
    WriterIdleTimeoutTask(ChannelHandlerContext ctx)
    {
      this.ctx = ctx;
    }
    
    public void run()
    {
      if (!ctx.channel().isOpen()) {
        return;
      }
      long lastWriteTime = IdleStateHandler.this.lastWriteTime;
      long nextDelay = writerIdleTimeNanos - (System.nanoTime() - lastWriteTime);
      if (nextDelay <= 0L)
      {
        writerIdleTimeout = ctx.executor().schedule(this, writerIdleTimeNanos, TimeUnit.NANOSECONDS);
        try
        {
          IdleStateEvent event = newIdleStateEvent(IdleState.WRITER_IDLE, firstWriterIdleEvent);
          if (firstWriterIdleEvent) {
            firstWriterIdleEvent = false;
          }
          channelIdle(ctx, event);
        }
        catch (Throwable t)
        {
          ctx.fireExceptionCaught(t);
        }
      }
      else
      {
        writerIdleTimeout = ctx.executor().schedule(this, nextDelay, TimeUnit.NANOSECONDS);
      }
    }
  }
  
  private final class AllIdleTimeoutTask
    implements Runnable
  {
    private final ChannelHandlerContext ctx;
    
    AllIdleTimeoutTask(ChannelHandlerContext ctx)
    {
      this.ctx = ctx;
    }
    
    public void run()
    {
      if (!ctx.channel().isOpen()) {
        return;
      }
      long nextDelay = allIdleTimeNanos;
      if (!reading) {
        nextDelay -= System.nanoTime() - Math.max(lastReadTime, lastWriteTime);
      }
      if (nextDelay <= 0L)
      {
        allIdleTimeout = ctx.executor().schedule(this, allIdleTimeNanos, TimeUnit.NANOSECONDS);
        try
        {
          IdleStateEvent event = newIdleStateEvent(IdleState.ALL_IDLE, firstAllIdleEvent);
          if (firstAllIdleEvent) {
            firstAllIdleEvent = false;
          }
          channelIdle(ctx, event);
        }
        catch (Throwable t)
        {
          ctx.fireExceptionCaught(t);
        }
      }
      else
      {
        allIdleTimeout = ctx.executor().schedule(this, nextDelay, TimeUnit.NANOSECONDS);
      }
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.handler.timeout.IdleStateHandler
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */