package io.netty.resolver.dns;

import io.netty.bootstrap.Bootstrap;
import io.netty.channel.AddressedEnvelope;
import io.netty.channel.Channel;
import io.netty.channel.ChannelFactory;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelFutureListener;
import io.netty.channel.ChannelHandler;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.ChannelInboundHandlerAdapter;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.ChannelOption;
import io.netty.channel.ChannelPipeline;
import io.netty.channel.EventLoop;
import io.netty.channel.FixedRecvByteBufAllocator;
import io.netty.channel.socket.DatagramChannel;
import io.netty.channel.socket.DatagramChannelConfig;
import io.netty.channel.socket.InternetProtocolFamily;
import io.netty.handler.codec.dns.DatagramDnsQueryEncoder;
import io.netty.handler.codec.dns.DatagramDnsResponse;
import io.netty.handler.codec.dns.DatagramDnsResponseDecoder;
import io.netty.handler.codec.dns.DnsQuestion;
import io.netty.handler.codec.dns.DnsRecord;
import io.netty.handler.codec.dns.DnsResponse;
import io.netty.resolver.HostsFileEntriesResolver;
import io.netty.resolver.InetNameResolver;
import io.netty.util.NetUtil;
import io.netty.util.ReferenceCountUtil;
import io.netty.util.concurrent.FastThreadLocal;
import io.netty.util.concurrent.Future;
import io.netty.util.concurrent.Promise;
import io.netty.util.internal.EmptyArrays;
import io.netty.util.internal.ObjectUtil;
import io.netty.util.internal.PlatformDependent;
import io.netty.util.internal.StringUtil;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;
import java.lang.reflect.Method;
import java.net.IDN;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

public class DnsNameResolver
  extends InetNameResolver
{
  private static final InternalLogger logger;
  private static final String LOCALHOST = "localhost";
  private static final InetAddress LOCALHOST_ADDRESS;
  static final InternetProtocolFamily[] DEFAULT_RESOLVE_ADDRESS_TYPES;
  static final String[] DEFAULT_SEACH_DOMAINS;
  
  static
  {
    logger = InternalLoggerFactory.getInstance(DnsNameResolver.class);
    if (NetUtil.isIpV4StackPreferred())
    {
      DEFAULT_RESOLVE_ADDRESS_TYPES = new InternetProtocolFamily[] { InternetProtocolFamily.IPv4 };
      LOCALHOST_ADDRESS = NetUtil.LOCALHOST4;
    }
    else
    {
      DEFAULT_RESOLVE_ADDRESS_TYPES = new InternetProtocolFamily[2];
      if (NetUtil.isIpV6AddressesPreferred())
      {
        DEFAULT_RESOLVE_ADDRESS_TYPES[0] = InternetProtocolFamily.IPv6;
        DEFAULT_RESOLVE_ADDRESS_TYPES[1] = InternetProtocolFamily.IPv4;
        LOCALHOST_ADDRESS = NetUtil.LOCALHOST6;
      }
      else
      {
        DEFAULT_RESOLVE_ADDRESS_TYPES[0] = InternetProtocolFamily.IPv4;
        DEFAULT_RESOLVE_ADDRESS_TYPES[1] = InternetProtocolFamily.IPv6;
        LOCALHOST_ADDRESS = NetUtil.LOCALHOST4;
      }
    }
    String[] searchDomains;
    try
    {
      Class<?> configClass = Class.forName("sun.net.dns.ResolverConfiguration");
      Method open = configClass.getMethod("open", new Class[0]);
      Method nameservers = configClass.getMethod("searchlist", new Class[0]);
      Object instance = open.invoke(null, new Object[0]);
      
      List<String> list = (List)nameservers.invoke(instance, new Object[0]);
      searchDomains = (String[])list.toArray(new String[list.size()]);
    }
    catch (Exception ignore)
    {
      String[] searchDomains;
      searchDomains = EmptyArrays.EMPTY_STRINGS;
    }
    DEFAULT_SEACH_DOMAINS = searchDomains;
  }
  
  private static final DatagramDnsResponseDecoder DECODER = new DatagramDnsResponseDecoder();
  private static final DatagramDnsQueryEncoder ENCODER = new DatagramDnsQueryEncoder();
  final DnsServerAddresses nameServerAddresses;
  final Future<Channel> channelFuture;
  final DatagramChannel ch;
  final DnsQueryContextManager queryContextManager = new DnsQueryContextManager();
  private final DnsCache resolveCache;
  private final FastThreadLocal<DnsServerAddressStream> nameServerAddrStream = new FastThreadLocal()
  {
    protected DnsServerAddressStream initialValue()
      throws Exception
    {
      return nameServerAddresses.stream();
    }
  };
  private final long queryTimeoutMillis;
  private final int maxQueriesPerResolve;
  private final boolean traceEnabled;
  private final InternetProtocolFamily[] resolvedAddressTypes;
  private final boolean recursionDesired;
  private final int maxPayloadSize;
  private final boolean optResourceEnabled;
  private final HostsFileEntriesResolver hostsFileEntriesResolver;
  private final String[] searchDomains;
  private final int ndots;
  
  public DnsNameResolver(EventLoop eventLoop, ChannelFactory<? extends DatagramChannel> channelFactory, DnsServerAddresses nameServerAddresses, final DnsCache resolveCache, long queryTimeoutMillis, InternetProtocolFamily[] resolvedAddressTypes, boolean recursionDesired, int maxQueriesPerResolve, boolean traceEnabled, int maxPayloadSize, boolean optResourceEnabled, HostsFileEntriesResolver hostsFileEntriesResolver, String[] searchDomains, int ndots)
  {
    super(eventLoop);
    ObjectUtil.checkNotNull(channelFactory, "channelFactory");
    this.nameServerAddresses = ((DnsServerAddresses)ObjectUtil.checkNotNull(nameServerAddresses, "nameServerAddresses"));
    this.queryTimeoutMillis = ObjectUtil.checkPositive(queryTimeoutMillis, "queryTimeoutMillis");
    this.resolvedAddressTypes = ((InternetProtocolFamily[])ObjectUtil.checkNonEmpty(resolvedAddressTypes, "resolvedAddressTypes"));
    this.recursionDesired = recursionDesired;
    this.maxQueriesPerResolve = ObjectUtil.checkPositive(maxQueriesPerResolve, "maxQueriesPerResolve");
    this.traceEnabled = traceEnabled;
    this.maxPayloadSize = ObjectUtil.checkPositive(maxPayloadSize, "maxPayloadSize");
    this.optResourceEnabled = optResourceEnabled;
    this.hostsFileEntriesResolver = ((HostsFileEntriesResolver)ObjectUtil.checkNotNull(hostsFileEntriesResolver, "hostsFileEntriesResolver"));
    this.resolveCache = resolveCache;
    this.searchDomains = ((String[])((String[])ObjectUtil.checkNotNull(searchDomains, "searchDomains")).clone());
    this.ndots = ObjectUtil.checkPositiveOrZero(ndots, "ndots");
    
    Bootstrap b = new Bootstrap();
    b.group(executor());
    b.channelFactory(channelFactory);
    b.option(ChannelOption.DATAGRAM_CHANNEL_ACTIVE_ON_REGISTRATION, Boolean.valueOf(true));
    final DnsResponseHandler responseHandler = new DnsResponseHandler(executor().newPromise());
    b.handler(new ChannelInitializer()
    {
      protected void initChannel(DatagramChannel ch)
        throws Exception
      {
        ch.pipeline().addLast(new ChannelHandler[] { DnsNameResolver.DECODER, DnsNameResolver.ENCODER, responseHandler });
      }
    });
    channelFuture = channelActivePromise;
    ch = ((DatagramChannel)b.register().channel());
    ch.config().setRecvByteBufAllocator(new FixedRecvByteBufAllocator(maxPayloadSize));
    
    ch.closeFuture().addListener(new ChannelFutureListener()
    {
      public void operationComplete(ChannelFuture future)
        throws Exception
      {
        resolveCache.clear();
      }
    });
  }
  
  public DnsCache resolveCache()
  {
    return resolveCache;
  }
  
  public long queryTimeoutMillis()
  {
    return queryTimeoutMillis;
  }
  
  public List<InternetProtocolFamily> resolvedAddressTypes()
  {
    return Arrays.asList(resolvedAddressTypes);
  }
  
  InternetProtocolFamily[] resolveAddressTypesUnsafe()
  {
    return resolvedAddressTypes;
  }
  
  final String[] searchDomains()
  {
    return searchDomains;
  }
  
  final int ndots()
  {
    return ndots;
  }
  
  public boolean isRecursionDesired()
  {
    return recursionDesired;
  }
  
  public int maxQueriesPerResolve()
  {
    return maxQueriesPerResolve;
  }
  
  public boolean isTraceEnabled()
  {
    return traceEnabled;
  }
  
  public int maxPayloadSize()
  {
    return maxPayloadSize;
  }
  
  public boolean isOptResourceEnabled()
  {
    return optResourceEnabled;
  }
  
  public HostsFileEntriesResolver hostsFileEntriesResolver()
  {
    return hostsFileEntriesResolver;
  }
  
  public void close()
  {
    if (ch.isOpen()) {
      ch.close();
    }
  }
  
  protected EventLoop executor()
  {
    return (EventLoop)super.executor();
  }
  
  private InetAddress resolveHostsFileEntry(String hostname)
  {
    if (hostsFileEntriesResolver == null) {
      return null;
    }
    InetAddress address = hostsFileEntriesResolver.address(hostname);
    if ((address == null) && (PlatformDependent.isWindows()) && ("localhost".equalsIgnoreCase(hostname))) {
      return LOCALHOST_ADDRESS;
    }
    return address;
  }
  
  protected void doResolve(String inetHost, Promise<InetAddress> promise)
    throws Exception
  {
    doResolve(inetHost, promise, resolveCache);
  }
  
  protected void doResolve(String inetHost, Promise<InetAddress> promise, DnsCache resolveCache)
    throws Exception
  {
    byte[] bytes = NetUtil.createByteArrayFromIpAddressString(inetHost);
    if (bytes != null)
    {
      promise.setSuccess(InetAddress.getByAddress(bytes));
      return;
    }
    String hostname = hostname(inetHost);
    
    InetAddress hostsFileEntry = resolveHostsFileEntry(hostname);
    if (hostsFileEntry != null)
    {
      promise.setSuccess(hostsFileEntry);
      return;
    }
    if (!doResolveCached(hostname, promise, resolveCache)) {
      doResolveUncached(hostname, promise, resolveCache);
    }
  }
  
  private boolean doResolveCached(String hostname, Promise<InetAddress> promise, DnsCache resolveCache)
  {
    List<DnsCacheEntry> cachedEntries = resolveCache.get(hostname);
    if ((cachedEntries == null) || (cachedEntries.isEmpty())) {
      return false;
    }
    InetAddress address = null;
    Throwable cause = null;
    synchronized (cachedEntries)
    {
      int numEntries = cachedEntries.size();
      assert (numEntries > 0);
      if (((DnsCacheEntry)cachedEntries.get(0)).cause() != null) {
        cause = ((DnsCacheEntry)cachedEntries.get(0)).cause();
      } else {
        for (InternetProtocolFamily f : resolvedAddressTypes) {
          for (int i = 0; i < numEntries; i++)
          {
            DnsCacheEntry e = (DnsCacheEntry)cachedEntries.get(i);
            if (f.addressType().isInstance(e.address()))
            {
              address = e.address();
              break;
            }
          }
        }
      }
    }
    if (address != null) {
      setSuccess(promise, address);
    } else if (cause != null)
    {
      if (!promise.tryFailure(cause)) {
        logger.warn("Failed to notify failure to a promise: {}", promise, cause);
      }
    }
    else {
      return false;
    }
    return true;
  }
  
  private static void setSuccess(Promise<InetAddress> promise, InetAddress result)
  {
    if (!promise.trySuccess(result)) {
      logger.warn("Failed to notify success ({}) to a promise: {}", result, promise);
    }
  }
  
  private void doResolveUncached(String hostname, Promise<InetAddress> promise, DnsCache resolveCache)
  {
    SingleResolverContext ctx = new SingleResolverContext(this, hostname, resolveCache);
    ctx.resolve(promise);
  }
  
  final class SingleResolverContext
    extends DnsNameResolverContext<InetAddress>
  {
    SingleResolverContext(DnsNameResolver parent, String hostname, DnsCache resolveCache)
    {
      super(hostname, resolveCache);
    }
    
    DnsNameResolverContext<InetAddress> newResolverContext(DnsNameResolver parent, String hostname, DnsCache resolveCache)
    {
      return new SingleResolverContext(DnsNameResolver.this, parent, hostname, resolveCache);
    }
    
    boolean finishResolve(Class<? extends InetAddress> addressType, List<DnsCacheEntry> resolvedEntries, Promise<InetAddress> promise)
    {
      int numEntries = resolvedEntries.size();
      for (int i = 0; i < numEntries; i++)
      {
        InetAddress a = ((DnsCacheEntry)resolvedEntries.get(i)).address();
        if (addressType.isInstance(a))
        {
          DnsNameResolver.setSuccess(promise, a);
          return true;
        }
      }
      return false;
    }
  }
  
  protected void doResolveAll(String inetHost, Promise<List<InetAddress>> promise)
    throws Exception
  {
    doResolveAll(inetHost, promise, resolveCache);
  }
  
  protected void doResolveAll(String inetHost, Promise<List<InetAddress>> promise, DnsCache resolveCache)
    throws Exception
  {
    byte[] bytes = NetUtil.createByteArrayFromIpAddressString(inetHost);
    if (bytes != null)
    {
      promise.setSuccess(Collections.singletonList(InetAddress.getByAddress(bytes)));
      return;
    }
    String hostname = hostname(inetHost);
    
    InetAddress hostsFileEntry = resolveHostsFileEntry(hostname);
    if (hostsFileEntry != null)
    {
      promise.setSuccess(Collections.singletonList(hostsFileEntry));
      return;
    }
    if (!doResolveAllCached(hostname, promise, resolveCache)) {
      doResolveAllUncached(hostname, promise, resolveCache);
    }
  }
  
  private boolean doResolveAllCached(String hostname, Promise<List<InetAddress>> promise, DnsCache resolveCache)
  {
    List<DnsCacheEntry> cachedEntries = resolveCache.get(hostname);
    if ((cachedEntries == null) || (cachedEntries.isEmpty())) {
      return false;
    }
    List<InetAddress> result = null;
    Throwable cause = null;
    synchronized (cachedEntries)
    {
      int numEntries = cachedEntries.size();
      assert (numEntries > 0);
      if (((DnsCacheEntry)cachedEntries.get(0)).cause() != null) {
        cause = ((DnsCacheEntry)cachedEntries.get(0)).cause();
      } else {
        for (InternetProtocolFamily f : resolvedAddressTypes) {
          for (int i = 0; i < numEntries; i++)
          {
            DnsCacheEntry e = (DnsCacheEntry)cachedEntries.get(i);
            if (f.addressType().isInstance(e.address()))
            {
              if (result == null) {
                result = new ArrayList(numEntries);
              }
              result.add(e.address());
            }
          }
        }
      }
    }
    if (result != null) {
      promise.trySuccess(result);
    } else if (cause != null) {
      promise.tryFailure(cause);
    } else {
      return false;
    }
    return true;
  }
  
  final class ListResolverContext
    extends DnsNameResolverContext<List<InetAddress>>
  {
    ListResolverContext(DnsNameResolver parent, String hostname, DnsCache resolveCache)
    {
      super(hostname, resolveCache);
    }
    
    DnsNameResolverContext<List<InetAddress>> newResolverContext(DnsNameResolver parent, String hostname, DnsCache resolveCache)
    {
      return new ListResolverContext(DnsNameResolver.this, parent, hostname, resolveCache);
    }
    
    boolean finishResolve(Class<? extends InetAddress> addressType, List<DnsCacheEntry> resolvedEntries, Promise<List<InetAddress>> promise)
    {
      List<InetAddress> result = null;
      int numEntries = resolvedEntries.size();
      for (int i = 0; i < numEntries; i++)
      {
        InetAddress a = ((DnsCacheEntry)resolvedEntries.get(i)).address();
        if (addressType.isInstance(a))
        {
          if (result == null) {
            result = new ArrayList(numEntries);
          }
          result.add(a);
        }
      }
      if (result != null)
      {
        promise.trySuccess(result);
        return true;
      }
      return false;
    }
  }
  
  private void doResolveAllUncached(String hostname, Promise<List<InetAddress>> promise, DnsCache resolveCache)
  {
    DnsNameResolverContext<List<InetAddress>> ctx = new ListResolverContext(this, hostname, resolveCache);
    ctx.resolve(promise);
  }
  
  private static String hostname(String inetHost)
  {
    String hostname = IDN.toASCII(inetHost);
    if ((StringUtil.endsWith(inetHost, '.')) && (!StringUtil.endsWith(hostname, '.'))) {
      hostname = hostname + ".";
    }
    return hostname;
  }
  
  public Future<AddressedEnvelope<DnsResponse, InetSocketAddress>> query(DnsQuestion question)
  {
    return query(nextNameServerAddress(), question);
  }
  
  public Future<AddressedEnvelope<DnsResponse, InetSocketAddress>> query(DnsQuestion question, Iterable<DnsRecord> additional)
  {
    return query(nextNameServerAddress(), question, additional);
  }
  
  public Future<AddressedEnvelope<DnsResponse, InetSocketAddress>> query(DnsQuestion question, Promise<AddressedEnvelope<? extends DnsResponse, InetSocketAddress>> promise)
  {
    return query(nextNameServerAddress(), question, Collections.emptyList(), promise);
  }
  
  private InetSocketAddress nextNameServerAddress()
  {
    return ((DnsServerAddressStream)nameServerAddrStream.get()).next();
  }
  
  public Future<AddressedEnvelope<DnsResponse, InetSocketAddress>> query(InetSocketAddress nameServerAddr, DnsQuestion question)
  {
    return query0(nameServerAddr, question, Collections.emptyList(), ch.eventLoop().newPromise());
  }
  
  public Future<AddressedEnvelope<DnsResponse, InetSocketAddress>> query(InetSocketAddress nameServerAddr, DnsQuestion question, Iterable<DnsRecord> additional)
  {
    return query0(nameServerAddr, question, additional, ch.eventLoop().newPromise());
  }
  
  public Future<AddressedEnvelope<DnsResponse, InetSocketAddress>> query(InetSocketAddress nameServerAddr, DnsQuestion question, Promise<AddressedEnvelope<? extends DnsResponse, InetSocketAddress>> promise)
  {
    return query0(nameServerAddr, question, Collections.emptyList(), promise);
  }
  
  public Future<AddressedEnvelope<DnsResponse, InetSocketAddress>> query(InetSocketAddress nameServerAddr, DnsQuestion question, Iterable<DnsRecord> additional, Promise<AddressedEnvelope<? extends DnsResponse, InetSocketAddress>> promise)
  {
    return query0(nameServerAddr, question, additional, promise);
  }
  
  private Future<AddressedEnvelope<DnsResponse, InetSocketAddress>> query0(InetSocketAddress nameServerAddr, DnsQuestion question, Iterable<DnsRecord> additional, Promise<AddressedEnvelope<? extends DnsResponse, InetSocketAddress>> promise)
  {
    Promise<AddressedEnvelope<DnsResponse, InetSocketAddress>> castPromise = cast((Promise)ObjectUtil.checkNotNull(promise, "promise"));
    try
    {
      new DnsQueryContext(this, nameServerAddr, question, additional, castPromise).query();
      return castPromise;
    }
    catch (Exception e)
    {
      return castPromise.setFailure(e);
    }
  }
  
  private static Promise<AddressedEnvelope<DnsResponse, InetSocketAddress>> cast(Promise<?> promise)
  {
    return promise;
  }
  
  private final class DnsResponseHandler
    extends ChannelInboundHandlerAdapter
  {
    private final Promise<Channel> channelActivePromise;
    
    DnsResponseHandler()
    {
      this.channelActivePromise = channelActivePromise;
    }
    
    public void channelRead(ChannelHandlerContext ctx, Object msg)
      throws Exception
    {
      try
      {
        DatagramDnsResponse res = (DatagramDnsResponse)msg;
        int queryId = res.id();
        if (DnsNameResolver.logger.isDebugEnabled()) {
          DnsNameResolver.logger.debug("{} RECEIVED: [{}: {}], {}", new Object[] { ch, Integer.valueOf(queryId), res.sender(), res });
        }
        DnsQueryContext qCtx = queryContextManager.get(res.sender(), queryId);
        if (qCtx == null) {
          DnsNameResolver.logger.warn("{} Received a DNS response with an unknown ID: {}", ch, Integer.valueOf(queryId));
        } else {
          qCtx.finish(res);
        }
      }
      finally
      {
        ReferenceCountUtil.safeRelease(msg);
      }
    }
    
    public void channelActive(ChannelHandlerContext ctx)
      throws Exception
    {
      super.channelActive(ctx);
      channelActivePromise.setSuccess(ctx.channel());
    }
    
    public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause)
      throws Exception
    {
      DnsNameResolver.logger.warn("{} Unexpected exception: ", ch, cause);
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.resolver.dns.DnsNameResolver
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */