package io.netty.util;

import io.netty.util.internal.PlatformDependent;
import java.util.concurrent.atomic.AtomicReference;
import java.util.concurrent.atomic.AtomicReferenceArray;
import java.util.concurrent.atomic.AtomicReferenceFieldUpdater;

public class DefaultAttributeMap
  implements AttributeMap
{
  private static final AtomicReferenceFieldUpdater<DefaultAttributeMap, AtomicReferenceArray> updater;
  private static final int BUCKET_SIZE = 4;
  private static final int MASK = 3;
  private volatile AtomicReferenceArray<DefaultAttribute<?>> attributes;
  
  static
  {
    AtomicReferenceFieldUpdater<DefaultAttributeMap, AtomicReferenceArray> referenceFieldUpdater = PlatformDependent.newAtomicReferenceFieldUpdater(DefaultAttributeMap.class, "attributes");
    if (referenceFieldUpdater == null) {
      referenceFieldUpdater = AtomicReferenceFieldUpdater.newUpdater(DefaultAttributeMap.class, AtomicReferenceArray.class, "attributes");
    }
    updater = referenceFieldUpdater;
  }
  
  public <T> Attribute<T> attr(AttributeKey<T> key)
  {
    if (key == null) {
      throw new NullPointerException("key");
    }
    AtomicReferenceArray<DefaultAttribute<?>> attributes = this.attributes;
    if (attributes == null)
    {
      attributes = new AtomicReferenceArray(4);
      if (!updater.compareAndSet(this, null, attributes)) {
        attributes = this.attributes;
      }
    }
    int i = index(key);
    DefaultAttribute<?> head = (DefaultAttribute)attributes.get(i);
    if (head == null)
    {
      head = new DefaultAttribute();
      DefaultAttribute<T> attr = new DefaultAttribute(head, key);
      next = attr;
      prev = head;
      if (attributes.compareAndSet(i, null, head)) {
        return attr;
      }
      head = (DefaultAttribute)attributes.get(i);
    }
    synchronized (head)
    {
      DefaultAttribute<?> curr = head;
      
      DefaultAttribute<?> next = next;
      if (next == null)
      {
        DefaultAttribute<T> attr = new DefaultAttribute(head, key);
        next = attr;
        prev = curr;
        return attr;
      }
      if ((key == key) && (!removed)) {
        return next;
      }
      curr = next;
    }
  }
  
  public <T> boolean hasAttr(AttributeKey<T> key)
  {
    if (key == null) {
      throw new NullPointerException("key");
    }
    AtomicReferenceArray<DefaultAttribute<?>> attributes = this.attributes;
    if (attributes == null) {
      return false;
    }
    int i = index(key);
    DefaultAttribute<?> head = (DefaultAttribute)attributes.get(i);
    if (head == null) {
      return false;
    }
    synchronized (head)
    {
      DefaultAttribute<?> curr = next;
      while (curr != null)
      {
        if ((key == key) && (!removed)) {
          return true;
        }
        curr = next;
      }
      return false;
    }
  }
  
  private static int index(AttributeKey<?> key)
  {
    return key.id() & 0x3;
  }
  
  private static final class DefaultAttribute<T>
    extends AtomicReference<T>
    implements Attribute<T>
  {
    private static final long serialVersionUID = -2661411462200283011L;
    private final DefaultAttribute<?> head;
    private final AttributeKey<T> key;
    private DefaultAttribute<?> prev;
    private DefaultAttribute<?> next;
    private volatile boolean removed;
    
    DefaultAttribute(DefaultAttribute<?> head, AttributeKey<T> key)
    {
      this.head = head;
      this.key = key;
    }
    
    DefaultAttribute()
    {
      head = this;
      key = null;
    }
    
    public AttributeKey<T> key()
    {
      return key;
    }
    
    public T setIfAbsent(T value)
    {
      while (!compareAndSet(null, value))
      {
        T old = get();
        if (old != null) {
          return old;
        }
      }
      return null;
    }
    
    public T getAndRemove()
    {
      removed = true;
      T oldValue = getAndSet(null);
      remove0();
      return oldValue;
    }
    
    public void remove()
    {
      removed = true;
      set(null);
      remove0();
    }
    
    private void remove0()
    {
      synchronized (head)
      {
        assert (head != null);
        if (prev == null) {
          return;
        }
        prev.next = next;
        if (next != null) {
          next.prev = prev;
        }
        prev = null;
        next = null;
      }
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.util.DefaultAttributeMap
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */