package io.netty.util;

import io.netty.util.concurrent.FastThreadLocal;
import io.netty.util.internal.MathUtil;
import io.netty.util.internal.SystemPropertyUtil;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;
import java.lang.ref.WeakReference;
import java.util.Arrays;
import java.util.Map;
import java.util.WeakHashMap;
import java.util.concurrent.atomic.AtomicInteger;

public abstract class Recycler<T>
{
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(Recycler.class);
  private static final Handle NOOP_HANDLE = new Handle()
  {
    public void recycle(Object object) {}
  };
  private static final AtomicInteger ID_GENERATOR = new AtomicInteger(Integer.MIN_VALUE);
  private static final int OWN_THREAD_ID = ID_GENERATOR.getAndIncrement();
  private static final int DEFAULT_INITIAL_MAX_CAPACITY_PER_THREAD = 32768;
  private static final int DEFAULT_MAX_CAPACITY_PER_THREAD;
  
  static
  {
    int maxCapacityPerThread = SystemPropertyUtil.getInt("io.netty.recycler.maxCapacityPerThread", SystemPropertyUtil.getInt("io.netty.recycler.maxCapacity", 32768));
    if (maxCapacityPerThread < 0) {
      maxCapacityPerThread = 32768;
    }
    DEFAULT_MAX_CAPACITY_PER_THREAD = maxCapacityPerThread;
    
    MAX_SHARED_CAPACITY_FACTOR = Math.max(2, SystemPropertyUtil.getInt("io.netty.recycler.maxSharedCapacityFactor", 2));
    
    MAX_DELAYED_QUEUES_PER_THREAD = Math.max(0, SystemPropertyUtil.getInt("io.netty.recycler.maxDelayedQueuesPerThread", Runtime.getRuntime().availableProcessors() * 2));
    
    LINK_CAPACITY = MathUtil.safeFindNextPositivePowerOfTwo(Math.max(SystemPropertyUtil.getInt("io.netty.recycler.linkCapacity", 16), 16));
    
    RATIO = MathUtil.safeFindNextPositivePowerOfTwo(SystemPropertyUtil.getInt("io.netty.recycler.ratio", 8));
    if (logger.isDebugEnabled()) {
      if (DEFAULT_MAX_CAPACITY_PER_THREAD == 0)
      {
        logger.debug("-Dio.netty.recycler.maxCapacityPerThread: disabled");
        logger.debug("-Dio.netty.recycler.maxSharedCapacityFactor: disabled");
        logger.debug("-Dio.netty.recycler.linkCapacity: disabled");
        logger.debug("-Dio.netty.recycler.ratio: disabled");
      }
      else
      {
        logger.debug("-Dio.netty.recycler.maxCapacityPerThread: {}", Integer.valueOf(DEFAULT_MAX_CAPACITY_PER_THREAD));
        logger.debug("-Dio.netty.recycler.maxSharedCapacityFactor: {}", Integer.valueOf(MAX_SHARED_CAPACITY_FACTOR));
        logger.debug("-Dio.netty.recycler.linkCapacity: {}", Integer.valueOf(LINK_CAPACITY));
        logger.debug("-Dio.netty.recycler.ratio: {}", Integer.valueOf(RATIO));
      }
    }
  }
  
  private static final int INITIAL_CAPACITY = Math.min(DEFAULT_MAX_CAPACITY_PER_THREAD, 256);
  private static final int MAX_SHARED_CAPACITY_FACTOR;
  private static final int MAX_DELAYED_QUEUES_PER_THREAD;
  private static final int LINK_CAPACITY;
  private static final int RATIO;
  private final int maxCapacityPerThread;
  private final int maxSharedCapacityFactor;
  private final int ratioMask;
  private final int maxDelayedQueuesPerThread;
  private final FastThreadLocal<Stack<T>> threadLocal = new FastThreadLocal()
  {
    protected Recycler.Stack<T> initialValue()
    {
      return new Recycler.Stack(Recycler.this, Thread.currentThread(), maxCapacityPerThread, maxSharedCapacityFactor, ratioMask, maxDelayedQueuesPerThread);
    }
  };
  
  protected Recycler()
  {
    this(DEFAULT_MAX_CAPACITY_PER_THREAD);
  }
  
  protected Recycler(int maxCapacityPerThread)
  {
    this(maxCapacityPerThread, MAX_SHARED_CAPACITY_FACTOR);
  }
  
  protected Recycler(int maxCapacityPerThread, int maxSharedCapacityFactor)
  {
    this(maxCapacityPerThread, maxSharedCapacityFactor, RATIO, MAX_DELAYED_QUEUES_PER_THREAD);
  }
  
  protected Recycler(int maxCapacityPerThread, int maxSharedCapacityFactor, int ratio, int maxDelayedQueuesPerThread)
  {
    ratioMask = (MathUtil.safeFindNextPositivePowerOfTwo(ratio) - 1);
    if (maxCapacityPerThread <= 0)
    {
      this.maxCapacityPerThread = 0;
      this.maxSharedCapacityFactor = 1;
      this.maxDelayedQueuesPerThread = 0;
    }
    else
    {
      this.maxCapacityPerThread = maxCapacityPerThread;
      this.maxSharedCapacityFactor = Math.max(1, maxSharedCapacityFactor);
      this.maxDelayedQueuesPerThread = Math.max(0, maxDelayedQueuesPerThread);
    }
  }
  
  public final T get()
  {
    if (maxCapacityPerThread == 0) {
      return (T)newObject(NOOP_HANDLE);
    }
    Stack<T> stack = (Stack)threadLocal.get();
    DefaultHandle<T> handle = stack.pop();
    if (handle == null)
    {
      handle = stack.newHandle();
      value = newObject(handle);
    }
    return (T)value;
  }
  
  @Deprecated
  public final boolean recycle(T o, Handle<T> handle)
  {
    if (handle == NOOP_HANDLE) {
      return false;
    }
    DefaultHandle<T> h = (DefaultHandle)handle;
    if (stack.parent != this) {
      return false;
    }
    h.recycle(o);
    return true;
  }
  
  final int threadLocalCapacity()
  {
    return threadLocal.get()).elements.length;
  }
  
  final int threadLocalSize()
  {
    return threadLocal.get()).size;
  }
  
  protected abstract T newObject(Handle<T> paramHandle);
  
  public static abstract interface Handle<T>
  {
    public abstract void recycle(T paramT);
  }
  
  static final class DefaultHandle<T>
    implements Recycler.Handle<T>
  {
    private int lastRecycledId;
    private int recycleId;
    boolean hasBeenRecycled;
    private Recycler.Stack<?> stack;
    private Object value;
    
    DefaultHandle(Recycler.Stack<?> stack)
    {
      this.stack = stack;
    }
    
    public void recycle(Object object)
    {
      if (object != value) {
        throw new IllegalArgumentException("object does not belong to handle");
      }
      stack.push(this);
    }
  }
  
  private static final FastThreadLocal<Map<Stack<?>, WeakOrderQueue>> DELAYED_RECYCLED = new FastThreadLocal()
  {
    protected Map<Recycler.Stack<?>, Recycler.WeakOrderQueue> initialValue()
    {
      return new WeakHashMap();
    }
  };
  
  private static final class WeakOrderQueue
  {
    static final WeakOrderQueue DUMMY = new WeakOrderQueue();
    private Link head;
    private Link tail;
    private WeakOrderQueue next;
    private final WeakReference<Thread> owner;
    
    private static final class Link
      extends AtomicInteger
    {
      private final Recycler.DefaultHandle<?>[] elements = new Recycler.DefaultHandle[Recycler.LINK_CAPACITY];
      private int readIndex;
      private Link next;
    }
    
    private final int id = Recycler.ID_GENERATOR.getAndIncrement();
    private final AtomicInteger availableSharedCapacity;
    
    private WeakOrderQueue()
    {
      owner = null;
      availableSharedCapacity = null;
    }
    
    private WeakOrderQueue(Recycler.Stack<?> stack, Thread thread)
    {
      head = (tail = new Link(null));
      owner = new WeakReference(thread);
      synchronized (stack)
      {
        next = head;
        head = this;
      }
      availableSharedCapacity = availableSharedCapacity;
    }
    
    static WeakOrderQueue allocate(Recycler.Stack<?> stack, Thread thread)
    {
      return reserveSpace(availableSharedCapacity, Recycler.LINK_CAPACITY) ? new WeakOrderQueue(stack, thread) : null;
    }
    
    private static boolean reserveSpace(AtomicInteger availableSharedCapacity, int space)
    {
      assert (space >= 0);
      for (;;)
      {
        int available = availableSharedCapacity.get();
        if (available < space) {
          return false;
        }
        if (availableSharedCapacity.compareAndSet(available, available - space)) {
          return true;
        }
      }
    }
    
    private void reclaimSpace(int space)
    {
      assert (space >= 0);
      availableSharedCapacity.addAndGet(space);
    }
    
    void add(Recycler.DefaultHandle<?> handle)
    {
      Recycler.DefaultHandle.access$1202(handle, id);
      
      Link tail = this.tail;
      int writeIndex;
      if ((writeIndex = tail.get()) == Recycler.LINK_CAPACITY)
      {
        if (!reserveSpace(availableSharedCapacity, Recycler.LINK_CAPACITY)) {
          return;
        }
        this.tail = (tail = next = new Link(null));
        
        writeIndex = tail.get();
      }
      elements[writeIndex] = handle;
      Recycler.DefaultHandle.access$502(handle, null);
      
      tail.lazySet(writeIndex + 1);
    }
    
    boolean hasFinalData()
    {
      return tail.readIndex != tail.get();
    }
    
    boolean transfer(Recycler.Stack<?> dst)
    {
      Link head = this.head;
      if (head == null) {
        return false;
      }
      if (readIndex == Recycler.LINK_CAPACITY)
      {
        if (next == null) {
          return false;
        }
        this.head = (head = next);
      }
      int srcStart = readIndex;
      int srcEnd = head.get();
      int srcSize = srcEnd - srcStart;
      if (srcSize == 0) {
        return false;
      }
      int dstSize = size;
      int expectedCapacity = dstSize + srcSize;
      if (expectedCapacity > elements.length)
      {
        int actualCapacity = dst.increaseCapacity(expectedCapacity);
        srcEnd = Math.min(srcStart + actualCapacity - dstSize, srcEnd);
      }
      if (srcStart != srcEnd)
      {
        Recycler.DefaultHandle[] srcElems = elements;
        Recycler.DefaultHandle[] dstElems = elements;
        int newDstSize = dstSize;
        for (int i = srcStart; i < srcEnd; i++)
        {
          Recycler.DefaultHandle element = srcElems[i];
          if (Recycler.DefaultHandle.access$1600(element) == 0) {
            Recycler.DefaultHandle.access$1602(element, Recycler.DefaultHandle.access$1200(element));
          } else if (Recycler.DefaultHandle.access$1600(element) != Recycler.DefaultHandle.access$1200(element)) {
            throw new IllegalStateException("recycled already");
          }
          srcElems[i] = null;
          if (!dst.dropHandle(element))
          {
            Recycler.DefaultHandle.access$502(element, dst);
            dstElems[(newDstSize++)] = element;
          }
        }
        if ((srcEnd == Recycler.LINK_CAPACITY) && (next != null))
        {
          reclaimSpace(Recycler.LINK_CAPACITY);
          
          this.head = next;
        }
        readIndex = srcEnd;
        if (size == newDstSize) {
          return false;
        }
        size = newDstSize;
        return true;
      }
      return false;
    }
    
    protected void finalize()
      throws Throwable
    {
      try
      {
        super.finalize();
      }
      finally
      {
        Link link;
        Link link = head;
        while (link != null)
        {
          reclaimSpace(Recycler.LINK_CAPACITY);
          link = next;
        }
      }
    }
  }
  
  static final class Stack<T>
  {
    final Recycler<T> parent;
    final Thread thread;
    final AtomicInteger availableSharedCapacity;
    final int maxDelayedQueues;
    private final int maxCapacity;
    private final int ratioMask;
    private Recycler.DefaultHandle<?>[] elements;
    private int size;
    private int handleRecycleCount = -1;
    private Recycler.WeakOrderQueue cursor;
    private Recycler.WeakOrderQueue prev;
    private volatile Recycler.WeakOrderQueue head;
    
    Stack(Recycler<T> parent, Thread thread, int maxCapacity, int maxSharedCapacityFactor, int ratioMask, int maxDelayedQueues)
    {
      this.parent = parent;
      this.thread = thread;
      this.maxCapacity = maxCapacity;
      availableSharedCapacity = new AtomicInteger(Math.max(maxCapacity / maxSharedCapacityFactor, Recycler.LINK_CAPACITY));
      elements = new Recycler.DefaultHandle[Math.min(Recycler.INITIAL_CAPACITY, maxCapacity)];
      this.ratioMask = ratioMask;
      this.maxDelayedQueues = maxDelayedQueues;
    }
    
    int increaseCapacity(int expectedCapacity)
    {
      int newCapacity = elements.length;
      int maxCapacity = this.maxCapacity;
      do
      {
        newCapacity <<= 1;
      } while ((newCapacity < expectedCapacity) && (newCapacity < maxCapacity));
      newCapacity = Math.min(newCapacity, maxCapacity);
      if (newCapacity != elements.length) {
        elements = ((Recycler.DefaultHandle[])Arrays.copyOf(elements, newCapacity));
      }
      return newCapacity;
    }
    
    Recycler.DefaultHandle<T> pop()
    {
      int size = this.size;
      if (size == 0)
      {
        if (!scavenge()) {
          return null;
        }
        size = this.size;
      }
      size--;
      Recycler.DefaultHandle ret = elements[size];
      elements[size] = null;
      if (Recycler.DefaultHandle.access$1200(ret) != Recycler.DefaultHandle.access$1600(ret)) {
        throw new IllegalStateException("recycled multiple times");
      }
      Recycler.DefaultHandle.access$1602(ret, 0);
      Recycler.DefaultHandle.access$1202(ret, 0);
      this.size = size;
      return ret;
    }
    
    boolean scavenge()
    {
      if (scavengeSome()) {
        return true;
      }
      prev = null;
      cursor = head;
      return false;
    }
    
    boolean scavengeSome()
    {
      Recycler.WeakOrderQueue cursor = this.cursor;
      if (cursor == null)
      {
        cursor = head;
        if (cursor == null) {
          return false;
        }
      }
      boolean success = false;
      Recycler.WeakOrderQueue prev = this.prev;
      do
      {
        if (cursor.transfer(this))
        {
          success = true;
          break;
        }
        Recycler.WeakOrderQueue next = Recycler.WeakOrderQueue.access$1800(cursor);
        if (Recycler.WeakOrderQueue.access$1900(cursor).get() == null)
        {
          if (cursor.hasFinalData()) {
            while (cursor.transfer(this)) {
              success = true;
            }
          }
          if (prev != null) {
            Recycler.WeakOrderQueue.access$1802(prev, next);
          }
        }
        else
        {
          prev = cursor;
        }
        cursor = next;
      } while ((cursor != null) && (!success));
      this.prev = prev;
      this.cursor = cursor;
      return success;
    }
    
    void push(Recycler.DefaultHandle<?> item)
    {
      Thread currentThread = Thread.currentThread();
      if (thread == currentThread) {
        pushNow(item);
      } else {
        pushLater(item, currentThread);
      }
    }
    
    private void pushNow(Recycler.DefaultHandle<?> item)
    {
      if ((Recycler.DefaultHandle.access$1600(item) | Recycler.DefaultHandle.access$1200(item)) != 0) {
        throw new IllegalStateException("recycled already");
      }
      Recycler.DefaultHandle.access$1602(item, Recycler.DefaultHandle.access$1202(item, Recycler.OWN_THREAD_ID));
      
      int size = this.size;
      if ((size >= maxCapacity) || (dropHandle(item))) {
        return;
      }
      if (size == elements.length) {
        elements = ((Recycler.DefaultHandle[])Arrays.copyOf(elements, Math.min(size << 1, maxCapacity)));
      }
      elements[size] = item;
      this.size = (size + 1);
    }
    
    private void pushLater(Recycler.DefaultHandle<?> item, Thread thread)
    {
      Map<Stack<?>, Recycler.WeakOrderQueue> delayedRecycled = (Map)Recycler.DELAYED_RECYCLED.get();
      Recycler.WeakOrderQueue queue = (Recycler.WeakOrderQueue)delayedRecycled.get(this);
      if (queue == null)
      {
        if (delayedRecycled.size() >= maxDelayedQueues)
        {
          delayedRecycled.put(this, Recycler.WeakOrderQueue.DUMMY);
          return;
        }
        if ((queue = Recycler.WeakOrderQueue.allocate(this, thread)) == null) {
          return;
        }
        delayedRecycled.put(this, queue);
      }
      else if (queue == Recycler.WeakOrderQueue.DUMMY)
      {
        return;
      }
      queue.add(item);
    }
    
    boolean dropHandle(Recycler.DefaultHandle<?> handle)
    {
      if (!hasBeenRecycled)
      {
        if ((++handleRecycleCount & ratioMask) != 0) {
          return true;
        }
        hasBeenRecycled = true;
      }
      return false;
    }
    
    Recycler.DefaultHandle<T> newHandle()
    {
      return new Recycler.DefaultHandle(this);
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.util.Recycler
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */