package io.netty.util;

import io.netty.util.internal.MathUtil;
import io.netty.util.internal.PlatformDependent;
import io.netty.util.internal.StringUtil;
import io.netty.util.internal.SystemPropertyUtil;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;
import java.lang.ref.PhantomReference;
import java.lang.ref.ReferenceQueue;
import java.util.ArrayDeque;
import java.util.Deque;
import java.util.EnumSet;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.atomic.AtomicBoolean;

public class ResourceLeakDetector<T>
{
  private static final String PROP_LEVEL_OLD = "io.netty.leakDetectionLevel";
  private static final String PROP_LEVEL = "io.netty.leakDetection.level";
  private static final Level DEFAULT_LEVEL = Level.SIMPLE;
  private static final String PROP_MAX_RECORDS = "io.netty.leakDetection.maxRecords";
  private static final int DEFAULT_MAX_RECORDS = 4;
  private static final int MAX_RECORDS;
  private static Level level;
  
  public static enum Level
  {
    DISABLED,  SIMPLE,  ADVANCED,  PARANOID;
    
    private Level() {}
  }
  
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(ResourceLeakDetector.class);
  static final int DEFAULT_SAMPLING_INTERVAL = 128;
  
  static
  {
    boolean disabled;
    if (SystemPropertyUtil.get("io.netty.noResourceLeakDetection") != null)
    {
      boolean disabled = SystemPropertyUtil.getBoolean("io.netty.noResourceLeakDetection", false);
      logger.debug("-Dio.netty.noResourceLeakDetection: {}", Boolean.valueOf(disabled));
      logger.warn("-Dio.netty.noResourceLeakDetection is deprecated. Use '-D{}={}' instead.", "io.netty.leakDetection.level", DEFAULT_LEVEL.name().toLowerCase());
    }
    else
    {
      disabled = false;
    }
    Level defaultLevel = disabled ? Level.DISABLED : DEFAULT_LEVEL;
    
    String levelStr = SystemPropertyUtil.get("io.netty.leakDetectionLevel", defaultLevel.name()).trim().toUpperCase();
    
    levelStr = SystemPropertyUtil.get("io.netty.leakDetection.level", levelStr).trim().toUpperCase();
    Level level = DEFAULT_LEVEL;
    for (Level l : EnumSet.allOf(Level.class)) {
      if ((levelStr.equals(l.name())) || (levelStr.equals(String.valueOf(l.ordinal())))) {
        level = l;
      }
    }
    MAX_RECORDS = SystemPropertyUtil.getInt("io.netty.leakDetection.maxRecords", 4);
    
    level = level;
    if (logger.isDebugEnabled())
    {
      logger.debug("-D{}: {}", "io.netty.leakDetection.level", level.name().toLowerCase());
      logger.debug("-D{}: {}", "io.netty.leakDetection.maxRecords", Integer.valueOf(MAX_RECORDS));
    }
  }
  
  @Deprecated
  public static void setEnabled(boolean enabled)
  {
    setLevel(enabled ? Level.SIMPLE : Level.DISABLED);
  }
  
  public static boolean isEnabled()
  {
    return getLevel().ordinal() > Level.DISABLED.ordinal();
  }
  
  public static void setLevel(Level level)
  {
    if (level == null) {
      throw new NullPointerException("level");
    }
    level = level;
  }
  
  public static Level getLevel()
  {
    return level;
  }
  
  private final ResourceLeakDetector<T>.DefaultResourceLeak head = new DefaultResourceLeak(null);
  private final ResourceLeakDetector<T>.DefaultResourceLeak tail = new DefaultResourceLeak(null);
  private final ReferenceQueue<Object> refQueue = new ReferenceQueue();
  private final ConcurrentMap<String, Boolean> reportedLeaks = PlatformDependent.newConcurrentHashMap();
  private final String resourceType;
  private final int samplingInterval;
  private final int mask;
  private final long maxActive;
  private long active;
  private final AtomicBoolean loggedTooManyActive = new AtomicBoolean();
  private long leakCheckCnt;
  
  @Deprecated
  public ResourceLeakDetector(Class<?> resourceType)
  {
    this(StringUtil.simpleClassName(resourceType));
  }
  
  @Deprecated
  public ResourceLeakDetector(String resourceType)
  {
    this(resourceType, 128, Long.MAX_VALUE);
  }
  
  public ResourceLeakDetector(Class<?> resourceType, int samplingInterval, long maxActive)
  {
    this(StringUtil.simpleClassName(resourceType), samplingInterval, maxActive);
  }
  
  @Deprecated
  public ResourceLeakDetector(String resourceType, int samplingInterval, long maxActive)
  {
    if (resourceType == null) {
      throw new NullPointerException("resourceType");
    }
    if (maxActive <= 0L) {
      throw new IllegalArgumentException("maxActive: " + maxActive + " (expected: 1+)");
    }
    this.resourceType = resourceType;
    this.samplingInterval = MathUtil.safeFindNextPositivePowerOfTwo(samplingInterval);
    
    mask = (this.samplingInterval - 1);
    this.maxActive = maxActive;
    
    head.next = tail;
    tail.prev = head;
  }
  
  public final ResourceLeak open(T obj)
  {
    Level level = level;
    if (level == Level.DISABLED) {
      return null;
    }
    if (level.ordinal() < Level.PARANOID.ordinal())
    {
      if ((leakCheckCnt++ & mask) == 0L)
      {
        reportLeak(level);
        return new DefaultResourceLeak(obj);
      }
      return null;
    }
    reportLeak(level);
    return new DefaultResourceLeak(obj);
  }
  
  private void reportLeak(Level level)
  {
    if (!logger.isErrorEnabled())
    {
      for (;;)
      {
        ResourceLeakDetector<T>.DefaultResourceLeak ref = (DefaultResourceLeak)refQueue.poll();
        if (ref == null) {
          break;
        }
        ref.close();
      }
      return;
    }
    int samplingInterval = level == Level.PARANOID ? 1 : this.samplingInterval;
    if ((active * samplingInterval > maxActive) && (loggedTooManyActive.compareAndSet(false, true))) {
      reportInstancesLeak(resourceType);
    }
    for (;;)
    {
      ResourceLeakDetector<T>.DefaultResourceLeak ref = (DefaultResourceLeak)refQueue.poll();
      if (ref == null) {
        break;
      }
      ref.clear();
      if (ref.close())
      {
        String records = ref.toString();
        if (reportedLeaks.putIfAbsent(records, Boolean.TRUE) == null) {
          if (records.isEmpty()) {
            reportUntracedLeak(resourceType);
          } else {
            reportTracedLeak(resourceType, records);
          }
        }
      }
    }
  }
  
  protected void reportTracedLeak(String resourceType, String records)
  {
    logger.error("LEAK: {}.release() was not called before it's garbage-collected. See http://netty.io/wiki/reference-counted-objects.html for more information.{}", resourceType, records);
  }
  
  protected void reportUntracedLeak(String resourceType)
  {
    logger.error("LEAK: {}.release() was not called before it's garbage-collected. Enable advanced leak reporting to find out where the leak occurred. To enable advanced leak reporting, specify the JVM option '-D{}={}' or call {}.setLevel() See http://netty.io/wiki/reference-counted-objects.html for more information.", new Object[] { resourceType, "io.netty.leakDetection.level", Level.ADVANCED.name().toLowerCase(), StringUtil.simpleClassName(this) });
  }
  
  protected void reportInstancesLeak(String resourceType)
  {
    logger.error("LEAK: You are creating too many " + resourceType + " instances.  " + resourceType + " is a shared resource that must be reused across the JVM," + "so that only a few instances are created.");
  }
  
  private final class DefaultResourceLeak
    extends PhantomReference<Object>
    implements ResourceLeak
  {
    private final String creationRecord;
    private final Deque<String> lastRecords = new ArrayDeque();
    private final AtomicBoolean freed;
    private ResourceLeakDetector<T>.DefaultResourceLeak prev;
    private ResourceLeakDetector<T>.DefaultResourceLeak next;
    private int removedRecords;
    
    DefaultResourceLeak(Object referent)
    {
      super(referent != null ? refQueue : null);
      ResourceLeakDetector.Level level;
      if (referent != null)
      {
        level = ResourceLeakDetector.getLevel();
        if (level.ordinal() >= ResourceLeakDetector.Level.ADVANCED.ordinal()) {
          creationRecord = ResourceLeakDetector.newRecord(null, 3);
        } else {
          creationRecord = null;
        }
        synchronized (head)
        {
          prev = head;
          next = head.next;
          head.next.prev = this;
          head.next = this;
          ResourceLeakDetector.access$408(ResourceLeakDetector.this);
        }
        freed = new AtomicBoolean();
      }
      else
      {
        creationRecord = null;
        freed = new AtomicBoolean(true);
      }
    }
    
    public void record()
    {
      record0(null, 3);
    }
    
    public void record(Object hint)
    {
      record0(hint, 3);
    }
    
    private void record0(Object hint, int recordsToSkip)
    {
      if (creationRecord != null)
      {
        String value = ResourceLeakDetector.newRecord(hint, recordsToSkip);
        synchronized (lastRecords)
        {
          int size = lastRecords.size();
          if ((size == 0) || (!((String)lastRecords.getLast()).equals(value))) {
            lastRecords.add(value);
          }
          if (size > ResourceLeakDetector.MAX_RECORDS)
          {
            lastRecords.removeFirst();
            removedRecords += 1;
          }
        }
      }
    }
    
    public boolean close()
    {
      if (freed.compareAndSet(false, true))
      {
        synchronized (head)
        {
          ResourceLeakDetector.access$410(ResourceLeakDetector.this);
          prev.next = next;
          next.prev = prev;
          prev = null;
          next = null;
        }
        return true;
      }
      return false;
    }
    
    public String toString()
    {
      if (creationRecord == null) {
        return "";
      }
      int removedRecords;
      synchronized (lastRecords)
      {
        Object[] array = lastRecords.toArray();
        removedRecords = this.removedRecords;
      }
      int removedRecords;
      Object[] array;
      StringBuilder buf = new StringBuilder(16384).append(StringUtil.NEWLINE);
      if (removedRecords > 0) {
        buf.append("WARNING: ").append(removedRecords).append(" leak records were discarded because the leak record count is limited to ").append(ResourceLeakDetector.MAX_RECORDS).append(". Use system property ").append("io.netty.leakDetection.maxRecords").append(" to increase the limit.").append(StringUtil.NEWLINE);
      }
      buf.append("Recent access records: ").append(array.length).append(StringUtil.NEWLINE);
      if (array.length > 0) {
        for (int i = array.length - 1; i >= 0; i--) {
          buf.append('#').append(i + 1).append(':').append(StringUtil.NEWLINE).append(array[i]);
        }
      }
      buf.append("Created at:").append(StringUtil.NEWLINE).append(creationRecord);
      
      buf.setLength(buf.length() - StringUtil.NEWLINE.length());
      return buf.toString();
    }
  }
  
  private static final String[] STACK_TRACE_ELEMENT_EXCLUSIONS = { "io.netty.util.ReferenceCountUtil.touch(", "io.netty.buffer.AdvancedLeakAwareByteBuf.touch(", "io.netty.buffer.AbstractByteBufAllocator.toLeakAwareBuffer(", "io.netty.buffer.AdvancedLeakAwareByteBuf.recordLeakNonRefCountingOperation(" };
  
  static String newRecord(Object hint, int recordsToSkip)
  {
    StringBuilder buf = new StringBuilder(4096);
    if (hint != null)
    {
      buf.append("\tHint: ");
      if ((hint instanceof ResourceLeakHint)) {
        buf.append(((ResourceLeakHint)hint).toHintString());
      } else {
        buf.append(hint);
      }
      buf.append(StringUtil.NEWLINE);
    }
    StackTraceElement[] array = new Throwable().getStackTrace();
    for (StackTraceElement e : array) {
      if (recordsToSkip > 0)
      {
        recordsToSkip--;
      }
      else
      {
        String estr = e.toString();
        
        boolean excluded = false;
        for (String exclusion : STACK_TRACE_ELEMENT_EXCLUSIONS) {
          if (estr.startsWith(exclusion))
          {
            excluded = true;
            break;
          }
        }
        if (!excluded)
        {
          buf.append('\t');
          buf.append(estr);
          buf.append(StringUtil.NEWLINE);
        }
      }
    }
    return buf.toString();
  }
}

/* Location:
 * Qualified Name:     io.netty.util.ResourceLeakDetector
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */