package io.netty.util.internal;

import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;
import java.io.ByteArrayOutputStream;
import java.io.Closeable;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.Method;
import java.net.URL;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Arrays;
import java.util.Locale;

public final class NativeLibraryLoader
{
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(NativeLibraryLoader.class);
  private static final String NATIVE_RESOURCE_HOME = "META-INF/native/";
  private static final String OSNAME = SystemPropertyUtil.get("os.name", "").toLowerCase(Locale.US).replaceAll("[^a-z0-9]+", "");
  private static final File WORKDIR;
  
  static
  {
    String workdir = SystemPropertyUtil.get("io.netty.native.workdir");
    if (workdir != null)
    {
      File f = new File(workdir);
      f.mkdirs();
      try
      {
        f = f.getAbsoluteFile();
      }
      catch (Exception localException) {}
      WORKDIR = f;
      logger.debug("-Dio.netty.native.workdir: " + WORKDIR);
    }
    else
    {
      WORKDIR = tmpdir();
      logger.debug("-Dio.netty.native.workdir: " + WORKDIR + " (io.netty.tmpdir)");
    }
  }
  
  private static File tmpdir()
  {
    try
    {
      File f = toDirectory(SystemPropertyUtil.get("io.netty.tmpdir"));
      if (f != null)
      {
        logger.debug("-Dio.netty.tmpdir: " + f);
        return f;
      }
      f = toDirectory(SystemPropertyUtil.get("java.io.tmpdir"));
      if (f != null)
      {
        logger.debug("-Dio.netty.tmpdir: " + f + " (java.io.tmpdir)");
        return f;
      }
      if (isWindows())
      {
        f = toDirectory(System.getenv("TEMP"));
        if (f != null)
        {
          logger.debug("-Dio.netty.tmpdir: " + f + " (%TEMP%)");
          return f;
        }
        String userprofile = System.getenv("USERPROFILE");
        if (userprofile != null)
        {
          f = toDirectory(userprofile + "\\AppData\\Local\\Temp");
          if (f != null)
          {
            logger.debug("-Dio.netty.tmpdir: " + f + " (%USERPROFILE%\\AppData\\Local\\Temp)");
            return f;
          }
          f = toDirectory(userprofile + "\\Local Settings\\Temp");
          if (f != null)
          {
            logger.debug("-Dio.netty.tmpdir: " + f + " (%USERPROFILE%\\Local Settings\\Temp)");
            return f;
          }
        }
      }
      else
      {
        f = toDirectory(System.getenv("TMPDIR"));
        if (f != null)
        {
          logger.debug("-Dio.netty.tmpdir: " + f + " ($TMPDIR)");
          return f;
        }
      }
    }
    catch (Exception localException) {}
    File f;
    File f;
    if (isWindows()) {
      f = new File("C:\\Windows\\Temp");
    } else {
      f = new File("/tmp");
    }
    logger.warn("Failed to get the temporary directory; falling back to: " + f);
    return f;
  }
  
  private static File toDirectory(String path)
  {
    if (path == null) {
      return null;
    }
    File f = new File(path);
    f.mkdirs();
    if (!f.isDirectory()) {
      return null;
    }
    try
    {
      return f.getAbsoluteFile();
    }
    catch (Exception ignored) {}
    return f;
  }
  
  private static boolean isWindows()
  {
    return OSNAME.startsWith("windows");
  }
  
  private static boolean isOSX()
  {
    return (OSNAME.startsWith("macosx")) || (OSNAME.startsWith("osx"));
  }
  
  public static void loadFirstAvailable(ClassLoader loader, String... names)
  {
    for (String name : names) {
      try
      {
        load(name, loader);
        return;
      }
      catch (Throwable t)
      {
        logger.debug("Unable to load the library: " + name + '.', t);
      }
    }
    throw new IllegalArgumentException("Failed to load any of the given libraries: " + Arrays.toString(names));
  }
  
  public static void load(String name, ClassLoader loader)
  {
    String libname = System.mapLibraryName(name);
    String path = "META-INF/native/" + libname;
    
    URL url = loader.getResource(path);
    if ((url == null) && (isOSX())) {
      if (path.endsWith(".jnilib")) {
        url = loader.getResource("META-INF/native/lib" + name + ".dynlib");
      } else {
        url = loader.getResource("META-INF/native/lib" + name + ".jnilib");
      }
    }
    if (url == null)
    {
      loadLibrary(loader, name, false);
      return;
    }
    int index = libname.lastIndexOf('.');
    String prefix = libname.substring(0, index);
    String suffix = libname.substring(index, libname.length());
    InputStream in = null;
    OutputStream out = null;
    File tmpFile = null;
    try
    {
      tmpFile = File.createTempFile(prefix, suffix, WORKDIR);
      in = url.openStream();
      out = new FileOutputStream(tmpFile);
      
      byte[] buffer = new byte['?'];
      int length;
      while ((length = in.read(buffer)) > 0) {
        out.write(buffer, 0, length);
      }
      out.flush();
      
      loadLibrary(loader, tmpFile.getPath(), true);
    }
    catch (Exception e)
    {
      throw ((UnsatisfiedLinkError)new UnsatisfiedLinkError("could not load a native library: " + name).initCause(e));
    }
    finally
    {
      closeQuietly(in);
      closeQuietly(out);
      if ((tmpFile != null) && (!tmpFile.delete())) {
        tmpFile.deleteOnExit();
      }
    }
  }
  
  private static void loadLibrary(ClassLoader loader, String name, boolean absolute)
  {
    try
    {
      Class<?> newHelper = tryToLoadClass(loader, NativeLibraryUtil.class);
      loadLibraryByHelper(newHelper, name, absolute);
    }
    catch (Exception e)
    {
      logger.debug("Unable to load the library: " + name + '.', e);
      NativeLibraryUtil.loadLibrary(name, absolute);
    }
  }
  
  private static void loadLibraryByHelper(Class<?> helper, final String name, final boolean absolute)
  {
    AccessController.doPrivileged(new PrivilegedAction()
    {
      public Object run()
      {
        try
        {
          Method method = val$helper.getMethod("loadLibrary", new Class[] { String.class, Boolean.TYPE });
          
          method.setAccessible(true);
          return method.invoke(null, new Object[] { name, Boolean.valueOf(absolute) });
        }
        catch (Exception e)
        {
          throw new IllegalStateException("Load library failed!", e);
        }
      }
    });
  }
  
  private static Class<?> tryToLoadClass(ClassLoader loader, final Class<?> helper)
    throws ClassNotFoundException
  {
    try
    {
      return loader.loadClass(helper.getName());
    }
    catch (ClassNotFoundException e)
    {
      final byte[] classBinary = classToByteArray(helper);
      (Class)AccessController.doPrivileged(new PrivilegedAction()
      {
        public Class<?> run()
        {
          try
          {
            Method defineClass = ClassLoader.class.getDeclaredMethod("defineClass", new Class[] { String.class, byte[].class, Integer.TYPE, Integer.TYPE });
            
            defineClass.setAccessible(true);
            return (Class)defineClass.invoke(val$loader, new Object[] { helper.getName(), classBinary, Integer.valueOf(0), Integer.valueOf(classBinary.length) });
          }
          catch (Exception e)
          {
            throw new IllegalStateException("Define class failed!", e);
          }
        }
      });
    }
  }
  
  private static byte[] classToByteArray(Class<?> clazz)
    throws ClassNotFoundException
  {
    String fileName = clazz.getName();
    int lastDot = fileName.lastIndexOf('.');
    if (lastDot > 0) {
      fileName = fileName.substring(lastDot + 1);
    }
    URL classUrl = clazz.getResource(fileName + ".class");
    if (classUrl == null) {
      throw new ClassNotFoundException(clazz.getName());
    }
    byte[] buf = new byte['?'];
    ByteArrayOutputStream out = new ByteArrayOutputStream(4096);
    InputStream in = null;
    try
    {
      in = classUrl.openStream();
      int r;
      while ((r = in.read(buf)) != -1) {
        out.write(buf, 0, r);
      }
      return out.toByteArray();
    }
    catch (IOException ex)
    {
      throw new ClassNotFoundException(clazz.getName(), ex);
    }
    finally
    {
      closeQuietly(in);
      closeQuietly(out);
    }
  }
  
  private static void closeQuietly(Closeable c)
  {
    if (c != null) {
      try
      {
        c.close();
      }
      catch (IOException localIOException) {}
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.util.internal.NativeLibraryLoader
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */