package io.netty.util.internal;

import io.netty.util.CharsetUtil;
import io.netty.util.internal.chmv8.ConcurrentHashMapV8;
import io.netty.util.internal.chmv8.LongAdderV8;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;
import io.netty.util.internal.shaded.org.jctools.queues.MpscArrayQueue;
import io.netty.util.internal.shaded.org.jctools.queues.MpscChunkedArrayQueue;
import io.netty.util.internal.shaded.org.jctools.queues.SpscLinkedQueue;
import io.netty.util.internal.shaded.org.jctools.queues.atomic.MpscAtomicArrayQueue;
import io.netty.util.internal.shaded.org.jctools.queues.atomic.MpscLinkedAtomicQueue;
import io.netty.util.internal.shaded.org.jctools.queues.atomic.SpscLinkedAtomicQueue;
import io.netty.util.internal.shaded.org.jctools.util.UnsafeAccess;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Deque;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Queue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedDeque;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.atomic.AtomicIntegerFieldUpdater;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.atomic.AtomicLongFieldUpdater;
import java.util.concurrent.atomic.AtomicReferenceFieldUpdater;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public final class PlatformDependent
{
  private static final InternalLogger logger;
  private static final Pattern MAX_DIRECT_MEMORY_SIZE_ARG_PATTERN;
  private static final boolean IS_ANDROID;
  private static final boolean IS_WINDOWS;
  private static volatile Boolean IS_ROOT;
  private static final int JAVA_VERSION;
  private static final boolean CAN_ENABLE_TCP_NODELAY_BY_DEFAULT;
  private static final boolean IS_EXPLICIT_NO_UNSAFE;
  private static final boolean HAS_UNSAFE;
  private static final boolean CAN_USE_CHM_V8;
  private static final boolean DIRECT_BUFFER_PREFERRED;
  private static final long MAX_DIRECT_MEMORY;
  private static final int MPSC_CHUNK_SIZE = 1024;
  private static final int MIN_MAX_MPSC_CAPACITY = 2048;
  private static final int DEFAULT_MAX_MPSC_CAPACITY = 1048576;
  private static final int MAX_ALLOWED_MPSC_CAPACITY = 1073741824;
  private static final long BYTE_ARRAY_BASE_OFFSET;
  private static final boolean HAS_JAVASSIST;
  private static final File TMPDIR;
  private static final int BIT_MODE;
  private static final int ADDRESS_SIZE;
  private static final boolean USE_DIRECT_BUFFER_NO_CLEANER;
  private static final AtomicLong DIRECT_MEMORY_COUNTER;
  private static final long DIRECT_MEMORY_LIMIT;
  public static final boolean BIG_ENDIAN_NATIVE_ORDER;
  
  static
  {
    logger = InternalLoggerFactory.getInstance(PlatformDependent.class);
    
    MAX_DIRECT_MEMORY_SIZE_ARG_PATTERN = Pattern.compile("\\s*-XX:MaxDirectMemorySize\\s*=\\s*([0-9]+)\\s*([kKmMgG]?)\\s*$");
    
    IS_ANDROID = isAndroid0();
    IS_WINDOWS = isWindows0();
    
    JAVA_VERSION = javaVersion0();
    
    CAN_ENABLE_TCP_NODELAY_BY_DEFAULT = !isAndroid();
    
    IS_EXPLICIT_NO_UNSAFE = explicitNoUnsafe0();
    HAS_UNSAFE = hasUnsafe0();
    CAN_USE_CHM_V8 = (HAS_UNSAFE) && (JAVA_VERSION < 8);
    DIRECT_BUFFER_PREFERRED = (HAS_UNSAFE) && (!SystemPropertyUtil.getBoolean("io.netty.noPreferDirect", false));
    
    MAX_DIRECT_MEMORY = maxDirectMemory0();
    
    BYTE_ARRAY_BASE_OFFSET = PlatformDependent0.byteArrayBaseOffset();
    
    HAS_JAVASSIST = hasJavassist0();
    
    TMPDIR = tmpdir0();
    
    BIT_MODE = bitMode0();
    
    ADDRESS_SIZE = addressSize0();
    
    BIG_ENDIAN_NATIVE_ORDER = ByteOrder.nativeOrder() == ByteOrder.BIG_ENDIAN;
    if (logger.isDebugEnabled()) {
      logger.debug("-Dio.netty.noPreferDirect: {}", Boolean.valueOf(!DIRECT_BUFFER_PREFERRED));
    }
    if ((!hasUnsafe()) && (!isAndroid()) && (!IS_EXPLICIT_NO_UNSAFE)) {
      logger.info("Your platform does not provide complete low-level API for accessing direct buffers reliably. Unless explicitly requested, heap buffer will always be preferred to avoid potential system unstability.");
    }
    long maxDirectMemory = SystemPropertyUtil.getLong("io.netty.maxDirectMemory", -1L);
    if ((maxDirectMemory == 0L) || (!hasUnsafe()) || (!PlatformDependent0.hasDirectBufferNoCleanerConstructor()))
    {
      USE_DIRECT_BUFFER_NO_CLEANER = false;
      DIRECT_MEMORY_COUNTER = null;
    }
    else
    {
      USE_DIRECT_BUFFER_NO_CLEANER = true;
      if (maxDirectMemory < 0L)
      {
        maxDirectMemory = maxDirectMemory0();
        if (maxDirectMemory <= 0L) {
          DIRECT_MEMORY_COUNTER = null;
        } else {
          DIRECT_MEMORY_COUNTER = new AtomicLong();
        }
      }
      else
      {
        DIRECT_MEMORY_COUNTER = new AtomicLong();
      }
    }
    DIRECT_MEMORY_LIMIT = maxDirectMemory;
    logger.debug("io.netty.maxDirectMemory: {} bytes", Long.valueOf(maxDirectMemory));
  }
  
  public static boolean isAndroid()
  {
    return IS_ANDROID;
  }
  
  public static boolean isWindows()
  {
    return IS_WINDOWS;
  }
  
  public static boolean isRoot()
  {
    if (IS_ROOT == null) {
      synchronized (PlatformDependent.class)
      {
        if (IS_ROOT == null) {
          IS_ROOT = Boolean.valueOf(isRoot0());
        }
      }
    }
    return IS_ROOT.booleanValue();
  }
  
  public static int javaVersion()
  {
    return JAVA_VERSION;
  }
  
  public static boolean canEnableTcpNoDelayByDefault()
  {
    return CAN_ENABLE_TCP_NODELAY_BY_DEFAULT;
  }
  
  public static boolean hasUnsafe()
  {
    return HAS_UNSAFE;
  }
  
  public static boolean isUnaligned()
  {
    return PlatformDependent0.isUnaligned();
  }
  
  public static boolean directBufferPreferred()
  {
    return DIRECT_BUFFER_PREFERRED;
  }
  
  public static long maxDirectMemory()
  {
    return MAX_DIRECT_MEMORY;
  }
  
  public static boolean hasJavassist()
  {
    return HAS_JAVASSIST;
  }
  
  public static File tmpdir()
  {
    return TMPDIR;
  }
  
  public static int bitMode()
  {
    return BIT_MODE;
  }
  
  public static int addressSize()
  {
    return ADDRESS_SIZE;
  }
  
  public static long allocateMemory(long size)
  {
    return PlatformDependent0.allocateMemory(size);
  }
  
  public static void freeMemory(long address)
  {
    PlatformDependent0.freeMemory(address);
  }
  
  public static void throwException(Throwable t)
  {
    if (hasUnsafe()) {
      PlatformDependent0.throwException(t);
    } else {
      throwException0(t);
    }
  }
  
  private static <E extends Throwable> void throwException0(Throwable t)
    throws Throwable
  {
    throw t;
  }
  
  public static <K, V> ConcurrentMap<K, V> newConcurrentHashMap()
  {
    if (CAN_USE_CHM_V8) {
      return new ConcurrentHashMapV8();
    }
    return new ConcurrentHashMap();
  }
  
  public static LongCounter newLongCounter()
  {
    if (HAS_UNSAFE) {
      return new LongAdderV8();
    }
    return new AtomicLongCounter(null);
  }
  
  public static <K, V> ConcurrentMap<K, V> newConcurrentHashMap(int initialCapacity)
  {
    if (CAN_USE_CHM_V8) {
      return new ConcurrentHashMapV8(initialCapacity);
    }
    return new ConcurrentHashMap(initialCapacity);
  }
  
  public static <K, V> ConcurrentMap<K, V> newConcurrentHashMap(int initialCapacity, float loadFactor)
  {
    if (CAN_USE_CHM_V8) {
      return new ConcurrentHashMapV8(initialCapacity, loadFactor);
    }
    return new ConcurrentHashMap(initialCapacity, loadFactor);
  }
  
  public static <K, V> ConcurrentMap<K, V> newConcurrentHashMap(int initialCapacity, float loadFactor, int concurrencyLevel)
  {
    if (CAN_USE_CHM_V8) {
      return new ConcurrentHashMapV8(initialCapacity, loadFactor, concurrencyLevel);
    }
    return new ConcurrentHashMap(initialCapacity, loadFactor, concurrencyLevel);
  }
  
  public static <K, V> ConcurrentMap<K, V> newConcurrentHashMap(Map<? extends K, ? extends V> map)
  {
    if (CAN_USE_CHM_V8) {
      return new ConcurrentHashMapV8(map);
    }
    return new ConcurrentHashMap(map);
  }
  
  public static void freeDirectBuffer(ByteBuffer buffer)
  {
    if ((hasUnsafe()) && (!isAndroid())) {
      PlatformDependent0.freeDirectBuffer(buffer);
    }
  }
  
  public static long directBufferAddress(ByteBuffer buffer)
  {
    return PlatformDependent0.directBufferAddress(buffer);
  }
  
  public static ByteBuffer directBuffer(long memoryAddress, int size)
  {
    if (PlatformDependent0.hasDirectBufferNoCleanerConstructor()) {
      return PlatformDependent0.newDirectBuffer(memoryAddress, size);
    }
    throw new UnsupportedOperationException("sun.misc.Unsafe or java.nio.DirectByteBuffer.<init>(long, int) not available");
  }
  
  public static Object getObject(Object object, long fieldOffset)
  {
    return PlatformDependent0.getObject(object, fieldOffset);
  }
  
  public static Object getObjectVolatile(Object object, long fieldOffset)
  {
    return PlatformDependent0.getObjectVolatile(object, fieldOffset);
  }
  
  public static int getInt(Object object, long fieldOffset)
  {
    return PlatformDependent0.getInt(object, fieldOffset);
  }
  
  public static long objectFieldOffset(Field field)
  {
    return PlatformDependent0.objectFieldOffset(field);
  }
  
  public static byte getByte(long address)
  {
    return PlatformDependent0.getByte(address);
  }
  
  public static short getShort(long address)
  {
    return PlatformDependent0.getShort(address);
  }
  
  public static int getInt(long address)
  {
    return PlatformDependent0.getInt(address);
  }
  
  public static long getLong(long address)
  {
    return PlatformDependent0.getLong(address);
  }
  
  public static byte getByte(byte[] data, int index)
  {
    return PlatformDependent0.getByte(data, index);
  }
  
  public static short getShort(byte[] data, int index)
  {
    return PlatformDependent0.getShort(data, index);
  }
  
  public static int getInt(byte[] data, int index)
  {
    return PlatformDependent0.getInt(data, index);
  }
  
  public static long getLong(byte[] data, int index)
  {
    return PlatformDependent0.getLong(data, index);
  }
  
  private static long getLongSafe(byte[] bytes, int offset)
  {
    if (BIG_ENDIAN_NATIVE_ORDER) {
      return bytes[offset] << 56 | (bytes[(offset + 1)] & 0xFF) << 48 | (bytes[(offset + 2)] & 0xFF) << 40 | (bytes[(offset + 3)] & 0xFF) << 32 | (bytes[(offset + 4)] & 0xFF) << 24 | (bytes[(offset + 5)] & 0xFF) << 16 | (bytes[(offset + 6)] & 0xFF) << 8 | bytes[(offset + 7)] & 0xFF;
    }
    return bytes[offset] & 0xFF | (bytes[(offset + 1)] & 0xFF) << 8 | (bytes[(offset + 2)] & 0xFF) << 16 | (bytes[(offset + 3)] & 0xFF) << 24 | (bytes[(offset + 4)] & 0xFF) << 32 | (bytes[(offset + 5)] & 0xFF) << 40 | (bytes[(offset + 6)] & 0xFF) << 48 | (bytes[(offset + 7)] & 0xFF) << 56;
  }
  
  private static int getIntSafe(byte[] bytes, int offset)
  {
    if (BIG_ENDIAN_NATIVE_ORDER) {
      return bytes[offset] << 24 | (bytes[(offset + 1)] & 0xFF) << 16 | (bytes[(offset + 2)] & 0xFF) << 8 | bytes[(offset + 3)] & 0xFF;
    }
    return bytes[offset] & 0xFF | (bytes[(offset + 1)] & 0xFF) << 8 | (bytes[(offset + 2)] & 0xFF) << 16 | bytes[(offset + 3)] << 24;
  }
  
  private static short getShortSafe(byte[] bytes, int offset)
  {
    if (BIG_ENDIAN_NATIVE_ORDER) {
      return (short)(bytes[offset] << 8 | bytes[(offset + 1)] & 0xFF);
    }
    return (short)(bytes[offset] & 0xFF | bytes[(offset + 1)] << 8);
  }
  
  private static int hashCodeAsciiCompute(CharSequence value, int offset, int hash)
  {
    return hash * 461845907 + hashCodeAsciiSanitizeInt(value, offset) * 461845907 + hashCodeAsciiSanitizeInt(value, offset + 4);
  }
  
  private static int hashCodeAsciiSanitizeInt(CharSequence value, int offset)
  {
    if (BIG_ENDIAN_NATIVE_ORDER) {
      return value.charAt(offset) & 0x1F | (value.charAt(offset + 2) & 0x1F) << '\b' | (value.charAt(offset + 1) & 0x1F) << '\020' | (value.charAt(offset) & 0x1F) << '\030';
    }
    return (value.charAt(offset + 3) & 0x1F) << '\030' | (value.charAt(offset + 2) & 0x1F) << '\020' | (value.charAt(offset + 1) & 0x1F) << '\b' | value.charAt(offset) & 0x1F;
  }
  
  private static int hashCodeAsciiSanitizeShort(CharSequence value, int offset)
  {
    if (BIG_ENDIAN_NATIVE_ORDER) {
      return value.charAt(offset + 1) & 0x1F | (value.charAt(offset) & 0x1F) << '\b';
    }
    return (value.charAt(offset + 1) & 0x1F) << '\b' | value.charAt(offset) & 0x1F;
  }
  
  private static int hashCodeAsciiSanitizsByte(char value)
  {
    return value & 0x1F;
  }
  
  public static void putOrderedObject(Object object, long address, Object value)
  {
    PlatformDependent0.putOrderedObject(object, address, value);
  }
  
  public static void putByte(long address, byte value)
  {
    PlatformDependent0.putByte(address, value);
  }
  
  public static void putShort(long address, short value)
  {
    PlatformDependent0.putShort(address, value);
  }
  
  public static void putInt(long address, int value)
  {
    PlatformDependent0.putInt(address, value);
  }
  
  public static void putLong(long address, long value)
  {
    PlatformDependent0.putLong(address, value);
  }
  
  public static void putByte(byte[] data, int index, byte value)
  {
    PlatformDependent0.putByte(data, index, value);
  }
  
  public static void putShort(byte[] data, int index, short value)
  {
    PlatformDependent0.putShort(data, index, value);
  }
  
  public static void putInt(byte[] data, int index, int value)
  {
    PlatformDependent0.putInt(data, index, value);
  }
  
  public static void putLong(byte[] data, int index, long value)
  {
    PlatformDependent0.putLong(data, index, value);
  }
  
  public static void copyMemory(long srcAddr, long dstAddr, long length)
  {
    PlatformDependent0.copyMemory(srcAddr, dstAddr, length);
  }
  
  public static void copyMemory(byte[] src, int srcIndex, long dstAddr, long length)
  {
    PlatformDependent0.copyMemory(src, BYTE_ARRAY_BASE_OFFSET + srcIndex, null, dstAddr, length);
  }
  
  public static void copyMemory(long srcAddr, byte[] dst, int dstIndex, long length)
  {
    PlatformDependent0.copyMemory(null, srcAddr, dst, BYTE_ARRAY_BASE_OFFSET + dstIndex, length);
  }
  
  public static void setMemory(byte[] dst, int dstIndex, long bytes, byte value)
  {
    PlatformDependent0.setMemory(dst, BYTE_ARRAY_BASE_OFFSET + dstIndex, bytes, value);
  }
  
  public static void setMemory(long address, long bytes, byte value)
  {
    PlatformDependent0.setMemory(address, bytes, value);
  }
  
  public static ByteBuffer allocateDirectNoCleaner(int capacity)
  {
    assert (USE_DIRECT_BUFFER_NO_CLEANER);
    
    incrementMemoryCounter(capacity);
    try
    {
      return PlatformDependent0.allocateDirectNoCleaner(capacity);
    }
    catch (Throwable e)
    {
      decrementMemoryCounter(capacity);
      throwException(e);
    }
    return null;
  }
  
  public static ByteBuffer reallocateDirectNoCleaner(ByteBuffer buffer, int capacity)
  {
    assert (USE_DIRECT_BUFFER_NO_CLEANER);
    
    int len = capacity - buffer.capacity();
    incrementMemoryCounter(len);
    try
    {
      return PlatformDependent0.reallocateDirectNoCleaner(buffer, capacity);
    }
    catch (Throwable e)
    {
      decrementMemoryCounter(len);
      throwException(e);
    }
    return null;
  }
  
  public static void freeDirectNoCleaner(ByteBuffer buffer)
  {
    assert (USE_DIRECT_BUFFER_NO_CLEANER);
    
    int capacity = buffer.capacity();
    PlatformDependent0.freeMemory(PlatformDependent0.directBufferAddress(buffer));
    decrementMemoryCounter(capacity);
  }
  
  private static void incrementMemoryCounter(int capacity)
  {
    if (DIRECT_MEMORY_COUNTER != null) {
      for (;;)
      {
        long usedMemory = DIRECT_MEMORY_COUNTER.get();
        long newUsedMemory = usedMemory + capacity;
        if (newUsedMemory > DIRECT_MEMORY_LIMIT) {
          throw new OutOfDirectMemoryError("failed to allocate " + capacity + " byte(s) of direct memory (used: " + usedMemory + ", max: " + DIRECT_MEMORY_LIMIT + ')');
        }
        if (DIRECT_MEMORY_COUNTER.compareAndSet(usedMemory, newUsedMemory)) {
          break;
        }
      }
    }
  }
  
  private static void decrementMemoryCounter(int capacity)
  {
    if (DIRECT_MEMORY_COUNTER != null)
    {
      long usedMemory = DIRECT_MEMORY_COUNTER.addAndGet(-capacity);
      assert (usedMemory >= 0L);
    }
  }
  
  public static boolean useDirectBufferNoCleaner()
  {
    return USE_DIRECT_BUFFER_NO_CLEANER;
  }
  
  public static boolean equals(byte[] bytes1, int startPos1, byte[] bytes2, int startPos2, int length)
  {
    return (!hasUnsafe()) || (!PlatformDependent0.unalignedAccess()) ? equalsSafe(bytes1, startPos1, bytes2, startPos2, length) : PlatformDependent0.equals(bytes1, startPos1, bytes2, startPos2, length);
  }
  
  public static int equalsConstantTime(byte[] bytes1, int startPos1, byte[] bytes2, int startPos2, int length)
  {
    return (!hasUnsafe()) || (!PlatformDependent0.unalignedAccess()) ? ConstantTimeUtils.equalsConstantTime(bytes1, startPos1, bytes2, startPos2, length) : PlatformDependent0.equalsConstantTime(bytes1, startPos1, bytes2, startPos2, length);
  }
  
  public static int hashCodeAscii(byte[] bytes, int startPos, int length)
  {
    return (!hasUnsafe()) || (!PlatformDependent0.unalignedAccess()) ? hashCodeAsciiSafe(bytes, startPos, length) : PlatformDependent0.hashCodeAscii(bytes, startPos, length);
  }
  
  public static int hashCodeAscii(CharSequence bytes)
  {
    int hash = -1028477387;
    int remainingBytes = bytes.length() & 0x7;
    switch (bytes.length())
    {
    case 24: 
    case 25: 
    case 26: 
    case 27: 
    case 28: 
    case 29: 
    case 30: 
    case 31: 
      hash = hashCodeAsciiCompute(bytes, bytes.length() - 24, hashCodeAsciiCompute(bytes, bytes.length() - 16, hashCodeAsciiCompute(bytes, bytes.length() - 8, hash)));
      
      break;
    case 16: 
    case 17: 
    case 18: 
    case 19: 
    case 20: 
    case 21: 
    case 22: 
    case 23: 
      hash = hashCodeAsciiCompute(bytes, bytes.length() - 16, hashCodeAsciiCompute(bytes, bytes.length() - 8, hash));
      
      break;
    case 8: 
    case 9: 
    case 10: 
    case 11: 
    case 12: 
    case 13: 
    case 14: 
    case 15: 
      hash = hashCodeAsciiCompute(bytes, bytes.length() - 8, hash);
      break;
    case 0: 
    case 1: 
    case 2: 
    case 3: 
    case 4: 
    case 5: 
    case 6: 
    case 7: 
      break;
    default: 
      for (int i = bytes.length() - 8; i >= remainingBytes; i -= 8) {
        hash = hashCodeAsciiCompute(bytes, i, hash);
      }
    }
    switch (remainingBytes)
    {
    case 7: 
      return ((hash * 461845907 + hashCodeAsciiSanitizsByte(bytes.charAt(0))) * 461845907 + hashCodeAsciiSanitizeShort(bytes, 1)) * 461845907 + hashCodeAsciiSanitizeInt(bytes, 3);
    case 6: 
      return (hash * 461845907 + hashCodeAsciiSanitizeShort(bytes, 0)) * 461845907 + hashCodeAsciiSanitizeInt(bytes, 2);
    case 5: 
      return (hash * 461845907 + hashCodeAsciiSanitizsByte(bytes.charAt(0))) * 461845907 + hashCodeAsciiSanitizeInt(bytes, 1);
    case 4: 
      return hash * 461845907 + hashCodeAsciiSanitizeInt(bytes, 0);
    case 3: 
      return (hash * 461845907 + hashCodeAsciiSanitizsByte(bytes.charAt(0))) * 461845907 + hashCodeAsciiSanitizeShort(bytes, 1);
    case 2: 
      return hash * 461845907 + hashCodeAsciiSanitizeShort(bytes, 0);
    case 1: 
      return hash * 461845907 + hashCodeAsciiSanitizsByte(bytes.charAt(0));
    }
    return hash;
  }
  
  public static <U, W> AtomicReferenceFieldUpdater<U, W> newAtomicReferenceFieldUpdater(Class<? super U> tclass, String fieldName)
  {
    if (hasUnsafe()) {
      try
      {
        return PlatformDependent0.newAtomicReferenceFieldUpdater(tclass, fieldName);
      }
      catch (Throwable localThrowable) {}
    }
    return null;
  }
  
  public static <T> AtomicIntegerFieldUpdater<T> newAtomicIntegerFieldUpdater(Class<? super T> tclass, String fieldName)
  {
    if (hasUnsafe()) {
      try
      {
        return PlatformDependent0.newAtomicIntegerFieldUpdater(tclass, fieldName);
      }
      catch (Throwable localThrowable) {}
    }
    return null;
  }
  
  public static <T> AtomicLongFieldUpdater<T> newAtomicLongFieldUpdater(Class<? super T> tclass, String fieldName)
  {
    if (hasUnsafe()) {
      try
      {
        return PlatformDependent0.newAtomicLongFieldUpdater(tclass, fieldName);
      }
      catch (Throwable localThrowable) {}
    }
    return null;
  }
  
  private static final class Mpsc
  {
    private static final boolean USE_MPSC_CHUNKED_ARRAY_QUEUE;
    
    static
    {
      Object unsafe = null;
      if (PlatformDependent.hasUnsafe()) {
        unsafe = AccessController.doPrivileged(new PrivilegedAction()
        {
          public Object run()
          {
            return UnsafeAccess.UNSAFE;
          }
        });
      }
      if (unsafe == null)
      {
        PlatformDependent.logger.debug("org.jctools-core.MpscChunkedArrayQueue: unavailable");
        USE_MPSC_CHUNKED_ARRAY_QUEUE = false;
      }
      else
      {
        PlatformDependent.logger.debug("org.jctools-core.MpscChunkedArrayQueue: available");
        USE_MPSC_CHUNKED_ARRAY_QUEUE = true;
      }
    }
    
    static <T> Queue<T> newMpscQueue(int maxCapacity)
    {
      if (USE_MPSC_CHUNKED_ARRAY_QUEUE)
      {
        int capacity = Math.max(Math.min(maxCapacity, 1073741824), 2048);
        
        return new MpscChunkedArrayQueue(1024, capacity, true);
      }
      return new MpscLinkedAtomicQueue();
    }
  }
  
  public static <T> Queue<T> newMpscQueue()
  {
    return newMpscQueue(1048576);
  }
  
  public static <T> Queue<T> newMpscQueue(int maxCapacity)
  {
    return Mpsc.newMpscQueue(maxCapacity);
  }
  
  public static <T> Queue<T> newSpscQueue()
  {
    return hasUnsafe() ? new SpscLinkedQueue() : new SpscLinkedAtomicQueue();
  }
  
  public static <T> Queue<T> newFixedMpscQueue(int capacity)
  {
    return hasUnsafe() ? new MpscArrayQueue(capacity) : new MpscAtomicArrayQueue(capacity);
  }
  
  public static ClassLoader getClassLoader(Class<?> clazz)
  {
    return PlatformDependent0.getClassLoader(clazz);
  }
  
  public static ClassLoader getContextClassLoader()
  {
    return PlatformDependent0.getContextClassLoader();
  }
  
  public static ClassLoader getSystemClassLoader()
  {
    return PlatformDependent0.getSystemClassLoader();
  }
  
  public static <C> Deque<C> newConcurrentDeque()
  {
    if (javaVersion() < 7) {
      return new LinkedBlockingDeque();
    }
    return new ConcurrentLinkedDeque();
  }
  
  private static boolean isAndroid0()
  {
    boolean android;
    try
    {
      Class.forName("android.app.Application", false, getSystemClassLoader());
      android = true;
    }
    catch (Throwable ignored)
    {
      boolean android;
      android = false;
    }
    if (android) {
      logger.debug("Platform: Android");
    }
    return android;
  }
  
  private static boolean isWindows0()
  {
    boolean windows = SystemPropertyUtil.get("os.name", "").toLowerCase(Locale.US).contains("win");
    if (windows) {
      logger.debug("Platform: Windows");
    }
    return windows;
  }
  
  private static boolean isRoot0()
  {
    if (isWindows()) {
      return false;
    }
    String[] ID_COMMANDS = { "/usr/bin/id", "/bin/id", "/usr/xpg4/bin/id", "id" };
    Pattern UID_PATTERN = Pattern.compile("^(?:0|[1-9][0-9]*)$");
    String idCmd;
    for (idCmd : ID_COMMANDS)
    {
      Process p = null;
      BufferedReader in = null;
      String uid = null;
      try
      {
        p = Runtime.getRuntime().exec(new String[] { idCmd, "-u" });
        in = new BufferedReader(new InputStreamReader(p.getInputStream(), CharsetUtil.US_ASCII));
        uid = in.readLine();
        in.close();
        for (;;)
        {
          try
          {
            int exitCode = p.waitFor();
            if (exitCode != 0) {
              uid = null;
            }
          }
          catch (InterruptedException localInterruptedException) {}
        }
        if (in != null) {
          try
          {
            in.close();
          }
          catch (IOException localIOException) {}
        }
        if (p != null) {
          try
          {
            p.destroy();
          }
          catch (Exception localException1) {}
        }
        if (uid == null) {
          continue;
        }
      }
      catch (Throwable ignored)
      {
        uid = null;
      }
      finally
      {
        if (in != null) {
          try
          {
            in.close();
          }
          catch (IOException localIOException2) {}
        }
        if (p != null) {
          try
          {
            p.destroy();
          }
          catch (Exception localException3) {}
        }
      }
      if (UID_PATTERN.matcher(uid).matches())
      {
        logger.debug("UID: {}", uid);
        return "0".equals(uid);
      }
    }
    logger.debug("Could not determine the current UID using /usr/bin/id; attempting to bind at privileged ports.");
    
    Pattern PERMISSION_DENIED = Pattern.compile(".*(?:denied|not.*permitted).*");
    for (int i = 1023; i > 0; i--)
    {
      ServerSocket ss = null;
      try
      {
        ss = new ServerSocket();
        ss.setReuseAddress(true);
        ss.bind(new InetSocketAddress(i));
        if (logger.isDebugEnabled()) {
          logger.debug("UID: 0 (succeded to bind at port {})", Integer.valueOf(i));
        }
        return 1;
      }
      catch (Exception e)
      {
        String message = e.getMessage();
        if (message == null) {
          message = "";
        }
        message = message.toLowerCase();
        if (PERMISSION_DENIED.matcher(message).matches())
        {
          if (ss == null) {
            break;
          }
          try
          {
            ss.close();
          }
          catch (Exception localException5) {}
        }
      }
      finally
      {
        if (ss != null) {
          try
          {
            ss.close();
          }
          catch (Exception localException7) {}
        }
      }
    }
    logger.debug("UID: non-root (failed to bind at any privileged ports)");
    return false;
  }
  
  private static int javaVersion0()
  {
    int majorVersion;
    int majorVersion;
    if (isAndroid()) {
      majorVersion = 6;
    } else {
      majorVersion = majorVersionFromJavaSpecificationVersion();
    }
    logger.debug("Java version: {}", Integer.valueOf(majorVersion));
    
    return majorVersion;
  }
  
  static int majorVersionFromJavaSpecificationVersion()
  {
    try
    {
      String javaSpecVersion = (String)AccessController.doPrivileged(new PrivilegedAction()
      {
        public String run()
        {
          return System.getProperty("java.specification.version");
        }
      });
      return majorVersion(javaSpecVersion);
    }
    catch (SecurityException e)
    {
      logger.debug("security exception while reading java.specification.version", e);
    }
    return 6;
  }
  
  static int majorVersion(String javaSpecVersion)
  {
    String[] components = javaSpecVersion.split("\\.");
    int[] version = new int[components.length];
    for (int i = 0; i < components.length; i++) {
      version[i] = Integer.parseInt(components[i]);
    }
    if (version[0] == 1)
    {
      assert (version[1] >= 6);
      return version[1];
    }
    return version[0];
  }
  
  private static boolean explicitNoUnsafe0()
  {
    boolean noUnsafe = SystemPropertyUtil.getBoolean("io.netty.noUnsafe", false);
    logger.debug("-Dio.netty.noUnsafe: {}", Boolean.valueOf(noUnsafe));
    if (noUnsafe)
    {
      logger.debug("sun.misc.Unsafe: unavailable (io.netty.noUnsafe)");
      return true;
    }
    boolean tryUnsafe;
    boolean tryUnsafe;
    if (SystemPropertyUtil.contains("io.netty.tryUnsafe")) {
      tryUnsafe = SystemPropertyUtil.getBoolean("io.netty.tryUnsafe", true);
    } else {
      tryUnsafe = SystemPropertyUtil.getBoolean("org.jboss.netty.tryUnsafe", true);
    }
    if (!tryUnsafe)
    {
      logger.debug("sun.misc.Unsafe: unavailable (io.netty.tryUnsafe/org.jboss.netty.tryUnsafe)");
      return true;
    }
    return false;
  }
  
  private static boolean hasUnsafe0()
  {
    if (isAndroid())
    {
      logger.debug("sun.misc.Unsafe: unavailable (Android)");
      return false;
    }
    if (IS_EXPLICIT_NO_UNSAFE) {
      return false;
    }
    try
    {
      boolean hasUnsafe = PlatformDependent0.hasUnsafe();
      logger.debug("sun.misc.Unsafe: {}", hasUnsafe ? "available" : "unavailable");
      return hasUnsafe;
    }
    catch (Throwable ignored) {}
    return false;
  }
  
  private static long maxDirectMemory0()
  {
    long maxDirectMemory = 0L;
    try
    {
      Class<?> vmClass = Class.forName("sun.misc.VM", true, getSystemClassLoader());
      Method m = vmClass.getDeclaredMethod("maxDirectMemory", new Class[0]);
      maxDirectMemory = ((Number)m.invoke(null, new Object[0])).longValue();
    }
    catch (Throwable localThrowable) {}
    if (maxDirectMemory > 0L) {
      return maxDirectMemory;
    }
    try
    {
      Class<?> mgmtFactoryClass = Class.forName("java.lang.management.ManagementFactory", true, getSystemClassLoader());
      
      Class<?> runtimeClass = Class.forName("java.lang.management.RuntimeMXBean", true, getSystemClassLoader());
      
      Object runtime = mgmtFactoryClass.getDeclaredMethod("getRuntimeMXBean", new Class[0]).invoke(null, new Object[0]);
      
      List<String> vmArgs = (List)runtimeClass.getDeclaredMethod("getInputArguments", new Class[0]).invoke(runtime, new Object[0]);
      for (int i = vmArgs.size() - 1; i >= 0; i--)
      {
        Matcher m = MAX_DIRECT_MEMORY_SIZE_ARG_PATTERN.matcher((CharSequence)vmArgs.get(i));
        if (m.matches())
        {
          maxDirectMemory = Long.parseLong(m.group(1));
          switch (m.group(2).charAt(0))
          {
          case 'K': 
          case 'k': 
            maxDirectMemory *= 1024L;
            break;
          case 'M': 
          case 'm': 
            maxDirectMemory *= 1048576L;
            break;
          case 'G': 
          case 'g': 
            maxDirectMemory *= 1073741824L;
          }
          break;
        }
      }
    }
    catch (Throwable localThrowable1) {}
    if (maxDirectMemory <= 0L)
    {
      maxDirectMemory = Runtime.getRuntime().maxMemory();
      logger.debug("maxDirectMemory: {} bytes (maybe)", Long.valueOf(maxDirectMemory));
    }
    else
    {
      logger.debug("maxDirectMemory: {} bytes", Long.valueOf(maxDirectMemory));
    }
    return maxDirectMemory;
  }
  
  private static boolean hasJavassist0()
  {
    if (isAndroid()) {
      return false;
    }
    boolean noJavassist = SystemPropertyUtil.getBoolean("io.netty.noJavassist", false);
    logger.debug("-Dio.netty.noJavassist: {}", Boolean.valueOf(noJavassist));
    if (noJavassist)
    {
      logger.debug("Javassist: unavailable (io.netty.noJavassist)");
      return false;
    }
    try
    {
      JavassistTypeParameterMatcherGenerator.generate(Object.class, getClassLoader(PlatformDependent.class));
      logger.debug("Javassist: available");
      return true;
    }
    catch (Throwable t)
    {
      logger.debug("Javassist: unavailable");
      logger.debug("You don't have Javassist in your class path or you don't have enough permission to load dynamically generated classes.  Please check the configuration for better performance.");
    }
    return false;
  }
  
  private static File tmpdir0()
  {
    try
    {
      File f = toDirectory(SystemPropertyUtil.get("io.netty.tmpdir"));
      if (f != null)
      {
        logger.debug("-Dio.netty.tmpdir: {}", f);
        return f;
      }
      f = toDirectory(SystemPropertyUtil.get("java.io.tmpdir"));
      if (f != null)
      {
        logger.debug("-Dio.netty.tmpdir: {} (java.io.tmpdir)", f);
        return f;
      }
      if (isWindows())
      {
        f = toDirectory(System.getenv("TEMP"));
        if (f != null)
        {
          logger.debug("-Dio.netty.tmpdir: {} (%TEMP%)", f);
          return f;
        }
        String userprofile = System.getenv("USERPROFILE");
        if (userprofile != null)
        {
          f = toDirectory(userprofile + "\\AppData\\Local\\Temp");
          if (f != null)
          {
            logger.debug("-Dio.netty.tmpdir: {} (%USERPROFILE%\\AppData\\Local\\Temp)", f);
            return f;
          }
          f = toDirectory(userprofile + "\\Local Settings\\Temp");
          if (f != null)
          {
            logger.debug("-Dio.netty.tmpdir: {} (%USERPROFILE%\\Local Settings\\Temp)", f);
            return f;
          }
        }
      }
      else
      {
        f = toDirectory(System.getenv("TMPDIR"));
        if (f != null)
        {
          logger.debug("-Dio.netty.tmpdir: {} ($TMPDIR)", f);
          return f;
        }
      }
    }
    catch (Throwable localThrowable) {}
    File f;
    File f;
    if (isWindows()) {
      f = new File("C:\\Windows\\Temp");
    } else {
      f = new File("/tmp");
    }
    logger.warn("Failed to get the temporary directory; falling back to: {}", f);
    return f;
  }
  
  private static File toDirectory(String path)
  {
    if (path == null) {
      return null;
    }
    File f = new File(path);
    f.mkdirs();
    if (!f.isDirectory()) {
      return null;
    }
    try
    {
      return f.getAbsoluteFile();
    }
    catch (Exception ignored) {}
    return f;
  }
  
  private static int bitMode0()
  {
    int bitMode = SystemPropertyUtil.getInt("io.netty.bitMode", 0);
    if (bitMode > 0)
    {
      logger.debug("-Dio.netty.bitMode: {}", Integer.valueOf(bitMode));
      return bitMode;
    }
    bitMode = SystemPropertyUtil.getInt("sun.arch.data.model", 0);
    if (bitMode > 0)
    {
      logger.debug("-Dio.netty.bitMode: {} (sun.arch.data.model)", Integer.valueOf(bitMode));
      return bitMode;
    }
    bitMode = SystemPropertyUtil.getInt("com.ibm.vm.bitmode", 0);
    if (bitMode > 0)
    {
      logger.debug("-Dio.netty.bitMode: {} (com.ibm.vm.bitmode)", Integer.valueOf(bitMode));
      return bitMode;
    }
    String arch = SystemPropertyUtil.get("os.arch", "").toLowerCase(Locale.US).trim();
    if (("amd64".equals(arch)) || ("x86_64".equals(arch))) {
      bitMode = 64;
    } else if (("i386".equals(arch)) || ("i486".equals(arch)) || ("i586".equals(arch)) || ("i686".equals(arch))) {
      bitMode = 32;
    }
    if (bitMode > 0) {
      logger.debug("-Dio.netty.bitMode: {} (os.arch: {})", Integer.valueOf(bitMode), arch);
    }
    String vm = SystemPropertyUtil.get("java.vm.name", "").toLowerCase(Locale.US);
    Pattern BIT_PATTERN = Pattern.compile("([1-9][0-9]+)-?bit");
    Matcher m = BIT_PATTERN.matcher(vm);
    if (m.find()) {
      return Integer.parseInt(m.group(1));
    }
    return 64;
  }
  
  private static int addressSize0()
  {
    if (!hasUnsafe()) {
      return -1;
    }
    return PlatformDependent0.addressSize();
  }
  
  private static boolean equalsSafe(byte[] bytes1, int startPos1, byte[] bytes2, int startPos2, int length)
  {
    int end = startPos1 + length;
    int i = startPos1;
    for (int j = startPos2; i < end; j++)
    {
      if (bytes1[i] != bytes2[j]) {
        return false;
      }
      i++;
    }
    return true;
  }
  
  static int hashCodeAsciiSafe(byte[] bytes, int startPos, int length)
  {
    int hash = -1028477387;
    int remainingBytes = length & 0x7;
    int end = startPos + remainingBytes;
    for (int i = startPos - 8 + length; i >= end; i -= 8) {
      hash = PlatformDependent0.hashCodeAsciiCompute(getLongSafe(bytes, i), hash);
    }
    switch (remainingBytes)
    {
    case 7: 
      return ((hash * 461845907 + PlatformDependent0.hashCodeAsciiSanitize(bytes[startPos])) * 461845907 + PlatformDependent0.hashCodeAsciiSanitize(getShortSafe(bytes, startPos + 1))) * 461845907 + PlatformDependent0.hashCodeAsciiSanitize(getIntSafe(bytes, startPos + 3));
    case 6: 
      return (hash * 461845907 + PlatformDependent0.hashCodeAsciiSanitize(getShortSafe(bytes, startPos))) * 461845907 + PlatformDependent0.hashCodeAsciiSanitize(getIntSafe(bytes, startPos + 2));
    case 5: 
      return (hash * 461845907 + PlatformDependent0.hashCodeAsciiSanitize(bytes[startPos])) * 461845907 + PlatformDependent0.hashCodeAsciiSanitize(getIntSafe(bytes, startPos + 1));
    case 4: 
      return hash * 461845907 + PlatformDependent0.hashCodeAsciiSanitize(getIntSafe(bytes, startPos));
    case 3: 
      return (hash * 461845907 + PlatformDependent0.hashCodeAsciiSanitize(bytes[startPos])) * 461845907 + PlatformDependent0.hashCodeAsciiSanitize(getShortSafe(bytes, startPos + 1));
    case 2: 
      return hash * 461845907 + PlatformDependent0.hashCodeAsciiSanitize(getShortSafe(bytes, startPos));
    case 1: 
      return hash * 461845907 + PlatformDependent0.hashCodeAsciiSanitize(bytes[startPos]);
    }
    return hash;
  }
  
  private static final class AtomicLongCounter
    extends AtomicLong
    implements LongCounter
  {
    private static final long serialVersionUID = 4074772784610639305L;
    
    public void add(long delta)
    {
      addAndGet(delta);
    }
    
    public void increment()
    {
      incrementAndGet();
    }
    
    public void decrement()
    {
      decrementAndGet();
    }
    
    public long value()
    {
      return get();
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.util.internal.PlatformDependent
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */