package io.netty.buffer;

import io.netty.util.Recycler.Handle;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

abstract class PooledByteBuf<T>
  extends AbstractReferenceCountedByteBuf
{
  private final Recycler.Handle<PooledByteBuf<T>> recyclerHandle;
  protected PoolChunk<T> chunk;
  protected long handle;
  protected T memory;
  protected int offset;
  protected int length;
  int maxLength;
  PoolThreadCache cache;
  private ByteBuffer tmpNioBuf;
  
  protected PooledByteBuf(Recycler.Handle<? extends PooledByteBuf<T>> recyclerHandle, int maxCapacity)
  {
    super(maxCapacity);
    this.recyclerHandle = recyclerHandle;
  }
  
  void init(PoolChunk<T> chunk, long handle, int offset, int length, int maxLength, PoolThreadCache cache)
  {
    assert (handle >= 0L);
    assert (chunk != null);
    
    this.chunk = chunk;
    this.handle = handle;
    memory = memory;
    this.offset = offset;
    this.length = length;
    this.maxLength = maxLength;
    tmpNioBuf = null;
    this.cache = cache;
  }
  
  void initUnpooled(PoolChunk<T> chunk, int length)
  {
    assert (chunk != null);
    
    this.chunk = chunk;
    handle = 0L;
    memory = memory;
    offset = 0;
    this.length = (maxLength = length);
    tmpNioBuf = null;
    cache = null;
  }
  
  final void reuse(int maxCapacity)
  {
    maxCapacity(maxCapacity);
    setRefCnt(1);
    setIndex0(0, 0);
    discardMarks();
  }
  
  public final int capacity()
  {
    return length;
  }
  
  public final ByteBuf capacity(int newCapacity)
  {
    ensureAccessible();
    if (chunk.unpooled)
    {
      if (newCapacity == length) {
        return this;
      }
    }
    else if (newCapacity > length)
    {
      if (newCapacity <= maxLength)
      {
        length = newCapacity;
        return this;
      }
    }
    else if (newCapacity < length)
    {
      if (newCapacity > maxLength >>> 1) {
        if (maxLength <= 512)
        {
          if (newCapacity > maxLength - 16)
          {
            length = newCapacity;
            setIndex(Math.min(readerIndex(), newCapacity), Math.min(writerIndex(), newCapacity));
            return this;
          }
        }
        else
        {
          length = newCapacity;
          setIndex(Math.min(readerIndex(), newCapacity), Math.min(writerIndex(), newCapacity));
          return this;
        }
      }
    }
    else {
      return this;
    }
    chunk.arena.reallocate(this, newCapacity, true);
    return this;
  }
  
  public final ByteBufAllocator alloc()
  {
    return chunk.arena.parent;
  }
  
  public final ByteOrder order()
  {
    return ByteOrder.BIG_ENDIAN;
  }
  
  public final ByteBuf unwrap()
  {
    return null;
  }
  
  public final ByteBuf retainedDuplicate()
  {
    return PooledDuplicatedByteBuf.newInstance(this, this, readerIndex(), writerIndex());
  }
  
  public final ByteBuf retainedSlice()
  {
    int index = readerIndex();
    return retainedSlice(index, writerIndex() - index);
  }
  
  public final ByteBuf retainedSlice(int index, int length)
  {
    return PooledSlicedByteBuf.newInstance(this, this, index, length);
  }
  
  protected final ByteBuffer internalNioBuffer()
  {
    ByteBuffer tmpNioBuf = this.tmpNioBuf;
    if (tmpNioBuf == null) {
      this.tmpNioBuf = (tmpNioBuf = newInternalNioBuffer(memory));
    }
    return tmpNioBuf;
  }
  
  protected abstract ByteBuffer newInternalNioBuffer(T paramT);
  
  protected final void deallocate()
  {
    if (this.handle >= 0L)
    {
      long handle = this.handle;
      this.handle = -1L;
      memory = null;
      chunk.arena.free(chunk, handle, maxLength, cache);
      recycle();
    }
  }
  
  private void recycle()
  {
    recyclerHandle.recycle(this);
  }
  
  protected final int idx(int index)
  {
    return offset + index;
  }
}

/* Location:
 * Qualified Name:     io.netty.buffer.PooledByteBuf
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */