package io.netty.channel.epoll;

import io.netty.buffer.ByteBuf;
import io.netty.buffer.CompositeByteBuf;
import io.netty.channel.ChannelOutboundBuffer.MessageProcessor;
import io.netty.util.internal.PlatformDependent;
import java.nio.ByteBuffer;

final class IovArray
  implements ChannelOutboundBuffer.MessageProcessor
{
  private static final int ADDRESS_SIZE = PlatformDependent.addressSize();
  private static final int IOV_SIZE = 2 * ADDRESS_SIZE;
  private static final int CAPACITY = Native.IOV_MAX * IOV_SIZE;
  private final long memoryAddress;
  private int count;
  private long size;
  
  IovArray()
  {
    memoryAddress = PlatformDependent.allocateMemory(CAPACITY);
  }
  
  void clear()
  {
    count = 0;
    size = 0L;
  }
  
  boolean add(ByteBuf buf)
  {
    if (count == Native.IOV_MAX) {
      return false;
    }
    int len = buf.readableBytes();
    if (len == 0) {
      return true;
    }
    long addr = buf.memoryAddress();
    int offset = buf.readerIndex();
    return add(addr, offset, len);
  }
  
  private boolean add(long addr, int offset, int len)
  {
    if (len == 0) {
      return true;
    }
    long baseOffset = memoryAddress(count++);
    long lengthOffset = baseOffset + ADDRESS_SIZE;
    if (Native.SSIZE_MAX - len < size) {
      return false;
    }
    size += len;
    if (ADDRESS_SIZE == 8)
    {
      PlatformDependent.putLong(baseOffset, addr + offset);
      PlatformDependent.putLong(lengthOffset, len);
    }
    else
    {
      assert (ADDRESS_SIZE == 4);
      PlatformDependent.putInt(baseOffset, (int)addr + offset);
      PlatformDependent.putInt(lengthOffset, len);
    }
    return true;
  }
  
  boolean add(CompositeByteBuf buf)
  {
    ByteBuffer[] buffers = buf.nioBuffers();
    if (count + buffers.length >= Native.IOV_MAX) {
      return false;
    }
    for (int i = 0; i < buffers.length; i++)
    {
      ByteBuffer nioBuffer = buffers[i];
      int offset = nioBuffer.position();
      int len = nioBuffer.limit() - nioBuffer.position();
      if (len != 0)
      {
        long addr = PlatformDependent.directBufferAddress(nioBuffer);
        if (!add(addr, offset, len)) {
          return false;
        }
      }
    }
    return true;
  }
  
  long processWritten(int index, long written)
  {
    long baseOffset = memoryAddress(index);
    long lengthOffset = baseOffset + ADDRESS_SIZE;
    if (ADDRESS_SIZE == 8)
    {
      long len = PlatformDependent.getLong(lengthOffset);
      if (len > written)
      {
        long offset = PlatformDependent.getLong(baseOffset);
        PlatformDependent.putLong(baseOffset, offset + written);
        PlatformDependent.putLong(lengthOffset, len - written);
        return -1L;
      }
      return len;
    }
    assert (ADDRESS_SIZE == 4);
    long len = PlatformDependent.getInt(lengthOffset);
    if (len > written)
    {
      int offset = PlatformDependent.getInt(baseOffset);
      PlatformDependent.putInt(baseOffset, (int)(offset + written));
      PlatformDependent.putInt(lengthOffset, (int)(len - written));
      return -1L;
    }
    return len;
  }
  
  int count()
  {
    return count;
  }
  
  long size()
  {
    return size;
  }
  
  long memoryAddress(int offset)
  {
    return memoryAddress + IOV_SIZE * offset;
  }
  
  void release()
  {
    PlatformDependent.freeMemory(memoryAddress);
  }
  
  public boolean processMessage(Object msg)
    throws Exception
  {
    if ((msg instanceof ByteBuf))
    {
      if ((msg instanceof CompositeByteBuf)) {
        return add((CompositeByteBuf)msg);
      }
      return add((ByteBuf)msg);
    }
    return false;
  }
}

/* Location:
 * Qualified Name:     io.netty.channel.epoll.IovArray
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */