package io.netty.handler.codec.http;

import io.netty.channel.Channel;
import io.netty.channel.ChannelHandler;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.ChannelMetadata;
import io.netty.channel.embedded.EmbeddedChannel;
import io.netty.handler.codec.compression.ZlibCodecFactory;
import io.netty.handler.codec.compression.ZlibWrapper;
import io.netty.util.AsciiString;

public class HttpContentCompressor
  extends HttpContentEncoder
{
  private final int compressionLevel;
  private final int windowBits;
  private final int memLevel;
  private ChannelHandlerContext ctx;
  
  public HttpContentCompressor()
  {
    this(6);
  }
  
  public HttpContentCompressor(int compressionLevel)
  {
    this(compressionLevel, 15, 8);
  }
  
  public HttpContentCompressor(int compressionLevel, int windowBits, int memLevel)
  {
    if ((compressionLevel < 0) || (compressionLevel > 9)) {
      throw new IllegalArgumentException("compressionLevel: " + compressionLevel + " (expected: 0-9)");
    }
    if ((windowBits < 9) || (windowBits > 15)) {
      throw new IllegalArgumentException("windowBits: " + windowBits + " (expected: 9-15)");
    }
    if ((memLevel < 1) || (memLevel > 9)) {
      throw new IllegalArgumentException("memLevel: " + memLevel + " (expected: 1-9)");
    }
    this.compressionLevel = compressionLevel;
    this.windowBits = windowBits;
    this.memLevel = memLevel;
  }
  
  public void handlerAdded(ChannelHandlerContext ctx)
    throws Exception
  {
    this.ctx = ctx;
  }
  
  protected HttpContentEncoder.Result beginEncode(HttpResponse headers, String acceptEncoding)
    throws Exception
  {
    String contentEncoding = headers.headers().get(HttpHeaderNames.CONTENT_ENCODING);
    if ((contentEncoding != null) && (!HttpHeaderValues.IDENTITY.contentEqualsIgnoreCase(contentEncoding))) {
      return null;
    }
    ZlibWrapper wrapper = determineWrapper(acceptEncoding);
    if (wrapper == null) {
      return null;
    }
    String targetContentEncoding;
    String targetContentEncoding;
    switch (wrapper)
    {
    case GZIP: 
      targetContentEncoding = "gzip";
      break;
    case ZLIB: 
      targetContentEncoding = "deflate";
      break;
    default: 
      throw new Error();
    }
    String targetContentEncoding;
    return new HttpContentEncoder.Result(targetContentEncoding, new EmbeddedChannel(ctx.channel().id(), ctx.channel().metadata().hasDisconnect(), ctx.channel().config(), new ChannelHandler[] { ZlibCodecFactory.newZlibEncoder(wrapper, compressionLevel, windowBits, memLevel) }));
  }
  
  protected ZlibWrapper determineWrapper(String acceptEncoding)
  {
    float starQ = -1.0F;
    float gzipQ = -1.0F;
    float deflateQ = -1.0F;
    for (String encoding : acceptEncoding.split(","))
    {
      float q = 1.0F;
      int equalsPos = encoding.indexOf('=');
      if (equalsPos != -1) {
        try
        {
          q = Float.parseFloat(encoding.substring(equalsPos + 1));
        }
        catch (NumberFormatException e)
        {
          q = 0.0F;
        }
      }
      if (encoding.contains("*")) {
        starQ = q;
      } else if ((encoding.contains("gzip")) && (q > gzipQ)) {
        gzipQ = q;
      } else if ((encoding.contains("deflate")) && (q > deflateQ)) {
        deflateQ = q;
      }
    }
    if ((gzipQ > 0.0F) || (deflateQ > 0.0F))
    {
      if (gzipQ >= deflateQ) {
        return ZlibWrapper.GZIP;
      }
      return ZlibWrapper.ZLIB;
    }
    if (starQ > 0.0F)
    {
      if (gzipQ == -1.0F) {
        return ZlibWrapper.GZIP;
      }
      if (deflateQ == -1.0F) {
        return ZlibWrapper.ZLIB;
      }
    }
    return null;
  }
}

/* Location:
 * Qualified Name:     io.netty.handler.codec.http.HttpContentCompressor
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */