package io.netty.handler.codec.http2;

import io.netty.buffer.ByteBuf;
import io.netty.buffer.Unpooled;
import io.netty.handler.codec.http2.internal.hpack.Encoder;
import io.netty.util.internal.ObjectUtil;
import java.util.Map.Entry;

public class DefaultHttp2HeadersEncoder
  implements Http2HeadersEncoder, Http2HeadersEncoder.Configuration
{
  private final Encoder encoder;
  private final Http2HeadersEncoder.SensitivityDetector sensitivityDetector;
  private final Http2HeaderTable headerTable;
  private final ByteBuf tableSizeChangeOutput = Unpooled.buffer();
  
  public DefaultHttp2HeadersEncoder()
  {
    this(4096, NEVER_SENSITIVE);
  }
  
  public DefaultHttp2HeadersEncoder(int maxHeaderTableSize, Http2HeadersEncoder.SensitivityDetector sensitivityDetector)
  {
    this.sensitivityDetector = ((Http2HeadersEncoder.SensitivityDetector)ObjectUtil.checkNotNull(sensitivityDetector, "sensitiveDetector"));
    encoder = new Encoder(maxHeaderTableSize);
    headerTable = new Http2HeaderTableEncoder(null);
  }
  
  public void encodeHeaders(Http2Headers headers, ByteBuf buffer)
    throws Http2Exception
  {
    try
    {
      if (headers.size() > headerTable.maxHeaderListSize()) {
        throw Http2Exception.connectionError(Http2Error.PROTOCOL_ERROR, "Number of headers (%d) exceeds maxHeaderListSize (%d)", new Object[] { Integer.valueOf(headers.size()), Integer.valueOf(headerTable.maxHeaderListSize()) });
      }
      if (tableSizeChangeOutput.isReadable())
      {
        buffer.writeBytes(tableSizeChangeOutput);
        tableSizeChangeOutput.clear();
      }
      for (Map.Entry<CharSequence, CharSequence> header : headers) {
        encodeHeader(buffer, (CharSequence)header.getKey(), (CharSequence)header.getValue());
      }
    }
    catch (Http2Exception e)
    {
      throw e;
    }
    catch (Throwable t)
    {
      throw Http2Exception.connectionError(Http2Error.COMPRESSION_ERROR, t, "Failed encoding headers block: %s", new Object[] { t.getMessage() });
    }
  }
  
  public Http2HeaderTable headerTable()
  {
    return headerTable;
  }
  
  public Http2HeadersEncoder.Configuration configuration()
  {
    return this;
  }
  
  private void encodeHeader(ByteBuf out, CharSequence key, CharSequence value)
  {
    encoder.encodeHeader(out, key, value, sensitivityDetector.isSensitive(key, value));
  }
  
  private final class Http2HeaderTableEncoder
    extends DefaultHttp2HeaderTableListSize
    implements Http2HeaderTable
  {
    private Http2HeaderTableEncoder() {}
    
    public void maxHeaderTableSize(int max)
      throws Http2Exception
    {
      if (max < 0) {
        throw Http2Exception.connectionError(Http2Error.PROTOCOL_ERROR, "Header Table Size must be non-negative but was %d", new Object[] { Integer.valueOf(max) });
      }
      try
      {
        encoder.setMaxHeaderTableSize(tableSizeChangeOutput, max);
      }
      catch (Throwable t)
      {
        throw new Http2Exception(Http2Error.PROTOCOL_ERROR, t.getMessage(), t);
      }
    }
    
    public int maxHeaderTableSize()
    {
      return encoder.getMaxHeaderTableSize();
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.handler.codec.http2.DefaultHttp2HeadersEncoder
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */