package io.netty.handler.codec.memcache.binary;

import io.netty.buffer.ByteBuf;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.CombinedChannelDuplexHandler;
import io.netty.handler.codec.PrematureChannelClosureException;
import io.netty.handler.codec.memcache.LastMemcacheContent;
import java.util.List;
import java.util.concurrent.atomic.AtomicLong;

public final class BinaryMemcacheClientCodec
  extends CombinedChannelDuplexHandler<BinaryMemcacheResponseDecoder, BinaryMemcacheRequestEncoder>
{
  private final boolean failOnMissingResponse;
  private final AtomicLong requestResponseCounter = new AtomicLong();
  
  public BinaryMemcacheClientCodec()
  {
    this(8192);
  }
  
  public BinaryMemcacheClientCodec(int decodeChunkSize)
  {
    this(decodeChunkSize, false);
  }
  
  public BinaryMemcacheClientCodec(int decodeChunkSize, boolean failOnMissingResponse)
  {
    this.failOnMissingResponse = failOnMissingResponse;
    init(new Decoder(decodeChunkSize), new Encoder(null));
  }
  
  private final class Encoder
    extends BinaryMemcacheRequestEncoder
  {
    private Encoder() {}
    
    protected void encode(ChannelHandlerContext ctx, Object msg, List<Object> out)
      throws Exception
    {
      super.encode(ctx, msg, out);
      if ((failOnMissingResponse) && ((msg instanceof LastMemcacheContent))) {
        requestResponseCounter.incrementAndGet();
      }
    }
  }
  
  private final class Decoder
    extends BinaryMemcacheResponseDecoder
  {
    Decoder(int chunkSize)
    {
      super();
    }
    
    protected void decode(ChannelHandlerContext ctx, ByteBuf in, List<Object> out)
      throws Exception
    {
      int oldSize = out.size();
      super.decode(ctx, in, out);
      if (failOnMissingResponse)
      {
        int size = out.size();
        for (int i = oldSize; i < size; i++)
        {
          Object msg = out.get(i);
          if ((msg instanceof LastMemcacheContent)) {
            requestResponseCounter.decrementAndGet();
          }
        }
      }
    }
    
    public void channelInactive(ChannelHandlerContext ctx)
      throws Exception
    {
      super.channelInactive(ctx);
      if (failOnMissingResponse)
      {
        long missingResponses = requestResponseCounter.get();
        if (missingResponses > 0L) {
          ctx.fireExceptionCaught(new PrematureChannelClosureException("channel gone inactive with " + missingResponses + " missing response(s)"));
        }
      }
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.handler.codec.memcache.binary.BinaryMemcacheClientCodec
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */