package io.netty.util.concurrent;

import io.netty.util.internal.ObjectUtil;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;

public class PromiseNotifier<V, F extends Future<V>>
  implements GenericFutureListener<F>
{
  private static final InternalLogger logger = InternalLoggerFactory.getInstance(PromiseNotifier.class);
  private final Promise<? super V>[] promises;
  private final boolean logNotifyFailure;
  
  @SafeVarargs
  public PromiseNotifier(Promise<? super V>... promises)
  {
    this(true, promises);
  }
  
  @SafeVarargs
  public PromiseNotifier(boolean logNotifyFailure, Promise<? super V>... promises)
  {
    ObjectUtil.checkNotNull(promises, "promises");
    for (Promise<? super V> promise : promises) {
      if (promise == null) {
        throw new IllegalArgumentException("promises contains null Promise");
      }
    }
    this.promises = ((Promise[])promises.clone());
    this.logNotifyFailure = logNotifyFailure;
  }
  
  public void operationComplete(F future)
    throws Exception
  {
    if (future.isSuccess())
    {
      V result = future.get();
      for (Promise<? super V> p : promises) {
        if ((!p.trySuccess(result)) && (logNotifyFailure)) {
          logger.warn("Failed to mark a promise as success because it is done already: {}", p);
        }
      }
    }
    else if (future.isCancelled())
    {
      for (Promise<? super V> p : promises) {
        if ((!p.cancel(false)) && (logNotifyFailure)) {
          logger.warn("Failed to cancel a promise because it is done already: {}", p);
        }
      }
    }
    else
    {
      Throwable cause = future.cause();
      for (Promise<? super V> p : promises) {
        if ((!p.tryFailure(cause)) && (logNotifyFailure)) {
          logger.warn("Failed to mark a promise as failure because it's done already: {}", p, cause);
        }
      }
    }
  }
}

/* Location:
 * Qualified Name:     io.netty.util.concurrent.PromiseNotifier
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */