package io.netty.util.concurrent;

import java.util.Queue;
import java.util.concurrent.Callable;
import java.util.concurrent.Delayed;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;

final class ScheduledFutureTask<V>
  extends PromiseTask<V>
  implements ScheduledFuture<V>
{
  private static final AtomicLong nextTaskId = new AtomicLong();
  private static final long START_TIME = System.nanoTime();
  
  static long nanoTime()
  {
    return System.nanoTime() - START_TIME;
  }
  
  static long deadlineNanos(long delay)
  {
    return nanoTime() + delay;
  }
  
  private final long id = nextTaskId.getAndIncrement();
  private long deadlineNanos;
  private final long periodNanos;
  
  ScheduledFutureTask(AbstractScheduledEventExecutor executor, Runnable runnable, V result, long nanoTime)
  {
    this(executor, toCallable(runnable, result), nanoTime);
  }
  
  ScheduledFutureTask(AbstractScheduledEventExecutor executor, Callable<V> callable, long nanoTime, long period)
  {
    super(executor, callable);
    if (period == 0L) {
      throw new IllegalArgumentException("period: 0 (expected: != 0)");
    }
    deadlineNanos = nanoTime;
    periodNanos = period;
  }
  
  ScheduledFutureTask(AbstractScheduledEventExecutor executor, Callable<V> callable, long nanoTime)
  {
    super(executor, callable);
    deadlineNanos = nanoTime;
    periodNanos = 0L;
  }
  
  protected EventExecutor executor()
  {
    return super.executor();
  }
  
  public long deadlineNanos()
  {
    return deadlineNanos;
  }
  
  public long delayNanos()
  {
    return Math.max(0L, deadlineNanos() - nanoTime());
  }
  
  public long delayNanos(long currentTimeNanos)
  {
    return Math.max(0L, deadlineNanos() - (currentTimeNanos - START_TIME));
  }
  
  public long getDelay(TimeUnit unit)
  {
    return unit.convert(delayNanos(), TimeUnit.NANOSECONDS);
  }
  
  public int compareTo(Delayed o)
  {
    if (this == o) {
      return 0;
    }
    ScheduledFutureTask<?> that = (ScheduledFutureTask)o;
    long d = deadlineNanos() - that.deadlineNanos();
    if (d < 0L) {
      return -1;
    }
    if (d > 0L) {
      return 1;
    }
    if (id < id) {
      return -1;
    }
    if (id == id) {
      throw new Error();
    }
    return 1;
  }
  
  public void run()
  {
    assert (executor().inEventLoop());
    try
    {
      if (periodNanos == 0L)
      {
        if (setUncancellableInternal())
        {
          V result = task.call();
          setSuccessInternal(result);
        }
      }
      else if (!isCancelled())
      {
        task.call();
        if (!executor().isShutdown())
        {
          long p = periodNanos;
          if (p > 0L) {
            deadlineNanos += p;
          } else {
            deadlineNanos = (nanoTime() - p);
          }
          if (!isCancelled())
          {
            Queue<ScheduledFutureTask<?>> scheduledTaskQueue = executorscheduledTaskQueue;
            
            assert (scheduledTaskQueue != null);
            scheduledTaskQueue.add(this);
          }
        }
      }
    }
    catch (Throwable cause)
    {
      setFailureInternal(cause);
    }
  }
  
  public boolean cancel(boolean mayInterruptIfRunning)
  {
    boolean canceled = super.cancel(mayInterruptIfRunning);
    if (canceled) {
      ((AbstractScheduledEventExecutor)executor()).removeScheduled(this);
    }
    return canceled;
  }
  
  boolean cancelWithoutRemove(boolean mayInterruptIfRunning)
  {
    return super.cancel(mayInterruptIfRunning);
  }
  
  protected StringBuilder toStringBuilder()
  {
    StringBuilder buf = super.toStringBuilder();
    buf.setCharAt(buf.length() - 1, ',');
    
    return buf.append(" id: ").append(id).append(", deadline: ").append(deadlineNanos).append(", period: ").append(periodNanos).append(')');
  }
}

/* Location:
 * Qualified Name:     io.netty.util.concurrent.ScheduledFutureTask
 * Java Class Version: 6 (50.0)
 * JD-Core Version:    0.7.1
 */