package org.apache.commons.io.output;

import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.List;

public class ByteArrayOutputStream
  extends OutputStream
{
  private List buffers = new ArrayList();
  private int currentBufferIndex;
  private int filledBufferSum;
  private byte[] currentBuffer;
  private int count;
  
  public ByteArrayOutputStream()
  {
    this(1024);
  }
  
  public ByteArrayOutputStream(int size)
  {
    if (size < 0) {
      throw new IllegalArgumentException("Negative initial size: " + size);
    }
    needNewBuffer(size);
  }
  
  private byte[] getBuffer(int index)
  {
    return (byte[])buffers.get(index);
  }
  
  private void needNewBuffer(int newcount)
  {
    if (currentBufferIndex < buffers.size() - 1)
    {
      filledBufferSum += currentBuffer.length;
      
      currentBufferIndex += 1;
      currentBuffer = getBuffer(currentBufferIndex);
    }
    else
    {
      int newBufferSize;
      if (currentBuffer == null)
      {
        newBufferSize = newcount;
        filledBufferSum = 0;
      }
      else
      {
        newBufferSize = Math.max(currentBuffer.length << 1, newcount - filledBufferSum);
        
        filledBufferSum += currentBuffer.length;
      }
      currentBufferIndex += 1;
      currentBuffer = new byte[newBufferSize];
      buffers.add(currentBuffer);
    }
  }
  
  public synchronized void write(byte[] b, int off, int len)
  {
    if ((off < 0) || (off > b.length) || (len < 0) || (off + len > b.length) || (off + len < 0)) {
      throw new IndexOutOfBoundsException();
    }
    if (len == 0) {
      return;
    }
    int newcount = count + len;
    int remaining = len;
    int inBufferPos = count - filledBufferSum;
    while (remaining > 0)
    {
      int part = Math.min(remaining, currentBuffer.length - inBufferPos);
      System.arraycopy(b, off + len - remaining, currentBuffer, inBufferPos, part);
      remaining -= part;
      if (remaining > 0)
      {
        needNewBuffer(newcount);
        inBufferPos = 0;
      }
    }
    count = newcount;
  }
  
  public synchronized void write(int b)
  {
    write(new byte[] { (byte)b }, 0, 1);
  }
  
  public int size()
  {
    return count;
  }
  
  public void close()
    throws IOException
  {}
  
  public synchronized void reset()
  {
    count = 0;
    filledBufferSum = 0;
    currentBufferIndex = 0;
    currentBuffer = getBuffer(currentBufferIndex);
  }
  
  public synchronized void writeTo(OutputStream out)
    throws IOException
  {
    int remaining = count;
    for (int i = 0; i < buffers.size(); i++)
    {
      byte[] buf = getBuffer(i);
      int c = Math.min(buf.length, remaining);
      out.write(buf, 0, c);
      remaining -= c;
      if (remaining == 0) {
        break;
      }
    }
  }
  
  public synchronized byte[] toByteArray()
  {
    int remaining = count;
    int pos = 0;
    byte[] newbuf = new byte[count];
    for (int i = 0; i < buffers.size(); i++)
    {
      byte[] buf = getBuffer(i);
      int c = Math.min(buf.length, remaining);
      System.arraycopy(buf, 0, newbuf, pos, c);
      pos += c;
      remaining -= c;
      if (remaining == 0) {
        break;
      }
    }
    return newbuf;
  }
  
  public String toString()
  {
    return new String(toByteArray());
  }
  
  public String toString(String enc)
    throws UnsupportedEncodingException
  {
    return new String(toByteArray(), enc);
  }
}

/* Location:
 * Qualified Name:     org.apache.commons.io.output.ByteArrayOutputStream
 * Java Class Version: 1.1 (45.3)
 * JD-Core Version:    0.7.1
 */